package org.eso.oca.backend;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.math.BigDecimal;
import java.net.InetAddress;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.TreeSet;

import org.apache.log4j.Logger;
import org.eso.oca.fits.OCAFile;

/**
 * This class implements the ABVisitor interface to create a DFO-style
 * association block.
 * 
 * @author szampier
 */
public class ABFileWriter implements ABVisitor {
    /** the directory where association blocks are created */
    private String dir;

    /** some useful constants */
    static final String SEPARATOR = "\t";

    static final String EOL = System.getProperty("line.separator");

    static final String UNDEFINED = "UNDEFINED";

    static final String NONE = "NONE";

    static final String RB_CONTENT = "RB_CONTENT";

    static final String SOF_CONTENT = "SOF_CONTENT";

    static final String RAWFILE = "RAWFILE";

    static final String RASSOC = "RASSOC";

    static final String MASSOC = "MASSOC";

    static final String MCALIB = "MCALIB";

    static Logger logger = Logger.getLogger(ABFileWriter.class);

    /**
     * Constructor
     * 
     * @param directory
     *            where ABs should be created
     */
    public ABFileWriter(String dir) {
        if (dir == null)
            throw new IllegalArgumentException("null AB dir");
        this.dir = dir;
    }

    /**
     * Create an association block as a text file in the AB directory
     */
    public void visit(AssociationBlock ab) throws ABVisitorException {
        try {
            OCAFile abExemplar = ab.getExemplar();
            String abEvent = ab.getGroupEvent();
            if (abEvent == null)
                abEvent = "SINGLE";

            String abRootName = ab.getABName();
            // if windows is used, make filename windows compliant
            String osname = System.getProperty("os.name");
            if (osname.indexOf("Windows") >= 0) {
                abRootName = abRootName.replace(':', '_');
            }

            String abFilename = dir + File.separator + abRootName + ".ab";
            File abFile = new File(abFilename.toString());
            FileWriter out = new FileWriter(abFile);

            /* General Info */
            String instrume;
            try {
                instrume = ab.getInstrument();
            } catch (Exception e3) {
                instrume = UNDEFINED;
            }

            String dfoDate = System.getProperty("DFO_DATE");
            String abDate = dfoDate == null ? NONE : dfoDate;
            String configVersion = System.getProperty("CONFIG_VERSION");
            String toolVersion = System.getProperty("VERSION");
            out.write("# general information" + EOL);
            out.write("TOOL_VERSION" + SEPARATOR + toolVersion + EOL);
            out.write("CONFIG_VERSION" + SEPARATOR + configVersion + EOL);
            out.write("INSTRUMENT" + SEPARATOR + instrume + EOL);
            out.write("DATE" + SEPARATOR + SEPARATOR + abDate + EOL);
            out.write(EOL);

            /* Classification Info */
            String dprCatg;
            String rawType;
            String packDir;
            try {
                dprCatg = abExemplar.getDprCatg();
            } catch (Exception e4) {
                dprCatg = NONE;
            }
            try {
                rawType = abExemplar.getRawType();
            } catch (Exception e5) {
                rawType = NONE;
            }
            try {
                packDir = abExemplar.getPackDir();
            } catch (Exception e5) {
                packDir = NONE;
            }
            out.write("DPR_CATG" + SEPARATOR + dprCatg + EOL);
            out.write("RAW_TYPE" + SEPARATOR + rawType + EOL);
            out.write("RECIPE" + SEPARATOR + SEPARATOR + ab.getRecipe() + EOL);
            out.write("DRS_TYPE" + SEPARATOR + System.getProperty("DRS_TYPE")
                    + EOL);
            out.write("PACK_DIR" + SEPARATOR + packDir + EOL);
            out.write(EOL);

            /* AB Info */
            out.write("AB_NAME" + SEPARATOR + SEPARATOR + abFile.getName()
                    + EOL);
            out.write("AB_EVENT" + SEPARATOR + abEvent + EOL);
            out.write("COMPLETENESS" + SEPARATOR + (ab.isValid() ? "" : "IN")
                    + "COMPLETE" + EOL);
            out.write("PROCESS_STATUS" + SEPARATOR + "CREATED" + EOL);
            out.write(EOL);

            /* OB Info */
            String obsProgId;
            try {
                obsProgId = abExemplar.getKeywordValue("OBS.PROG.ID");
            } catch (Exception e6) {
                obsProgId = UNDEFINED;
            }
            String obsId;
            try {
                obsId = abExemplar.getKeywordValue("OBS.ID");
            } catch (Exception e7) {
                obsId = UNDEFINED;
            }
            out.write("OBS_PROG_ID" + SEPARATOR + obsProgId + EOL);
            out.write("OBS_ID" + SEPARATOR + SEPARATOR + obsId + EOL);
            out.write(EOL);

            /* MJD Info */
            BigDecimal mjdObs;
            BigDecimal modifiedMjdObs;
            try {
                mjdObs = abExemplar.getMjdObs();
            } catch (Exception e8) {
                mjdObs = new BigDecimal(0.0);
            }
            try {
                modifiedMjdObs = ab.getModifiedMjdObs();
            } catch (Exception e9) {
                modifiedMjdObs = new BigDecimal(0.0);
            }
            out.write("MJD-OBS" + SEPARATOR + SEPARATOR
                    + mjdObs.setScale(8, BigDecimal.ROUND_HALF_EVEN) + EOL);
            out.write("MJD-OBS_MOD" + SEPARATOR
                    + modifiedMjdObs.setScale(8, BigDecimal.ROUND_HALF_EVEN)
                    + EOL);
            out.write(EOL);

            /* General Info */
            String prodRootName = "r." + abRootName;
            // String logName = abRootName + ".rblog";

            out.write("PROD_ROOT_NAME" + SEPARATOR + prodRootName + EOL);
            out.write(EOL);
            out.write("RB_NAME" + SEPARATOR + SEPARATOR + NONE + EOL);
            out.write("LOG_NAME" + SEPARATOR + NONE + EOL);
            out.write(EOL);

            String prodDate = dfoDate == null ? "" : dfoDate;
            String prodRoot = ab.getProductDir();
            String prodPath = prodRoot + File.separator + rawType
                    + File.separator + prodDate;
            out.write("PROD_PATH" + SEPARATOR + prodPath + EOL);
            out.write(EOL);

            /* Raw Match Keywords */
            out.write("# raw match key" + EOL);
            String[] matchKeys = ab.getMatchKeywords();
            if (matchKeys.length == 0)
                out.write("RAW_MATCH_KEY" + SEPARATOR + NONE + EOL);
            for (int i = 0; i < matchKeys.length; i++) {
                String matchKey = matchKeys[i];
                String matchVal;
                try {
                    matchVal = abExemplar.getKeywordValue(matchKey);
                } catch (Exception e11) {
                    matchVal = UNDEFINED;
                }
                out.write("RAW_MATCH_KEY" + SEPARATOR + matchKey + "="
                        + matchVal + EOL);
            }
            out.write(EOL);

            /* Input (Raw) Files */
            StringBuffer rawFileBuffer = new StringBuffer();
            StringBuffer rassocBuffer = new StringBuffer();
            StringBuffer rbRawBuffer = new StringBuffer();
            StringBuffer rbCalibBuffer = new StringBuffer();
            StringBuffer sofBuffer = new StringBuffer();

            OCAFile[] rawFiles = ab.getInputFiles();
            for (int i = 0; i < rawFiles.length; i++) {
                String rawFilename;
                String doClass;
                try {
                    rawFilename = rawFiles[i].getFilename();
                } catch (Exception e11) {
                    rawFilename = UNDEFINED;
                }
                try {
                    doClass = rawFiles[i].getDoClass();
                } catch (Exception e12) {
                    doClass = UNDEFINED;
                }
                sofBuffer.append(SOF_CONTENT + SEPARATOR + rawFilename + " "
                        + doClass + " RAW" + EOL);
                rawFileBuffer.append("RAWFILE" + SEPARATOR + SEPARATOR
                        + rawFilename + SEPARATOR + doClass + EOL);
                rbRawBuffer.append(RB_CONTENT + SEPARATOR + rawFilename + " "
                        + doClass + EOL);
            }

            /* Associated (Calib) Files */
            OCAFile[] calibFiles = ab.getAssociatedFiles();
            StringBuffer mcalibBuffer = new StringBuffer();
            StringBuffer massocBuffer = new StringBuffer();
            TreeSet virtualABs = new TreeSet();
            for (int i = 0; i < calibFiles.length; i++) {
                String qcCalibFilename;
                try {
                    qcCalibFilename = calibFiles[i].getFilename();
                } catch (Exception e11) {
                    qcCalibFilename = UNDEFINED;
                }

                String calibDoClass;
                try {
                    calibDoClass = calibFiles[i].getDoClass();
                } catch (Exception e12) {
                    calibDoClass = UNDEFINED;
                }

                String calibProCatg;
                try {
                    calibProCatg = calibFiles[i].getProCatg();
                } catch (Exception e12) {
                    calibProCatg = UNDEFINED;
                }

                String calibRawType;
                try {
                    calibRawType = calibFiles[i].getRawType();
                } catch (Exception e12) {
                    calibRawType = UNDEFINED;
                }

                String calibPipefile;
                try {
                    calibPipefile = calibFiles[i].getPipefile();
                } catch (Exception e12) {
                    calibPipefile = UNDEFINED;
                }

                String calibABName;
                try {
                    calibABName = calibFiles[i].getABName();
                    if (calibFiles[i].isVirtual()) {
                        virtualABs.add(calibABName);
                    }
                } catch (Exception e12) {
                    calibABName = UNDEFINED;
                }

                BigDecimal deltaT = null;
                try {
                    deltaT = ab.getModifiedMjdObs().subtract(
                            calibFiles[i].getMjdObs());
                } catch (Exception e) {
                    deltaT = new BigDecimal(-9999.0);
                }

                String realOrVirtual = "REAL";
                if (calibFiles[i].isVirtual()) {
                    realOrVirtual = "VIRTUAL";
                    String calibDate;
                    try {
                        calibDate = calibFiles[i].getDate();
                    } catch (Exception e12) {
                        calibDate = UNDEFINED;
                    }
                    qcCalibFilename = ab.getProductDir() + File.separator
                            + calibRawType + File.separator + calibDate
                            + File.separator + calibPipefile;
                }

                boolean isMassoc = false;
                boolean isRassoc = false;
                try {
                    isMassoc = calibFiles[i].getAlias().toUpperCase().indexOf(
                            MASSOC) > 0;
                } catch (Exception e13) {
                }
                try {
                    isRassoc = calibFiles[i].getAlias().toUpperCase().indexOf(
                            RASSOC) > 0;
                } catch (Exception e13) {
                }
                if (isRassoc == true) {
                    rassocBuffer.append(RASSOC + SEPARATOR + SEPARATOR
                            + qcCalibFilename + SEPARATOR + calibDoClass + EOL);
                } else if (isMassoc == true) {
                    massocBuffer.append(MASSOC + SEPARATOR + SEPARATOR
                            + realOrVirtual + SEPARATOR + qcCalibFilename
                            + SEPARATOR + calibProCatg + SEPARATOR
                            + calibABName + SEPARATOR
                            + deltaT.setScale(3, BigDecimal.ROUND_HALF_EVEN)
                            + EOL);
                } else {
                    rbCalibBuffer.append(RB_CONTENT + SEPARATOR
                            + qcCalibFilename + " " + calibProCatg + EOL);

                    sofBuffer.append(SOF_CONTENT + SEPARATOR + qcCalibFilename
                            + " " + calibProCatg + " CALIB" + EOL);

                    mcalibBuffer.append("MCALIB" + SEPARATOR + SEPARATOR
                            + realOrVirtual + SEPARATOR + qcCalibFilename
                            + SEPARATOR + calibProCatg + SEPARATOR
                            + calibABName + SEPARATOR
                            + deltaT.setScale(3, BigDecimal.ROUND_HALF_EVEN)
                            + EOL);
                }
            }

            /* write RAW files */
            out.write("# input file(s)" + EOL);
            if (rawFileBuffer.length() == 0)
                out.write(RAWFILE + SEPARATOR + SEPARATOR + NONE + EOL);
            else
                out.write(rawFileBuffer.toString());
            out.write(EOL);

            /* write RASSOC files */
            out
                    .write("# associated raw file(s) (only for SCIENCE; taken from same night only!)"
                            + EOL);
            if (rassocBuffer.length() == 0)
                out.write(RASSOC + SEPARATOR + SEPARATOR + NONE + EOL);
            else
                out.write(rassocBuffer.toString());
            out.write(EOL);

            /* write products */
            out.write("# product file(s)" + EOL);
            out.write("PRODUCTS" + SEPARATOR + NONE + SEPARATOR + NONE + EOL);
            out.write(EOL);

            /* write MCALIB files */
            out.write("# associated mcalib file(s), used for processing" + EOL);
            if (mcalibBuffer.length() == 0)
                out.write(MCALIB + SEPARATOR + SEPARATOR + NONE + EOL);
            else
                out.write(mcalibBuffer.toString());
            out.write(EOL);

            /* write MASSOC files */
            out.write("# associated mcalib file(s), used for packing" + EOL);
            if (massocBuffer.length() == 0)
                out.write(MASSOC + SEPARATOR + SEPARATOR + NONE + EOL);
            else
                out.write(massocBuffer.toString());
            out.write(EOL);

            /* write recipe parameters */
            String params[] = ab.getRecipeParams();
            out.write("# parameters for processing" + EOL);
            if (params.length == 0) {
                out.write("PARAM" + SEPARATOR + SEPARATOR + NONE + EOL);
            } else {
                for (int i = 0; i < params.length; ++i) {
                    out
                            .write("PARAM" + SEPARATOR + SEPARATOR + params[i]
                                    + EOL);
                }
            }
            out.write(EOL);

            /* waitfor */
            out.write("# defined waitfors (for CONDOR)" + EOL);
            if (virtualABs.size() == 0) {
                out.write("WAITFOR" + SEPARATOR + SEPARATOR + NONE + EOL);
            } else {
                for (Iterator it = virtualABs.iterator(); it.hasNext();) {
                    out.write("WAITFOR" + SEPARATOR + SEPARATOR + it.next()
                            + EOL);
                }
            }
            out.write(EOL);

            out
                    .write("# associated QC1 parameters are stored in the following QC1 database table"
                            + EOL);
            out.write("QC1_DB" + SEPARATOR + SEPARATOR + NONE + EOL);
            out.write(EOL);

            out.write("# further associated information" + EOL);
            String packAdd = System.getProperty("PACK_ADD");
            String packAddArr[] = null;
            if (packAdd != null)
                packAddArr = packAdd.split(" ");
            if (packAddArr == null)
                out.write("FURTHER" + SEPARATOR + NONE + EOL);
            else
                for (int i = 0; i < packAddArr.length; i++) {
                    out.write("FURTHER_" + packAddArr[i] + SEPARATOR + NONE
                            + EOL);
                }
            out.write(EOL);

            String username = System.getProperty("user.name");
            String hostname = InetAddress.getLocalHost().getHostName();
            Date now = new Date(Calendar.getInstance().getTimeInMillis());
            String nowAsString = now.toLocaleString();
            String createdStatus = "created by '" + toolVersion + "' on "
                    + nowAsString + " by " + username + " on " + hostname;

            out.write("# status of AB" + EOL);
            out.write("TEXEC" + SEPARATOR + SEPARATOR + NONE + EOL);
            out.write("AB_STATUS" + SEPARATOR + createdStatus + EOL);
            out.write(EOL);

            /* Reduction Block */
            out.write("# ========== RB section starts here ==========" + EOL);
            String prodName = prodPath + File.separator + prodRootName;
            out.write(RB_CONTENT + SEPARATOR + "recipe: " + ab.getRecipe()
                    + EOL);
            out.write(RB_CONTENT + EOL);
            out.write(RB_CONTENT + SEPARATOR + "instrument: " + instrume + EOL);
            out.write(RB_CONTENT + EOL);
            out.write(RB_CONTENT + SEPARATOR + prodName + EOL);
            out.write(RB_CONTENT + EOL);
            out.write(RB_CONTENT + SEPARATOR + "{" + EOL);
            // write RB raw files
            out.write(rbRawBuffer.toString());
            out.write(RB_CONTENT + SEPARATOR + "}" + EOL);
            out.write(RB_CONTENT + EOL);
            out.write(RB_CONTENT + SEPARATOR + "{" + EOL);
            // write RB calib files
            out.write(rbCalibBuffer.toString());
            out.write(RB_CONTENT + SEPARATOR + "}" + EOL);
            out.write(RB_CONTENT + EOL);

            // write recipe parameters
            for (int i = 0; i < params.length; ++i) {
                out.write(RB_CONTENT + SEPARATOR + params[i] + EOL);
            }
            out.write(RB_CONTENT + EOL);
            out.write(EOL);

            /* write SOF files */
            out.write("# ========== SOF section starts here ==========" + EOL);
            out.write(sofBuffer.toString());
            out.write(EOL);

            /* close output stream */
            out.close();
        } catch (IOException e) {
            throw new ABVisitorException(e);
        }
    }
}