package org.eso.oca.backend;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.oca.fits.KeywordNotFoundException;
import org.eso.oca.fits.OCAFile;
import org.eso.oca.fits.UndefinedValueException;

/**
 * This class represents an Association Block (AB).
 * 
 * @author szampier
 */

public class AssociationBlock {
    /** The instrument. */
    private String instrument = null;

    /** The input files. */
    private OCAFile[] inputFiles = new OCAFile[0];

    /** The associated files. */
    private OCAFile[] associatedFiles = new OCAFile[0];
    
    /** The results of the individual association rules are stored
     * in SelectionBlock objects
     */
    Vector<SelectionBlock> selections = new Vector<SelectionBlock>();

    /** The virtual product files. */
    private OCAFile[] productFiles = new OCAFile[0];

    /** The recipe name. */
    private String recipe;

    /** The recipe parameters. */
    private Vector recipeParams = new Vector();

    /** The action name (as in the OCA rules). */
    private String action;

    /** The group event. */
    private String groupEvent;

    /** The group event extension. */
    private String groupEventExt;

    /** The raw match keywords. */
    private String[] matchKeywords = new String[0];

    /** The modified mjd-obs: MJD-OBS of first raw file + SUM(EXPTIME)/2. */
    private BigDecimal modifiedMjdObs = null;

    /** The exemplar frame is used as a reference for searching 
     * associated frames.
     * For each group of input files created by the organisation rules,
     * an exemplar frame is selected and passed as input file to
     * the association rules.
     */
    private OCAFile exemplar = null;

    /** This flag tells whether the AB is valid or not. For example an AB is
     * not valid if not enough calibrations are found.
     */
    private boolean isValid = true;

    /** The name of this AB. */
    private String abName;

    /** The Apache Lo4j logger. */
    static Logger logger = Logger.getLogger(AssociationBlock.class);

    /** The directory where pipeline products should be stored. */
    private String productDir = "/default/product/dir";

    public void setValid(boolean flag) {
        isValid = flag;
    }

    public boolean isValid() {
        return isValid;
    }

    public String getAction() {
        return action;
    }

    public void setAction(String a) {
        action = a;
    }

    public OCAFile[] getAssociatedFiles() {
        return associatedFiles;
    }

    public void setAssociatedFiles(OCAFile[] aFiles) {
        associatedFiles = aFiles;
    }

    public void addAssociatedFiles(OCAFile[] aFiles) {
        int prevLen = associatedFiles.length;
        if (prevLen == 0) {
            setAssociatedFiles(aFiles);
        } else {
            OCAFile[] aFilesNew = new OCAFile[prevLen + aFiles.length];
            for (int i = 0; i < associatedFiles.length; i++) {
                aFilesNew[i] = associatedFiles[i];
            }
            for (int i = 0; i < aFiles.length; i++) {
                aFilesNew[i + prevLen] = aFiles[i];
            }
            associatedFiles = aFilesNew;
        }
    }
    
    /**
     * Add a selection block to this AB.
     * @param selectionBlock the object to add.
     */
    public void addSelectionBlock(SelectionBlock selectionBlock) {
    	if (selectionBlock == null) {
    		throw new IllegalArgumentException("selectionBlock must not be null");
    	}
    	selections.add(selectionBlock);
    }

    public OCAFile[] getInputFiles() {
        return inputFiles;
    }

    public void setInputFiles(OCAFile[] iFiles) {
        inputFiles = iFiles;
    }

    public OCAFile[] getProductFiles() {
        return productFiles;
    }

    public void setABName(String aName) {
        abName = aName;
    }

    public String getABName() {
//        if (abName != null)
//            return abName;

        String tmpName;
        try {
            tmpName = getExemplar().getDpId();
        } catch (Exception e1) {
            tmpName = new Long(Calendar.getInstance().getTimeInMillis())
                    .toString();
            logger.error("Can't use exemplar dp_id to build AB name."
                    + " Using timestamp : " + tmpName);
        }

        String abExt = getGroupEventExt();
        if (abExt == null)
            abExt = "";
        else
            abExt = "_" + abExt;

        abName = tmpName + abExt;

        return abName;
    }

    public void setProductFiles(OCAFile[] pFiles) {
        productFiles = pFiles;
        String abName = getABName() + ".ab";
        String dfoDate = System.getProperty("DFO_DATE");
        for (int i = 0; i < productFiles.length; ++i) {
            productFiles[i].addMetaCard(OCAFile.AB_NAME, abName);
            if (dfoDate != null) {
                productFiles[i].addMetaCard(OCAFile.DATE, dfoDate);
            }
        }
    }

    public String getRecipe() {
        return recipe;
    }

    public void setRecipe(String r) {
        recipe = r;
    }

    public String[] getRecipeParams() {
        return (String[]) recipeParams.toArray(new String[0]);
    }

    public void setRecipeParams(Vector p) {
        recipeParams = p;
    }

    /**
     * the representative input frame (exemplar) is the "oldest" one in the
     * group, according to the observation date (MJD-OBS)
     * 
     * @return the exemplar frame
     * @throws KeywordNotFoundException
     */
    public OCAFile getExemplar() {
        if (exemplar != null) {
            return exemplar;
        }

        /* default exemplar is the first file with valid MJD-OBS */
        for (int i = 0; i < inputFiles.length; i++) {
            try {
                inputFiles[i].getMjdObs();
            } catch (Exception e) {
                continue;
            }
            // ok we found the default exemplar
            exemplar = inputFiles[i];
            break;
        }

        /*
         * if this happens it means that MJD-OBS is undefined for all input
         * files
         */
        if (exemplar == null) {
            logger
                    .error("None of the input files has a valid MJD-OBS. Invalid AB.");
            exemplar = inputFiles[0];
            setValid(false);
            return exemplar;
        }

        /*
         * We have a valid default exemplar. Now we can search for an input file
         * with smaller MJD-OBS.
         */
        for (int i = 0; i < inputFiles.length; i++) {
            try {
                if (inputFiles[i].getMjdObs().compareTo(exemplar.getMjdObs()) < 0) {
                    exemplar = inputFiles[i];
                }
            } catch (Exception e) {
                try {
                    logger.info(inputFiles[i].getFilename()
                            + ". Can't compare MJD-OBS with exemplar.");
                } catch (Exception e1) {
                }
            }
        }
        return exemplar;
    }

    protected void setExemplar(OCAFile exemplarVal) {
        exemplar = exemplarVal;
    }

    public String getInstrument() throws KeywordNotFoundException,
            UndefinedValueException {
        if (instrument != null)
            return instrument;
        else
            return getExemplar().getInstrument();
    }

    public void setInstrument(String i) {
        if (i == null)
            throw new IllegalArgumentException("null instrument");
        instrument = i;
    }

    /**
     * Accept the AB visitor
     * 
     * @param visitor
     * @throws ABVisitorException
     */
    public void accept(ABVisitor visitor) throws ABVisitorException {
        visitor.visit(this);
    }

    /**
     * @return String representation of the AB
     */
    public String toString() {
        StringBuffer out = new StringBuffer();
        out.append((isValid ? "GOOD" : "BAD") + " AB\n");
        out.append("recipe : " + recipe);
        out.append("\n");
        out.append("input files {\n");
        for (int i = 0; i < inputFiles.length; ++i) {
            try {
                out.append(inputFiles[i].getFilename() + "\t");
            } catch (Exception e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            try {
                out.append(inputFiles[i].getDoClass());
            } catch (Exception e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
            out.append("\n");
        }
        out.append("}\n");
        out.append("associated files {\n");
        for (int i = 0; i < associatedFiles.length; ++i) {
            try {
                out.append(associatedFiles[i].getFilename() + "\t");
            } catch (Exception e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            try {
                out.append(associatedFiles[i].getAlias() + "\t");
            } catch (Exception e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
            out.append(associatedFiles[i].isVirtual() ? "VIRTUAL" : "REAL");
            out.append("\n");
        }
        out.append("}\n");
        out.append("product files {\n");
        for (int i = 0; i < productFiles.length; ++i) {
            try {
                out.append(productFiles[i].getFilename() + "\t");
            } catch (Exception e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            try {
                out.append(productFiles[i].getProCatg());
            } catch (Exception e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
            out.append("\n");
        }
        out.append("}\n");
        return out.toString();
    }

    /**
     * @param matchKeywords
     */
    public void setMatchKeywords(String[] mKeywords) {
        if (mKeywords == null)
            throw new IllegalArgumentException("null match keywords");
        matchKeywords = mKeywords;
    }

    public String[] getMatchKeywords() {
        return matchKeywords;
    }

    protected void setModifiedMjdObs(BigDecimal modifiedMjdObsVal) {
        modifiedMjdObs = modifiedMjdObsVal;
    }

    public BigDecimal getModifiedMjdObs() throws NumberFormatException,
            KeywordNotFoundException, UndefinedValueException {
        if (modifiedMjdObs == null) {
            double mjdStart = getExemplar().getMjdObs().doubleValue();
            double totExptime = 0.0;
            for (int i = 0; i < inputFiles.length; i++) {
                double exptime = 0.0;
                try {
                    exptime = new Float(inputFiles[i]
                            .getKeywordValue(OCAFile.EXPTIME)).doubleValue();
                } catch (Exception e) {
                    logger.warn("Can't use EXPTIME of file "
                            + inputFiles[i].getKeywordValue(OCAFile.FILENAME)
                            + " to compute MODIFIED MJD-OBS");
                }
                totExptime += exptime;
            }
            modifiedMjdObs = new BigDecimal(mjdStart + totExptime / 86400 / 2);
        }
        return modifiedMjdObs;
    }

    public void setProductDir(String dir) {
        if (dir == null)
            throw new IllegalArgumentException("null product dir");
        productDir = dir;
    }

    public String getProductDir() {
        return productDir;
    }

    public void setGroupEvent(String event) {
        groupEvent = event;
    }

    public String getGroupEvent() {
        return groupEvent;
    }

    public void setGroupEventExt(String ext) {
        groupEventExt = ext;
    }

    public String getGroupEventExt() {
        return groupEventExt;
    }

    /**
     * Returns the association types contained in this AB. 
     * @return array of association types.
     */
    public String[] getAssociationTypes() {
    	String[] types = new String[selections.size()];
    	for (int i = 0; i < selections.size(); ++i) {
    		types[i] = selections.elementAt(i).getAlias();
    	}
    	return types;
    }

    /**
     * Returns the selection block corresponding to the specified association type.  
     * @param type the association type
     * @return selection block corresponding to the specified association type.
     */
    public SelectionBlock getSelectionBlockForType(String type) {
        if (type == null) {
            throw new IllegalArgumentException("type must not be null");
        }
    	for (int i = 0; i < selections.size(); ++i) {
    		if (selections.elementAt(i).getAlias().equals(type)) {
    			return selections.elementAt(i);
    		}
    	}
    	logger.error("No selection block found for type " + type);
    	// To avoid null pointer exceptions we return a dummy selection block. 
    	return new SelectionBlock(type, 0, 0, new OCAFile[0]);
    }
    
    /**
     * Returns the files corresponding to the specified association type.  
     * @param type the association type.
     * @return array of OCAFile objects corresponding to the specified association type.
     */
    public OCAFile[] getAssociatedFilesForType(String type) {
        if (type == null) {
            throw new IllegalArgumentException("type must not be null");
        }
    	for (int i = 0; i < selections.size(); ++i) {
    		if (selections.elementAt(i).getAlias().equals(type)) {
    			return selections.elementAt(i).getSelectedFiles();
    		}
    	}
    	logger.error("No associations found for type " + type);
    	// To avoid null pointer exceptions we return a dummy 
		return new OCAFile[0];
    }
}