package org.eso.oca.backend;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.eso.oca.fits.OCAFile;
import org.eso.util.misc.Conversions;
import org.eso.util.filesystem.FileSystemOperations;

/**
 * This class implements the ABVisitor interface to create a Reduction Block
 * 
 * @author szampier
 */
public class RBFileWriter implements ABVisitor {
    /** the directory where reduction blocks are created */
    private String rbDir;

    /** some useful constants */
    static final String EOL = System.getProperty("line.separator");

    static final String UNDEFINED = "UNDEFINED";

    /** acceptable characters for indicating the cascade level of an RB */ 
    private static final String RB_LEVEL_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    /** the maximum level of an RB within a pipeline cascade */ 
    private static final int MAX_RB_LEVEL = RB_LEVEL_CHARS.length() - 1; 

    static Logger logger = Logger.getLogger(RBFileWriter.class);

    /**
     * Constructor
     * 
     * @param directory
     *            where ABs should be created
     */
    public RBFileWriter(String dir) {
        if (dir == null)
            throw new IllegalArgumentException("null AB dir");
        rbDir = dir;
    }

    /** <p>Creates a Reduction Block (RB) as a text file in the RB directory 
     * from the specified Association Block (AB). The RB file will be given 
     * a name in this format:</p> 
     * 
     * <p>{OLAS-ID}.{MJD-OBS}{TPL}-{RB-LEVEL}{RB-NUMBER}.rb{TEMP}</p>
     * 
     * <p>
     *   <table border=1> 
     *     <tbody>
     *       <tr>
     *         <td>{OLAS-ID}</td>
     *         <td>The OLAS ID of the instrument this RB was generated for.
     *             </td>
     *       </tr> 
     *       <tr>
     *         <td>{MJD-OBS}</td>
     *         <td>The MJD-OBS of the first input file used in generating 
     *             this RB.</td>
     *       </tr> 
     *       <tr>
     *         <td>{TPL}</td>
     *         <td>The string <b>_tpl</b> if this RB was generated for a 
     *             template, or the empty string otherwise.</td>
     *       </tr> 
     *       <tr>
     *         <td>{RB-LEVEL}</td>
     *         <td>A letter ranging from <b>A</b> to <b>Z</b> indicating 
     *             the level of this RB in a pipeline cascade. A cascade 
     *             can have up to 26 levels, with <b>A</b> being the first 
     *             and <b>Z</b> being the last.</td>
     *       </tr> 
     *       <tr>
     *         <td>{RB-NUMBER}</td>
     *         <td>A two-digit number ranging from <b>01</b> to <b>99</b>, 
     *             used to help distinguish this RB from any others within 
     *             this level of the pipeline cascade. The RB number is 
     *             incremented every time a new RB is generated within a 
     *             cascade level for a given set of input files.</td> 
     *       </tr> 
     *       <tr>
     *         <td>{TEMP}</td>
     *         <td>The string <b>.tmp</b> if this RB is still being written 
     *             to disk; the empty string once the RB file is complete.
     *             </td> 
     *       </tr> 
     *     </tbody>
     *   </table> 
     * </p>
     * 
     * <p>In practice, the RB names look like this:
     *   <ul>
     *     <li>FORS1.2003-05-23T03:04:53.712-A01.rb</li> 
     *     <li>VIMOS.2005-02-12T09:31:25.199_tpl-A01.rb</li> 
     *     <li>UVES.2000-02-10T20:54:25.506-A01.rb.tmp</li> 
     *   </ul>
     * </p> 
     * 
     * @param ab the AB to create the RB from. 
     * 
     * @throws ABVisitorException in case of error. 
     */
    public void visit(AssociationBlock ab) throws ABVisitorException {
        try {
            String rbBaseName = ab.getABName();

            /* The RB level can range from 0 to MAX_RB_LEVEL. 
             * 
             * Determine the level of this RB. 
             */ 
            int currentIndex = rbBaseName.lastIndexOf('-'); 
            int rbLevel = 0; 

            if(currentIndex >= 0) 
            {
                /* This RB's basename may contain a level letter. */ 
                char rbLevelChar = rbBaseName.charAt(currentIndex + 1); 
                rbLevel = RB_LEVEL_CHARS.indexOf(rbLevelChar); 

                if(rbLevel < 0) 
                {
                    /* This RB's basename does not contain a level letter - 
                     * set the level to zero. 
                     */ 
                    rbLevel = 0; 
                }
                else 
                {
                    /* This RB's basename already contains a level letter - 
                     * increment the RB level. 
                     */
                    ++rbLevel; 

                    if(rbLevel > MAX_RB_LEVEL) 
                    {
                        final String logMessage = 
                            "Processing RB basename [" + rbBaseName 
                            + "] will exceed the maximum RB level [" 
                            + RB_LEVEL_CHARS.charAt(MAX_RB_LEVEL) + "]."; 

                        logger.error(logMessage); 
                        throw new ABVisitorException(logMessage); 
                    }

                    /* Strip everything from the '-' character and the old 
                     * level letter onwards from the RB basename. 
                     */
                    rbBaseName = rbBaseName.substring(0, currentIndex); 
                }
            }

            /* Add the RB level letter to the basename. */
            rbBaseName = rbBaseName + '-' + RB_LEVEL_CHARS.charAt(rbLevel); 

            /* The RB number within a level can range from '01' to '99'. */ 
            int rbNumber = 0; 
            DecimalFormat rbNumberFormat = new DecimalFormat("00"); 

            /* If the RB basename has any "r." prefixes, strip them. */ 
            while(rbBaseName.startsWith("r.")) 
            {
                rbBaseName = rbBaseName.substring(2); 
            }

            /* Check the RB directory for existing RB files. */ 
            final HashSet<String> existingFilenames = new HashSet<String>(); 

            if(FileSystemOperations
               .listFilenamesInDirectory(new File(rbDir), 
                                         null, 
                                         existingFilenames) == false) 
            {
                final String logMessage = 
                    "Unable to list contents of RB directory [" 
                    + rbDir + "]."; 

                logger.error(logMessage); 
                throw new ABVisitorException(logMessage); 
            }

            /* Reduce the existing RB filenames to RB basenames. */ 
            final HashSet<String> existingBasenames = new HashSet<String>(); 
            final Iterator<String> 
                existingFilenameIterator = existingFilenames.iterator(); 
            String currentString = null; 

            while(existingFilenameIterator.hasNext()) 
            {
                currentString = existingFilenameIterator.next(); 
                currentIndex = currentString.lastIndexOf('.'); 

                if(currentIndex < 0) 
                {
                    existingBasenames.add(currentString); 
                }
                else
                {
                    existingBasenames.add(
                        currentString.substring(0, currentIndex)); 
                }
            }

            /* Generate an RB file name which is not already in use. */ 
            do 
            {
                ++rbNumber; 

                currentString = rbBaseName + rbNumberFormat.format(rbNumber); 

                if(rbNumber > 99) 
                {
                    final String logMessage = 
                        "RB basename [" + currentString 
                        + "] will exceed the maximum RB number [99]."; 

                    logger.error(logMessage); 
                    throw new ABVisitorException(logMessage); 
                }
            }
            while(existingBasenames.contains(currentString)); 

            /* Adopt the new basename. */ 
            rbBaseName = currentString; 

            /* To prevent the RBS from reading the RB before the DO has 
             * finished writing it, the RB is written to a temporary file 
             * and then renamed.
             */
            String rbFilename = rbDir + File.separator + rbBaseName + ".rb"; 
            File rbFile = new File(rbFilename); 
            String tmpRbFilename = rbFilename + ".tmp"; 
            File tmpRbFile = new File(tmpRbFilename); 
            FileWriter out = new FileWriter(tmpRbFile); 
			
            /* build RB's product name */
            String prodRootName = "r." + rbBaseName; 
            Date rbDate = new Date();
            SimpleDateFormat isoFormat = new SimpleDateFormat("yyyy-MM-dd");
            try {
                rbDate = 
                    Conversions.convertMjdObsToJavaDate(
                        ab.getExemplar().getMjdObs().doubleValue()); 
            } catch (Exception e) {
                logger.error("Using default date");
            }
            String rbDateStr = isoFormat.format(rbDate);
            String rbProductName = ab.getProductDir() 
                + File.separator + rbDateStr
                + File.separator + prodRootName; 

            /* get instrument ID */
            String instrume;
            try {
                instrume = ab.getInstrument().toLowerCase();
            } catch (Exception e3) {
                instrume = UNDEFINED;
            }

            /* get raw files */
            StringBuffer rbRawBuffer = new StringBuffer();
            OCAFile[] rawFiles = ab.getInputFiles();
            for (int i = 0; i < rawFiles.length; i++) {
                String rawFilename;
                String doCatg;
                try {
                    rawFilename = rawFiles[i].getFilename();
                } catch (Exception e11) {
                    rawFilename = UNDEFINED;
                }
                try {
                    doCatg = rawFiles[i].getDoCatg();
                } catch (Exception e12) {
                    doCatg = UNDEFINED;
                }
                rbRawBuffer.append(rawFilename + " " + doCatg + EOL);
            }

            /* get calib files */
            StringBuffer rbCalibBuffer = new StringBuffer();
            OCAFile[] calibFiles = ab.getAssociatedFiles();
            for (int i = 0; i < calibFiles.length; i++) {
                String calibFilename;
                String doCatg;
                try {
                    calibFilename = calibFiles[i].getLinkname();
                } catch (Exception e11) {
                    calibFilename = UNDEFINED;
                }
                try {
                    doCatg = calibFiles[i].getDoCatg();
                } catch (Exception e12) {
                    doCatg = UNDEFINED;
                }
                rbCalibBuffer.append(calibFilename + " " + doCatg + EOL);
            }

            /* get recipe parameters */
            String params[] = ab.getRecipeParams();

            /* write reduction block */
            out.write("recipe: " + ab.getRecipe() + EOL + EOL);
            out.write("instrument: " + instrume + EOL + EOL);
            out.write(rbProductName + EOL + EOL);
            out.write("{" + EOL);
            out.write(rbRawBuffer.toString());
            out.write("}" + EOL + EOL);
            out.write("{" + EOL);
            out.write(rbCalibBuffer.toString());
            out.write("}" + EOL + EOL);
            for (int i = 0; i < params.length; ++i) {
                out.write(params[i] + EOL);
            }

            /* close output stream */
            out.close();

            /* rename temp file to RB file */
            tmpRbFile.renameTo(rbFile);
            
            logger.debug("RBFileWriter::visit() - moved ["
            		+ tmpRbFile.getAbsolutePath() + "] to [" 
            		+ rbFile.getAbsolutePath() + "].");
            
            // Log the path of the newly-written RB. 
            logger.info("RBFileWriter::visit()"
                        + " - wrote RB [" + rbFilename + "]."); 
        } catch (IOException e) {
            throw new ABVisitorException(e);
        }
    }
}