package org.eso.oca.dal;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * This class contains static methods to access the database
 * meta-data.
 * 
 * @author szampier
 *
 */
public class DBMetaData {
	/** The Apache Log4j logger. */
	private static Logger logger = Logger.getLogger(DBMetaData.class);
	
	// Some useful constants
	final static String TABLE_NAME = "TABLE_NAME";
	
	final static String COLUMN_NAME = "COLUMN_NAME";
	
	final static String DATA_TYPE = "DATA_TYPE";

	final static String TYPE_NAME = "TYPE_NAME";
	
	/**
	 * Returns the meta data object for the specified table.
	 * 
	 * @param conn the JDBC connection
	 * @param tableName the table name
	 * @return the meta data object for the specified table or
	 *          null if the meta-data could not be retrieved.
	 * @throws SQLException
	 */
	public static TableMetaData getTableMetaData(Connection conn, String tableName)
			throws SQLException {
		final String thisMethod = "DBMetaData::getTableMetaData(" + tableName + ")";
		logger.trace(thisMethod);
		
		if (tableName == null) {
			throw new IllegalArgumentException(thisMethod
					+ " - tableName must not be null");
		}

		if (conn == null) {
			throw new IllegalArgumentException(thisMethod
					+ " - connection must not be null");
		}

		// Get database meta-data
		DatabaseMetaData dbmd = conn.getMetaData();
		// Get meta-data of specified table
		ResultSet rs = dbmd.getColumns(null, null, tableName, null);

		// Initialise meta-data objects
		TableMetaData result = new TableMetaData();
		Vector<ColumnMetaData> columns = new Vector<ColumnMetaData>();

		// Parse results
		while (rs.next()) {
			ColumnMetaData column = new ColumnMetaData();
			for (int i = 1; i <= rs.getMetaData().getColumnCount(); ++i) {
				if (rs.getMetaData().getColumnName(i).equals(TABLE_NAME)) {
					result.name = rs.getString(i);
				} else if (rs.getMetaData().getColumnName(i).equals(COLUMN_NAME)) {
					column.columnName = rs.getString(i);
				} else if (rs.getMetaData().getColumnName(i).equals(DATA_TYPE)) {
					column.type = rs.getInt(i);
				} else if (rs.getMetaData().getColumnName(i).equals(TYPE_NAME)) {
					column.typeName = rs.getString(i);
				}
			}
			logger.debug("[ColumnName=" + column.columnName
					+ "] [ColumnType=" + column.typeName + "]");
			
			columns.add(column);
		}

		if (columns.size() == 0) {
			result = null;
		} else {
			result.columns = columns.toArray(new ColumnMetaData[columns.size()]);
		}
		logger.debug(thisMethod + " - " + result);
		return result;
	}

	/**
	 * Returns the meta-data object for the specified column. 
	 * 
	 * @param tableMetaData the table meta-data object
	 * @param columnName the name of the column whose meta-data must be returned.
	 * @return the meta-data object for the specified column, or null
	 *         if the meta-data could not be retrieved.
	 */
	public static ColumnMetaData getColumnMetaData(TableMetaData tableMetaData,
			String columnName) {
		final String thisMethod = "DBMetaData::getColumnMetaData(" + columnName + ")";
		logger.trace(thisMethod);
		
		ColumnMetaData result = null;
		if (tableMetaData != null) {
			ColumnMetaData[] columns = tableMetaData.columns;
			for (int i = 0; i < columns.length; ++i) {
				if (columns[i].columnName.equals(columnName)) {
					result = columns[i];
					break;
				}
			}
		}
		logger.debug(thisMethod + " - " + result);
		return result;
	}

	/**
	 * Returns the null value casted to the type of the specified column.
	 * Example: convert(int, null)
	 * This is needed in some special cases, e.g. in connection with views.
	 * 
	 * @param tableMetaData the table meta-data object.
	 * @param columnName the column name.
	 * @return the null value casted to the type of the specified column,
	 *         or just "null" if the column type could not be retrieved.
	 */
	public static String getNullValueForColumn(TableMetaData tableMetaData,
			String columnName) {
		final String thisMethod = "DBMetaData::getNullValueForColumn(" + columnName + ")";
		logger.trace(thisMethod);
		
		String result = "null";
		ColumnMetaData column = getColumnMetaData(tableMetaData, columnName);
		if (column != null) {
			String typeName = column.typeName;
			result = "convert(" + typeName + ", null)";
		}
		logger.debug(thisMethod + " - " + result);
		return result;
	}
}
