package org.eso.oca.fits;

import java.io.File;
import java.util.Arrays;
import java.util.ArrayList; 
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.util.filesystem.DirectoryFileSelectionModel;
import org.eso.util.filesystem.FileSelectionModel;
import org.eso.util.filesystem.RecursiveFileSelectionModel;

/** This class provides a convenient way to extract the keywords from all of 
 * the FITS/TFITS/HDR files stored in a given directory, returning the results
 * as a list of OCAFile objects. 
 * 
 * @author szampier
 * @author nferguso 
 */
public class OCAFileFactory 
{
    /** The list of directories to search for FITS files. */
    private String rootDir = ".";
	
    /** Provides the list of files stored under a given directory. */
    private FileSelectionModel fileSelectionModel = null;
    
    static Logger logger = Logger.getLogger(OCAFileFactory.class);
	
    /** Creates a new OCAFileFactory to work with files under the specified 
     * directory. 
     * 
     * @param directory the root directory this OCAFileFactory should 
     *                  work with. 
     */
    public OCAFileFactory(String directory) 
    {
        logger.trace("OCAFileFactory::OCAFileFactory()"); 

        rootDir = directory;
        fileSelectionModel = new DirectoryFileSelectionModel(rootDir);
    }

    /**
     * Creates a new instance that will create OCAFiles from the given list
     * of files and/or directories.
     * 
     * @param filesAndDirectories The list of files and/or directories.
     */
    public OCAFileFactory(String[] filesAndDirectories) 
    {
        logger.trace("OCAFileFactory::OCAFileFactory(String[])"); 

        fileSelectionModel = new DirectoryFileSelectionModel(filesAndDirectories);
    }

    
    /** Enable recursive file selection, i.e. select files from the root 
     * directory and its subdirectories. 
     */
    public void setRecursiveFileSelection() 
    {
        logger.trace("OCAFileFactory::setRecursiveFileSelection()"); 

        fileSelectionModel = new RecursiveFileSelectionModel(rootDir);
    }
    
    /** Returns the root directory this OCAFileFactory was constructed 
     * to work with. 
     *
     * @return the root directory as a String. 
     */ 
    public String getRootDir() 
    {
        logger.trace("OCAFileFactory::getRootDir()"); 

        return rootDir;
    }
    
    /** Extracts the specified keywords from the given FITS extension of 
     * each HDR/FITS/TFITS file found. The resulting OCAFiles are filtered 
     * with filter (if specified), then returned as an array. 
     * 
     * @throws NullPointerException if keywords is null. 
     * 
     * @param extension the FITS extension from which keywords should be read. 
     * @param keywords the list of keywords to be extracted. 
     * @param filter an optional filter (if null, a DefaultFileFilter is used).
     */
    public OCAFile[] createOCAFiles(int extension, 
                                    String[] keywords, 
                                    OCAFileFilter filter) 
    {
        logger.trace("OCAFileFactory::createOCAFiles" 
                     + "(int, String[], OCAFileFilter)"); 

        // The list of OCAFiles that were read. 
        final ArrayList<OCAFile> readFiles = new ArrayList<OCAFile>(); 

        // The list of Files that were skipped (this data will be discarded). 
        final ArrayList<File> skippedFiles = new ArrayList<File>(); 

        // Delegate to overloaded peer method. 
        createOCAFiles(extension, keywords, filter, readFiles, skippedFiles); 

        // Convert the list of OCAFiles to an array and return it. 
        final OCAFile[] fileArray = new OCAFile[readFiles.size()]; 

        return(readFiles.toArray(fileArray)); 
    }

    /** Extracts the specified keywords from the given FITS extension of 
     * each HDR/FITS/TFITS file found. The resulting OCAFiles are filtered 
     * with filter (if specified), then appended to readFiles. Files which 
     * were skipped or filtered out are appended to skippedFiles. 
     * 
     * @throws NullPointerException if keywords, readFiles or skippedFiles 
     *                              is null. 
     * 
     * @param extension the FITS extension from which keywords should be read. 
     * @param keywords the list of keywords to be extracted. 
     * @param filter an optional filter (if null, a DefaultFileFilter is used).
     * @param readFiles read files will be appended to this list. 
     * @param skippedFiles skipped files will be appended to this list. 
     */
    public void createOCAFiles(final int extension, 
                               final String[] keywords, 
                               OCAFileFilter filter, 
                               final ArrayList<OCAFile> readFiles, 
                               final ArrayList<File> skippedFiles) 
    {
        final String methodName = 
            "OCAFileFactory::createOCAFiles(int, String[], " 
            + "OCAFileFilter, ArrayList<OCAFile>, ArrayList<File>)"; 

        logger.trace(methodName); 

        // The keyword array and file lists must not be null. 
        if((keywords == null) || (readFiles == null) || (skippedFiles == null))
        {
            final String message = 
                methodName + " - null parameter(s):" 
                + (keywords == null ? " keywords" : "") 
                + (readFiles == null ? " readFiles" : "") 
                + (skippedFiles == null ? " skippedFiles" : "") 
                + "."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(filter == null) 
        {
            // No filter was specified; accept all files. 
            filter = new DefaultFileFilter();
        }
        
        // Supplement the requested keywords with the standard set that 
        // virtually all applications require. 
        TreeSet tmpData = new TreeSet(Arrays.asList(keywords)); 
        tmpData.add(OCAFile.INSTRUME);
        tmpData.add(OCAFile.MJD_OBS);
        tmpData.add(OCAFile.OBS_PROG_ID);
        tmpData.add(OCAFile.OBS_ID);
        tmpData.add(OCAFile.EXPTIME);
        tmpData.add(OCAFile.TPL_START);
        tmpData.add(OCAFile.TPL_EXPNO);
        tmpData.add(OCAFile.TPL_NEXP);

        String[] requestedData = new String[tmpData.size()];
        requestedData = (String[]) tmpData.toArray(new String[0]);

        // Get the list of files in the current directory. 
        Vector files = fileSelectionModel.getFileList();

        // Loop over the list of files. 
        for (Iterator it = files.iterator(); it.hasNext();) 
        {
            // Get the current filename in canonical form. 
            String linkname = (String) it.next();
            String filename = OCAFile.getCanonicalPath(linkname);

            if (filename == null) 
            {
                // This filename does not have a canonical form - skip it. 
                logger.warn(methodName + " - skipping invalid pathname [" 
                            + linkname + "]."); 

                skippedFiles.add(new File(linkname)); 
                continue;
            }

            // Handler for a given file type. 
            DataTransportFormatHandler handler = null;

            // Check whether the current filename has a file extension. 
            int indexOfLastDot = filename.lastIndexOf('.');

            if (indexOfLastDot < 0) 
            {
                // This filename has no extension; it cannot be identified. 
                logger.debug(methodName + " - skipping unidentifiable file [" 
                             + linkname + "]."); 

                skippedFiles.add(new File(linkname)); 
                continue;
            }

            // Extract the file extension of the current filename. 
            String ext = filename.substring(indexOfLastDot).toLowerCase(); 

            if (ext.equals(".hdr")) 
            {
                // This looks like an HDR file. Try to open it with the 
                // HDR file handler.
                logger.debug(methodName + " - invoking HDR file handler " 
                             + "for file [" + linkname + "]."); 

                try 
                {
                    handler = new HDRHandler(filename);
                } 
                catch (FileHandlerException e) 
                {
                    logger.error(methodName + " - could not handle file [" 
                                 + linkname + "] as an HDR file [" 
                                 + e.getMessage() + "]."); 

                    skippedFiles.add(new File(linkname)); 
                    continue;
                }
            } 
            else if(ext.equals(".fits") || ext.equals(".tfits")) 
            {
                // This looks like a FITS file. Try to open it with the 
                // FITS file handler. 
                logger.debug(methodName + " - invoking FITS file handler " 
                             + "for file [" + linkname + "].");

                try 
                {
                    handler = new FITSHandler(filename);
                } 
                catch (FileHandlerException e) 
                {
                    logger.error(methodName + " - could not handle file [" 
                                 + linkname + "] as a FITS file [" 
                                 + e.getMessage() + "].");

                    skippedFiles.add(new File(linkname)); 
                    continue;
                }
            } 
            else 
            {
                // Files with this extension are not supported. 
                logger.debug(methodName + " - skipping file [" + linkname 
                             + "] of unsupported type."); 

                skippedFiles.add(new File(linkname)); 
                continue;
            }

            // Get the requested keywords from the FITS header. 
            TypedHeaderCard[] cards = null;

            try 
            {
                cards = handler.getFITSCards(extension, requestedData);
            } 
            catch (Exception e) 
            {
                logger.error(methodName + " - unable to retrieve FITS cards " 
                             + "for file [" + linkname + "] [" 
                             + e.getMessage() + "].");

                skippedFiles.add(new File(linkname)); 
                continue;
            }

            // Convert the list of FITS cards into a Hashtable 
            // (to be stored in an OCAFile). 
            Hashtable keywordValues;

            try 
            {
                keywordValues = convertArrayOfCardsToHashtable(cards);
            } 
            catch (TypedHeaderCardException e) 
            {
                logger.error(methodName + " - unable to convert FITS cards " 
                             + "to hashtable for file [" + linkname + "] [" 
                             + e.getMessage() + "]."); 

                skippedFiles.add(new File(linkname)); 
                continue;
            }

            // Add the meta-keywords FILENAME and LINKNAME. 
            keywordValues.put(OCAFile.FILENAME, filename);
            keywordValues.put(OCAFile.LINKNAME, linkname);

            // Generate a value for the meta-keyword OLAS-ID. 
            // 
            // The OLAS ID comprises the characters between the last slash 
            // (if any) in the filename and the first dot after that point. 
            String olasID = null; 

            // Find the last slash (if any). 
            int indexOfLastSlash = filename.lastIndexOf(File.separatorChar); 

            if (indexOfLastSlash < 0) 
            {
                // There are no slashes - think of this as a 'virtual slash' 
                // immediately preceding the filename... 
                indexOfLastSlash = -1; 
            }

            // Find the first dot. 
            int indexOfFirstDot = filename.indexOf('.', indexOfLastSlash); 

            if (indexOfFirstDot < 0) 
            {
                logger.warn(methodName + " - could not determine the "
                            + "OLAS ID for file [" + linkname + "]."); 
                olasID = ""; 
            } 
            else 
            { 
                olasID = filename.substring(indexOfLastSlash + 1, 
                                            indexOfFirstDot); 
            }

            // Add the meta-keyword OLAS-ID. 
            keywordValues.put(OCAFile.OLAS_ID, olasID.toUpperCase());

            // Add the meta-keyword DP.ID. 
            keywordValues.put(OCAFile.DP_ID, makeDpId(filename)); 

            // Create a new OCAFile from the card Hashtable. 
            OCAFile ocaFile = new OCAFile(keywordValues);

            // Check the new OCAFile against the file filter. 
            if (filter.filter(ocaFile) == true)
            {
                logger.debug(methodName + " - file filter accepted file [" 
                             + linkname + "]."); 
                
                readFiles.add(ocaFile);
            }
            else
            {
                logger.debug(methodName + " - file filter rejected file [" 
                             + linkname + "]."); 
                
                skippedFiles.add(new File(linkname)); 
            }
        }
    }

    /** This method converts an array of TypedHeaderCards to a Hashtable of 
     * keyword-value pairs. 
     * 
     * @throws NullPointerException if cards is null. 
     * @throws TypedHeaderCardException if an unsupported FITS type is 
     *                                  encountered. 
     *
     * @param cards array of TypedHeaderCards. 
     *
     * @return Hashtable of keyword-value pairs. 
    */
    public static Hashtable convertArrayOfCardsToHashtable(
        TypedHeaderCard[] cards) 
        throws TypedHeaderCardException 
    {
        logger.trace("OCAFileFactory::convertArrayOfCardsToHashtable()"); 

        if(cards == null)
        {
            final String message = 
                "OCAFileFactory::convertArrayOfCardsToHashtable()" 
                + " - cards must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        } 

        TypedHeaderCard currentCard;
        String currentKey;
        Hashtable keyValues = new Hashtable();

        /* Loop through the TypedHeaderCards and store their values. */
        for (int i = 0; i < cards.length; i++) 
        {
            currentCard = cards[i];
            currentKey = currentCard.getKey(TypedHeaderCard.SHORT_DOTTED);

            switch (currentCard.getType()) 
            {
              case TypedHeaderCard.BOOLEAN:
              case TypedHeaderCard.FLOAT:
              case TypedHeaderCard.INTEGER:
              case TypedHeaderCard.STRING:
              case TypedHeaderCard.UNTYPED:
              case TypedHeaderCard.UNDEFINED:
                keyValues.put(currentKey, currentCard.getValueObject());
                break;

              default:
                /* This FITS type is not supported by the OCA interpreter. */
                final String message = 
                    "OCAFileFactory::convertArrayOfCardsToHashtable()" 
                    + " - type error for keyword '" + currentKey
                    + "' with type '" + currentCard.getTypeString() + "'."; 

                logger.error(message);
                throw new TypedHeaderCardException(message);
            }
        }

        return(keyValues); 
    }
    
    /** Generates a DPID from the specified filename. 
     * 
     * @throws NullPointerException if filename is null. 
     * 
     * @param filename the filename to generate the DPID from. 
     * 
     * @return the DPID string. 
     */ 
    public static String makeDpId(String filename)
    {
        logger.trace("OCAFileFactory::makeDpId()"); 

        if(filename == null)
        {
            final String message = 
                "OCAFileFactory::makeDpId() - filename must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        } 

        String basename = new File(filename).getName();
        String dpId = basename;

        int indexOfLastDot = basename.lastIndexOf('.');

        if(indexOfLastDot > 0)
        {
            dpId = basename.substring(0, indexOfLastDot);
        }

        return dpId;
    }
}
