package org.eso.oca.fits;

import java.lang.String;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.lang.NumberFormatException;

import nom.tam.fits.HeaderCard;
import nom.tam.fits.HeaderCardException;

import org.apache.log4j.Logger;

/** <P>This class extends the HEASARC JavaFITS HeaderCard class. It provides 
 * a set of methods allowing the user to obtain information about the type of 
 * the [Typed]HeaderCard's value, if it has one. TypedHeaderCard instances 
 * can be constructed from an existing HeaderCard, or from separate 
 * key/value/comment parts. As with HeaderCard itself, the resulting objects 
 * are immutable.</P>
 *
 * <P>Supported value types are UNDEFINED, UNTYPED, STRING, BOOLEAN, INTEGER 
 * and FLOAT. If the value type cannot be deduced, a TypeDeterminationException
 * will be thrown. The complex number types defined by the FITS standard are 
 * not supported at present, so a TypeDeterminationException will also be 
 * thrown if a value has one of these types.</P> 
 *
 * <P><B>WARNING:</B> The FITS standard effectively defines the INTEGER value 
 * type as a subset of the FLOAT value type. As a result, there is no way 
 * for a TypedHeaderCard to decide whether a numeric value with no decimal 
 * point and no exponent is an INTEGER or a FLOAT. In such cases, the type 
 * is arbitrarily set to INTEGER.</P>
 * 
 * <P><B>WARNING:</B> There is a minor inconsistency in the way HeaderCard 
 * handles leading spaces in comments. Normally, leading spaces are removed, 
 * but this may not occur if the comment is attached to a non-hierarchical 
 * keyword having a STRING value type. For example, the comment in:</P>
 *
 * <P><PRE><CODE>
 * ORIGIN  = 'ESO     '           / European Southern Observatory
 * </CODE></PRE></P>
 *
 * <P>will be parsed as "<SAMP> European Southern Observatory</SAMP>", rather 
 * than "<SAMP>European Southern Observatory</SAMP>" without the leading 
 * space. If a [Typed]HeaderCard like this is copied many times, the comment 
 * can gradually acquire more and more leading spaces, until it finally gets 
 * pushed out of the 80 byte HeaderCard space and disappears.</P> 
 *
 * <P><B>WARNING:</B> The HeaderCard class has a bug in its component-part 
 * constructors, where the keyword, value & comment are specified separately. 
 * If the value is given as a String, and that String is empty (length zero), 
 * a StringIndexOutOfBoundsException will occur. The workaround for this is 
 * to submit the String with FITS quotes, so that the empty string is given 
 * as ''. The quotes are automatically stripped by the HeaderCard constructor 
 * before the String value is stored. 
 * 
 * @author Neil Ferguson ESO/DMD/DFS <nferguso@eso.org> 
 */ 
public class TypedHeaderCard extends HeaderCard 
{
    static Logger logger = Logger.getLogger(TypedHeaderCard.class);

    /* *** BEGIN VALUE TYPE CONSTANTS *** */ 

    /** This is a special type used only by the {@link #findFITSType(Object)} 
     * method. It indicates that the TypedHeaderCard class is not able to map 
     * a particular Java object type to a FITS value type. 
     */ 
    public static final int UNKNOWN = 0; 

    /** This is a special type indicating that the keyword in question is 
     * completely undefined. For example, it may not have been present 
     * within a given FITS file. 
     */
    public static final int UNDEFINED = 1;

    /** The keyword is defined, but does not have a value type, which means 
     * that no value is present, or that the keyword is a comment keyword. 
     */
    public static final int UNTYPED = 2;

    /** The keyword is defined and has a value of type STRING. */
    public static final int STRING = 3;

    /** The keyword is defined and has a value of type BOOLEAN. */
    public static final int BOOLEAN = 4;

    /** The keyword is defined and has a value of type INTEGER. */
    public static final int INTEGER = 5;

    /** The keyword is defined and has a value of type FLOAT. */
    public static final int FLOAT = 6;

    /* COMPLEX NUMBER TYPES GO HERE. */ 

    /* *** END VALUE TYPE CONSTANTS *** */ 


    /* *** BEGIN KEY FORMATTING CONSTANTS *** */ 

    /** The ESO HIERARCH prefix with dots. */
    public static final String KEY_PREFIX_DOTTED = new String("HIERARCH.ESO.");

    /** The ESO HIERARCH prefix with spaces. */
    public static final String KEY_PREFIX_SPACED = new String("HIERARCH ESO ");

    /** Denotes a key formatted in short form (without prefix) 
     * with dot separators. 
     */
    public static final int SHORT_DOTTED = 0;

    /** Denotes a key formatted in short form (without prefix) 
     * with space separators. 
     */
    public static final int SHORT_SPACED = 1; 

    /** Denotes a key formatted in long form (with prefix) 
     * with dot separators. 
     */
    public static final int LONG_DOTTED = 2; 

    /** Denotes a key formatted in long form (with prefix) 
     * with space separators. 
     */
    public static final int LONG_SPACED = 3; 

    /* *** END KEY FORMATTING CONSTANTS *** */ 


    /** The value type of this TypedHeaderCard. */
    private int valueType; 

    /** The actual value of this TypedHeaderCard, stored in an instance of 
     * the appropriate Java class. 
     * 
     * @see #establishType()
     * @see #getValueObject()
     */
    private Object valueObject; 

    /** Creates a TypedHeaderCard based on the given HeaderCard. 
     *
     * @throws HeaderCardException this should never occur in practice. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(HeaderCard card) 
        throws HeaderCardException, TypeDeterminationException 
    {
        super(card.toString()); 
        establishType();
    }

    /** Creates a TypedHeaderCard based on the given FITS card image. 
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String cardImage) 
        throws HeaderCardException, TypeDeterminationException 
    {
        super(cardImage); 
        establishType();
    }

    /** Creates a TypedHeaderCard with the given keyword, but no value or 
     * comment. This is the only way to construct a TypedHeaderCard with 
     * the UNDEFINED type. 
     *
     * @param key keyword 
     * @param isDefined whether the keyword should be considered defined 
     *                  or not. Set this to false to consider the keyword 
     *                  UNDEFINED, or true to consider the keyword defined 
     *                  but UNTYPED. If isDefined is true, this call will 
     *                  be equivalent to TypedHeaderCard(key, null, null). 
     *
     * @see #UNDEFINED
     * @see #UNTYPED
     *
     * @throws HeaderCardException for an invalid keyword. 
     * @throws TypeDeterminationException this should never occur in practice.
     */
    public TypedHeaderCard(String key, boolean isDefined)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, null, null);

        if(isDefined)
        {
            valueType = UNTYPED;
            valueObject = new UntypedFITSValue(); 
        }
        else
        {
            valueType = UNDEFINED;
            valueObject = new UndefinedFITSValue(); 
        }
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values)
     * @param comment comment (set this to null if no comment is needed)
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, boolean value, String comment)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment); 
        establishType();
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values)
     * @param comment comment (set this to null if no comment is needed)
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, double value, String comment)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment); 
        establishType();
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values)
     * @param comment comment (set this to null if no comment is needed)
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, int value, String comment)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment); 
        establishType();
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values)
     * @param comment comment (set this to null if no comment is needed)
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, long value, String comment)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment); 
        establishType();
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values) 
     * @param comment comment (set this to null if no comment is needed)
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, String value, String comment)
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment); 
        establishType();
    }

    /** Creates a TypedHeaderCard from separate key/value/comment parts. 
     *
     * @param key keyword (set this to null for a comment)
     * @param value value 
     *              (set this to null for comments or keywords without values) 
     * @param comment comment (set this to null if no comment is needed)
     * @param nullable true if this card's value is nullable
     *
     * @throws HeaderCardException for any invalid keyword or value. 
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */
    public TypedHeaderCard(String key, 
                           String value, 
                           String comment, 
                           boolean nullable) 
        throws HeaderCardException, TypeDeterminationException 
    {
        super(key, value, comment, nullable); 
        establishType();
    }

    /** <P>Returns the key in the specified format. Permitted formats are 
     * SHORT_DOTTED, SHORT_SPACED, LONG_DOTTED and LONG_SPACED. If the 
     * format is not recognised, the unmodified key will be returned.</P> 
     *
     * <P><I>NOTE:</I> It is assumed that hierarchical keywords with only 
     * one level are never used. For example, it makes more sense to define 
     * a custom key called "WIDGET" than one called "HIERARCH ESO WIDGET". 
     * Single-level keywords have no need of hierarchical constructs.</P> 
     *
     * <P>This assumption allows formatKey() to heuristically determine 
     * which keys should be given a hierarchical prefix when reformatting 
     * them into a long form - any short-form key containing separators 
     * must be a hierarchical key, all others must be non-hierarchical.</P> 
     *
     * @see #SHORT_DOTTED
     * @see #SHORT_SPACED
     * @see #LONG_DOTTED
     * @see #LONG_SPACED
     */
    public static String formatKey(String key, int format)
    {
        String prefix = KEY_PREFIX_DOTTED; 
        String result = key; 

        /* Unrecognised format specifications will fall through both switch 
         * statements, causing this method to return the unmodified key as
         * its result. 
         */
        switch(format)
        {
          case SHORT_DOTTED:
          case LONG_DOTTED:
            /* Requested format uses dots - replace all spaces. */ 
            result = key.replace(' ', '.');
            /* Dotted key prefix already selected. */ 
            break; 

          case SHORT_SPACED:
          case LONG_SPACED:
            /* Requested format uses spaces - replace all dots. */ 
            result = key.replace('.', ' ');
            /* Select the spaced key prefix. */ 
            prefix = KEY_PREFIX_SPACED;
            break;
        }

        switch(format)
        {
          case SHORT_DOTTED:
          case SHORT_SPACED:
            if(result.indexOf(prefix) == 0)
            {
                /* Short format requested, but the prefix is present 
                 * - remove it. 
                 */
                result = result.substring(prefix.length());
            }
            break; 

          case LONG_DOTTED:
            if((result.indexOf('.') > 0) && (result.indexOf(prefix) < 0))
            {
                /* Long dotted format requested, dot separators present, 
                 * but no prefix - prepend it. 
                 */
                result = prefix + result;
            }
            break; 

          case LONG_SPACED:
            if((result.indexOf(' ') > 0) && (result.indexOf(prefix) < 0))
            {
                /* Long spaced format requested, space separators present, 
                 * but no prefix - prepend it. 
                 */
                result = prefix + result;
            }
            break; 
        }

        return(result);
    }

    /** Returns this object's key in the specified format. Permitted formats are 
     * SHORT_DOTTED, SHORT_SPACED, LONG_DOTTED and LONG_SPACED. 
     *
     * @see #formatKey(String,int)
     * @see #SHORT_DOTTED
     * @see #SHORT_SPACED
     * @see #LONG_DOTTED
     * @see #LONG_SPACED
     */
    public String getKey(int format)
    {
        return(formatKey(getKey(), format));
    }

    /** If the HeaderCard's value is a FITS INTEGER, sets valueType and 
     * valueObject appropriately, then returns true. Otherwise, no changes 
     * are made and false is returned. 
     */
    private boolean tryAsFITSInteger()
    {
        try
        {
            String value = getValue();

            /* FITS permits a leading '+' character for positive integers, 
             * but the Java 1.4 API does not require wrapper objects (such 
             * as Integer or BigInteger) to accept this. As a result, any 
             * leading '+' character must be stripped prior to parsing. 
             */ 
            if((value.length() > 0) && (value.charAt(0) == '+')) 
            {
                value = value.substring(1); 
            }

            /* Attempt to parse the value as a integer - throws a 
             * NumberFormatException in the event of a failure. 
             */ 
            valueObject = new BigInteger(value); 

            /* Parsing succeeded - this value is a FITS INTEGER. */
            valueType = INTEGER;
            return true;
        }
        catch(NumberFormatException nfe) {}
        
        /* Parsing failed - this value is not a FITS INTEGER. */
        return false;
    }

    /** If the HeaderCard's value is a FITS FLOAT, sets valueType and 
     * valueObject appropriately, then returns true. Otherwise, no 
     * changes are made and false is returned. 
     */
    private boolean tryAsFITSFloat()
    {
        try
        {
            String value = getValue();

            /* Reject constructs which are valid in Java but not FITS: 
             * - Java allows floating point type suffixes, FITS does not. 
             */
            if(value.matches("[^\\d\\.]\\z"))
            {
                throw new NumberFormatException();
            }

            /* Attempt to parse the value as a floating point number - throws 
             * a NumberFormatException in the event of a failure. Note that 
             * FITS allows 'D' for exponents, but Java does not, so any 'D' 
             * characters must be substituted with an 'E'. 
             */
            valueObject = new BigDecimal(value.replace('D', 'E'));

            /* Parsing succeeded - this value is a FITS FLOAT. */
            valueType = FLOAT;
            return true;
        }
        catch(NumberFormatException nfe) {}

        /* Parsing failed - this value is not a FITS FLOAT. */
        return false;
    }

    /** This method tries to deduce the HeaderCard's value type. 
     *
     * @throws TypeDeterminationException if the value type could not be 
     *                                    determined. 
     */ 
    private void establishType() throws TypeDeterminationException
    {
        String value = getValue();

        if(value == null)
        {
            /* The HeaderCard has no value, so no type can be set. */ 
            valueType = UNTYPED;
            valueObject = new UntypedFITSValue();
            return;
        }
        
        if(isStringValue())
        {
            /* The HeaderCard contains a STRING value. */
            valueType = STRING;
            valueObject = value.trim(); 
            return;
        }
        
        if(value.equals("T"))
        {
            /* The HeaderCard contains the BOOLEAN 'true' value. */
            valueType = BOOLEAN;
            valueObject = new Boolean(true);
            return;
        }
        else if(value.equals("F"))
        {
            /* The HeaderCard contains the BOOLEAN 'false' value. */
            valueType = BOOLEAN;
            valueObject = new Boolean(false);
            return;
        }
        
        if(tryAsFITSInteger())
        {
            /* The HeaderCard contains an INTEGER value. */
            /* The valueType and valueObject are set by tryAsFITSInteger(). */ 
            return;
        }

        if(tryAsFITSFloat())
        {
            /* The HeaderCard contains a FLOAT value. */
            /* The valueType and valueObject are set by tryAsFITSFloat(). */
            return;
        }

        /* COMPLEX NUMBER TYPE DETERMINATION GOES HERE. */ 

        /* No type could be deduced for the HeaderCard's value. */ 
        String message = 
            "No value type could be deduced for '" 
            + getKey() + " = " + getValue() + " / " + getComment() + "'."; 

        logger.error(message); 
        throw new TypeDeterminationException(message); 
    }

    /** Returns the numeric constant corresponding to this object's value 
     * type. 
     *
     * @see #UNDEFINED
     * @see #UNTYPED
     * @see #STRING
     * @see #BOOLEAN
     * @see #INTEGER
     * @see #FLOAT
     */
    public int getType()
    {
        return(valueType); 
    }

    /** Returns a String corresponding to this object's value type. */ 
    public String getTypeString()
    {
        String result; 

        switch(valueType)
        {
          case UNDEFINED:
            result = "UNDEFINED";
            break;

          case UNTYPED:
            result = "UNTYPED";
            break;

          case STRING:
            result = "STRING";
            break;

          case BOOLEAN:
            result = "BOOLEAN";
            break;

          case INTEGER:
            result = "INTEGER";
            break;

          case FLOAT:
            result = "FLOAT";
            break;

          /* COMPLEX NUMBER TYPES GO HERE. */ 

          default:
            result = "UNKNOWN(" + valueType + ")"; 
            break;
        }

        return(result);
    }

    /** Returns a numeric constant giving the FITS type that the candidate 
     * object would correspond to, if that object had been returned by the 
     * {@link #getValueObject()} method. Returns {@link #UNKNOWN} if the 
     * object's runtime type does not correspond to a FITS type known to 
     * the TypedHeaderCard class. 
     *
     * @see #UNKNOWN 
     * @see #UNDEFINED
     * @see #UNTYPED
     * @see #STRING
     * @see #BOOLEAN
     * @see #INTEGER
     * @see #FLOAT
     */ 
    public static int findFITSType(final Object candidate) 
    {
        final Class type = candidate.getClass(); 
        int result; 

        if(type.equals(UndefinedFITSValue.class))
        {
            result = UNDEFINED;
        }
        else if(type.equals(UntypedFITSValue.class))
        {
            result = UNTYPED;
        }
        else if(type.equals(String.class))
        {
            result = STRING;
        }
        else if(type.equals(Boolean.class))
        {
            result = BOOLEAN;
        }
        else if(type.equals(BigInteger.class))
        {
            result = INTEGER;
        }
        else if(type.equals(BigDecimal.class))
        {
            result = FLOAT;
        }
        /* COMPLEX NUMBER TYPES GO HERE. */ 
        else
        {
            result = UNKNOWN; 
        }

        return(result); 
    }

    /** Returns a String giving the FITS type that the candidate object would 
     * correspond to, if that object had been returned by the 
     * {@link #getValueObject()} method. Returns "UNKNOWN(<I>classname</I>)" 
     * if the object's runtime type does not correspond to a FITS type known 
     * to the TypedHeaderCard class, where <I>classname</I> is the Java class 
     * name. 
     */ 
    public static String findFITSTypeString(final Object candidate)
    {
        final Class type = candidate.getClass();
        String result; 

        if(type.equals(UndefinedFITSValue.class))
        {
            result = "UNDEFINED";
        }
        else if(type.equals(UntypedFITSValue.class))
        {
            result = "UNTYPED";
        }
        else if(type.equals(String.class))
        {
            result = "STRING";
        }
        else if(type.equals(Boolean.class))
        {
            result = "BOOLEAN";
        }
        else if(type.equals(BigInteger.class))
        {
            result = "INTEGER";
        }
        else if(type.equals(BigDecimal.class))
        {
            result = "FLOAT";
        }
        /* COMPLEX NUMBER TYPES GO HERE. */ 
        else
        {
            result = "UNKNOWN(" + type.getName() + ")"; 
        }

        return(result); 
    }

    /** Returns an Object corresponding to this object's value. The mapping of 
     * FITS types to Java classes is: 
     * 
     * <P><DL>
     * <DT>UNDEFINED</DT><DD>org.eso.oca.fits.UndefinedFITSValue</DD>
     * <DT>UNTYPED</DT><DD>org.eso.oca.fits.UntypedFITSValue</DD>
     * <DT>STRING</DT><DD>java.lang.String</DD>
     * <DT>BOOLEAN</DT><DD>java.lang.Boolean</DD>
     * <DT>INTEGER</DT><DD>java.lang.BigInteger</DD>
     * <DT>FLOAT</DT><DD>java.lang.BigDecimal</DD>
     * </DL></P>
     */ 
    public Object getValueObject()
    {
        return(valueObject);
    }

    /** Returns true if the keyword is not defined.  */
    public boolean isUndefined()
    {
        return(valueType == UNDEFINED); 
    }

    /** Returns true if the keyword is defined.  */
    public boolean isDefined()
    {
        return(valueType != UNDEFINED); 
    }

    /** Returns true if the keyword does not have a value type.  */
    public boolean isUntyped()
    {
        return(valueType == UNTYPED); 
    }

    /** Returns true if the keyword has a STRING value type. */
    public boolean isString()
    {
        return(valueType == STRING); 
    }

    /** Returns true if the keyword has a BOOLEAN value type. */
    public boolean isBoolean()
    {
        return(valueType == BOOLEAN); 
    }

    /** Returns true if the keyword has an INTEGER value type. */
    public boolean isInteger()
    {
        return(valueType == INTEGER); 
    }

    /** Returns true if the keyword has a FLOAT value type. */
    public boolean isFloat()
    {
        return(valueType == FLOAT); 
    }

    /* COMPLEX NUMBER TYPES GO HERE. */ 
}
