package org.eso.oca.fitsmapper;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.eso.oca.data.MetaColumn;
import org.eso.oca.data.MetaTable;
import org.eso.oca.data.ObsColumnsDAO;
import org.eso.oca.data.ObsTablesDAO;

/**
 * Interface DbFitsMapper
 *
 * This is the abstract class for a DbFitsMapper:
 * 
 * A class which maps between FITS keywords and columns in a database.
 *  
 * The configuration assumptions are the following:
 * 
 *  - The configuration is stored in a database. For the observations database the configuration
 * is stored in "obs_tables" and "obs_columns" table
 *  
 * - The "obs_tables" table contains pairs (instrument,table), which means that for a certain instrument,
 * its data may be stored in different tables. A data product may be stored in one or more tables.
 * 
 * - The "obs_columns" contains information about the keywords. Each keyword is assumed to have only on row in this tables. 
 * Thus, a keyword is assumed to correspond only to one column in a table.
 *  
 * 
 * Each implementation of this abstract class will store the proper business objects in the hashtables
 * All methods throw a DbFitsMapperException if something unexpected occurs 
 *
 * @version $Revision: 1.4 $
 * @author pnunes, Apr 1, 2005
 */

// $Id: DbFitsMapper.java,v 1.4 2005/06/06 10:31:16 szampier Exp $

public abstract class DbFitsMapper {
    

    /**
     * Hashtable containing a Vector of objects from calib or raw   
     * and the instrument_name as key
     */
    protected Hashtable tables;
    
    /**
     * Hashtable containing  objects  from calib or raw  
     * and the table_name+keyword as key
     */
    protected Hashtable columnsFromKW;
    
    /**
     * Hashtable containing table objects from calib or raw  
     * and the table_name+column_name as key
     */
    protected Hashtable columnsFromColumnName;
    
    
            
    /**
     * Returns the name of a column, given an instrument and a keyword
     * 
     * @param instrument
     * @param keyword
     * @return a String containing the name of a column
     * @throws DbFitsMapperException
     */
    
    //public abstract String getColumn(String instrument, String keyword) throws DbFitsMapperException;

    public String getColumn(String instrument, String keyword) throws DbFitsMapperException {
        String columnName = null;
        String tableName = null;
        Vector metaTables;
        MetaTable metaTable;
        MetaColumn metaColumn = null;
        
        
        //      find instrument table
        metaTables = (Vector)tables.get(instrument);
        
        if (metaTables == null)
            throw new DbFitsMapperException("No table found for instrument " + instrument + " and keyword " + keyword);
        
        // iterate tables and find the column
                       
        for (Iterator iter = metaTables.iterator(); iter.hasNext();) {
            metaTable = (MetaTable) iter.next();                	
            tableName = metaTable.getName();
            
            //   find column
            metaColumn = (MetaColumn)columnsFromKW.get(tableName+keyword);
            
            if ( metaColumn != null) 
                break;
        }
                                                       
        //  if there's no column, throw the exception 
        if (metaColumn == null)
            throw new DbFitsMapperException("No column found for instrument " + instrument + " and keyword " + keyword);
        
        columnName = metaColumn.getName();
        
        return columnName;
    }
       
    
    /**
     * Returns the name of a table, given an instrument and a keyword
     * 
     * @param instrument
     * @param keyword
     * @return
     * @throws DbFitsMapperException
     */               
    public String getTable(String instrument, String keyword) throws DbFitsMapperException {
        String tableName = null;
        MetaTable metaTable = null;
        MetaColumn metaColumn = null;
        Vector metaTables = null;  
        
        //      find instrument table
        metaTables = (Vector)tables.get(instrument);
        
        if (metaTables == null)
            throw new DbFitsMapperException("No table found for instrument " + instrument + " and keyword " + keyword);
        
        for (Iterator iter = metaTables.iterator(); iter.hasNext();) {
            metaTable = (MetaTable) iter.next();
        
            tableName = metaTable.getName();
        
            // check if the table names matches
            metaColumn = (MetaColumn)columnsFromKW.get(tableName+keyword);
            
            if ( metaColumn != null) 
                break;
 
        }
                
        if (metaColumn == null)
            throw new DbFitsMapperException("No column found for keyword " + keyword + " and instrument " + instrument);
                                       
        return tableName;
    }
    
    
    /**
     * Returns the name of a column, given an instrument and a keyword, with sybase sintaxe
     * like the following: table_name..column_name
     * 
     * @param instrument
     * @param keyword
     * @return
     * @throws DbFitsMapperException
     */
    public String getTableAndColumn(String instrument, String keyword) throws DbFitsMapperException {
        String columnName = null;
        String tableName = null;
        
        tableName = getTable(instrument,keyword);
        columnName = getColumn(instrument,keyword);        
        
        if (tableName==null || columnName==null)
            throw new DbFitsMapperException("keyword "+keyword+" does not exist in DB");
            
        return tableName + "." + columnName;
    }
           
    
    /**
     * Adds an association between an instrument and a table
     * If this association already exists, it will be overwritten
     * 
     * @param table
     * @param instrument
     */     
            
    public void addTable(String table, String instrument) {
        Vector metaTables = (Vector)tables.get(instrument);
        ObsTablesDAO obsTablesDAO = ObsTablesDAO.getInstance();
        
        MetaTable metaTable = obsTablesDAO.createObsTablesBean();
        metaTable.setName(table);
        metaTable.setInstrume(instrument);
        
        if ( metaTables == null) 
            metaTables = new Vector();
        
        metaTables.add(metaTable);
        tables.put(instrument,metaTables);
        
    }
    
    

    
    /**
     * Adds an association between a keyword and a column
     * If this association already exists, it will be overwritten
     * 
     * @param keyword
     * @param table
     * @param name
     */
    //public abstract void addColumn(String keyword , String table , String name);
        
    public void addColumn(String keyword, String table, String name) {
        
        ObsColumnsDAO obsColumnsDAO = ObsColumnsDAO.getInstance();
        
        
        MetaColumn metaColumn = obsColumnsDAO.createObsColumnsBean();        
        metaColumn.setKeyword(keyword);
        metaColumn.setTname(table);
        metaColumn.setName(name);        
        columnsFromKW.put( metaColumn.getTname() + metaColumn.getKeyword()  , metaColumn );
        columnsFromColumnName.put ( metaColumn.getTname()+ metaColumn.getName() , metaColumn );
    }
    
    
    
    
}





