package org.eso.oca.parser;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.oca.fits.OCAFileFactory;
import org.eso.oca.fits.TypedHeaderCard;
import org.eso.oca.fits.TypedHeaderCardException;

/** This class acts purely as a container for OCA state that may not be 
 * shared across threads. There is one OCAState object for any given 
 * evaluation of OCA rules. Most fields have package visibility so that 
 * the OCA interpreter can reach them directly, but user code is required 
 * to use accessor methods. 
 *
 * @author Neil Ferguson ESO/DMD/DFS <nferguso@eso.org> 
 */
public class OCAState
{
    static Logger logger = Logger.getLogger(OCAState.class);

    /** The following variables are updated at interpretation time */ 

    /** symbol table used by the interpreter. It is loaded
     * with the keyword values of the file being classified
     */
    Hashtable symtab = new Hashtable();
    
    /** this symbol table is used during the interpretation
     * of association rules, to hold the keyword values
     * of the input (raw) file 
     */
    Hashtable inputFileSymtab = new Hashtable();
    
    /** the calibration files which can be selected by the
     * association rules. This list must be loaded before
     * invoking the association process. New (virtual) products
     * can be added through the product rules.
     */
    Vector globCalibFiles = new Vector();
 
    /** the input files which can be selected by the
     * association rules. This list must be loaded before
     * invoking the association process.
     */
    Vector globInputFiles = new Vector();

    /** whether the current file was selected by the organization
     * rule being interpreted or not.
     */
    boolean isSelected = false;

    /** whether the Association Block being built is valid or not. E.g.
     * an AB is not valid if not enough associated files were found.
     */
    boolean isABValid = true;

    /** the minimum number of requested products */
    Integer minRet = new Integer(1);
    
    /** the maximum number of requested products */
    Integer maxRet = new Integer(1);
    
    /** the name of the recipe, as specified in the action
     * being interpreted
     */
    String globRecipe = "";

    /** the recipe parameters */
    Vector globRecipeParams = new Vector();
    
    /** the associated files */
    Vector associatedFiles = new Vector();
    
    /** the list of (virtual) products,  as specified in the action
     * being interpreted
     */
    Vector globVirtualProducts = new Vector();

    /** calculation stack for the internal interpreter. */
    Object[] stack = new Object[1024];

    /** top of the calculation stack */
    int top = -1;

    /** Used to count interpretation operations for logging purposes. */ 
    long opCounter = 0;

    /** Dump the symbol table from this OCAState to the standard output. */ 
    public void dumpSymtab()
    {
        Enumeration keys = symtab.keys();
        while (keys.hasMoreElements())
        {
            String key = (String) keys.nextElement();
            System.out.println(key + " = " + symtab.get(key));
        }
    }

    /** Stores keywords and values from the given TypedHeaderCards into this 
     * OCAState for use by the OCA interpreter. 
     * 
     * @param requestedData an array of TypedHeaderCards whose contents are 
     *                      to be stored. 
     * 
     * @throws TypedHeaderCardException if a TypedHeaderCard is encountered 
     *                                  whose FITS value type is not supported 
     *                                  by the OCA interpreter. 
     */
    public void updateSymtab(TypedHeaderCard[] requestedData) 
        throws TypedHeaderCardException
    {
        symtab = OCAFileFactory.convertArrayOfCardsToHashtable(requestedData);
    }

    /** Returns a list of all metakeywords and any values they have, as an 
     * array of TypedHeaderCards. This method assumes that all metakeyword 
     * values are of type STRING. 
     *
     * @param node a node in the OCA parse tree containing the metakeyword 
     *             corpus to use. 
     *
     * @throws TypedHeaderCardException if a metakeyword and its value could 
     *                                  not be stored as a TypedHeaderCard. 
     */
    public TypedHeaderCard[] getMetaKeywords(SimpleNode node) 
        throws TypedHeaderCardException
    {
        Collection metaKeywordCorpus = node.getMetaKeywordCorpus(); 

        TypedHeaderCard[] ret = 
            new TypedHeaderCard[metaKeywordCorpus.size()];

        int i = 0;              /* Current position in metaKeywords[] */ 

        for (Iterator it = metaKeywordCorpus.iterator(); it.hasNext();)
        {
            /* The metakeyword. */
            String metaKey = (String) it.next();

            /* The metakeyword value (assumed to be of STRING type). 
             *
             * NOTE: The value string needs to be quoted in order to 
             * sidestep a HEASARC FITS library bug. 
             */
            Object metaVal = symtab.get(metaKey); 
            String metaValString; 

            if(metaVal == null)
                metaValString = "''"; 
            else
                metaValString = "'" + metaVal.toString() + "'"; 

            /* Ensure that the metakeyword is in long format with dot 
             * separators, as required by the HEASARC nom.tam.fits 
             * library classes. 
             */
            metaKey = TypedHeaderCard.formatKey(metaKey, 
                                                TypedHeaderCard.LONG_DOTTED); 

            try 
            {
                /* Store the metakeyword and its value as a TypedHeaderCard 
                 * for return. 
                 */
                ret[i++] = new TypedHeaderCard(metaKey, metaValString, null); 
            }
            catch(Exception e)
            {
                String message = 
                    "Could not use metakeyword '" 
                    + metaKey + " = " + metaValString 
                    + "' to construct a TypedHeaderCard object."; 

                logger.error(message); 
                throw new TypedHeaderCardException(message, e); 
            }
        }
        return ret;
    }

}
