package org.eso.util.configuration; 

import java.io.File;

import org.apache.log4j.Logger;

/** <p>An instance of this class defines a single configuration option, as 
 * handled by the {@link Configuration} class. In order to ensure appropriate 
 * error checking, ConfigurationOption objects should only be created by the 
 * {@link MetaConfigurator#metaConfigureOption(Properties, String)} method, 
 * or an equivalent to it.</p> 
 *
 * <p>For that reason, this class is final, and its constructor can only be 
 * called from within the configuration package. The only public components 
 * of this class are the constants and convenience methods which allow an 
 * application to work with option value types and sources.</p> 
 * 
 * <p>The following constants are used to define the data type of a 
 * configuration option's value: 
 * <ul>
 *   <li>{@link #NO_TYPE}</li>
 *   <li>{@link #STRING}</li>
 *   <li>{@link #BOOLEAN}</li>
 *   <li>{@link #INTEGER}</li>
 * </ul>
 * An option can have only one data type for its value; the type is fixed at 
 * meta-configuration time.</p> 
 * 
 * <p>Further constants are provided for defining the source of a value: 
 * <ul>
 *   <li>{@link #NO_SOURCE}</li>
 *   <li>{@link #DEFAULT_VALUE}</li>
 *   <li>{@link #APPLICATION_PROPERTY}</li>
 *   <li>{@link #SYSTEM_PROPERTY}</li>
 *   <li>{@link #COMMAND_LINE}</li>
 *   <li>{@link #INTERNAL_CONFIGURATION}</li>
 *   <li>{@link #ALL_SOURCES}</li>
 * </ul>
 * Although the actual value of an option can only come from one specific 
 * source, it can be updated with a value supplied by any valid source for 
 * that option. The set of valid sources is fixed at meta-configuration 
 * time and stored as a bitmask.</p> 
 *
 * <p><b>NOTE:</b> A ConfigurationOption's value may only be set if the 
 * new value's source has a higher priority than the old source, and the 
 * new source is allowed by the option's valid source bitmask.</p> 
 *
 * <p>The priority of a value's source is determined by the corresponding
 * constant. For example, a {@link #DEFAULT_VALUE} has priority 1 and can 
 * be overridden by a value from any other source. However, a value from 
 * the {@link #COMMAND_LINE} (priority 8) can only be overridden if the 
 * application decides to use an {@link #INTERNAL_CONFIGURATION} (priority 
 * 16).</p> 
 *
 * <p><i>The following paragraph is not relevant to application developers.
 * </i></p> 
 * 
 * <p><b>NOTE:</b> Some fields of a ConfigurationOption are only relevant 
 * under specific circumstances. For example: 
 * <ul> 
 *   <li>the propertyName need only be set if this option can be set from a 
 *       Java property.</li>
 *   <li>the propertyFile need only be set if the option's value has actually 
 *       been set from that file.</li>
 * </ul>
 * Unused or irrelevant fields may be set to arbitrary values such as null, 
 * empty string or zero.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public final class ConfigurationOption 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(ConfigurationOption.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "ConfigurationOption"; 

    /** The option name - the application uses this name as a key when 
     * retrieving the option's value. 
     */
    final String optionName; 

    /** Designates an uninitialised (and invalid) option type. */ 
    public static final int NO_TYPE = 0; 

    /** Designates a string option. */
    public static final int STRING = 1; 

    /** Designates a boolean option. */
    public static final int BOOLEAN = 2; 

    /** Designates an integer option. */
    public static final int INTEGER = 3; 

    /** The option type - valid values are {@link STRING}, {@link BOOLEAN} or 
     * or {@link INTEGER}. 
     */ 
    final int optionType; 

    /** Designates an option whose source has not been set. */ 
    public static final int NO_SOURCE = 0x0000; 

    /** Designates an option with a default value. */
    public static final int DEFAULT_VALUE = 0x0001; 

    /** Designates an application property. */
    public static final int APPLICATION_PROPERTY = 0x0002; 

    /** Designates a system property. */ 
    public static final int SYSTEM_PROPERTY = 0x0004; 

    /** Designates a command line option. */ 
    public static final int COMMAND_LINE = 0x0008; 

    /** Designates an internally configured option. */
    public static final int INTERNAL_CONFIGURATION = 0x0010; 

    /** Designates an option whose value can come from any of the known 
     * sources - combines all of the constants {@link #DEFAULT_VALUE}, 
     * {@link #APPLICATION_PROPERTY}, {@link #SYSTEM_PROPERTY}, 
     * {@link #COMMAND_LINE} and {@link #INTERNAL_CONFIGURATION}. 
     */
    public static final int ALL_SOURCES = 
        DEFAULT_VALUE | APPLICATION_PROPERTY | SYSTEM_PROPERTY | 
        COMMAND_LINE | INTERNAL_CONFIGURATION; 

    /** A bitmask controlling the permissible sources for this option's value 
     * - valid mask bits are: {@link DEFAULT_VALUE}, 
     * {@link APPLICATION_PROPERTY}, {@link SYSTEM_PROPERTY}, 
     * {@link COMMAND_LINE} and {@link INTERNAL_CONFIGURATION}. For 
     * convenience, all bits can be set using {@link ALL_SOURCES}. Bits set 
     * to <tt>true</tt> indicate that the corresponding source is valid. 
     */ 
    final int validValueSources; 

    /** The actual source of this option's value - valid values are 
     * {@link NO_SOURCE}, {@link DEFAULT_VALUE}, {@link APPLICATION_PROPERTY}, 
     * {@link SYSTEM_PROPERTY}, {@link COMMAND_LINE} or 
     * {@link INTERNAL_CONFIGURATION}. 
     */
    private int actualValueSource = NO_SOURCE; 

    /** The long name for this option on the command line. */ 
    final String commandLineName; 

    /** The single-character name for this option on the command line. */ 
    final char commandLineChar; 

    /** The name of this option when given as a Java property. */ 
    final String propertyName; 

    /** A property file providing this option's value. */ 
    private File propertyFile = null; 

    /** The default value of this option. */ 
    final Object defaultValue; 

    /** The actual value of this option. */ 
    private Object actualValue = null; 

    /** The help text describing this option. */
    final String helpText; 

    /** Constructs a new ConfigurationOption. All parameters are immutable 
     * after construction, except for the <tt>actualValueSource</tt>, 
     * <tt>propertyFile</tt> and <tt>actualValue</tt>, which may be 
     * changed at any time. 
     *
     * @param optionName the option name. 
     * @param optionType the option type - see {@link optionType}. 
     * @param validValueSources a bitmask controlling the permissible 
     *                           sources for this option's value - see 
     *                           {@link validValueSources}. 
     * @param actualValueSource the actual source of this option's value 
     *                           - see {@link actualValueSource}. 
     * @param commandLineName the long name for this option on the 
     *                        command line. 
     * @param commandLineChar the single-character name for this option 
     *                        on the command line. 
     * @param propertyName the name of this option when given as a Java 
     *                     property. 
     * @param propertyFile a property file providing this option's value. 
     * @param defaultValue the default value of this option. 
     * @param actualValue the actual value of this option. 
     * @param helpText the help text describing this option. 
     * 
     * @throws NullPointerException if optionName, commandLineName, 
     *                              propertyName, or helpText is null. 
     * @throws IllegalArgumentException if: 
     *   <ul> 
     *     <li>optionName is the empty string.</li> 
     *     <li>optionType is not valid (see {@link #optionType}).</li> 
     *     <li>validValueSources is not valid (see 
     *         {@link #validValueSources}).</li> 
     *     <li>validValueSources includes {@link #COMMAND_LINE}, but 
     *         neither commandLineName nor commandLineChar is valid.</li> 
     *     <li>validValueSources includes {@link #APPLICATION_PROPERTY} or 
     *         {@link #SYSTEM_PROPERTY}, but propertyName is empty.</li> 
     *     <li>validValueSources includes {@link #DEFAULT_VALUE}, but 
     *         defaultValue is null.</li> 
     *     <li>actualValueSource is {@link #DEFAULT_VALUE}, but 
     *         defaultValue and actualValue do not compare equal.</li> 
     *     <li>any of 
     *         {@link #setActualValue(Object, int, File) setActualValue()}'s 
     *         error checks are violated.</li> 
     *   </ul> 
     */ 
    ConfigurationOption(final String optionName,
                        final int optionType,
                        final int validValueSources, 
                        final int actualValueSource, 
                        final String commandLineName, 
                        final char commandLineChar, 
                        final String propertyName, 
                        final File propertyFile, 
                        final Object defaultValue, 
                        final Object actualValue, 
                        final String helpText)
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::ConfigurationOption()"; 

        logger.trace(methodLogName); 

        if((optionName == null) || (commandLineName == null)
           || (propertyName == null) || (helpText == null)) 
        {
            final String message = 
                methodLogName + " - optionName [" + optionName 
                + "], commandLineName [" + commandLineName 
                + "], propertyName [" + propertyName 
                + "], actualValue [" + actualValue 
                + "] and helpText [" + helpText + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(optionName.equals("")) 
        {
            final String message = 
                methodLogName + " - optionName must not be empty."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if((optionType == NO_TYPE) || (!isDefinedOptionType(optionType))) 
        {
            final String message = 
                methodLogName + " - optionType [" 
                + getTypeString(optionType) + "] is invalid."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if((validValueSources == NO_SOURCE) 
           || (!allValueSourcesDefined(validValueSources))) 
        {
            final String message = 
                methodLogName + " - validValueSources [" 
                + getValueSourceString(validValueSources) + "] is invalid."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(isValidValueSource(COMMAND_LINE, validValueSources) 
           && commandLineName.equals("") 
           && (commandLineChar == 0)) 
        {
            final String message = 
                methodLogName + " - validValueSources [" 
                + getValueSourceString(validValueSources) + "] includes " 
                + getValueSourceString(COMMAND_LINE) 
                + ", but neither commandLineName [" + commandLineName 
                + "] nor commandLineChar [" + commandLineChar + "] is valid."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(isValidValueSource(APPLICATION_PROPERTY, validValueSources) 
           && isValidValueSource(SYSTEM_PROPERTY, validValueSources) 
           && propertyName.equals("")) 
        {
            final String message = 
                methodLogName + " - validValueSources [" 
                + getValueSourceString(validValueSources) + "] includes " 
                + getValueSourceString(APPLICATION_PROPERTY) + " or " 
                + getValueSourceString(SYSTEM_PROPERTY) 
                + ", but propertyName [" + propertyName + "] is empty."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(isValidValueSource(DEFAULT_VALUE, validValueSources) 
           && (defaultValue == null)) 
        {
            final String message = 
                methodLogName + " - validValueSources [" 
                + getValueSourceString(validValueSources) + "] includes " 
                + getValueSourceString(DEFAULT_VALUE) 
                + ", but defaultValue [" + defaultValue + "] is null."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if((actualValueSource == DEFAULT_VALUE) 
           && (defaultValue != null) 
           && !defaultValue.equals(actualValue)) 
        {
            final String message = 
                methodLogName + " - actualValueSource is " 
                + getValueSourceString(DEFAULT_VALUE) 
                + ", but actualValue [" + actualValue 
                + "] and defaultValue [" + defaultValue + "] are not equal."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        this.optionName = optionName; 
        this.optionType = optionType; 
        this.validValueSources = validValueSources; 
        this.commandLineName = commandLineName; 
        this.commandLineChar = commandLineChar; 
        this.propertyName = propertyName; 
        this.defaultValue = defaultValue; 
        this.helpText = helpText; 

        // Delegate this part (to reuse setActualValue()'s error checking). 
        setActualValue(actualValue, actualValueSource, propertyFile); 
    }

    /** Returns the source of this ConfigurationOption's current value. */ 
    int getActualValueSource()
    {
        final String methodLogName = 
            classLogName + "::getActualValueSource()"; 

        logger.trace(methodLogName); 

        return(actualValueSource); 
    }

    /** Returns the current value of this ConfigurationOption. */
    Object getActualValue()
    {
        final String methodLogName = classLogName + "::getActualValue()"; 

        logger.trace(methodLogName); 

        // This is safe - all Objects accepted by setValue() are immutable. 
        return(actualValue); 
    }

    /** Returns the property file this ConfigurationOption's current value 
     * came from, or null if the value did not come from a property file. 
     */ 
    File getPropertyFile()
    {
        final String methodLogName = classLogName + "::getPropertyFile()"; 

        logger.trace(methodLogName); 

        // This is safe - File objects are immutable. 
        return(propertyFile); 
    }

    /** Sets the current value of this ConfigurationOption to optionValue. 
     * The value source is updated to match valueSource, and the property 
     * file is updated to match propertyFile. 
     * 
     * @param optionValue the new value. 
     * @param valueSource the source of the new value. 
     * @param propertyFile the property file this value came from. 
     * 
     * @throws IllegalArgumentException if: 
     *   <ul> 
     *     <li>if optionValue has the wrong data type.</li>
     *     <li>if valueSource is not a single, defined value source.</li> 
     *     <li>if this option cannot take values from valueSource.</li> 
     *     <li>if valueSource cannot override the source of this option's 
     *         current value.</li>
     *     <li>if valueSource is {@link #NO_SOURCE} and optionValue is 
     *         not null.<li> 
     *     <li>if valueSource is not {@link #NO_SOURCE} and optionValue 
     *         is null.<li> 
     *     <li>if valueSource is {@link #APPLICATION_PROPERTY} and 
     *         propertyFile is null.</li>
     *     <li>if valueSource is not {@link #APPLICATION_PROPERTY} and 
     *         propertyFile is not null.</li>
     *   </ul>
     */ 
    void setActualValue(final Object optionValue, 
                        final int valueSource, 
                        final File propertyFile) 
        throws IllegalArgumentException 
    {
        final String methodLogName = classLogName + "::setActualValue()"; 

        logger.trace(methodLogName); 

        // Ensure that a single, defined source was given for the new value. 
        if(!isSingleDefinedValueSource(valueSource)) 
        {
            final String message = 
                methodLogName + " - valueSource ["
                + getValueSourceString(valueSource) 
                + "] must specify a single, defined source."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Ensure that this option is allowed to take a value from the 
        // specified source. 
        if(!isValidValueSource(valueSource, validValueSources)) 
        {
            final String message = 
                methodLogName + " - the specified valueSource ["
                + getValueSourceString(valueSource) 
                + "] is not a valid value source [" 
                + getValueSourceString(validValueSources) 
                + "] for option [" + optionName + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Ensure that the specified valueSource can override that of the 
        // option's current value. 
        if(!canOverrideValueSource(valueSource, actualValueSource)) 
        {
            final String message = 
                methodLogName + " - option [" + optionName + "] has already" 
                + " been set with a value from a higher priority source ["
                + getValueSourceString(actualValueSource) 
                + "] than the specified valueSource [" 
                + getValueSourceString(valueSource) + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Ensure that the optionValue is only null when the value source 
        // is NO_SOURCE. 
        if(!((valueSource == NO_SOURCE) ^ (optionValue != null))) 
        {
            final String message = 
                methodLogName + " - valueSource [" 
                + getValueSourceString(valueSource) 
                + "] requires optionValue to be" 
                + (valueSource == NO_SOURCE ? "null." : "non-null."); 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
        }

        // Ensure that the propertyFile is only given when the value 
        // source is APPLICATION_PROPERTY. 
        if(!((valueSource == APPLICATION_PROPERTY) ^ (propertyFile == null))) 
        {
            final String message = 
                methodLogName + " - valueSource [" 
                + getValueSourceString(valueSource) 
                + "] requires propertyFile to be" 
                + (valueSource == APPLICATION_PROPERTY ? 
                   "non-null." : "null."); 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
        }

        // If a value source was given - implying that the value itself 
        // is meaningful - ensure that the value data type is correct. 
        if(valueSource != NO_SOURCE) 
        {
            boolean correctValueDataType = false; 

            switch(optionType) 
            {
              case STRING: 
                correctValueDataType = (optionValue instanceof String); 
                break; 

              case BOOLEAN: 
                correctValueDataType = (optionValue instanceof Boolean); 
                break; 

              case INTEGER: 
                correctValueDataType = (optionValue instanceof Integer); 
                break; 

              default: 
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - invalid optionType [" 
                    + getTypeString(optionType) 
                    + "] encountered whilst processing configuration option [" 
                    + optionName + "]."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            if(!correctValueDataType) 
            {
                final String message = 
                    methodLogName + " - optionValue [" 
                    + optionValue.getClass().getName() 
                    + "][" + getTypeString(optionValue) 
                    + "] does not match the value data type ["
                    + getTypeString(optionType) 
                    + "] of option [" + optionName + "]."; 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
            }
        }

        // All checks have passed - set the option's new value, value 
        // source and property file. 
        this.actualValue = optionValue; 
        this.actualValueSource = valueSource; 
        this.propertyFile = propertyFile; 

        logger.debug(methodLogName + " - set value [" + optionValue 
                     + "][" + getValueSourceString(valueSource) + "]" 
                     + (valueSource == APPLICATION_PROPERTY ?
                        "[property file: " + propertyFile + "]" : "") 
                     + " for option [" + optionName + "]."); 
    }

    /** Returns true if the optionType is defined. */ 
    public static boolean isDefinedOptionType(final int optionType) 
    {
        final String methodLogName = classLogName + "::isDefinedOptionType()"; 

        logger.trace(methodLogName); 

        return((optionType == NO_TYPE) 
               || (optionType == STRING) 
               || (optionType == BOOLEAN) 
               || (optionType == INTEGER)); 
    }

    /** Returns a String naming the optionType. */ 
    public static String getTypeString(final int optionType) 
    {
        final String methodLogName = classLogName + "::getTypeString(int)"; 

        logger.trace(methodLogName); 

        String result = null; 

        switch(optionType)
        {
          case NO_TYPE:
            result = "NO_TYPE";
            break;

          case STRING:
            result = "STRING";
            break;

          case BOOLEAN:
            result = "BOOLEAN";
            break;

          case INTEGER:
            result = "INTEGER";
            break;

          default:
            result = "UNKNOWN(" + optionType + ")"; 
            break;
        }

        return(result); 
    }

    /** Returns a String naming the option type which the candidate object 
     * could be used for. 
     */ 
    public static String getTypeString(final Object candidate) 
    {
        final String methodLogName = classLogName + "::getTypeString(Object)"; 

        logger.trace(methodLogName); 

        final Class type = (candidate == null ? null : candidate.getClass()); 
        String result = null; 

        if(type == null) 
        {
            result = "NO_TYPE"; 
        }
        else if(type.equals(String.class)) 
        {
            result = "STRING"; 
        }
        else if(type.equals(Boolean.class)) 
        {
            result = "BOOLEAN"; 
        }
        else if(type.equals(Integer.class)) 
        {
            result = "INTEGER"; 
        }
        else 
        {
            result = "UNKNOWN(" + type.getName() + ")"; 
        }

        return(result); 
    }

    /** Returns true if sourceBitmask specifies a single value source defined 
     * by this class. 
     */ 
    public static boolean isSingleDefinedValueSource(final int sourceBitmask) 
    {
        final String methodLogName = 
            classLogName + "::isSingleDefinedValueSource()"; 

        logger.trace(methodLogName); 

        return((sourceBitmask == NO_SOURCE) 
               || (sourceBitmask == DEFAULT_VALUE) 
               || (sourceBitmask == APPLICATION_PROPERTY) 
               || (sourceBitmask == SYSTEM_PROPERTY) 
               || (sourceBitmask == COMMAND_LINE) 
               || (sourceBitmask == INTERNAL_CONFIGURATION)); 
    }

    /** Returns true if only value source bits defined by this class are 
     * set within the sourceBitmask. 
     */ 
    public static boolean allValueSourcesDefined(final int sourceBitmask) 
    {
        final String methodLogName = 
            classLogName + "::allValueSourcesDefined()"; 

        logger.trace(methodLogName); 

        return((sourceBitmask & ~ALL_SOURCES) == 0); 
    }

    /** Returns a String naming the value sources set by sourceBitmask. */ 
    public static String getValueSourceString(final int sourceBitmask) 
    {
        final String methodLogName = 
            classLogName + "::getValueSourceString()"; 

        logger.trace(methodLogName); 

        String result = ""; 
        
        if(sourceBitmask == NO_SOURCE) 
        {
            result = "NO_SOURCE"; 
        }
        else 
        { 
            boolean useSeparator = false; 

            if((sourceBitmask & DEFAULT_VALUE) == DEFAULT_VALUE) 
            {
                result = "DEFAULT_VALUE"; 
                useSeparator = true; 
            }

            if((sourceBitmask & APPLICATION_PROPERTY) == APPLICATION_PROPERTY) 
            {
                result += (useSeparator ? "|" : "") + "APPLICATION_PROPERTY"; 
                useSeparator = true; 
            }

            if((sourceBitmask & SYSTEM_PROPERTY) == SYSTEM_PROPERTY) 
            {
                result += (useSeparator ? "|" : "") + "SYSTEM_PROPERTY"; 
                useSeparator = true; 
            }

            if((sourceBitmask & COMMAND_LINE) == COMMAND_LINE) 
            {
                result += (useSeparator ? "|" : "") + "COMMAND_LINE"; 
                useSeparator = true; 
            }

            if((sourceBitmask & INTERNAL_CONFIGURATION) 
               == INTERNAL_CONFIGURATION) 
            {
                result += (useSeparator ? "|" : "") + "INTERNAL_CONFIGURATION";
                useSeparator = true; 
            }

            if(!allValueSourcesDefined(sourceBitmask))
            {
                result += 
                    (useSeparator ? "|" : "") 
                    + "UNKNOWN(" + (sourceBitmask & ~ALL_SOURCES) + ")"; 
            }
            else if(sourceBitmask == ALL_SOURCES)
            {
                result += " -> ALL_SOURCES"; 
            }
        }

        return(result); 
    }

    /** Returns true if the valueSource is valid for a configuration option 
     * with the specified validValueSources. 
     * 
     * @throws IllegalArgumentException if valueSource or validValueSources 
     *                         specifies a source not defined by this class, 
     *                         or valueSource specifies more than one source. 
     */ 
    public static boolean isValidValueSource(final int valueSource, 
                                             final int validValueSources) 
    {
        final String methodLogName = classLogName + "::isValidValueSource()"; 

        logger.trace(methodLogName); 

        if(!allValueSourcesDefined(valueSource | validValueSources)) 
        {
            final String message = 
                methodLogName + " - valueSource [" 
                + getValueSourceString(valueSource) 
                + "] and validValueSources [" 
                + getValueSourceString(validValueSources) 
                + "] must specify defined value sources."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(!isSingleDefinedValueSource(valueSource)) 
        {
            final String message = 
                methodLogName + " - valueSource ["
                + getValueSourceString(valueSource) 
                + "] may only specify one source."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        return((valueSource == NO_SOURCE)
               || ((valueSource & validValueSources) != 0)); 
    }

    /** Returns true if a value from newValueSource can override a value 
     * from oldValueSource. 
     * 
     * @throws IllegalArgumentException if either value source is not 
     *                         defined by this class, or either value 
     *                         source specifies more than one source. 
     */ 
    public static boolean canOverrideValueSource(final int newValueSource, 
                                                 final int oldValueSource) 
    {
        final String methodLogName = 
            classLogName + "::canOverrideValueSource()"; 

        logger.trace(methodLogName); 

        if(!allValueSourcesDefined(newValueSource | oldValueSource)) 
        {
            final String message = 
                methodLogName + " - newValueSource [" 
                + getValueSourceString(newValueSource) 
                + "] and oldValueSource [" 
                + getValueSourceString(oldValueSource) 
                + "] must specify defined value sources."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(!isSingleDefinedValueSource(newValueSource) 
           || !isSingleDefinedValueSource(oldValueSource)) 
        {
            final String message = 
                methodLogName + " - newValueSource [" 
                + getValueSourceString(newValueSource) 
                + "] and oldValueSource [" 
                + getValueSourceString(oldValueSource) 
                + "] may only specify one source each."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        return(newValueSource >= oldValueSource); 
    }
}
