package org.eso.util.dal;

import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.sql.Timestamp;

public interface ArchiveStatisticsDAO {
	
	/**
	 * This method returns a list of files that did not reach the DataTransfer machine in 
	 * Chile from the OLAS machine, ordered from oldest to newest
	 * 
	 * @param days consider only files not older than this number of days
	 * @return a list of maps containing information about the files
     * @throws DAOException
	 */
	List<Map<String, String>>	fromOLASToDT(int days) throws DAOException;

	/**
	 * This method returns a list of files that did not reach the staging area in 
	 * Garching from the DataTransfer machine, ordered from oldest to newest
	 * 
	 * @param days consider only files not older than this number of days
	 * @return a list of maps containing information about the files 
	 * @throws DAOException
	 */
	List<Map<String, String>>	fromDTToPAStage(int days) throws DAOException;

	/**
	 * This method returns a list of files that did not reach the Primary Archive in 
	 * Garching from the DataTransfer staging machine, ordered from oldest to newest
	 * 
	 * @param days consider only files not older than this number of days
	 * @return a list of maps containing information about the files 
	 * @throws DAOException
	 */
	List<Map<String, String>>	fromPAStageToPA(int days) throws DAOException;
	
	/**
	 * This method returns a list of files that did not reach the Secondary Archive 
	 * from the Primary Archive, ordered from oldest to newest
	 * 
	 * @param days consider only files not older than this number of days
	 * @return a list of maps containing information about the files
	 * @throws DAOException
	 */
	List<Map<String, String>>	fromPAToSA(int days) throws DAOException;
	
	/**
	 * This method returns a list of files whose transmission has been completed in  
	 * the last days, ordered by submission time
	 * 
	 * @param days consider only files not older than this number of days
	 * @return a list of maps containing information about the files
	 * @throws DAOException
	 */
	List<Map<String, String>>	transmissionTime(Timestamp start, Timestamp stop) throws DAOException;
	
	/**
	 * This method returns the bandwidth usage of the DataTransfer over the last hours
	 *  
	 * @param interval the period (in hours) to be evaluated
	 * @param step the duration (in seconds) of the time intervals over which the bandwidth is calculated 
	 * @return an ordered Map of time intervals and bandwidth usage
	 * @throws DAOException
	 */
	SortedMap<Timestamp, Double> bandwidthUsage(Timestamp start, Timestamp stop, int step) throws DAOException;
	
	/**
	 * Search for a file in the secondary archive
	 * @param fileId
	 * @return true if found
	 * @throws DAOException
	 */
	public boolean searchInSA(String fileId) throws DAOException;
	
	/**
	 * Search for a file in the primary archive
	 * @param fileId 
	 * @return true if found
	 * @throws DAOException
	 */
	public boolean searchInPA(String fileId) throws DAOException;
	
	/**
	 * Search for a file in Garching staging area
	 * @param fileId
	 * @return true if the file reached the staging area (it doesn't mean it's still there)
	 * @throws DAOException
	 */
	public boolean searchInStagingGarching(String fileId) throws DAOException;
	
	/**
	 * Search for a file in Paranal/LaSilla staging area
	 * @param fileId
	 * @return true if the file reached the staging area (it doesn't mean it's still there)
	 * @throws DAOException
	 */
	public boolean searchInStagingChile(String fileId) throws DAOException;
	
	/**
	 * Search for a file in the OLAS machine
	 * @param fileId
	 * @return true if the file reached the OLAS machine (it doesn't mean it's still there)
	 * @throws DAOException
	 */
	public boolean searchInOLAS(String fileId) throws DAOException;
	
	/**
	 * Search for a file in USB disks
	 * @param fileId
	 * @return a String containing the USB disks on which the file is stored
	 * @throws DAOException
	 */
	public String searchInUSBDisks(String fileId) throws DAOException;
		
	/**
	 * Converts the origname generated on the instrument workstation in a file_id/dp_id generated by OLAS
	 * @param origname the filename generated on the instrument workstation
	 * @return the filename generated by OLAS
	 * @throws DAOException
	 */
	public String  orignameToFileId(String origname) throws DAOException;
}
