package org.eso.util.dal;

import java.util.Date;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * This class represents a file transfer request, ie a set of metadata
 * associated to a file to transfer.
 * 
 * @author cmoins
 * 
 */
public class TransferRequest implements Cloneable {

    /** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(TransferRequest.class);

    /* available state for a transfer request */
    public enum Status {
        INCOMING, PROCESSING, PROCESSED, REJECTED, SUBMITTED, TRANSFERRED, ARCHIVED, UNKNOWN, ALL
    };

    /* available fields in a transfer request */
    public enum FIELDS {
        REQUEST_ID, FILE_ID, FILE_NAME, UNCOMPRESSED_FILE_NAME, TRANSFER_METHOD, TRANSFER_PRIORITY, COMPRESSION_METHOD, SIZE, UNCOMPRESSED_SIZE, CHECKSUM, TRANSFER_CATG, USER, DEST_DIR, MULTI_FILES_COMPONENT, SUBMISSION_TIME, CREATION_TIME, STATUS,LAST_UPDATE_TIME
    };

    /* Available categories of file to be transfered */
    public static final String RAW_FILE_CATG = "rawfile";
    public static final String TOO_CATG = "too";
    public static final String MULTI_FILE_CATG = "multifile";
    public static final String HEADER_FILE_CATG = "header";

    /* Available transfer methods */
    public static final String TRANSFER_METHOD_DISK = "disk";
    public static final String TRANSFER_METHOD_NETWORK = "network";

    /* Available compression methods */
    public static final String COMPRESSION_METHOD_NONE = "none";
    public static final String COMPRESSION_METHOD_UNIXCOMPRESS = "unixcompress";

    /** transfer request identifier */
    private String requestId = null;

    /** file identifier */
    private String fileId = null;
    /** file name */
    private String fileName = null;
    /** name of the uncompressed file */
    private String uncompressedFileName = null;
    /** file transfer method */
    private String transferMethod = null;
    /** transfer priority */
    private Integer transferPriority = null;
    /** compression method */
    private String compressionMethod = null;
    /** size of the file */
    private Long size = null;
    /** size of the uncompressed file */
    private Long uncompressedSize = null;
    /** file checksum */
    private String checksum = null;
    /** request category : raw files, too, package */
    private String transferCatg = null;
    /** user account for too files */
    private String user = null;
    /** destination directory for too files */
    private String destDir = null;
    /** components of the file if it is a package */
    private String multiFileComponents = null;
    /**
     * creation date expressed in millisecond since January 1, 1970, 00:00:00
     * UTC
     */
    private Long creationTime = null;

    /**
     * last update expressed in millisecond since January 1, 1970, 00:00:00 UTC
     */
    private Long lastUpdateTime = null;

    /** status of the transfer request */
    private Status status = null;

    /** returns the list of all available categories separated by a blank */
    public static String getAllCatg() {
        String trCatg =
                RAW_FILE_CATG + " " + TOO_CATG + " " + MULTI_FILE_CATG + " "
                        + HEADER_FILE_CATG;
        return trCatg;
    }

    /* Getters and setters for data members */
    public String getRequestId() {
        return requestId;
    }

    public void setRequestId(String requestId) {
        this.requestId = requestId;
    }

    public String getFileId() {
        return fileId;
    }

    public void setFileId(String fileId) {
        this.fileId = fileId;
    }

    public String getFileName() {
        return fileName;
    }

    public void setFileName(String fileName) {
        this.fileName = fileName;
    }

    public String getUncompressedFileName() {
        return uncompressedFileName;
    }

    public void setUncompressedFileName(String uncompressedFileName) {
        this.uncompressedFileName = uncompressedFileName;
    }

    public String getTransferMethod() {
        return transferMethod;
    }

    public void setTransferMethod(String transferMethod) {
        this.transferMethod = transferMethod;
    }

    public Integer getTransferPriority() {
        return transferPriority;
    }

    public void setTransferPriority(Integer transferPriority) {
        this.transferPriority = transferPriority;
    }

    public String getCompressionMethod() {
        return compressionMethod;
    }

    public void setCompressionMethod(String compressionMethod) {
        this.compressionMethod = compressionMethod;
    }

    public Long getSize() {
        return size;
    }

    public void setSize(Long size) {
        this.size = size;
    }

    public Long getUncompressedSize() {
        return uncompressedSize;
    }

    public void setUncompressedSize(Long uncompressedSize) {
        this.uncompressedSize = uncompressedSize;
    }

    public String getChecksum() {
        return checksum;
    }

    public void setChecksum(String checksum) {
        this.checksum = checksum;
    }

    public String getTransferCatg() {
        return transferCatg;
    }

    public void setTransferCatg(String transferCatg) {
        this.transferCatg = transferCatg;
    }

    public String getUser() {
        return user;
    }

    public void setUser(String user) {
        this.user = user;
    }

    public String getDestDir() {
        return destDir;
    }

    public void setDestDir(String destDir) {
        this.destDir = destDir;
    }

    public String getMultiFileComponents() {
        return multiFileComponents;
    }

    public void setMultiFileComponents(String multiFileComponents) {
        this.multiFileComponents = multiFileComponents;
    }

    
    public Status getStatus() {
        return (this.status);
    }

    
    public void setStatus(Status status) {
        this.status = status;
    }

    /**
     * return creation time expressed in millisecond since January 1, 1970,
     * 00:00:00 UTC
     */
    public Long getCreationTime() {
        return creationTime;
    }

    /**
     * set the creation time
     * 
     * @param creationTime
     *            creation time expressed in millisecond since January 1, 1970,
     *            00:00:00 UTC
     * 
     */
    public void setCreationTime(Long creationTime) {
        this.creationTime = creationTime;
    }
    
    public Long getLastUpdateTime() {
        return lastUpdateTime;
    }

    public void setLastUpdateTime(Long lastUpdateTime) {
        this.lastUpdateTime = lastUpdateTime;
    }

    public String toString() {
        String res = "Transfer request instance: ";
        res += "\n\tfileId " + fileId;
        res += "\n\tfileName " + fileName;
        res += "\n\tuncompressedFileName " + uncompressedFileName;
        res += "\n\ttransferMethod " + transferMethod;
        res += "\n\ttransferPriority " + transferPriority;
        res += "\n\tcompressionMethod " + compressionMethod;
        res += "\n\tsize " + size;
        res += "\n\tuncompressedSize " + uncompressedSize;
        res += "\n\tchecksum " + checksum;
        res += "\n\ttransferCatg " + transferCatg;
        res += "\n\tuser " + user;
        res += "\n\tdestDir " + destDir;
        res += "\n\tmultiFileComponents " + multiFileComponents;
        res += "\n\tcreation  time " + creationTime;
        return res;
    }

    /**
     * Performs a field to field copy of the transfer request
     */
    public TransferRequest clone() {
        TransferRequest trCopy = new TransferRequest();

        if (this.requestId != null)
            trCopy.requestId = this.requestId;
        if (this.fileId != null)
            trCopy.fileId = this.fileId;
        if (this.fileName != null)
            trCopy.fileName = this.fileName;
        if (this.uncompressedFileName != null)
            trCopy.uncompressedFileName = this.uncompressedFileName;
        if (this.transferMethod != null)
            trCopy.transferMethod = this.transferMethod;
        if (this.transferPriority != null)
            trCopy.transferPriority = new Integer(transferPriority);
        if (this.compressionMethod != null)
            trCopy.compressionMethod = trCopy.compressionMethod;
        if (this.size != null)
            trCopy.size = new Long(this.size);
        if (this.uncompressedSize != null)
            trCopy.uncompressedSize = new Long(uncompressedSize);
        if (this.checksum != null)
            trCopy.checksum = this.checksum;
        if (this.transferCatg != null)
            trCopy.transferCatg = this.transferCatg;
        if (this.user != null)
            trCopy.user = this.user;
        if (this.destDir != null)
            trCopy.destDir = this.destDir;
        if (this.multiFileComponents != null)
            trCopy.multiFileComponents = this.multiFileComponents;
        if (this.status != null)
            trCopy.status = this.status;
        if (this.creationTime != null)
            trCopy.creationTime = new Long(creationTime);

        return trCopy;
    }

    public static void main(String[] args) {
        // Basic Log4j configuration
        BasicConfigurator.configure();
        TransferRequestValidator trValidator =
                new DefaultTransferRequestValidator();

        logger.info("Set of tests on the TransferRequest class");
        TransferRequest transferRequest = new TransferRequest();
        // Case 1 : No transfer request validator
        logger.info("Case 1 : Is transfer request without validator valid : "
                + trValidator.isValid(transferRequest));
        // Case 2 : transfer request with default validator but no field set
        logger.info("Case 2 : Is transfer request with validator but no field set valid : "
                + trValidator.isValid(transferRequest));
        // Case 3 : transfer request with default validator and required field
        // set
        transferRequest.setFileId("FILEID");
        transferRequest.setFileName("FILENAME");
        transferRequest.setSize(new Long(1000));
        logger.info("Case 3 : Is transfer request with validator and required field set valid : "
                + trValidator.isValid(transferRequest));

        logger.info("Transfer Request fileId " + transferRequest.getFileId()
                + ",filename " + transferRequest.getFileName() + ",size "
                + transferRequest.getSize());

        // Case 4 : test of the transfer request clone method
        TransferRequest tr = new TransferRequest();
        tr.setFileId("FILEID");
        tr.setFileName("FILENAME");
        tr.setSize(new Long(1000));

        TransferRequest trCopy = tr.clone();
        if ((trCopy.getFileId().compareTo(tr.getFileId()) == 0)
                && (trCopy.getFileName().compareTo(tr.getFileName()) == 0)
                && (trCopy.getSize().compareTo(tr.getSize()) == 0)) {
            logger.info("SUCCESSFUL Test clone ");
        } else {
            logger.error("FAILED testClone ");
        }
    }
}
