package org.eso.util.dal;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.eso.util.dal.TransferRequest.Status;

/**
 * Implementation of a transfer request dao in the case where the transfer
 * request repository is a sybase database
 * 
 * @author cmoins
 * 
 */
public class TransferRequestDAOSyb implements TransferRequestDAO {

    /** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(TransferRequestDAOSyb.class);
    /** The name of this class for logging purposes. */
    private static final String classLogName = "TransferRequestDAOSyb";

    private static final String REQUEST_ID_FIELD = "request_id";
    private static final String FILE_ID_FIELD = "file_id";
    private static final String FILE_NAME_FIELD = "file_name";
    private static final String SIZE_FIELD = "size";
    private static final String UNCOMPRESSED_SIZE_FIELD = "uncompressed_size";
    private static final String CHECKSUM_FIELD = "checksum";
    private static final String COMPRESSION_METHOD_FIELD = "compression_method";
    private static final String TRANSFER_METHOD_FIELD = "transfer_method";
    private static final String TRANSFER_PRIORITY_FIELD = "transfer_priority";
    private static final String TRANSFER_CATG_FIELD = "transfer_catg";
    private static final String DEST_USER_FIELD = "dest_user";
    private static final String DEST_DIR_FIELD = "dest_dir";
    private static final String TRANSFER_STATUS_FIELD = "transfer_status";
    private static final String CREATION_TIME_FIELD = "submission_time";
    private static final String LAST_UPDATE_TIME_FIELD = "last_update_time";
    /** Date format for date parameter in the database */
    private static final String DATE_FORMAT = "MMM dd yyyy HH:mm:ss.SSS";

    /** mapping between transfer requests fields and database table ones */
    private static final Map<TransferRequest.FIELDS, String> TR_TABLE_FIELD_MAP =
            new HashMap<TransferRequest.FIELDS, String>();
    static {
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.CHECKSUM, CHECKSUM_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.COMPRESSION_METHOD,
                COMPRESSION_METHOD_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.SUBMISSION_TIME,
                CREATION_TIME_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.DEST_DIR, DEST_DIR_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.FILE_ID, FILE_ID_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.FILE_NAME,
                FILE_NAME_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.REQUEST_ID,
                REQUEST_ID_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.SIZE, SIZE_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.STATUS,
                TRANSFER_STATUS_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.TRANSFER_CATG,
                TRANSFER_CATG_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.TRANSFER_METHOD,
                TRANSFER_METHOD_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.TRANSFER_PRIORITY,
                TRANSFER_PRIORITY_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.UNCOMPRESSED_SIZE,
                UNCOMPRESSED_SIZE_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.USER, DEST_USER_FIELD);
        TR_TABLE_FIELD_MAP.put(TransferRequest.FIELDS.LAST_UPDATE_TIME,
                LAST_UPDATE_TIME_FIELD);
    }

    // mapping between DAO and SQL sorting order keyword
    private static final Map<SortingOrder, String> SORTING_ORDER_MAP =
            new HashMap<SortingOrder, String>();
    static {
        SORTING_ORDER_MAP.put(SortingOrder.ASCENDING, "ASC");
        SORTING_ORDER_MAP.put(SortingOrder.DESCENDING, "DESC");
    }

    private final String dbUrl;
    private final String dbUsername;
    private final String dbPassword;
    private final String dbTable;
    private final TransferRequestValidator transferRequestValidator;

    public TransferRequestDAOSyb(String dbUrl, String dbUsername,
            String dbPassword, String dbTable,
            TransferRequestValidator transferRequestValidator) {
        if ((dbUrl == null) || (dbUsername == null) || (dbPassword == null)
                || (dbTable == null) || (transferRequestValidator == null)) {
            String message =
                    "Missing parameter to instantiate a TransferRequestDAOSyb object";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        this.dbUrl = dbUrl;
        this.dbUsername = dbUsername;
        this.dbPassword = dbPassword;
        this.dbTable = dbTable;
        this.transferRequestValidator = transferRequestValidator;
    }

    /**
     * Deletes all the transfer requests having the same request id field
     * 
     * @param transferRequest
     *            transfer request to be deleted
     * @return the number of deleted rows if succeeded, -1 otherwise
     * @throws DAOException
     */
    public int delete(TransferRequest transferRequest) throws DAOException {
        String methodLogName = classLogName + "::delete()";
        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing transfer request argument");
        }
        if (transferRequest.getRequestId() == null) {
            String message =
                    methodLogName + " - Mandatory field " + REQUEST_ID_FIELD
                            + " is not available";
            logger.error(message);
            throw new DAOException(message);
        }
        String sqlStatement =
                "delete from " + dbTable + " where " + REQUEST_ID_FIELD + "='"
                        + transferRequest.getRequestId() + "'";
        logger.debug(methodLogName + " - Execution of query " + sqlStatement);
        int nbRows = executeUpdateQuery(sqlStatement);
        if (nbRows < 0) {
            logger.error(methodLogName + " - " + sqlStatement + " failed");
        }
        return nbRows;
    }

    /**
     * find all the transfer request being in the specified status
     * 
     * @param status
     *            status of the transfer requests to be retrieved
     * @return a list of transfer request if successful, null otherwise
     * @throws DAOException
     */
    public List<TransferRequest> findByStatus(Status status)
            throws DAOException {
        String methodLogName = classLogName + "::findByStatus()";

        if (status == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing status argument");
        }
        List<TransferRequest> transferRequestList = null;
        final String sqlStatement =
                "select * from " + dbTable + " where " + TRANSFER_STATUS_FIELD
                        + "=" + String.valueOf(status.ordinal());
        logger.debug(methodLogName + " - Execution of query " + sqlStatement);
        transferRequestList = executeQuery(sqlStatement);

        return transferRequestList;
    }

    /**
     * checks if a transfer request is valid
     * 
     * @param transferRequest
     *            transfer request to be validated
     * @return true if the request is valid, false otherwise
     */
    public Boolean isValid(TransferRequest transferRequest) {
        String methodLogName = classLogName + "::isValid()";
        // Check input parameter
        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - transferRequest must not be null");
        }
        return transferRequestValidator.isValid(transferRequest);
    }

    /**
     * saves a transfer request in the database
     * 
     * @param transferRequest
     *            transfer request to be saved
     * @return the number of inserted rows if succeeded, -1 otherwise
     * @throws DAOException
     */
    public int save(TransferRequest transferRequest) throws DAOException {
        String methodLogName = classLogName + "::save()";

        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing transfer request argument");
        }
        if (transferRequest.getRequestId() == null) {
            String message =
                    methodLogName + " - Mandatory field " + REQUEST_ID_FIELD
                            + " is not available";
            logger.error(message);
            throw new DAOException(message);
        }
        if (transferRequest.getFileId() == null) {
            String message =
                    methodLogName + " - Mandatory field " + FILE_ID_FIELD
                            + " is not available";
            logger.error(message);
            throw new DAOException(message);
        }

        String columnList = REQUEST_ID_FIELD;
        String valueList = "'" + transferRequest.getRequestId() + "'";
        columnList += "," + FILE_ID_FIELD;
        valueList += ",'" + transferRequest.getFileId() + "'";

        if (transferRequest.getFileName() != null) {
            columnList += "," + FILE_NAME_FIELD;
            valueList += ",'" + transferRequest.getFileName() + "'";
        }
        if (transferRequest.getSize() != null) {
            columnList += "," + SIZE_FIELD;
            valueList += "," + transferRequest.getSize();
        }
        if (transferRequest.getUncompressedSize() != null) {
            columnList += "," + UNCOMPRESSED_SIZE_FIELD;
            valueList += "," + transferRequest.getUncompressedSize();
        }
        if (transferRequest.getChecksum() != null) {
            columnList += "," + CHECKSUM_FIELD;
            valueList += ",'" + transferRequest.getChecksum() + "'";
        }
        if (transferRequest.getCompressionMethod() != null) {
            columnList += "," + COMPRESSION_METHOD_FIELD;
            valueList += ",'" + transferRequest.getCompressionMethod() + "'";
        }
        if (transferRequest.getTransferMethod() != null) {
            columnList += "," + TRANSFER_METHOD_FIELD;
            valueList += ",'" + transferRequest.getTransferMethod() + "'";
        }
        if (transferRequest.getTransferPriority() != null) {
            columnList += "," + TRANSFER_PRIORITY_FIELD;
            valueList += "," + transferRequest.getTransferPriority();
        }
        if (transferRequest.getTransferCatg() != null) {
            columnList += "," + TRANSFER_CATG_FIELD;
            valueList += ",'" + transferRequest.getTransferCatg() + "'";
        }
        if (transferRequest.getUser() != null) {
            columnList += "," + DEST_USER_FIELD;
            valueList += ",'" + transferRequest.getUser() + "'";
        }
        if (transferRequest.getDestDir() != null) {
            columnList += "," + DEST_DIR_FIELD;
            valueList += ",'" + transferRequest.getDestDir() + "'";
        }
        if (transferRequest.getStatus() != null) {
            columnList += "," + TRANSFER_STATUS_FIELD;
            valueList += "," + transferRequest.getStatus().ordinal();
        }
        columnList += "," + CREATION_TIME_FIELD;
        if (transferRequest.getCreationTime() != null) {
            SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
            dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
            valueList +=
                    ",'"
                            + dateFormat.format(new Date(transferRequest.getCreationTime()))
                            + "'";
        } else {
            valueList += ",getUTCdate()";
        }

        columnList += "," + LAST_UPDATE_TIME_FIELD;
        valueList += ",getUTCdate()";

        String sqlStatement =
                "insert into " + dbTable + " (" + columnList + ") values("
                        + valueList + ")";

        logger.debug(methodLogName + " - Execution of query " + sqlStatement);
        int nbRows = executeUpdateQuery(sqlStatement);
        if (nbRows < 0) {
            logger.error(methodLogName + " - " + sqlStatement + " failed");
        }
        return nbRows;
    }

    /**
     * Update the status of a specified transfer request
     * 
     * @param transferRequest
     *            transfer request to be updated
     * @param status
     *            new status of the transfer request
     * @return the number of inserted rows if succeeded, -1 otherwise
     * @throws DAOException
     */
    public int updateStatus(TransferRequest transferRequest, Status status)
            throws DAOException {
        String methodLogName = classLogName + "::updateStatus()";
        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing transfer request argument");
        }
        if (status == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing status argument");
        }
        if (transferRequest.getRequestId() == null) {
            String message =
                    methodLogName + " - Mandatory field " + REQUEST_ID_FIELD
                            + " is not available";
            logger.error(message);
            throw new DAOException(message);
        }
        Integer intStatus = status.ordinal();
        // First updates the status then the date
        String sqlStatement =
                "update " + dbTable + " set transfer_status = " + intStatus
                        + ",last_update_time=getUTCdate() where request_id='"
                        + transferRequest.getRequestId() + "'";
        logger.debug(sqlStatement);
        int nbRows = executeUpdateQuery(sqlStatement);
        if (nbRows < 0) {
            logger.error(methodLogName + " - " + sqlStatement + " failed");
        }
        return nbRows;
    }

    /**
     * gets transfer requests associated to a list of instruments, in a given
     * status, starting from a given date ordered according to a given field in
     * a given order. If no instrument is given, all the instruments will be
     * selected
     * 
     * @param startDate
     *            date the research should start from
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space
     * @param status
     *            transfer request status
     * @param orderBy
     *            results sorting criteria
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDateGreaterThan(Date startDate,
            String instrumentList, TransferRequest.Status status,
            TransferRequest.FIELDS orderBy, SortingOrder sortingOrder)
            throws DAOException {
        String methodLogName = classLogName + "::findByDateGreaterThan()";

        if (startDate == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing night argument");
        }
        if (orderBy == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing sort by argument");
        }
        if (sortingOrder == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing sorting order argument");
        }

        // The instrument condition is implemented by checking if a given
        // instrument is contained in the file identifier
        // Each instrument of the instrument list is separated by a space
        String instrumentCondition = new String();
        String instrumentArray[];
        if (instrumentList != null && (instrumentList.length() != 0)) {
            instrumentCondition += " AND (";
            instrumentArray = instrumentList.split(" ");
            for (int index = 0; index < instrumentArray.length; index++) {
                instrumentCondition =
                        instrumentCondition + "file_id like '"
                                + instrumentArray[index] + "%'";
                if (index != (instrumentArray.length - 1)) {
                    instrumentCondition += " OR ";
                }
            }
            instrumentCondition += ") ";
        }

        // builds the status condition
        String statusCondition = new String();
        if ((status != null) && (status != TransferRequest.Status.ALL)) {
            statusCondition = " AND transfer_status=" + status.ordinal() + " ";
        }

        List<TransferRequest> transferRequestList = null;

        SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
        dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        final String sqlStatement =
                "select * from " + dbTable + " where " + CREATION_TIME_FIELD
                        + ">'" + dateFormat.format(startDate) + "'"
                        + instrumentCondition + statusCondition + " order by "
                        + TR_TABLE_FIELD_MAP.get(orderBy) + " "
                        + SORTING_ORDER_MAP.get(sortingOrder);
        logger.debug(methodLogName + " - Execution of query " + sqlStatement);
        transferRequestList = executeQuery(sqlStatement);

        return transferRequestList;

    }

    /**
     * gets transfer requests associated to a list of instruments of a given
     * observation date, in a given status, ordered according to some given
     * fields in a given order. If no instrument is given, all the instruments
     * will be selected
     * 
     * @param obsDate
     *            observation date of the files to be retrieved, starting from
     *            noon of the given day till noon of the following day
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space
     * @param statusList
     *            list of transfer request status
     * @param fileType
     *            type of the file to be transfered. Non mandatory
     * @param orderByList
     *            results sorting criteria
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDate(Date obsDate,
            String instrumentList, List<TransferRequest.Status> statusList,
            String fileType, List<TransferRequest.FIELDS> orderByList,
            SortingOrder sortingOrder) throws DAOException {

        String methodLogName = classLogName + "::findByDate";

        if (obsDate == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing night argument");
        }
        if (statusList == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing status argument");
        }
        if (orderByList == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing order by argument");
        }
        if (sortingOrder == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - Missing sorting order argument");
        }

        // The instrument condition is implemented by checking if a given
        // instrument is contained in the file identifier
        // Each instrument of the instrument list is separated by a space
        String instrumentCondition = new String();
        String instrumentArray[];
        if (instrumentList != null && (instrumentList.length() != 0)) {
            instrumentCondition += " AND (";
            instrumentArray = instrumentList.split(" ");
            for (int index = 0; index < instrumentArray.length; index++) {

                instrumentCondition =
                        instrumentCondition + "file_id like '"
                                + instrumentArray[index] + "%'";
                if (index != (instrumentArray.length - 1)) {
                    instrumentCondition += " OR ";
                }
            }
            instrumentCondition += ") ";
        }

        // builds the file name condition with the file pattern given by the
        // user. Checks also if the file is compressed
        // In case the file type is too, only fits file of the too category
        // will be retrieved
        String fileNameExtCondition = new String();
        String fileCatgCondition = new String();
        final String TOO_FILE_TYPE = "too";
        if ((fileType != null) && (fileType.length() != 0)) {
            if (fileType.compareToIgnoreCase(TOO_FILE_TYPE) == 0) {
                fileType = "fits";
                fileCatgCondition =
                        " AND transfer_catg='" + TransferRequest.TOO_CATG
                                + "' ";
            }
            fileNameExtCondition +=
                    " AND  (file_name like '%" + fileType.toLowerCase()
                            + "' OR file_name like '%" + fileType.toLowerCase()
                            + ".Z')";

        }

        // builds the status condition
        String statusCondition = new String();
        if (statusList.contains(TransferRequest.Status.ALL) == false) {
            Iterator<TransferRequest.Status> statusIterator =
                    statusList.iterator();
            if (statusIterator.hasNext()) {
                statusCondition = " AND (";
            }
            while (statusIterator.hasNext()) {
                statusCondition +=
                        " (transfer_status=" + statusIterator.next().ordinal()
                                + ") ";
                if (statusIterator.hasNext()) {
                    statusCondition += " OR ";
                } else {
                    statusCondition += ") ";
                }
            }
        }

        // Time range computation
        SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
        dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        // time of the start date will be set to 12:00pm
        Calendar startDate = new GregorianCalendar();
        startDate.setTime(obsDate);
        startDate.setTimeZone(TimeZone.getTimeZone("UTC"));
        startDate.set(Calendar.HOUR_OF_DAY, 12);
        startDate.set(Calendar.MINUTE, 0);
        startDate.set(Calendar.SECOND, 0);
        startDate.set(Calendar.MILLISECOND, 0);
        // the end date is one day later
        Calendar endDate = new GregorianCalendar();
        endDate.setTime(startDate.getTime());
        endDate.add(Calendar.DAY_OF_MONTH, 1);
        endDate.setTimeZone(TimeZone.getTimeZone("UTC"));

        String sqlStatement =
                "select * from " + dbTable + " where " + CREATION_TIME_FIELD
                        + " between '" + dateFormat.format(startDate.getTime())
                        + "' AND " + "'" + dateFormat.format(endDate.getTime())
                        + "' " + instrumentCondition + statusCondition
                        + fileCatgCondition + fileNameExtCondition
                        + " order by ";

        Iterator<TransferRequest.FIELDS> fieldsIterator =
                orderByList.iterator();
        while (fieldsIterator.hasNext()) {
            sqlStatement += TR_TABLE_FIELD_MAP.get(fieldsIterator.next());
            if (!fieldsIterator.hasNext()) {
                sqlStatement += " ";
            } else {
                sqlStatement += ",";
            }
        }
        sqlStatement += SORTING_ORDER_MAP.get(sortingOrder);
        logger.debug(methodLogName + " - Execution of query " + sqlStatement);
        List<TransferRequest> transferRequestList = null;
        transferRequestList = executeQuery(sqlStatement);
        return transferRequestList;

    }

    /**
     * Executes a query given as an argument. If the query does not imply any
     * returned records, am empty list of tranfer requests will be returned.
     * 
     * @param sqlStatement
     *            query to be executed
     * @return a list of transfer requests if successful, null otherwise
     * @throws DAOException
     */
    private List<TransferRequest> executeQuery(String sqlStatement)
            throws DAOException {
        Connection connection = null;
        List<TransferRequest> transferRequestList = null;
        try {
            connection =
                    SybaseConnectionService.getConnection(dbUrl,
                            dbUsername,
                            dbPassword);
            ResultSet rs =
                    SybaseConnectionService.executeStatement(connection,
                            sqlStatement);
            transferRequestList = toTransferRequestList(rs);
        } catch (Exception e) {
            throw new DAOException(e.getMessage());
        } finally {
            try {
                SybaseConnectionService.closeConnection(connection);
            } catch (ConnectionServiceException e) {
                throw new DAOException(e.getMessage());
            }
        }
        return transferRequestList;
    }

    /**
     * Executes a query given as an argument that changes database content. It
     * can be either an update or a delete query.
     * 
     * @param sqlStatement
     *            query to be executed
     * @return the number of updated or deleted rows if successful, -1 otherwise
     * @throws DAOException
     */
    private int executeUpdateQuery(String sqlStatement) throws DAOException {
        Connection connection = null;
        int nbUpdatedRows = 0;
        try {
            connection =
                    SybaseConnectionService.getConnection(dbUrl,
                            dbUsername,
                            dbPassword);
            nbUpdatedRows =
                    SybaseConnectionService.executeUpdateStatement(connection,
                            sqlStatement);
        } catch (Exception e) {
            nbUpdatedRows = -1;
            throw new DAOException(e.getMessage());
        } finally {
            try {
                SybaseConnectionService.closeConnection(connection);
            } catch (ConnectionServiceException e) {
                throw new DAOException(e.getMessage());
            }
        }
        return nbUpdatedRows;
    }

    /**
     * Creates a list of TransferRequest instances out of a result set. For each
     * row of the result set a TransferRequest instance is added to the list.
     * 
     * @param rs
     *            The result set.
     * @return The list of TransferRequest instances.
     * @throws DAOException
     *             in case of error.
     * @throws ParseException
     */
    private List<TransferRequest> toTransferRequestList(ResultSet rs)
            throws DAOException {

        Calendar utcCal = new GregorianCalendar();
        utcCal.setTimeZone(TimeZone.getTimeZone("UTC"));

        final List<TransferRequest> transferRequestList =
                new ArrayList<TransferRequest>();
        Status statusArray[] = Status.class.getEnumConstants();

        try {

            while (rs.next()) {
                TransferRequest transferRequest = new TransferRequest();
                transferRequest.setRequestId(rs.getString(REQUEST_ID_FIELD));
                transferRequest.setFileId(rs.getString(FILE_ID_FIELD));
                transferRequest.setFileName(rs.getString(FILE_NAME_FIELD));
                transferRequest.setSize(rs.getLong(SIZE_FIELD));
                transferRequest.setUncompressedSize(rs.getLong(UNCOMPRESSED_SIZE_FIELD));
                transferRequest.setChecksum(rs.getString(CHECKSUM_FIELD));
                transferRequest.setCompressionMethod(rs.getString(COMPRESSION_METHOD_FIELD));
                transferRequest.setTransferMethod(rs.getString(TRANSFER_METHOD_FIELD));
                transferRequest.setTransferPriority(rs.getInt(TRANSFER_PRIORITY_FIELD));
                transferRequest.setTransferCatg(rs.getString(TRANSFER_CATG_FIELD));
                transferRequest.setUser(rs.getString(DEST_USER_FIELD));
                transferRequest.setDestDir(rs.getString(DEST_DIR_FIELD));
                if (statusArray[rs.getInt(TRANSFER_STATUS_FIELD)] != null) {
                    transferRequest.setStatus(statusArray[rs.getInt(TRANSFER_STATUS_FIELD)]);
                } else {
                    transferRequest.setStatus(TransferRequest.Status.UNKNOWN);
                }
                TimeZone oldDefaultTz = TimeZone.getDefault();
                TimeZone.setDefault(TimeZone.getTimeZone("UTC"));
                java.sql.Timestamp curTimeStamp =
                        rs.getTimestamp(CREATION_TIME_FIELD, utcCal);
                transferRequest.setCreationTime(curTimeStamp.getTime());
                curTimeStamp = rs.getTimestamp(LAST_UPDATE_TIME_FIELD, utcCal);
                transferRequest.setLastUpdateTime(curTimeStamp.getTime());
                TimeZone.setDefault(oldDefaultTz);
                transferRequestList.add(transferRequest);
            }

        } catch (SQLException e) {
            throw new DAOException("Error while parsing result set. SQLException : "
                    + e.getMessage());
        }
        return transferRequestList;
    }

    /**
     * Test program
     * 
     * @param args
     */
    public static void main(String[] args) {
        BasicConfigurator.configure();

        TransferRequestValidator trValidator =
                new DefaultTransferRequestValidator();
        TransferRequestDAO transferRequestDAO =
                new TransferRequestDAOSyb("acdbdev.hq.eso.org:6789",
                        "sara",
                        "sara_ps",
                        "sara..transfer_requests",
                        trValidator);
        TransferRequest transferRequest = new TransferRequest();
        transferRequest.setRequestId("RequestId");
        transferRequest.setFileId("FileId");
        transferRequest.setFileName("FileName");
        transferRequest.setSize(new Long(1000));
        transferRequest.setUncompressedSize(new Long(1000000));
        transferRequest.setChecksum("Checksum");
        transferRequest.setCompressionMethod("CompressionMethod");
        transferRequest.setTransferMethod("TransferMethod");
        transferRequest.setTransferPriority(new Integer(3));
        transferRequest.setTransferCatg("TransferCatg");
        transferRequest.setUser("DestUser");
        transferRequest.setDestDir("DestDir");
        transferRequest.setStatus(Status.PROCESSED);
        transferRequest.setCreationTime(new Long(3600000 * 37 + 59115));

        logger.info("Set of tests on the TransferRequestDAOSyb class");
        logger.info("Case 1 : Set of tests on the save method");
        logger.info("\t Case 1_1 : Nominal case ");
        try {
            int nbInsertedRows = transferRequestDAO.save(transferRequest);
            if (nbInsertedRows == 1) {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getFileId() + " has been inserted ");
            } else {
                logger.error("\t FAILED Number of inserted transfer requests is "
                        + nbInsertedRows);
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId() + " has not been saved : "
                    + e.getMessage());
        }
        try {
            transferRequestDAO.delete(transferRequest);
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId()
                    + " has not been deleted : " + e.getMessage());
        }

        logger.info("\t Case 1_2 : Try to insert twice a transfer request ");
        try {
            transferRequestDAO.save(transferRequest);
            transferRequestDAO.save(transferRequest);
            logger.error("\t FAILED No DAOException has been raised");

        } catch (DAOException e) {
            logger.error("\t SUCCESSFUL DAOException has been raised : "
                    + e.getMessage());
        } catch (Exception e) {
            logger.error("\t FAILED Unknown exception has been raised "
                    + e.getMessage());
        }
        try {
            transferRequestDAO.delete(transferRequest);
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getFileId() + " has not been deleted : "
                    + e.getMessage());
        }

        logger.info("\t Case 1_3 : request id is null ");
        String formerRequestId = transferRequest.getRequestId();
        transferRequest.setRequestId(null);
        try {
            transferRequestDAO.save(transferRequest);
            logger.error("\t FAILED No DAOException has been raised");

        } catch (DAOException e) {
            logger.info("\t SUCCESSFUL DAOException has been raised : "
                    + e.getMessage());
        }
        transferRequest.setRequestId(formerRequestId);

        logger.info("\t Case 1_4 : file id is null ");
        String formerFileId = transferRequest.getFileId();
        transferRequest.setFileId(null);
        try {
            transferRequestDAO.save(transferRequest);
            logger.error("\t FAILED No DAOException has been raised");

        } catch (DAOException e) {
            logger.info("\t SUCCESSFUL DAOException has been raised : "
                    + e.getMessage());
        }
        transferRequest.setFileId(formerFileId);

        logger.info("\t Case 1_4 :  No transfer request is provided ");
        try {
            transferRequestDAO.save(null);
            logger.error("\t FAILED No DAOException has been raised");

        } catch (DAOException e) {
            logger.error("\t FAILED DAOException has been raised : "
                    + e.getMessage());
        } catch (IllegalArgumentException e) {
            logger.info("\t SUCCESSFUL IllegalArgumentException has been raised : "
                    + e.getMessage());
        }

        logger.info("\t Case 1_5 : Transfer request with no creation time stamp ");
        Long oldTimeStamp = transferRequest.getCreationTime();
        try {

            transferRequest.setCreationTime(null);

            int nbInsertedRows = transferRequestDAO.save(transferRequest);
            if (nbInsertedRows == 1) {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getFileId() + " has been inserted ");
            } else {
                logger.error("\t FAILED Number of inserted transfer requests is "
                        + nbInsertedRows);
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId() + " has not been saved : "
                    + e.getMessage());
        }
        try {
            transferRequestDAO.delete(transferRequest);
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId()
                    + " has not been deleted : " + e.getMessage());
        }
        transferRequest.setCreationTime(oldTimeStamp);

        logger.info("Case 2 : Set of tests on the delete method");
        logger.info("\t Case 2_1 : Nominal case ");
        try {
            transferRequestDAO.save(transferRequest);
            int nbDeletedRows = transferRequestDAO.delete(transferRequest);
            if (nbDeletedRows == 1) {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getFileId() + " has been deleted ");
            } else {
                logger.error("\t FAILED Number of deleted transfer requests is "
                        + nbDeletedRows);
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getFileId() + " has not been deleted : "
                    + e.getMessage());
        }
        logger.info("\t Case 2_2 : try to delete a non existing transferRequest ");
        formerRequestId = transferRequest.getRequestId();
        transferRequest.setRequestId("NonExistingRid");
        try {
            int nbDeletedRows = transferRequestDAO.delete(transferRequest);
            if (nbDeletedRows == 1) {
                logger.error("\t FAILED  transferRequest "
                        + transferRequest.getRequestId() + " has been deleted ");
            } else if (nbDeletedRows == 0) {
                logger.info("\t SUCCESSFUL  Number of deleted transfer requests is 0");

            } else {
                logger.info("\t FAILED  Number of deleted transfer requests is "
                        + nbDeletedRows);
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId()
                    + " has not been deleted : " + e.getMessage());
        }
        transferRequest.setRequestId(formerRequestId);

        logger.info("\t Case 2_3 : Request id is null ");
        formerRequestId = transferRequest.getRequestId();
        transferRequest.setRequestId(null);
        try {
            transferRequestDAO.delete(transferRequest);
            logger.error("\t FAILED No exception raised");

        } catch (DAOException e) {
            logger.error("\t SUCCESSFUL DAOException has been raised "
                    + e.getMessage());
        }
        transferRequest.setRequestId(formerRequestId);

        logger.info("\t Case 2_4 : no transfer request is provided ");
        try {
            transferRequestDAO.delete(null);
            logger.error("\t FAILED No exception raised");

        } catch (IllegalArgumentException e) {
            logger.info("\t SUCCESSFUL  IllegalArgumentException  has been raised : "
                    + e.getMessage());
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException  has raised : "
                    + e.getMessage());
        }

        logger.info("Case 3 : Set of tests on the findbystatus method");
        logger.info("\t Case 3_1 : Nominal case ");
        Status status = transferRequest.getStatus();
        try {
            transferRequestDAO.save(transferRequest);

            List<TransferRequest> trList =
                    transferRequestDAO.findByStatus(status);
            Iterator<TransferRequest> trIterator = trList.iterator();
            TransferRequest curTransferRequest;
            boolean trRetrieved = false;
            while (trIterator.hasNext()) {
                curTransferRequest = trIterator.next();
                String curTrFileId = curTransferRequest.getFileId();
                if (curTrFileId.compareTo(transferRequest.getFileId()) == 0) {
                    if (curTransferRequest.getStatus() == status) {
                        trRetrieved = true;
                    }
                    break;
                }
            }
            if (trRetrieved == true) {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getFileId() + " in status  " + status
                        + "  has been retrieved ");
            } else {
                logger.error("\t FAILED  transferRequest "
                        + transferRequest.getFileId() + " in status  " + status
                        + "  has not been retrieved ");
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getFileId() + " has not been deleted : "
                    + e.getMessage());
        }

        logger.info("\t Case 3_2 : Looking for a status different from the status of the test record");
        try {
            Status retrievedStatus =
                    Status.values()[(transferRequest.getStatus().ordinal() + 1)];
            List<TransferRequest> trList =
                    transferRequestDAO.findByStatus(retrievedStatus);
            Iterator<TransferRequest> trIterator = trList.iterator();
            TransferRequest curTransferRequest;
            boolean trRetrieved = false;
            while (trIterator.hasNext()) {
                curTransferRequest = trIterator.next();
                String curTrRequestId = curTransferRequest.getRequestId();
                if (curTrRequestId.compareTo(transferRequest.getRequestId()) == 0) {
                    if (curTransferRequest.getStatus() == status) {
                        trRetrieved = true;
                    }
                    break;
                }
            }
            if (trRetrieved == true) {
                logger.error("\t FAILED  transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + retrievedStatus + "  has been retrieved ");
            } else {
                logger.info("\t SUCCESSFUL  No transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + retrievedStatus + "  has been retrieved ");
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getFileId()
                    + " has not been retrieved : " + e.getMessage());
        }

        logger.info("\t Case 3_3 : No status is provided");
        try {
            transferRequestDAO.findByStatus(null);
            logger.error("\t FAILED No IllegalArgumentException has been raised");
        } catch (IllegalArgumentException e) {
            logger.info("\t SUCCESSFUL IllegalArgumentException exception has been raised");
        } catch (Exception e) {
            logger.error("\t FAILED Unknown exception has been raised");
        }

        // clean up
        try {
            transferRequestDAO.delete(transferRequest);
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getFileId() + " has not been deleted : "
                    + e.getMessage());
        }

        logger.info("Case 4 : Set of tests on the updateStatus method");
        logger.info("\t Case 4_1 : Nominal case ");
        try {
            transferRequestDAO.save(transferRequest);
            Status newStatus =
                    Status.values()[transferRequest.getStatus().ordinal() + 1];
            int nbUpdated =
                    transferRequestDAO.updateStatus(transferRequest, newStatus);
            if (nbUpdated == 1) {
                logger.info("\tSUCCESSFUL Number of updated transfer request is 1");
            } else {
                logger.error("\tFAILED Number of updated transfer request is "
                        + nbUpdated);
            }

            List<TransferRequest> trList =
                    transferRequestDAO.findByStatus(newStatus);
            Iterator<TransferRequest> trIterator = trList.iterator();
            TransferRequest curTransferRequest;
            boolean trRetrieved = false;
            while (trIterator.hasNext()) {
                curTransferRequest = trIterator.next();
                String curTrRequestId = curTransferRequest.getRequestId();
                if (curTrRequestId.compareTo(transferRequest.getRequestId()) == 0) {
                    if (curTransferRequest.getStatus() == newStatus) {
                        trRetrieved = true;
                    }
                    break;
                }
            }
            if (trRetrieved == true) {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + newStatus + "  has been retrieved ");
            } else {
                logger.error("\t FAILED  transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + newStatus + "  has not been retrieved ");
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId()
                    + " has not been updated : " + e.getMessage());
        }

        logger.info("\t Case 4_2 : The transfer request is not available ");
        try {
            transferRequestDAO.delete(transferRequest);
            status = transferRequest.getStatus();
            int nbUpdated =
                    transferRequestDAO.updateStatus(transferRequest, status);
            if (nbUpdated == 0) {
                logger.info("\tSUCCESSFUL No transfer request has been retrieved");
            } else {
                logger.error("\tFAILED Number of updated transfer request is "
                        + nbUpdated);
            }

            List<TransferRequest> trList =
                    transferRequestDAO.findByStatus(status);
            Iterator<TransferRequest> trIterator = trList.iterator();
            TransferRequest curTransferRequest;
            boolean trRetrieved = false;
            while (trIterator.hasNext()) {
                curTransferRequest = trIterator.next();
                String curTrRequestId = curTransferRequest.getRequestId();
                if (curTrRequestId.compareTo(transferRequest.getRequestId()) == 0) {
                    if (curTransferRequest.getStatus() == status) {
                        trRetrieved = true;
                    }
                    break;
                }
            }
            if (trRetrieved == true) {
                logger.error("\t FAILED  transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + status + "  has been retrieved ");
            } else {
                logger.info("\t SUCCESSFUL  transferRequest "
                        + transferRequest.getRequestId() + " in status  "
                        + status + "  has not been retrieved ");
            }
        } catch (DAOException e) {
            logger.error("\t FAILED DAOException "
                    + transferRequest.getRequestId()
                    + " has not been updated : " + e.getMessage());
        }

        logger.info("\t Case 4_3 : No transfer request is provided ");
        try {
            Status newStatus = Status.INCOMING;
            transferRequestDAO.updateStatus(null, newStatus);
            logger.error("\tFAILED No exception has been raised");

        } catch (IllegalArgumentException e) {
            logger.info("\t SUCCESSFUL IllegalArgumentException has been raised : "
                    + e.getMessage());
        } catch (Exception e) {
            logger.error("\t FAILED Unknown Exception has been raised : "
                    + e.getMessage());
        }
        logger.info("\t Case 4_4 : No status is provided ");
        try {
            transferRequestDAO.updateStatus(transferRequest, null);
            logger.error("\tFAILED No exception has been raised");

        } catch (IllegalArgumentException e) {
            logger.info("\t SUCCESSFUL IllegalArgumentException has been raised : "
                    + e.getMessage());
        } catch (Exception e) {
            logger.error("\t FAILED Unknown Exception has been raised : "
                    + e.getMessage());
        }
        // some clean up
        try {
            transferRequestDAO.delete(transferRequest);
        } catch (DAOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

    }

    public void deleteByStatus(TransferRequest.Status status)
            throws DAOException {
        throw new DAOException("Method not implemented yet");
    }

}
