package org.eso.util.database;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * ConnectionPool maintains pools of database connections, including
 * enforcing an upper limit and closing connections that have not recently
 * been used.  If there are no connections left in the pool, new ones are
 * allocated upto the given limit of MAX_CONNECTIONS.  If all permitted
 * connections have been created and are currently allocated, the client
 * must wait for one to be released.
 * 
 * <p>
 * An inner class, Reaper, checks every so often for idle connections.
 * Those that have not been used in the last MAX_IDLE_LIFETIME milliseconds
 * are removed from the pool and closed.
 * 
 * <p>
 * Note: some methods and fields have been made protected so that
 *       the test driver can access them more easily.
 * 
 * @author Bob Kemp
 * @version $Revision: 1.1 $
 */
public class ConnectionPool {

    private static final String rcsid = "$Id: ConnectionPool.java,v 1.1 2007/03/07 09:20:43 szampier Exp $";

    /**
     * <br>ChangeLog: amchavan, 18-Dec-2001: added
     */
    private static Logger stdlog_ = 
	Logger.getLogger( ConnectionPool.class );

    /**
     * the number of open connections in the pool (1 or more).
     */
    public static final String PoolSize = "conn.pool.size";

    /**
     * the timeout for unused connections - 0 means never timeout.
     */
    public static final String Timeout = "conn.pool.timeout";

    /**
     * reuse connections if "T": unused connections maybe
     * closed (if unused) and re-opened to a different URL.
     */
    public static final String ReopenAllowed = "conn.pool.reopen";

    /** Maximum number of connections we can allocate */
    protected static int MAX_CONNECTIONS = 20;

    /** Milliseconds between checks of idle/old connections */
    protected static long REAPER_SLEEPTIME = 5*1000;    // in millisecs

    /** Milliseconds a connection can remain idle before being closed */
    protected static long MAX_IDLE_LIFETIME = 60*1000;  // in millisecs

    /** Milliseconds an unallocated connection lasts before being closed */
    protected static int MAX_LIFETIME = 600*1000; // in millisecs

    /** is re-open allowed on unused connections */
    protected static boolean reopenAllowed_ = true;

    protected Vector<Element> pool_ = new Vector<Element>();

    private static ConnectionPool conPool_;

    private int maxPoolSize_;

    private Reaper reaper_;

    private String userName_;
 
    private String password_;

    static {
        /**
         * Register the JDBC driver,
    	 * as described in Sybase jConnect documentation.
    	 * XXX NOTE: using Class.forName() _doesn't_ work properly !!!
             */
    
    	Properties sysProps = System.getProperties();
    	StringBuffer drivers = new StringBuffer("com.sybase.jdbc2.jdbc.SybDriver");
    	String oldDrivers = sysProps.getProperty("jdbc.drivers");
    	if (oldDrivers != null)
    	    drivers.append(":" + oldDrivers);
    	sysProps.put("jdbc.drivers", drivers.toString());
    }

    public static long getMaxIdleLifetime() {
        return MAX_IDLE_LIFETIME;
    }

    public static long getReaperSleepTime() {
        return REAPER_SLEEPTIME;
    }

    /**
     * Set the initialization properties of the connection pool.
     * NOTE: this must be done prior to the first call to getPool().
     * @return false if the pool singleton has already been created.
     */

    public synchronized static boolean setInitProperties(Properties props) {

        String prop = props.getProperty(PoolSize);
        if (prop != null) {
            try {
                MAX_CONNECTIONS = Integer.parseInt(prop);
            } catch (NumberFormatException e) {
                e.printStackTrace();
            }
            stdlog_.info( "MAX_CONNECTIONS set to: " + MAX_CONNECTIONS );
        }

        prop = props.getProperty(Timeout);
        if (prop != null) {
            try {
                MAX_IDLE_LIFETIME = Integer.parseInt(prop);
            } catch (NumberFormatException e) {
                e.printStackTrace();
            }
            stdlog_.info( "MAX_IDLE_LIFETIME set to: " + MAX_IDLE_LIFETIME );
        }

        prop = props.getProperty(ReopenAllowed);
        if (prop != null) {
            if (prop.equalsIgnoreCase("T")) {
                reopenAllowed_ = true;
            } else {
                reopenAllowed_ = false;
            }
            stdlog_.info( "reopenAllowed_ set to: " + reopenAllowed_ );
        }

        return (conPool_ == null);
    }
    
    protected ConnectionPool(int maxPoolSize) {
    	maxPoolSize_ 	= maxPoolSize;
    
            if (MAX_IDLE_LIFETIME > 0) {
    	reaper_		= new Reaper(this, pool_);
    	reaper_.start();
            } else {
                reaper_ = null;
            }
    
            stdlog_.debug( "MAX_IDLE_LIFETIME: " + MAX_IDLE_LIFETIME );
            	stdlog_.debug( "MAX_CONNECTIONS: " + MAX_CONNECTIONS );
    	stdlog_.debug( "reopenAllowed_: " + reopenAllowed_ );
    }
    
    /**
     * Set the userName_ property - can be overridden by acquireConnection.
     */
    public void setUserName(String un) {
        userName_ = un;
    }
 
    /**
     * Set the password_ property - can be overridden by acquireConnection.
     */   
    public void setPassword(String pw) {
        this.password_ = pw;
    }

    public synchronized Connection acquireConnection(String dbUrl,
					String userName, String password)
					    throws SQLException {	
    	Element elem = null; 
        userName_ = userName;
        setPassword(password); 
        if (stdlog_.isDebugEnabled()){
        stdlog_.debug( "URL: "+dbUrl+ " user name: " + userName_ );
        }
        String dbName 
            = dbUrl.substring(dbUrl.lastIndexOf('/')+1,dbUrl.length());
    
    	for ( ; ; ) {
    	    int poolSize = pool_.size();
    
    	    for (int j=0; j < poolSize; j++) {
    		elem = pool_.elementAt(j);
    
                    // only compare the dbName if reopenAllowed_ is true.
    		if (elem.available
                            && (!reopenAllowed_ || dbName.equals(elem.dbaseName)))
                    {
    		    elem.available = false;
    		    stdlog_.debug("@@ acquireConnection: found " + j);
                //debugging: (new Exception()).printStackTrace();
    		    return elem.connection;
    		}
    	    }
    
                if (reopenAllowed_) {
                    // search for a possible connection to re-use.
    	        for (int j=0; j < poolSize; j++) {
    		    elem = pool_.elementAt(j);
    
    		    if (elem.available) {
    		        elem.available = false;
                stdlog_.debug( "@@ acquireConnection: found " + j);
                // debugging: (new Exception()).printStackTrace();
    		        reuseConnection(elem,dbUrl,dbName);
    		        return elem.connection;
    		    }
    	        }
                }
    
    	    if (pool_.size() < maxPoolSize_) {
    		break;		// allocate new connection
    	    }
    
    	    try {
    		stdlog_.debug("@@ acquireConnection: waiting");
    		wait();		// for someone to release a connection
    	    }
    	    catch (InterruptedException e) {
    		e.printStackTrace();
    	    }
    	}
    
    	stdlog_.debug("@@ acquireConnection: adding conn " + pool_.size());
    	elem = new Element();
    	elem.available = false;
    	elem.creationTime = System.currentTimeMillis();
    	elem.connection = openConnection(dbUrl);
    	elem.dbaseName = dbName;
    
    	pool_.addElement(elem);
    
    	return elem.connection;
    }

    /**
     * Release all connections.
     * @throws SQLException 
     */
    public synchronized void releaseAllConnections() throws SQLException {

        for( Iterator<Element> iter = pool_.iterator(); iter.hasNext(); ) {
            Element element = iter.next();
            closeConnection( element.connection );
        }
        pool_.removeAllElements();
    }

    /**
     * Release the given connection back into the pool.
     * <p>
     * 
     * <b>NOTE:</b> This should rollback the returned connection
     *              or something similar, so that it can check for
     *              SQL exceptions on the connection and discard it.
     *              Currently the connection is simply returned to
     *              the pool and could cause further errors later.
     */
    public synchronized void releaseConnection(Connection con) {
    	if (con == null) {
    	    return;
    	}
    
    	int poolSize = pool_.size();
    
    	for (int j=0; j < poolSize; j++) {
    	    Element elem = pool_.elementAt(j);
    
    	    if (elem.connection == con) {
                    boolean removeElem = false;
                    try {
                        if (elem.connection.isClosed()) {
                            removeElem = true;
                        }
                    } catch (SQLException e) {
                        // problem with this connection....
                        // remove it.
                        removeElem = true;
                    }
                    if (removeElem) {
                        pool_.removeElementAt(j);
                    } else {
    		    elem.available = true;
    		    elem.last_released = System.currentTimeMillis();
                    }
    		stdlog_.debug("@@ releaseConnection: releasing " + j);
            // debugging: (new Exception()).printStackTrace();
    
    			    // Notify one waiting thread that
    			    //    a connection is available.
    		notify();
    		return;
    	    }
    	}
    
    	stdlog_.warn("@@ releaseConnection: NOT FOUND");
    	throw new IllegalArgumentException(
    			    "internal error: unknown connection");
    }


    /**
     * Closed the element connection and re-open it to the new URL.
     * NOTE: this method does nothing unless reopenAllowed_ is true.
     */

    protected void reuseConnection(Element elem,String dbUrl,String dbName)
				    throws SQLException {
        if (reopenAllowed_) {
    	    closeConnection(elem.connection);
    	    elem.connection = openConnection(dbUrl);
    	    elem.dbaseName = dbName;
        }
    }

    // NB: this method must be protected, so that the test driver
    //     can override it.
    protected Connection openConnection(String url)
                throws SQLException {
        return openConnection(url, null, null);
    }

    protected Connection openConnection(String url, String un, String pw) throws SQLException
				 {

    	/*
    	stdlog_.debug( "Trying: [" + url + "] [" 
    		       + userName_ + "] [" 
    		       + password_ + "]" ); 
    	*/
//		To get the driver to work with jConnect 5.x to work with 12.5 on HP
//		do the following
//
//		Replace in ConnectionPool.java
//
//              conn = DriverManager.getConnection(url,un,pw) ;
//
//		with
//
//		Properties props = new Properties();
//		props.put("CHARSET","iso_1");
//		props.put("user",un);
//		props.put("password",pw);
//		conn = DriverManager.getConnection(url,props);


        if (un == null) {
            un = userName_;
        }
        if (pw == null) {
            pw = getPassword();
        }
        Properties props = new Properties();
        props.put("CHARSET","iso_1");
        props.put("user",un); 
        props.put("password",pw); 
        Connection conn;
		try {
			conn = DriverManager.getConnection(url,props);
		} catch (SQLException e) {
			stdlog_.error("url: "+ url + " user: "+ un );
			throw e;
		}
        
    	return conn;
    }
    
    /**
     * This method provides a means to open a connnection
     * that is NOT associated with the connection pool.
     * The connection must be closed with conn.close()
     * and not releaseConnection().
     */
    public Connection openExtraPoolConnection(String url, String un, String pw) 
            throws SQLException {
        return openConnection(url, un, pw);
    }


    // NB: this method must be protected, so that the test driver
    //     can override it.
    protected void closeConnection(Connection con)
				throws SQLException {
	   con.close();
    }

    // ----------- Inner classes ----------

    protected static class Element {
    	public String dbaseName;
    	public boolean available;
    	public Connection connection;
    	public long last_released;
    	public long creationTime;

    	public String toString() {
    	    return
    		"Element["
    		+ "dbaseName=" + dbaseName
    		+ ",available=" + available
    		+ ",last_released=" + last_released
    		+ ",creationTime=" + creationTime
    		+ "]";
    	}
    }

    /**
     * The reaper class that scans the pool 
     * and closes old (un-needed) connnections.
     */
    private class Reaper extends Thread {
    	private ConnectionPool pool_;
    	private Vector<Element> poolVec_;
    	private boolean finished_ = false;
    
    	public Reaper(ConnectionPool pool, Vector<Element> poolVec) {
    	    super("ConnectionPool - Dead Connection Reaper");
    	    pool_	= pool;
    	    poolVec_	= poolVec;
    
    	    setDaemon(true);	// background daemon thread.
    	}
    
    	public void shutdown() {
    	    finished_ = true;
    	    synchronized (this) {
    		try {
    		    wait();
    		}
    		catch (InterruptedException e) {
    		    // don't care
    		}
    	    }
    	}
    
    	public void run() {
    	    Element elem = null;
    
    	    while ( ! finished_ ) {
    		long now = System.currentTimeMillis();
    
    		synchronized (pool_) {
    
    		    for (int j=poolVec_.size()-1; j >= 0; j--) {
    			elem = poolVec_.elementAt(j);
    
    			if (elem.available &&
    			    (elem.last_released+MAX_IDLE_LIFETIME < now
    				|| elem.creationTime+MAX_LIFETIME < now)) {
    					// Connection unused for too long
    					// Close it.
    			    stdlog_.debug("@@ reaper: last release "
    					  + elem.last_released);
    			    poolVec_.removeElementAt(j);
    			    forciblyCloseConnection(elem);
    			}
    		    }
    		}
    
    		try {
    		    Thread.sleep(REAPER_SLEEPTIME);
    		}
    		catch (InterruptedException e) {
    		    // don't care
    		}
    
    	    }
    
    	    synchronized (this) {
    		notifyAll();
    	    }
    
    	}
    
    	public void forciblyCloseConnection(Element elem) {
    	    stdlog_.debug("@@ reaper: closing " + elem);
    	    try {
    		  closeConnection(elem.connection);
    		  elem.connection = null;
    	    }
    	    catch (SQLException e) {
    		  e.printStackTrace();
    	    }
    	}

    }


    // ----------- Static methods ----------

    public static synchronized ConnectionPool getPool() {
    	if (conPool_ == null) {
    	    conPool_ = new ConnectionPool(MAX_CONNECTIONS);
    	}
    	return conPool_;
    }

    public static synchronized void shutdown() {
    	if (conPool_ != null) {
                if (conPool_.reaper_ != null) {
    	        conPool_.reaper_.shutdown();
                }
    	    conPool_ = null;
    	}
    }

    /**
     * @return Returns the password_.
     */
    private String getPassword() {
        return password_;
    }

}

