package org.eso.util.datatransfer;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.concurrent.TimeoutException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.eso.util.misc.SystemCommand;

public class FileCopyCommand extends SystemCommand {

    /** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(FileCopyCommand.class);

    /** Set of available keyword in a command format */
    private final String HOST_KEYWORD = "\\$HOST";

    private final String PORT_KEYWORD = "\\$PORT";

    private final String USER_KEYWORD = "\\$USER";

    private final String DIR_KEYWORD = "\\$DIR";

    private final String FILENAME_KEYWORD = "\\$FILENAME";

    private final String FILEID_KEYWORD = "\\$FILEID";

    private final String FILEVERSION_KEYWORD = "\\$FILEVERSION";

    /** The name of this class, for logging. */
    private static final String thisClass = "FileCopyCommand";

    /**
     * Constructor with filename.
     * 
     * @param copyCommand
     *            The copy command.
     * @param filename
     *            The name of the file to copy.
     * @param destUrl
     *            The destination URL.
     * @param timeout
     *            command timeout
     */
    public FileCopyCommand(String copyCommand, String filename, URL destUrl,
            long timeout) {
        super();
        String thisMethod = thisClass + "::FileCopyCommand()";
        if ((copyCommand == null) || (filename == null) || (destUrl == null)) {
            String message = thisMethod + " - Some arguments are missing";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        logger.trace(thisMethod);
        String command = copyCommand;
        if (destUrl.getHost() != null)
            command = command.replaceAll(HOST_KEYWORD, destUrl.getHost());
        command =
                command.replace(PORT_KEYWORD, String.valueOf(destUrl.getPort()));
        if (destUrl.getPath() != null)
            command = command.replaceAll(DIR_KEYWORD, destUrl.getPath());
        command = command.replaceAll(FILENAME_KEYWORD, filename);
        setCommandLine(command);
        setTimeout(timeout);
    }

    /**
     * Constructor with filename, fileid, fileversion.
     * 
     * @param copyCommand
     *            The copy command.
     * @param filename
     *            The name of the file to copy.
     * @param fileid
     *            The id of the file to copy.
     * @param fileversion
     *            The version of the file to copy.
     * @param destUrl
     *            The destination URL.
     * @param timeout
     *            command timeout
     */
    public FileCopyCommand(String copyCommand, String filename, String fileid,
            String fileversion, URL destUrl, long timeout) {
        super();
        String thisMethod = thisClass + "::FileCopyCommand()";
        logger.trace(thisMethod);
        if ((copyCommand == null) || (filename == null) || (fileid == null)
                || (fileversion == null) || (destUrl == null)) {
            String message = thisMethod + " - Some arguments are missing";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        String command = copyCommand;
        if (destUrl.getHost() != null)
            command = command.replaceAll(HOST_KEYWORD, destUrl.getHost());
        command =
                command.replaceAll(PORT_KEYWORD,
                        String.valueOf(destUrl.getPort()));
        if (destUrl.getPath() != null)
            command = command.replaceAll(DIR_KEYWORD, destUrl.getPath());
        command = command.replaceAll(FILENAME_KEYWORD, filename);
        command = command.replaceAll(FILEID_KEYWORD, fileid);
        command = command.replaceAll(FILEVERSION_KEYWORD, fileversion);
        setCommandLine(command);
        setTimeout(timeout);
    }

    /**
     * Constructor with filename, fileid, fileversion, destination url, name of
     * the user associated to the copy, destination directory
     * 
     * @param copyCommand
     *            The copy command.
     * @param filename
     *            The name of the file to copy.
     * @param fileid
     *            The id of the file to copy.
     * @param fileversion
     *            The version of the file to copy.
     * @param destUrl
     *            The destination URL.
     * @param copyUuser
     *            name of the user associated to the copy
     * @param destDir
     *            destination of the copy 
     * @param timeout
     *            command timeout
     */
    public FileCopyCommand(String copyCommand, String filename, String fileid,
            String fileversion, String copyUser, URL destUrl, long timeout) {
        super(copyCommand);
        String thisMethod = thisClass + "::FileCopyCommand()";
        logger.trace(thisMethod);
        if ((copyCommand == null) || (filename == null) || (fileid == null)
                || (fileversion == null) || (copyUser == null)
                || (destUrl == null)) {
            String message = thisMethod + " - Some arguments are missing";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        String command = copyCommand;
        if (destUrl.getHost() != null)
            command = command.replaceAll(HOST_KEYWORD, destUrl.getHost());
        command =
                command.replaceAll(PORT_KEYWORD,
                        String.valueOf(destUrl.getPort()));
        if (destUrl.getPath() != null)
            command = command.replaceAll(DIR_KEYWORD, destUrl.getPath());

        command = command.replaceAll(FILENAME_KEYWORD, filename);

        command = command.replaceAll(FILEID_KEYWORD, fileid);

        command = command.replaceAll(FILEVERSION_KEYWORD, fileversion);

        command = command.replaceAll(USER_KEYWORD, copyUser);
        setCommandLine(command);
        setTimeout(timeout);
    }

    /**
     * Test program
     * 
     * @param args
     */
    public static void main(String[] args) {
        // Basic Log4j configuration
        BasicConfigurator.configure();
        String FS = System.getProperty("file.separator");
        // Copy command
        String copyCmd = "rcp $FILENAME $HOST:$DIR";
        // String filename = "temp" + FS + "pippo.trq";
        String filename = "/tmp/toto.txt";
        URL destUrl = null;
        try {
            // destUrl = new URL("ftp://dfsdev1.hq.eso.org/diskb/data");
            destUrl = new URL("http://dmdlinb/tmp");
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
        FileCopyCommand cmd =
                new FileCopyCommand(copyCmd, filename, destUrl, 1800);
        logger.debug(cmd.getCommandLine());

        try {
            cmd.execute();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (IllegalStateException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (TimeoutException e) {
            e.printStackTrace();
        }

        finally {
            logger.debug("Exit status = " + cmd.getExitStatus());
            if (cmd.getExitStatus() != 0) {
                if (cmd.getStandardErrorString() != null)
                    logger.error(cmd.getStandardErrorString());
            } else {
                if (cmd.getStandardOutputString() != null)
                    logger.info(cmd.getStandardOutputString());
            }
        }
    }
}