package org.eso.util.datatransfer;

import java.io.IOException;
import java.util.Calendar;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * This class encapsulates the execution of a shell command from a Java
 * application. In particular it provides easy access to the output and error
 * streams of the command, and to its exit status.
 * 
 * @author szampier <szampier@eso.org>
 * 
 */
// RIL : CM would not it be more generic and meaningful to call this class
// ExternalCommand rather than ShellCommand ?
// REPLY: I don't see a big difference. What about SystemCommand ?
public class ShellCommand {

	/** The Log4j logger */
	static Logger logger = Logger.getLogger(ShellCommand.class);

	// RIL : CM should not we add an _ at the end of a data member ?
	// REPLY: it's a matter of personal preferences. I find the code more readable
	// without underscores.
	/** The command */
	String command = null;

	/** The command output */
	String output = null;

	/** The command error */
	String error = null;

	/** The command exit status */
	int exitStatus = 0;

	/** The name of this class */
	static final String thisClass = "ShellCommand";

	/**
	 * Default constructor.
	 */
	protected ShellCommand() {
	}

	/**
	 * Constructor.
	 * 
	 * @param command
	 *            The command to execute.
	 */
	public ShellCommand(String command) {
		String thisMethod = thisClass + "::Command()";
		logger.trace(thisMethod);

		if (command == null) {
			String message = thisMethod + " - command must not be null.";
			throw new IllegalArgumentException(message);
		}
		this.command = command;
	}

	// RIL : CM redundant comments in the header
	// REPLY: SZA ok, will fix.
	// RIM : the accessor to the output and error streams are declared 
	// as synchronized. Should not the exec method be declared also
	// as synchronized as it modifies these streams ?
	// REPLY: SZA agreed.
	/**
	 * Executes the command and returns its exit status.
	 * 
	 * @return the exit status of the command.
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws IOException
	 */
	public int exec() throws InterruptedException, IOException {
		String thisMethod = thisClass + "::exec()";
		logger.trace(thisMethod);

		Runtime rt = Runtime.getRuntime();
		Process proc;
		try {
		    logger.debug(command);
			proc = rt.exec(command);
		} catch (IOException e) {
			exitStatus = -1;
			logger.error(e.getMessage());
			throw e;
		}

		// Instantiate a StreamConsumer to read the error stream
		StreamConsumer errorConsumer = new StreamConsumer(proc.getErrorStream());
		// Instantiate a StreamConsumer to read the output stream
		StreamConsumer outputConsumer = new StreamConsumer(proc
				.getInputStream());
		// Spawn a new thread to read the error stream
		errorConsumer.start();
		// Spawn a new thread to read the output stream
		outputConsumer.start();
		// Wait for command to terminate
		try {
			exitStatus = proc.waitFor();
		} catch (InterruptedException e) {
			exitStatus = -1;
			logger.error(e.getMessage());
			throw e;
		}
		/// RIM : CM should not it be waited that the two threads are completed 
		// with a join ?
		// Store output and error buffers into class members
		// REPLY: agreed. I will probably invoke join() from toString() and invoke start() from the constructor.
		output = outputConsumer.toString();
		error = errorConsumer.toString();

		proc.destroy();
		
		// Return the exit status
		return exitStatus;
	}

	/* Getters and Setters */

	public synchronized String getCommand() {
		return command;
	}

	protected synchronized void setCommand(String c) {
		this.command = c;
	}

	public synchronized String getError() {
		return error;
	}

	public synchronized int getExitStatus() {
		return exitStatus;
	}

	public synchronized String getOutput() {
		return output;
	}

	/**
	 * Test program
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		// Basic Log4j configuration
		BasicConfigurator.configure();

		long objectCount = 10000000;
		long startTime = Calendar.getInstance().getTimeInMillis();
		for (int i = 0; i < objectCount; i++) {
			ShellCommand sc = new ShellCommand("dummy");
		}
		long elapsedTime = Calendar.getInstance().getTimeInMillis() - startTime;
		double timeToCreateOneObject = (double) elapsedTime
				/ (double) objectCount;
		logger.debug("Created " + objectCount + " objects in " + elapsedTime
				+ " ms");
		logger.debug("timeToCreateOneObject: " + timeToCreateOneObject + " ms");

		String goodOne = "cmd.exe /C dir";
		String badOne = "cmd.exe /C dira";
		ShellCommand goodCmd = new ShellCommand(goodOne);
		ShellCommand badCmd = new ShellCommand(badOne);
		try {
			if (goodCmd.exec() == 0) {
				logger.debug("Command '" + goodOne + "' succeeded.");
				logger.debug(goodCmd.getOutput());
			} else {
				logger.debug("Command '" + goodOne + "' failed.");
				logger.debug(goodCmd.getError());
			}
			if (badCmd.exec() == 0) {
				logger.debug("Command '" + badOne + "' succeeded.");
				logger.debug(badCmd.getOutput());
			} else {
				logger.debug("Command '" + badOne + "' failed.");
				logger.debug(badCmd.getError());
			}
		} catch (IOException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
