package org.eso.util.filesystem; 

import java.io.File; 
import java.io.FilenameFilter; 

import org.apache.log4j.Logger;

/** This class implements the {@link java.io.FilenameFilter} interface in a 
 * way which allows its users to filter filenames by their extension. For 
 * example, text files with names like "filename.txt" can be filtered for by 
 * specifying the ".txt" extension at construction time. See the Java 
 * {@link java.io.File#listFiles(java.io.FilenameFilter) File.listFiles()} 
 * method for a means of using these filters. 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class FileExtensionFilter implements FilenameFilter
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(FileExtensionFilter.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "FileExtensionFilter"; 

    /** The file extension accepted by this filter. */ 
    private final String fileExtension; 

    /** Constructs a new FileExtensionFilter. 
     * 
     * @throws NullPointerException if fileExtension is null. 
     * 
     * @param fileExtension the file extension this filter should accept. If 
     *                      the extension should follow a dot ('.' character), 
     *                      include the dot here. For example, to accept files 
     *                      with names like "filename.txt", specify ".txt". 
     */ 
    public FileExtensionFilter(String fileExtension) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::FileExtensionFilter()"; 

        logger.trace(methodLogName); 

        if(fileExtension == null)
        {
            final String message = 
                methodLogName + " - fileExtension must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Store the file extension to filter for. 
        this.fileExtension = fileExtension; 
    }

    /** Returns the file extension this FileExtensionFilter filters for. */ 
    public String getFileExtension() 
    {
        final String methodLogName = classLogName + "::getFileExtension()"; 

        logger.trace(methodLogName); 

        return(fileExtension); 
    }

    /** Returns true if the filename has the extension accepted by this 
     * FileExtensionFilter. The directory parameter is required by the 
     * FilenameFilter interface, but it is not used in this implementation. 
     * 
     * @throws NullPointerException if filename is null. 
     * 
     * @param directory the directory containing the file (not used). 
     * @param filename the name of the file to filter. 
     * 
     * @return true if the filename has the extension accepted by this 
     *         FileExtensionFilter, false otherwise. 
     */ 
    public boolean accept(File directory, String filename)
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::accept()"; 

        logger.trace(methodLogName); 

        if(filename == null)
        {
            final String message = 
                methodLogName + " - filename must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Check whether this filename has the appropriate extension or not, 
        // and return the result. 
        return(filename.endsWith(fileExtension)); 
    }
}
