package org.eso.util.misc;

import java.util.concurrent.Executors; 
import java.util.concurrent.ThreadFactory; 

import org.apache.log4j.Logger;

/** <p>Instances of this class can be used to create new threads in exactly 
 * the same way as a ThreadFactory obtained from 
 * {@link java.util.concurrent.Executors#defaultThreadFactory()} would, 
 * except that the new threads will be daemon threads. This can be useful 
 * in cases where, for example, a 
 * {@link java.util.concurrent.ScheduledThreadPoolExecutor}'s threads should 
 * not keep the JVM alive after an application's main thread has exited.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public class DefaultDaemonThreadFactory implements ThreadFactory 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = 
        Logger.getLogger(DefaultDaemonThreadFactory.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "DefaultDaemonThreadFactory"; 

    /** The default thread factory this instance wraps. */ 
    private final ThreadFactory defaultThreadFactory = 
        Executors.defaultThreadFactory(); 

    /** Constructs a new DefaultDaemonThreadFactory. */ 
    public DefaultDaemonThreadFactory()
    {
        final String methodLogName = 
            classLogName + "::DefaultDaemonThreadFactory()"; 

        logger.trace(methodLogName); 

        // This constructor does not need to do anything. 
    }

    /** Creates a new Thread to execute the specified Runnable, in 
     * exactly the same way as a ThreadFactory obtained from 
     * {@link java.util.concurrent.Executors#defaultThreadFactory()} 
     * would, except that the new thread will be a daemon thread. 
     * 
     * @param runnable the Runnable to be executed by the new thread. 
     * 
     * @returns a new daemon thread which will execute the specified 
     *          Runnable. 
     */ 
    public Thread newThread(Runnable runnable) 
    {
        final String methodLogName = classLogName + "::newThread()"; 

        logger.trace(methodLogName); 

        // Get the new thread from the default thread factory. 
        final Thread thread = defaultThreadFactory.newThread(runnable); 

        // Mark it as a daemon thread and return it. 
        thread.setDaemon(true); 
        return(thread); 
    }
}
