package org.eso.util.misc;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.eso.util.stream.Message;
import org.eso.util.stream.ProcessingTask;
import org.eso.util.stream.TerminatingTask;

/**
 * This task maintains a history of processed messages and computes some
 * statistical information (e.g. throughput) at regular intervals.
 * 
 * @author szampier/vforchi
 * 
 */

public class StatsAccumulatorTask extends ProcessingTask {

	/** Apache Log4J logger. */
	static final Logger logger = Logger.getLogger(StatsAccumulatorTask.class);

	/** The name of this class for logging purposes. */
	private static final String classLogName = "StatsAccumulatorTask";

	/** When this task was created. Used to display the application uptime. */
	final private long creationTime = System.currentTimeMillis();

	/** The total size of processed data, since the application was started. */
	private long totalSize = 0;

	/**
	 * The total number of messages processed, since the application was
	 * started.
	 */
	private int totalMessages = 0;

	/**
	 * The total number of elements processed, since the application was
	 * started.
	 */
	private int totalElements = 0;

	/**
	 * This class is a storage for some statistical information about each
	 * processed message.
	 */
	private class HistoryElement {
		/** When the message was received by the application. */
		final long startTime;
		/** When the message was processed by the application. */
		final long endTime;
		/** The message size (in bytes) */
		final long size;
		/** The number of elements contained in this message. */
		final int numElements;

		/** Constructor with startTime and size. */
		public HistoryElement(long startTime, long size) {
			this.startTime = startTime;
			this.endTime = System.currentTimeMillis();
			this.size = size;
			this.numElements = 1;
		}

		/** Constructor with startTime, size and numElements. */
		public HistoryElement(long startTime, long size, int numElements) {
			this.startTime = startTime;
			this.endTime = System.currentTimeMillis();
			this.size = size;
			this.numElements = numElements;
		}
	}

	/**
	 * This list maintains statistical information about the last N processed
	 * messages.
	 */
	LinkedList<HistoryElement> history = new LinkedList<HistoryElement>();

	/** The max history size. */
	private int historyDepth;

	/** The next time statistics should be computed and displayed. */
	private Date nextComputingTime;

	/** The time in seconds between stats computation. */
	private int computingPeriod;

	/** A list of time intervals over which the statistics are calculated */
	private List<Integer> intervals;

	/** A map used to store the max throughput calculated for each interval */
	private Map<Integer, Float> maxThroughput = new HashMap<Integer, Float>();

	private String hostname = null;

	/** Used to format compute stats times in log messages. */
	private final SimpleDateFormat dateFormat = new SimpleDateFormat(
			"yyyy-MM-dd'T'HH:mm:ss.SSS");

	// Some useful constants.
	final static int MILLIS_DAY = 24 * 60 * 60 * 1000;
	final static int MILLIS_HOUR = 60 * 60 * 1000;
	final static int MILLIS_MIN = 60 * 1000;
	final static int MILLIS_SEC = 1 * 1000;

	final static int KILOBYTE = 1024;
	final static int MEGABYTE = 1024 * 1024;
	final static int GIGABYTE = 1024 * 1024 * 1024;
	final static long TERABYTE = (long) 1024 * 1024 * 1024 * 1024;

	final static int MAX_HISTORY_DEPTH = 100000;

	/**
	 * Constructor.
	 * 
	 * @param computingPeriod
	 *            the time interval between two statistics computations
	 * @param computingIntervals
	 *            the intervals (in seconds) over which the statistics are
	 *            computed
	 * @throws NullPointerException
	 */
	public StatsAccumulatorTask(int computingPeriod,
			List<Integer> computingIntervals) throws NullPointerException {
		super("StatsAccumulatorTask", 1, computingPeriod);

		init(computingPeriod, computingIntervals, MAX_HISTORY_DEPTH);
	}

	/**
	 * Constructor
	 * 
	 * @param computingPeriod
	 *            the time interval between two statistics computations
	 * @param computingIntervals
	 *            the intervals (in seconds) over which the statistics are
	 *            computed
	 * @param historyDepth
	 *            the max number of messages stored
	 * @throws NullPointerException
	 */
	public StatsAccumulatorTask(int computingPeriod,
			List<Integer> computingIntervals, int historyDepth)
			throws NullPointerException {
		super("StatsAccumulatorTask", 1, computingPeriod);

		init(computingPeriod, computingIntervals, historyDepth);
	}

	/**
	 * real constructor
	 * 
	 * @param computingPeriod
	 * @param computingIntervals
	 * @param historyDepth
	 * @throws NullPointerException
	 */
	private void init(int computingPeriod, List<Integer> computingIntervals,
			int historyDepth) throws NullPointerException {
		final String methodLogName = classLogName + "::init()";
		logger.trace(methodLogName);

		// Check input parameters.
		if (computingPeriod <= 0) {
			throw new IllegalArgumentException(methodLogName
					+ " - computingPeriod must be greater than 0.");
		}

		if (historyDepth <= 0) {
			throw new IllegalArgumentException(methodLogName
					+ " - historyDepth must be greater than 0.");
		}

		try {
			hostname = InetAddress.getLocalHost().getHostName();
		} catch (UnknownHostException e) {
			logger.warn(methodLogName + " - could not get hostname : "
					+ e.getMessage());
			hostname = "UnknownHost";
		}

		nextComputingTime = new Date();
		this.computingPeriod = computingPeriod;
		this.historyDepth = historyDepth;
		this.intervals = computingIntervals;

		// Initialize max throughput
		for (int interval : intervals) {
			this.maxThroughput.put(interval, (float) 0.0);
		}
	}

	@Override
	protected boolean canProcessMessage(Message message) {
		final String methodLogName = classLogName + "::canProcessMessage()";
		logger.trace(methodLogName);

		// Check if the message implements the StatsProvider interface.
		try {
			StatsProvider info = (StatsProvider) message;
			return info.canProvideStats();
		} catch (ClassCastException e) {
			logger.warn(methodLogName + " - Cannot convert message: "
					+ e.getMessage());
			return false;
		}
	}

	@Override
	protected void process(List<Message> msgList) {
		final String methodLogName = classLogName + "::process()";
		logger.trace(methodLogName);

		// Check if it is time to compute statistics.
		if (System.currentTimeMillis() >= nextComputingTime.getTime()) {

			computeOverallStats();
			// Compute recent stats for each available interval.
			if (intervals != null) {
				for (int interval : intervals) {
					computeRecentStats(interval);
				}
			}

			// Set the next polling time.
			nextComputingTime
					.setTime((System.currentTimeMillis() + (computingPeriod * 1000)));

			logger.debug(methodLogName + " - next computing time: "
					+ dateFormat.format(nextComputingTime) + ".");

		}

		for (Iterator<Message> msgIter = msgList.iterator(); msgIter.hasNext();) {
			StatsProvider info = (StatsProvider) msgIter.next();

			// Update history.
			history.add(new HistoryElement(info.getStartTime(), info.getSize(),
					info.getNumElements()));
			// Prevent history to grow indefinitely.
			if (history.size() > historyDepth) {
				history.removeFirst();
			}
			// Update global statistics.
			totalMessages++;
			totalElements += info.getNumElements();
			totalSize += info.getSize();
		}
	}

	/**
	 * Compute some statistics for a given time period
	 * 
	 * @param seconds
	 *            statistics are evaluated over the last [seconds] secs
	 */
	private void computeRecentStats(int seconds) {
		final String methodLogName = classLogName + "::computeRecentStats()";
		logger.trace(methodLogName);

		if (history.isEmpty()) {
			return;
		}
		/* Statistics are evaluated backwards from the last history element until
		 * the starting time, computed as (NOW - seconds)
		 */
		long startingTime = System.currentTimeMillis() - seconds * MILLIS_SEC;

		if (startingTime < creationTime) {
			startingTime = creationTime;
		}

		long size = 0;
		int msgCount = 0;
		int elCount = 0;

		// Convert history list to array for more efficient navigation.
		HistoryElement[] historyArray = history
				.toArray(new HistoryElement[history.size()]);

		// start iteration from last inserted message
		for (int i = historyArray.length - 1; i >= 0; i--) {
			HistoryElement el = historyArray[i];

			/* check if the message is inside the computation interval */
			if (el.endTime > startingTime) {
				msgCount++;
				if (el.startTime > startingTime)
					size += el.size;
				else
				{
					/* take into account only the part of transmission during the computation period */
					long partialSize = (long) (((float)(el.endTime-startingTime)) / (el.endTime-el.startTime) * el.size);
					logger.debug("Total size: " + el.size + ", actual size: " + partialSize);
					size += partialSize;
				}
				elCount += el.numElements;
			} else {
				break;
			}
		}

		/* format message header */
		String tag = "last ";
		if (seconds == (3600 * 24))
			tag += "day stats";
		else if (seconds == 3600)
			tag += "hr stats";
		else if (seconds == 60)
			tag += "min stats";
		else if (seconds == 1)
			tag += "sec stats";
		else if (seconds % (3600 * 24) == 0)
			tag += seconds / (3600 * 24) + " days stats";
		else if (seconds % 3600 == 0)
			tag += seconds / 3600 + " hrs stats";
		else if (seconds % 60 == 0)
			tag += seconds / 60 + " mins stats";
		else
			tag += seconds + " secs stats";

		/* evaluate max throughput for this interval */
		float th = ((float) size) / KILOBYTE / seconds;
		if (th > maxThroughput.get(seconds)) {
			maxThroughput.put(seconds, th);
		}
		/* Format log message. */
		logger.info(format(tag, null, msgCount, elCount, size, seconds,
				maxThroughput.get(seconds)));
	}

	/**
	 * Compute statistics from startup
	 */
	private void computeOverallStats() {
		final String methodLogName = classLogName + "::computeOverallStats()";
		logger.trace(methodLogName);

		// Compute uptime and format it as : up <N> days <HH>:<MM>:<SS>
		long uptime = System.currentTimeMillis() - creationTime;
		int nDays = (int) uptime / MILLIS_DAY;
		int remainder = (int) uptime % MILLIS_DAY;
		int nHours = remainder / MILLIS_HOUR;
		remainder = remainder % MILLIS_HOUR;
		int nMinutes = remainder / MILLIS_MIN;
		remainder = remainder % MILLIS_MIN;
		int nSeconds = remainder / MILLIS_SEC;

		DecimalFormat df = new DecimalFormat("00");

		/* format uptime string */
		String uptimeStr = "host " + hostname + ", up " + nDays + " days "
				+ df.format(nHours) + ":" + df.format(nMinutes) + ":"
				+ df.format(nSeconds);

		String info = format("overall stats", uptimeStr, totalMessages,
				totalElements, totalSize, uptime / 1000, null);
		logger.info(info);
	}

	/***************************************************************************
	 * Format statistics in a string
	 * 
	 * @param tag
	 *            a tag representing the period over which the stats are
	 *            calculated
	 * @param custom
	 *            a custom string (used e.g. for uptime)
	 * @param messages
	 *            nummber of messages
	 * @param elements
	 *            number of items
	 * @param size
	 *            total size
	 * @param period
	 *            the period used for calculation
	 * @param max
	 *            max throughput
	 * @return
	 */
	private String format(String tag, String custom, int messages,
			int elements, long size, long period, Float max) {
		DecimalFormat df = new DecimalFormat("0.###");
		// Format the total number of processed messages.
		String totalMessagesStr = messages + " msgs";
		// Format the total number of processed Elements.
		String totalElementsStr = elements + " items";

		// Human readable representation of total processed size.
		String totalSizeStr;
		if (size > TERABYTE) {
			totalSizeStr = "" + df.format((float) size / TERABYTE) + " TB";
		} else if (size > GIGABYTE) {
			totalSizeStr = "" + df.format((float) size / GIGABYTE) + " GB";
		} else if (size > MEGABYTE) {
			totalSizeStr = "" + df.format((float) size / MEGABYTE) + " MB";
		} else {
			totalSizeStr = "" + df.format((float) size / KILOBYTE) + " KB";
		}

		// Compute overall throughput in KB/s.
		String throughputStr = "" + df.format((float) size / KILOBYTE / period)
				+ " KB/s";

		// Assemble statistical info in one line message.
		String infoStr = "[" + tag + "] ";
		if (custom != null) {
			infoStr += custom + ", ";
		}
		infoStr += totalMessagesStr + ", " + totalElementsStr + ", "
				+ totalSizeStr + ", " + throughputStr;
		if (max != null) {
			infoStr += ", max " + df.format(max) + " KB/s";
		}
		return infoStr;
	}

	/**
	 * @return the history depth.
	 */
	public int getHistoryDepth() {
		return historyDepth;
	}

	/**
	 * Set history depth.
	 * 
	 * @param historyDepth
	 */
	public void setHistoryDepth(int historyDepth) {
		this.historyDepth = historyDepth;
	}

	/**
	 * Test program.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		// Basic Log4j config
		BasicConfigurator.configure();
		Logger utilLogger = Logger.getLogger("org.eso.util");
		utilLogger.setLevel(Level.INFO);

		class TestMessage extends Message implements StatsProvider {
			long creationTime = System.currentTimeMillis() - 500;
			final long size;
			final int numElements;

			public TestMessage(int numElements, long size) {
				super(MSG_USER);
				this.numElements = numElements;
				this.size = size;
			}

			public boolean canProvideStats() {
				return true;
			}

			public int getNumElements() {
				return numElements;
			}

			public long getSize() {
				return size;
			}

			public long getStartTime() {
				return creationTime;
			}
		}

		List<Integer> p = new LinkedList<Integer>();
		p.add(2);
		//p.add(60);
		//p.add(3600);
		StatsAccumulatorTask task = new StatsAccumulatorTask(1, p);
		task.setNextTask(new TerminatingTask());
		task.open();
		for (int i = 0; i < 2000; i++) {
			TestMessage msg = new TestMessage(1, 100);
			task.put(msg);
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

}
