package org.eso.util.stream; 

import java.util.concurrent.LinkedBlockingQueue; 
import java.util.concurrent.TimeUnit; 

import org.apache.log4j.Logger;

/** A HoldingTask accepts messages into a queue, but does not process or 
 * forward them. It provides a way of storing stream messages for other 
 * parts of an application to retrieve at their convenience (using the 
 * get() method). As this task does not forward messages, the next task 
 * need not be set. 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class HoldingTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(HoldingTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "HoldingTask"; 

    /** A queue of messages to be held by this task. */
    private final LinkedBlockingQueue<Message> msgQueue = 
        new LinkedBlockingQueue<Message>(); 

    /** Constructs a new HoldingTask with the name "HoldingTask". */ 
    public HoldingTask()
    {
        super("HoldingTask"); 

        final String methodLogName = 
            classLogName + "::HoldingTask() [for task HoldingTask]"; 

        logger.trace(methodLogName); 
    }

    /** Constructs a new HoldingTask with the specified name. 
     *
     * @param taskName the name of this task. 
     */
    public HoldingTask(final String taskName)
    {
        super(taskName); 

        final String methodLogName = 
            classLogName 
            + "::HoldingTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
    }
    
    /** {@inheritDoc} Delegates to {@link #get(long timeout)} with the timeout 
     * set to zero. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to get(long). 
        return(get(0)); 
    }

    /** Removes the next message from this task's message queue and returns 
     * it. If this thread is interrupted whilst waiting for a message, its 
     * interrupt status will be set and null will be returned. 
     *
     * @param timeout the time this method should wait for a message to arrive 
     *                on the message queue. Specified in seconds. Set to zero 
     *                if this method should block until a message arrives. 
     *
     * @return the next message from this task's message queue, or null if the 
     *         queue was empty and the timeout has expired, or null if this 
     *         thread was interrupted whilst waiting for a message. 
     */
    public final Message get(final long timeout)
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName);

        Message message = null;         // A newly received message. 

        try
        {
            if(timeout == 0) 
            {
                // Block until a message arrives. 
                message = msgQueue.take(); 
            }
            else
            {
                // Block until a message arrives or the timeout expires. 
                message = msgQueue.poll(timeout, TimeUnit.SECONDS); 
            }
        }
        catch(InterruptedException ie)
        {
            // This thread was interrupted whilst waiting for a message; 
            // set its interrupt status. 
            logger.debug(
                methodLogName 
                + " - interrupt received whilst waiting for message."); 
            
            Thread.currentThread().interrupt(); 
        }

        if(message == null)
        {
            logger.debug(
                methodLogName + " - no message retrieved."); 
        }
        else
        {
            logger.debug(
                methodLogName + " - message retrieved."); 
        }

        return(message); 
    }

    /** Queues a message in this task for retrieval with get(). This method 
     * will not return until the message has been successfully queued. Thread 
     * interrupts will be noted, but ignored until the message is queued. If 
     * an interrupt occurs, this thread's interrupt status will be set before 
     * this method returns. 
     *
     * @param message a message to be queued in this task. 
     */ 
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // True if this thread has been interrupted. 
        boolean interruptDetected = Thread.interrupted(); 

        // The message must be successfully queued. 
        while(true)
        {
            try
            {
                // Try to queue the message. 
                msgQueue.put(message); 

                logger.debug(
                    methodLogName + " - added message to queue."); 
            }
            catch(InterruptedException ie)
            {
                // This thread was interrupted whilst waiting to queue the 
                // message - note the interrupt and try again. 
                logger.debug(methodLogName 
                             + " - interrupt received whilst waiting" 
                             + " to queue a message."); 

                interruptDetected = true; 
                continue; 
            }

            // The message was successfully queued. 
            break;
        } 

        // If an interrupt was detected, set this thread's interrupt status. 
        if(interruptDetected) 
        {
            Thread.currentThread().interrupt(); 
        }
    }
    
    /** HoldingTasks do not perform putNext() operations as such, as they 
     * have no next task to use as a message destination. This method is an 
     * alias for the {@link #put(Message message)} method. 
     * 
     * @param message the message. 
     */
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - delegating to put()."); 

        put(message); 
    }
}
