package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** <p>Instances of this class represent messages which can be sent along a 
 * stream. All messages have a type which tasks can check when deciding what 
 * to do with them. Standard message types are defined here; subclasses have 
 * the option of defining further types starting from {@link #MSG_USER}.</p> 
 * 
 * <p>NOTE: not all tasks will support all of the actions defined by the 
 * standard message types. However, tasks which have their own task threads 
 * must comply with messages of type {@link #MSG_SHUTDOWN}.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class Message 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(Message.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "Message"; 

    /** Indicates that the receiving task must stop processing. {@link Task}s 
     * which have their own task threads are required to forcibly terminate 
     * those threads upon receiving a shutdown message. Messages of this type 
     * must be forwarded to subsequent tasks in a stream. 
     */
    public static final int MSG_SHUTDOWN = 0; 

    /** Indicates that the receiving task should complete any pending 
     * operations and enter a quiescent state. The receiving task may 
     * allow reactivation (see {@link #MSG_CONTINUE}), or it may choose 
     * to close itself (by analogy to {@link #MSG_SHUTDOWN}). Messages 
     * of this type should be forwarded to subsequent tasks in a stream. 
     */
    public static final int MSG_FINISH = MSG_SHUTDOWN + 1; 

    /** Indicates that the receiving task should immediately enter a 
     * quiescent state. The receiving task should allow reactivation 
     * (see {@link #MSG_CONTINUE}). Messages of this type should be 
     * forwarded to subsequent tasks in a stream. 
     */ 
    public static final int MSG_STANDBY = MSG_FINISH + 1; 

    /** Indicates that the receiving task should enter an active state and 
     * continue processing. See {@link #MSG_FINISH} and {@link #MSG_STANDBY}. 
     * Messages of this type should be forwarded to subsequent tasks in a 
     * stream. 
     */ 
    public static final int MSG_CONTINUE = MSG_STANDBY + 1; 

    /** Indicates that the originating task has encountered a recoverable 
     * error. Messages of this type should be forwarded to subsequent tasks 
     * in a stream, so that they can eventually be returned to the stream's 
     * controlling application. 
     */
    public static final int MSG_ERROR = MSG_CONTINUE + 1; 

    /** Indicates that the originating task has suffered an irretrievable 
     * error. Messages of this type should be forwarded to subsequent tasks 
     * in a stream, so that they can eventually be returned to the stream's 
     * controlling application. 
     */
    public static final int MSG_FATAL_ERROR = MSG_ERROR + 1; 

    /** Subclasses of Message are free to define message types starting from 
     * this value. Note that two subclasses are free to use the same message 
     * type values to mean different things; only the 
     * non-{@link Message#MSG_USER MSG_USER} values defined in the Message 
     * class are considered reserved. 
     */
    public static final int MSG_USER = MSG_FATAL_ERROR + 1; 

    /** The type of this message. */ 
    private final int messageType; 

    /** Constructs a new Message with the specified type. 
     * 
     * @param messageType the type this message should have. 
     */
    public Message(final int messageType) 
    {
        final String methodLogName = classLogName + "::Message()"; 

        logger.trace(methodLogName); 

        this.messageType = messageType; 
    }

    /** Returns the type of this message. 
     *
     * @return the type of this message. 
     */ 
    public final int getMessageType() 
    {
        final String methodLogName = classLogName + "::getMessageType()"; 

        logger.trace(methodLogName); 

        return(messageType); 
    }

    /** Returns a string describing the type of this message. 
     *
     * @return a string describing the type of this message. 
     */ 
    public String getMessageTypeString() 
    {
        final String methodLogName = 
            classLogName + "::getMessageTypeString()"; 

        logger.trace(methodLogName); 

        String result; 

        switch(messageType) 
        {
          case MSG_SHUTDOWN: 
            result = "SHUTDOWN"; 
            break; 

          case MSG_FINISH:
            result = "FINISH";
            break; 

          case MSG_STANDBY:
            result = "STANDBY"; 
            break; 

          case MSG_CONTINUE:
            result = "CONTINUE"; 
            break; 

          case MSG_ERROR: 
            result = "ERROR"; 
            break; 

          case MSG_FATAL_ERROR: 
            result = "FATAL ERROR"; 
            break; 

          case MSG_USER: 
            result = "USER"; 
            break; 

          default: 
            result = "UNKNOWN (" + messageType + ")"; 
            break; 
        }

        return(result); 
    }

    /** <p>Returns true if this message permits fast forwarding.</p> 
     * 
     * <p>In order to prevent ordering problems within a stream, system 
     * messages never permit fast forwarding. For example, a message of 
     * type MSG_SHUTDOWN is normally the last message to pass through 
     * a stream; as such, it should not be allowed to overtake other 
     * messages by way of fast forwarding. Subclasses are free to 
     * override this method if they do permit fast forwarding.</p> 
     *
     * @return false. 
     */ 
    public boolean permitsFastForwarding() 
    {
        final String methodLogName = 
            classLogName + "::permitsFastForwarding()"; 

        logger.trace(methodLogName); 

        logger.debug(methodLogName 
                     + " - message does not permit fast forwarding."); 

        return(false); 
    }
}
