package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** <p>Modules are small objects which group pairs of {@link Task} objects 
 * together, corresponding to the downstream and upstream directions in a 
 * stream. A module provides facilities for connecting its tasks to the 
 * tasks of adjacent modules in a stream; the module's tasks can also 
 * discover each other. The latter capability is used when a task needs 
 * to pass a message to its counterpart on the other side of a stream.</p> 
 *
 * <p>It is unlikely that Module will need to be subclassed, or used 
 * directly by a stream-based application; subclasses of {@link Stream} 
 * and {@link Task} should be able to provide most specialisations that 
 * might be needed.</p> 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class Module
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(Module.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "Module"; 

    /** True if this module has been opened by calling open(), and close() 
     * has not been called since the last time open() was called. 
     */ 
    private boolean moduleOpen = false; 

    /** The downstream task for this module. */ 
    private final Task downstreamTask; 

    /** The upstream task for this module. */ 
    private final Task upstreamTask; 

    /** Constructs a new Module having the specified upstream and downstream 
     * tasks. If either task is null, a PassThroughTask will be created for 
     * that stream direction. 
     *
     * @param downstreamTask   the downstream task for this module. 
     * @param upstreamTask     the upstream task for this module. 
     */ 
    public Module(Task downstreamTask, Task upstreamTask) 
    {
        final String methodLogName = classLogName + "::Module()"; 

        logger.trace(methodLogName); 

        // Module tasks may not be null. If no task was specified for a given 
        // stream direction, that task becomes a PassThroughTask. 
        if(downstreamTask == null)
        {
            downstreamTask = new PassThroughTask(); 
        }

        if(upstreamTask == null)
        {
            upstreamTask = new PassThroughTask(); 
        }

        // Inform both tasks that they now have a containing module. 
        downstreamTask.setContainingModule(this); 
        upstreamTask.setContainingModule(this); 

        // Having set the tasks up, store references to them. 
        this.downstreamTask = downstreamTask; 
        this.upstreamTask = upstreamTask; 
    }

    /** Performs any operations required to start/open/prepare this module
     * for use after it has been constructed. For example, open() will be 
     * called on both of this module's tasks. 
     */
    public final synchronized void open()
    {
        final String methodLogName = classLogName + "::open()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - opening module."); 

        // Open both tasks. 
        downstreamTask.open(); 
        upstreamTask.open(); 

        moduleOpen = true; 
    }

    /** Returns true if this module is open. 
     *
     * @return true if this module is open, false otherwise. */
    public final boolean isOpen()
    {
        final String methodLogName = classLogName + "::isOpen()"; 

        logger.trace(methodLogName); 

        return(moduleOpen); 
    }

    /** Performs any operations required to stop/close/prevent use of this 
     * module. For example, close() will be called on both of this module's 
     * tasks. 
     */
    public final synchronized void close()
    {
        final String methodLogName = classLogName + "::close()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - closing module."); 

        // Close both tasks. 
        downstreamTask.close(); 
        upstreamTask.close(); 

        moduleOpen = false; 
    }

    /** Restarts this module by calling close(), then open(). */ 
    public final synchronized void restart()
    {
        final String methodLogName = classLogName + "::restart()"; 

        logger.trace(methodLogName);
        logger.debug(methodLogName + " - restarting module."); 

        close(); 
        open(); 
    }

    /** Connects this module to the specified downstream module. 
     * 
     * @throws AssertionError if called while this module is open. 
     * 
     * @param module the downstream module this module should be connected to. 
     */ 
    public final void connectDownstreamModule(Module module) 
    {
        final String methodLogName = 
            classLogName + "::connectDownstreamModule()"; 

        logger.trace(methodLogName);

        // Ensure that the module is closed. 
        if(moduleOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - module must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        logger.debug(methodLogName + " - connecting."); 

        // This module's downstream task will forward messages to the 
        // downstream module's downstream task. 
        downstreamTask.setNextTask(module.downstreamTask); 
    }

    /** Connects this module to the specified upstream module. 
     * 
     * @throws AssertionError if called while this module is open. 
     * 
     * @param module the upstream module this module should be connected to. 
     */
    public final void connectUpstreamModule(Module module) 
    {
        final String methodLogName = 
            classLogName + "::connectUpstreamModule()"; 

        logger.trace(methodLogName);

        // Ensure that the module is closed. 
        if(moduleOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - module must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        logger.debug(methodLogName + " - connecting."); 

        // This module's upstream task will forward messages to the upstream 
        // module's upstream task. 
        upstreamTask.setNextTask(module.upstreamTask); 
    }

    /** Disconnect the downstream module, if any, and return it. 
     *
     * @throws AssertionError if called while this module is open. 
     * 
     * @return the former downstream module, or null if no downstream module 
     *         was set. 
     */
    public final Module disconnectDownstreamModule()
    {
        final String methodLogName = 
            classLogName + "::disconnectDownstreamModule()"; 

        logger.trace(methodLogName); 

        // Ensure that the module is closed. 
        if(moduleOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - module must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // The current downstream module, if any. 
        Module oldDownstreamModule = null;

        // Get the next downstream task, if any. 
        final Task oldDownstreamNextTask = downstreamTask.getNextTask(); 

        if(oldDownstreamNextTask != null)
        {
            // The next task downstream is set; get its containing module. 
            oldDownstreamModule = oldDownstreamNextTask.getContainingModule(); 
        }

        // Perform the actual disconnection. 
        logger.debug(methodLogName + " - disconnecting."); 

        downstreamTask.setNextTask(null); 

        // Return the former downstream module. 
        return(oldDownstreamModule); 
    }

    /** Disconnect the upstream module, if any, and return it. 
     *
     * @throws AssertionError if called while this module is open. 
     * 
     * @return the former upstream module, or null if no upstream module 
     *         was set. 
     */
    public final Module disconnectUpstreamModule()
    {
        final String methodLogName = 
            classLogName + "::disconnectUpstreamModule()"; 

        logger.trace(methodLogName); 

        // Ensure that the module is closed. 
        if(moduleOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - module must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // The current upstream module, if any. 
        Module oldUpstreamModule = null;

        // Get the next upstream task, if any. 
        final Task oldUpstreamNextTask = upstreamTask.getNextTask(); 

        if(oldUpstreamNextTask != null)
        {
            // The next upstream task is set; get its containing module. 
            oldUpstreamModule = oldUpstreamNextTask.getContainingModule(); 
        }

        // Perform the actual disconnection. 
        logger.debug(methodLogName + " - disconnecting."); 

        upstreamTask.setNextTask(null); 

        // Return the former upstream module. 
        return(oldUpstreamModule); 
    }

    /** Returns the downstream task of this module. 
     *
     * @return the downstream task of this module. 
     */
    public final Task getDownstreamTask()
    {
        final String methodLogName = classLogName + "::getDownstreamTask()"; 

        logger.trace(methodLogName); 

        return(downstreamTask); 
    }

    /** Returns the upstream task of this module. 
     *
     * @return the upstream task of this module. 
     */
    public final Task getUpstreamTask()
    {
        final String methodLogName = classLogName + "::getUpstreamTask()"; 

        logger.trace(methodLogName); 

        return(upstreamTask); 
    }

    /** Returns the sibling task of the specified task in this module. If 
     * task is the downstream task of this module, the upstream task will 
     * be returned, and vice versa. If the specified task is not one of 
     * this module's tasks, a null reference will be returned. 
     *
     * @param task the task whose sibling task should be returned. 
     * 
     * @return the sibling task, or null if the specified task is not 
     *         actually one of this module's tasks. 
     */
    public final Task getSiblingTask(Task task)
    {
        final String methodLogName = classLogName + "::getSiblingTask()"; 

        logger.trace(methodLogName);

        if(task == downstreamTask)
        {
            return(upstreamTask); 
        }

        if(task == upstreamTask)
        {
            return(downstreamTask); 
        }

        // The specified task is not one of this module's tasks. 
        return(null); 
    }
}
