package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** RoutingTask is an abstract class providing common functionality 
 * to subclasses which perform stream connection functions. Please 
 * note that RoutingTasks are not supposed to process message content. 
 * Tasks which process data carried by messages should subclass the 
 * {@link ProcessingTask} class, not this one. 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public abstract class RoutingTask implements Task 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(RoutingTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "RoutingTask"; 

    /** The module containing this task (if any). */
    private Module containingModule = null; 

    /** The name of this task. */ 
    private final String taskName; 

    /** True if this task has been opened by calling open(), and close() 
     * has not been called since the last time open() was called. 
     */ 
    private boolean taskOpen = false; 

    /** The task which should receive any messages generated by this task. */ 
    private Task nextTask = null; 

    /** Constructs a new RoutingTask with the name "RoutingTask". */ 
    public RoutingTask()
    {
        final String methodLogName = 
            classLogName + "::RoutingTask() [for task RoutingTask]"; 

        logger.trace(methodLogName); 

        taskName = "RoutingTask"; 
    }

    /** Constructs a new RoutingTask with the specified name. 
     *
     * @throws NullPointerException if taskName is null. 
     *
     * @param taskName the name of this task. 
     */
    public RoutingTask(final String taskName) throws NullPointerException 
    {
        final String methodLogName = 
            classLogName 
            + "::RoutingTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        if(taskName == null) 
        {
            final String message = 
                methodLogName + " - taskName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        this.taskName = taskName; 
    }

    // Performs any operations required to start/open/prepare this task for 
    // use after it has been constructed.
    // 
    // Documentation is inherited from the Task interface. 
    public final void open() 
    {
        final String methodLogName = 
            classLogName + "::open() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - opening task."); 

        taskOpen = true; 
    }
    
    // Returns true if this task is open. 
    // 
    // Documentation is inherited from the Task interface. 
    public final boolean isOpen()
    {
        final String methodLogName = 
            classLogName + "::isOpen() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        return(taskOpen); 
    }

    // Performs any operations required to stop/close/prevent use of this 
    // task. 
    // 
    // Documentation is inherited from the Task interface. 
    public final void close() 
    {
        final String methodLogName = 
            classLogName + "::close() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - closing task."); 

        taskOpen = false; 
    } 

    // Restarts this task by calling close(), then open(). 
    // 
    // Documentation is inherited from the Task interface. 
    public final void restart() 
    {
        final String methodLogName = 
            classLogName + "::restart() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - restarting task."); 

        close(); 
        open(); 
    }

    // These methods inherit documentation from the Task interface. 
    public abstract Message get(); 
    public abstract Message get(final long timeout); 
    public abstract void put(final Message message); 
    public abstract void putNext(final Message message); 

    // Returns the name of this task. 
    // 
    // Documentation is inherited from the Task interface. 
    public final String getTaskName() 
    {
        final String methodLogName = 
            classLogName + "::getTaskName() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        return(taskName); 
    }

    // Returns a reference to the task following this one in a stream. 
    // 
    // Documentation is inherited from the Task interface. 
    public final Task getNextTask() 
    {
        final String methodLogName = 
            classLogName + "::getNextTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        return(nextTask); 
    }
    
    /** {@inheritDoc} This method may not be called while this task is open. 
     *
     * @throws AssertionError if called while this task is open. 
     */ 
    public final Task setNextTask(final Task task)
    {
        final String methodLogName = 
            classLogName + "::setNextTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        // Ensure that the task is closed. 
        if(isOpen()) 
        {
            final String logMessage = 
                methodLogName 
                + " - task must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // Store the old next task to return, then set the new next task. 
        final Task oldTask = nextTask; 
        nextTask = task; 

        logger.debug(methodLogName + " - next task set [previously: "
                     + (oldTask == null 
                        ? "null][now: " : oldTask.getTaskName() + "][now: ") 
                     + (nextTask == null 
                        ? "null]." : nextTask.getTaskName() + "].")); 

        return(oldTask); 
    }

    // Returns a reference to the module containing this task. 
    // 
    // Documentation is inherited from the Task interface. 
    public final Module getContainingModule()
    {
        final String methodLogName = 
            classLogName 
            + "::getContainingModule() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        return(containingModule); 
    }

    /** {@inheritDoc} This method may not be called while this task is open. 
     *
     * @throws AssertionError if called while this task is open. 
     */ 
    public final Module setContainingModule(final Module module) 
    {
        final String methodLogName = 
            classLogName 
            + "::setContainingModule() [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        // Ensure that the task is closed. 
        if(isOpen()) 
        {
            final String logMessage = 
                methodLogName 
                + " - task must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // Store the old containing module to return, then set the new 
        // containing module. 
        final Module oldModule = containingModule; 
        containingModule = module; 

        logger.debug(methodLogName + " - containing module set."); 

        return(oldModule); 
    }
}
