package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** <p>A SystemPassThroughTask acts as a filter in a stream. All messages 
 * having a non-{@link Message#MSG_USER MSG_USER} type which is defined in 
 * {@link Message} are immediately forwarded to the next task. All other 
 * messages are dropped and will be garbage collected if no other references 
 * to them exist.</p> 
 *
 * <p>The next task must be set before using an instance of this class.</p> 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class SystemPassThroughTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = 
        Logger.getLogger(SystemPassThroughTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "SystemPassThroughTask"; 

    /** Constructs a new SystemPassThroughTask with the name 
     * "SystemPassThroughTask". 
     */ 
    public SystemPassThroughTask()
    {
        super("SystemPassThroughTask");

        final String methodLogName = 
            classLogName 
            + "::SystemPassThroughTask() [for task SystemPassThroughTask]"; 

        logger.trace(methodLogName); 
    }

    /** Constructs a new SystemPassThroughTask with the specified name. 
     *
     * @param taskName the name of this task. 
     */
    public SystemPassThroughTask(final String taskName)
    {
        super(taskName);

        final String methodLogName = 
            classLogName 
            + "::SystemPassThroughTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get(final long timeout) 
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** {@inheritDoc} All messages are handed off to a 
     * {@link #putNext(Message message)} call. 
     */
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to putNext(). 
        putNext(message); 
    }

    /** Messages having a non-{@link Message#MSG_USER MSG_USER} type which 
     * is defined in {@link Message} are immediately forwarded to the next 
     * task. All other messages are dropped and will be garbage collected 
     * if no other references to them exist. 
     * 
     * @param message the message to be dropped or sent to the next task. 
     */ 
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        if(message.getMessageType() < Message.MSG_USER) 
        {
            logger.debug(methodLogName 
                         + " - forwarding system message to next task."); 

            getNextTask().put(message); 
        }
        else
        {
            logger.debug(methodLogName + " - dropping non-system message."); 
        }
    }
}
