package org.eso.util.stream; 

import org.apache.log4j.Logger; 

/** A TerminatingTask does not hold, modify or forward messages; it simply 
 * lets them fall out of scope and be garbage collected (assuming no other 
 * references to them exist). It can be used to terminate a stream where 
 * no messages are returned to the controlling application. As this task 
 * does not forward messages, the next task need not be set. 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class TerminatingTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(TerminatingTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "TerminatingTask"; 

    /** Constructs a new TerminatingTask with the name "TerminatingTask". */ 
    public TerminatingTask()
    {
        super("TerminatingTask"); 

        final String methodLogName = 
            classLogName + "::TerminatingTask() [for task TerminatingTask]"; 

        logger.trace(methodLogName); 
    }

    /** Constructs a new TerminatingTask with the specified name. 
     *
     * @param taskName the name of this task. 
     */
    public TerminatingTask(final String taskName)
    {
        super(taskName); 

        final String methodLogName = 
            classLogName 
            + "::TerminatingTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
    }
    
    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get(final long timeout) 
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** {@inheritDoc} Messages are immediately dropped using a 
     * {@link #putNext(Message message)} call. 
     */
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to putNext(). 
        putNext(message); 
    }

    /** Drops the specified message, allowing it to be garbage collected 
     * if no other references to it exist. 
     * 
     * @param message the message to be dropped. 
     */
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - dropping message."); 
    }
}
