package org.eso.util.dal;

import java.io.File;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.BasicConfigurator;
import org.eso.util.dal.TransferRequest.Status;
import org.eso.util.filesystem.FileSystemOperations;

import junit.framework.TestCase;

public class TransferRequestDAOFileTest extends TestCase {

    final static String TR_FILE_EXT = ".trq";

    private File incomingDirectory = null;
    private File processingDirectory = null;
    private File processedDirectory = null;
    private File rejectedDirectory = null;
    private TransferRequest transferRequest = null;
    private TransferRequestDAO trDAO = null;
    private TransferRequestDAO oneStateDAO = null;

    public TransferRequestDAOFileTest(String name) {
        super(name);

        // Basic Log4j configuration
        BasicConfigurator.configure();

        // Directory structure set up
        incomingDirectory = new File("/tmp/incoming");
        if (!incomingDirectory.exists()) {
            incomingDirectory.mkdirs();
        }
        processingDirectory = new File("/tmp/processing");
        if (!processingDirectory.exists()) {
            processingDirectory.mkdirs();
        }
        processedDirectory = new File("/tmp/processed");
        if (!processedDirectory.exists()) {
            processedDirectory.mkdirs();
        }
        rejectedDirectory = new File("/tmp/rejected");
        if (!rejectedDirectory.exists()) {
            rejectedDirectory.mkdirs();
        }
    }

    protected void finalize() throws Throwable {
        super.finalize(); // not necessary if extending Object.

        // removes directory structure
        if (incomingDirectory.exists()) {
            incomingDirectory.delete();
        }
        if (processingDirectory.exists()) {
            processingDirectory.delete();
        }
        if (processedDirectory.exists()) {
            processedDirectory.delete();
        }
        if (rejectedDirectory.exists()) {
            rejectedDirectory.delete();
        }

    }

    protected void setUp() throws Exception {
        super.setUp();

        // DAO set up
        trDAO =
                new TransferRequestDAOFile(incomingDirectory,
                        processingDirectory,
                        processedDirectory,
                        rejectedDirectory,
                        TR_FILE_EXT,
                        new DefaultTransferRequestValidator());

        oneStateDAO =
                new TransferRequestDAOFile(incomingDirectory,
                        TR_FILE_EXT,
                        new DefaultTransferRequestValidator());

        // Transfer request set up
        transferRequest = new TransferRequest();
        transferRequest.setFileId("FILEID");
        transferRequest.setFileName("2008/FILENAME.fits.Z");
        transferRequest.setSize(new Long(1000));
        transferRequest.setChecksum("1000xA");
        transferRequest.setCompressionMethod("TESTMETHOD");
        transferRequest.setDestDir("TESTDESTDIR");
        transferRequest.setMultiFileComponents("COMP1,COMP2,COMP3");
        transferRequest.setTransferCatg("CATGTEST");
        transferRequest.setTransferMethod("TRMETHTEST");
        transferRequest.setTransferPriority(new Integer(1));
        transferRequest.setUncompressedFileName("UNCOMPTEST");
        transferRequest.setUncompressedSize(new Long(1000));
        transferRequest.setUser("TESTUSER");
    }

    protected void tearDown() throws Exception {
        super.tearDown();
    }

    /** Nominal case : retrieves a set of existing files in the incoming status */
    public void testFindByStatus_01() {

        // a created request is in the INCOMING state
        // checks afterwards if it can be retrieved in the associated
        // directory
        try {
            trDAO.save(transferRequest);

            List<TransferRequest> trList = trDAO.findByStatus(Status.INCOMING);
            Iterator<TransferRequest> itTr = trList.iterator();
            Boolean trFound = new Boolean(false);
            while (itTr.hasNext()) {
                TransferRequest curTr = itTr.next();
                if (curTr.getFileId().compareTo(transferRequest.getFileId()) == 0) {
                    trFound = new Boolean(true);
                    break;
                }
            }
            assertTrue("File in the INCOMING status successfully retrieved",
                    trFound);
            trDAO.delete(transferRequest);
        } catch (DAOException e) {
            // TODO Auto-generated catch block
            fail("A DAOException has been raised");
        }
    }

    /**
     * Looking for files in a non handled state
     */
    public void testFindByStatus_02() {

        // Only the incoming state is managed
        try {
            oneStateDAO.findByStatus(Status.PROCESSING);
            fail("No exception has been raised");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /**
     * No transfer request is provided
     */
    public void testFindByStatus_03() {
        try {
            trDAO.findByStatus(null);
            fail("FAILED No exception has been raised");
        } catch (IllegalArgumentException ie) {
            // expected behaviour
        } catch (DAOException daoe) {
            fail("a DAOException has been raised " + daoe.getMessage());
        }
    }

    /**
     * Nominal case : updating the status of an incoming transfer request to
     * processing
     */
    public void testUpdateStatus_01() {
        try {
            trDAO.save(transferRequest);
            // The file will be moved to the PROCESSING DIRECTORY
            trDAO.updateStatus(transferRequest,
                    TransferRequest.Status.PROCESSING);
            // Checks if the file is available in the PROCESSING status and not
            // in the INCOMING status
            List<TransferRequest> trList =
                    trDAO.findByStatus(Status.PROCESSING);
            Iterator<TransferRequest> itTr = trList.iterator();
            Boolean trFound = new Boolean(false);
            while (itTr.hasNext()) {
                TransferRequest curTr = itTr.next();
                if (curTr.getFileId().compareTo(transferRequest.getFileId()) == 0) {
                    trFound = new Boolean(true);
                    break;
                }
            }
            assertTrue("Transfer request is in the PROCESSING status", trFound);

            trList = trDAO.findByStatus(Status.INCOMING);
            itTr = trList.iterator();
            trFound = new Boolean(false);
            while (itTr.hasNext()) {
                TransferRequest curTr = itTr.next();
                if (curTr.getFileId().compareTo(transferRequest.getFileId()) == 0) {
                    trFound = new Boolean(true);
                    break;
                }
            }
            assertTrue("Transfer request is no more in the INCOMING status",
                    (trFound == false));
            trDAO.delete(transferRequest);
        } catch (DAOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /** Invalid transfer request : file id is not set */
    public void testUpdateStatus_02() {
        transferRequest.setFileId(null);
        try {
            trDAO.updateStatus(transferRequest,
                    TransferRequest.Status.PROCESSING);
            fail("no exception was raised");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /** transfer request without status */
    public void testUpdateStatus_03() {
        try {
            trDAO.save(transferRequest);
            transferRequest.setStatus(null);
            trDAO.updateStatus(transferRequest,
                    TransferRequest.Status.PROCESSING);
            fail("no exception was raised");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /** transfer request status not handled */
    public void testUpdateStatus_04() {
        try {
            oneStateDAO.save(transferRequest);
            transferRequest.setStatus(Status.PROCESSING);
            oneStateDAO.updateStatus(transferRequest,
                    TransferRequest.Status.INCOMING);
            fail("no exception was raised");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /** Transfer request file does not exist */
    public void testUpdateStatus_05() {
        try {
            trDAO.updateStatus(transferRequest, TransferRequest.Status.INCOMING);
            fail("No exception was raised");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /** destination status does not exist */
    public void testUpdateStatus_06() {
        try {
            transferRequest = new TransferRequest();
            oneStateDAO.save(transferRequest);
            oneStateDAO.updateStatus(transferRequest,
                    TransferRequest.Status.PROCESSING);
            fail(" no exception was raised");
        } catch (DAOException daoe) {
            // expected behaviour
        } finally {
            try {
                oneStateDAO.delete(transferRequest);
            } catch (DAOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }

    /**
     * No transfer request provided
     * 
     * @throws DAOException
     */
    public void testUpdateStatus_07() {
        try {
            trDAO.updateStatus(null, Status.INCOMING);
            fail("no exception has been raised");
        } catch (IllegalArgumentException ie) {
            // expected behaviour
        } catch (DAOException e) {
            fail("A DAOException has been raised");
        }
    }

    /** No status provided */
    public void testUpdateStatus_08() {
        try {
            trDAO.save(transferRequest);
            trDAO.updateStatus(transferRequest, null);
            fail("no exception has been raised");
        } catch (IllegalArgumentException ie) {
            // expected behaviour
        } catch (DAOException e) {
            fail("A DAOException has been raised");
        }
    }

    /**
     * Nominal case : creation of a transfer request file
     * 
     */
    public void testSave_01() {
        int nbSavedFile = 0;
        try {
            nbSavedFile = trDAO.save(transferRequest);
        } catch (DAOException e) {
            e.printStackTrace();
        }
        // checks that a file has been created
        File proviFile = new File(transferRequest.getFileName());
        File trFile =
                new File(incomingDirectory.getAbsolutePath()
                        + File.separatorChar + proviFile.getName()
                        + TR_FILE_EXT);
        // Checks that the file exists
        assertEquals(true, trFile.exists());
        // Checks that the number of saved file is equal to one
        assertEquals(1, nbSavedFile);
        if (trFile.exists() == true) {
            FileSystemOperations.delete(trFile);
        }
    }

    /**
     * Mandatory field file size is not set
     */
    public void testSave_02() {
        transferRequest.setSize(null);
        try {
            trDAO.save(transferRequest);
            fail("no exception has been thrown");
        } catch (DAOException daoe) {
            // Expected behaviour
        }

    }

    /** No transfer request available */
    public void testSave_03() {
        try {
            trDAO.save(null);
            fail("no exception has been thrown");
        } catch (IllegalArgumentException ie) {
            // Expected behaviour
        } catch (Exception e) {
            fail("a DAOException exception has been thrown : " + e.getMessage());
        }

    }

    /**
     * Nominal case : removal of an existing transfer request file
     */
    public void testDelete_01() {
        try {
            trDAO.save(transferRequest);
            trDAO.delete(transferRequest);

            File trFile =
                    new File(incomingDirectory.getAbsolutePath()
                            + File.separatorChar
                            + transferRequest.getFileName() + TR_FILE_EXT);

            assertEquals(false, trFile.exists());
        } catch (DAOException e) {
            // TODO Auto-generated catch block
            fail("DAOException has been raised " + e.getMessage());
        }

    }

    /**
     * The transfer request is not valid (size field is missing)
     */
    public void testDelete_02() {

        transferRequest.setSize(null);
        try {
            trDAO.delete(transferRequest);
            fail("No exception has been thrown");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /**
     * A valid request is in a non valid state
     */
    public void testDelete_03() {
        transferRequest.setStatus(null);

        try {
            trDAO.delete(transferRequest);
            fail("No exception has been thrown");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /**
     * A valid request is associated to a non existing file
     */
    public void testDelete_04() {
        try {
            trDAO.delete(transferRequest);
            fail("FAILED no exception has been thrown");
        } catch (DAOException daoe) {
            // expected behaviour
        }
    }

    /**
     * No transfer request is provided
     */
    public void testDelete_05() {
        try {
            trDAO.delete(null);
            fail("no exception has been thrown");
        } catch (IllegalArgumentException ie) {
            // expected behaviour
        } catch (DAOException daoe) {
            fail("a DAOException has been raised " + daoe.getMessage());
        }
    }

}
