package org.eso.util.dal.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.TimeZone;

import org.eso.util.dal.PackageReport;
import org.eso.util.dal.PackageReportsDAO;
import org.eso.util.dal.PackageReportsDAOSyb;
import org.eso.util.dal.SybaseConnectionService;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;


public class PackageReportsDAOSybTest {
	
	private static PackageReportsDAO dao = null;
	private static Statement stmt = null;
	
	private static final String REPORTS_TABLE = "rhandler..pi_pkg_reports";
	
	private static final String PKGS_TABLE = "rhandler..pi_pkgs";
	
	private static PackageReport testPackage;
	
	private static PackageReport testPackage2;
	
	private static Connection conn1;
	private static Connection conn2;
	
	@BeforeClass
	public static void setUpBeforeClass() throws Exception {

		testPackage = new PackageReport();
		testPackage.setFileUrl("someUrl");
		testPackage.setPackageId(123456789);
		testPackage.setReport("some report".getBytes());
		
		testPackage2 = new PackageReport();
		testPackage2.setFileUrl(testPackage.getFileUrl());
		testPackage2.setPackageId(testPackage.getPackageId());
		testPackage2.setReport("some report - updated".getBytes());
		
		TimeZone.setDefault(TimeZone.getTimeZone("UTC"));
		
		conn1 = SybaseConnectionService.getConnection("acdbdev:6789", "pipack", "pipack_ps");
		dao = new PackageReportsDAOSyb(conn1, REPORTS_TABLE);
		conn2 = SybaseConnectionService.getConnection("acdbdev:6789", "pipack", "pipack_ps");
		stmt = conn2.createStatement();
		
		// insert into archive..pi_pkgs values (123456789, 'RUN1', 'INSTRU1', 'OPEN', 'DATA')
		// the ID must be in the pi_pkgs table too or the test will not run
		final String query =
			"insert into " + PKGS_TABLE + " values ("
			+ testPackage.getPackageId()
			+ ", 'RUN182712613', 'INSTRU75764563', 'OPEN6755436', 'DATA877655443')";
		stmt.execute(query);
	}
	
	@AfterClass
	public static void tearDownAfterClass() throws Exception {

		truncateTable();

		final String query = "delete from " + PKGS_TABLE + " where pkg_id=" + testPackage.getPackageId();
		stmt.execute(query);
		
		conn1.close();
		conn2.close();
	}

	@Before
	public void setUp() throws Exception {
	}

	@Test
	public void testInsert() {
		try {
			truncateTable();
						
			/***************************************************************/
			dao.insert(testPackage);
			/****************************************************************/
			
			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE
							+ " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage.getPackageId()
							)
					);
			
			ResultSet rs = stmt.getResultSet();
			assertTrue(rs.next());
			assertEquals(testPackage.getFileUrl(), rs.getString(PackageReportsDAOSyb.FILEURL_KW));
			
			byte[] reportData = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertTrue(compareByteArray(reportData, testPackage.getReport()));
			assertFalse(rs.next());
				
		} catch (Exception e) {
			fail(e.getMessage());
		}
	}

	@Test
	public void testDelete() {
		try {
			truncateTable();
			insertTestEntry(testPackage);
			
			/***************************************************************/
			dao.delete(testPackage);
			/***************************************************************/
			
			// check that there is no entry with the test package id
			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE + " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage.getPackageId()
							)
					);
			ResultSet rs = stmt.getResultSet();
			assertFalse(rs.next());
			
		} catch (Exception e) {
			fail(e.getMessage());
		}
	}
	
	@Test
	public void testFind() {
		try {
			truncateTable();
			insertTestEntry(testPackage);
			
			/***************************************************************/
			PackageReport report = dao.find(testPackage.getPackageId(), testPackage.getFileUrl());
			/***************************************************************/
			
			assertTrue(report != null);
			assertEquals(testPackage.getFileUrl(), report.getFileUrl());
			assertEquals(testPackage.getPackageId(), report.getPackageId());
			assertTrue(compareByteArray(report.getReport(), testPackage.getReport()));
			
		} catch (Exception e) {
			fail(e.getMessage());
		}
	}
	
	
	@Test
	public void testInsertOrUpdate() {
		try {
			truncateTable();

			// because the table is empty, the first insertOrUpdate() should
			// perform an insertion
			
			/***************************************************************/
			dao.insertOrUpdate(testPackage);
			/***************************************************************/
			
			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE + " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage.getPackageId()
							)
					);
			ResultSet rs = stmt.getResultSet();
			assertTrue(rs.next());
			assertEquals(testPackage.getFileUrl(), rs.getString(PackageReportsDAOSyb.FILEURL_KW));
			
			byte[] reportData = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertTrue(compareByteArray(reportData, testPackage.getReport()));
			assertFalse(rs.next());
			
			// now it should perform an update because the record is already present
			/***************************************************************/
			dao.insertOrUpdate(testPackage2);
			/***************************************************************/

			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE + " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage2.getPackageId()
							)
					);
			rs = stmt.getResultSet();
			assertTrue(rs.next());
			assertEquals(testPackage2.getFileUrl(), rs.getString(PackageReportsDAOSyb.FILEURL_KW));
			byte[] reportData2 = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertTrue(compareByteArray(reportData2, testPackage2.getReport()));
			assertFalse(rs.next());

		} catch (Exception e) {
			fail(e.getMessage());
		}
	}
	
	@Test
	public void testUpdate() {
		try {
			truncateTable();
			insertTestEntry(testPackage);

			/***************************************************************/
			dao.update(testPackage2);
			/***************************************************************/
			
			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE + " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage2.getPackageId()
							)
					);
			ResultSet rs = stmt.getResultSet();
			assertTrue(rs.next());
			assertEquals(testPackage2.getFileUrl(), rs.getString(PackageReportsDAOSyb.FILEURL_KW));
			byte[] reportData2 = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertTrue(compareByteArray(reportData2, testPackage2.getReport()));
			assertFalse(rs.next());

		} catch (Exception e) {
			fail(e.getMessage());
		}
	}
	
	
	@Test
	public void testFindMultiplePackages() {
		try {
			truncateTable();
			
			insertTestEntry(testPackage);

			// file with the same package id
			testPackage2.setFileUrl(testPackage.getFileUrl() + "2");
			
			insertTestEntry(testPackage2);
			
			/***************************************************************/
			dao.find(testPackage.getPackageId());
			/***************************************************************/

			// it should return 2 entries
			assertTrue(
					stmt.execute(
							"SELECT * FROM " + REPORTS_TABLE + " WHERE "
							+ PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage.getPackageId()
							)
					);
			ResultSet rs = stmt.getResultSet();
			assertTrue(rs.next());
			final Integer id1 = rs.getInt(PackageReportsDAOSyb.PACKAGEID_KW);
			final String url1 = rs.getString(PackageReportsDAOSyb.FILEURL_KW);
			final byte[] reportData1 = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertTrue(rs.next());
			final Integer id2 = rs.getInt(PackageReportsDAOSyb.PACKAGEID_KW);
			final String url2 = rs.getString(PackageReportsDAOSyb.FILEURL_KW);
			final byte[] reportData2 = PackageReportsDAOSyb.readAll(rs.getBinaryStream(PackageReportsDAOSyb.REPORT_KW));
			assertFalse(rs.next());
			
			assertTrue(id1.equals(testPackage.getPackageId()) && id1.equals(id2));
			assertTrue(
					(url1.equals(testPackage.getFileUrl()) && url2.equals(testPackage2.getFileUrl())) ||
					(url2.equals(testPackage.getFileUrl()) && url1.equals(testPackage2.getFileUrl()))
					);
			assertTrue(compareByteArray(reportData1, testPackage.getReport()));
			assertTrue(compareByteArray(reportData2, testPackage2.getReport()));
		} catch (Exception e) {
			fail(e.getMessage());
		}
	}

	public static void truncateTable() throws Exception {
		stmt.execute("DELETE FROM " + REPORTS_TABLE + " WHERE " + PackageReportsDAOSyb.PACKAGEID_KW + "=" + testPackage.getPackageId());
	}
	
	public void insertTestEntry(PackageReport packageReport) throws Exception {
		byte[] reportData = packageReport.getReport();
		String hexData = "0x";
		for (int i = 0; i < reportData.length; i++)
			hexData += Integer.toHexString(reportData[i]);
		final String columns ="(" + PackageReportsDAOSyb.PACKAGEID_KW + "," + PackageReportsDAOSyb.FILEURL_KW + "," + PackageReportsDAOSyb.REPORT_KW + ")";
		final String values = "(" + packageReport.getPackageId() + ",'" + packageReport.getFileUrl() + "', " + hexData + ")";
		final String query = "INSERT INTO " + REPORTS_TABLE + " " + columns + " VALUES " + values;
		stmt.execute(query);
	}
	
	public static boolean compareByteArray(byte[] o1, byte[] o2) {
		boolean result = true;
		for (int i = 0; i < Math.max(o1.length, o2.length); i++) {
			result &= (o1[i] == o2[i]);
		}
		return result;
	}
}
