package org.eso.dfs.datatransfer;

/*
 * Abstract class for handling FileList flavour data transfers. from java 
 * application sources and external sources (desktop file explorer).
 * 
 * Implementing classes must complete the exportFileList() and importFileList()
 * methods.
 */
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.InputEvent;
import java.io.IOException;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.TransferHandler;

public abstract class FileListTransferHandler extends TransferHandler {

    /**
     * Export will only occur if export is enabled.
     */
    private boolean exportEnabled = true;
    
    /**
     * Import will only occur if import is enabled.
     */
    private boolean importEnabled = true;
    
    /**
     * Create a FileListTransferHandler with import and export enabled.
     * @see java.lang.Object#Object()
     */
    public FileListTransferHandler() {
    }
    
    /**
     * Create a FileListTransferHandler with the import and export
     * enabled as given.
     * @param importOK
     * @param exportOK
     */
    public FileListTransferHandler(boolean importOK, boolean exportOK) {
        setImportEnabled(importOK);
        setExportEnabled(exportOK);
    }
    
    /**
     * @param c
     * @return List of File objects to be exported, or NULL if no files
     * can be exported - in which case the drag will fail.
     */
    protected abstract List exportFileList(JComponent c);
    
    /**
     * Import the List of File objects.
     * @param c
     * @param fileList
     */
    protected abstract void importFileList(JComponent c, List fileList);

    /**
     * Create a Transferable from the specified component. This called
     * exportFileList().
     * @see javax.swing.TransferHandler#createTransferable(JComponent)
     */
    protected Transferable createTransferable(JComponent c) {
        Transferable rc = null;
        if (isExportEnabled()) {
            List exported = exportFileList(c);
            
            if (exported == null) {
                rc = null;
            } else {
                rc = new FileListTransferable(exportFileList(c));
            }
        }
        return rc;
    }

    /**
     * Only COPY is supported in this handler - not MOVE.
     * @see javax.swing.TransferHandler#getSourceActions(JComponent)
     */
    public int getSourceActions(JComponent c) {
        return COPY;
    }

    /**
     * Extract the file list and call importFileList().
     * 
     * @see javax.swing.TransferHandler#importData(JComponent, Transferable)
     */
    public boolean importData(JComponent c, Transferable t) {
        if (canImport(c, t.getTransferDataFlavors())) {
            try {
                List files =
                    (List)t.getTransferData(DataFlavor.javaFileListFlavor);
                importFileList(c, files);
                return true;
            } catch (UnsupportedFlavorException ufe) {
            } catch (IOException ioe) {
            }
        }

        return false;
    }

    protected void exportDone(JComponent c, Transferable data, int action) {
        // nothing to do here.
    }

    /**
     * @return true of the passed flavours contains javaFileListFlavor, else
     * false.
     * @see javax.swing.TransferHandler#canImport(JComponent, DataFlavor[])
     */
    public boolean canImport(JComponent c, DataFlavor[] flavors) {
        if (!isImportEnabled()) {
            return false;   
        }
        for (int i = 0; i < flavors.length; i++) {
            if (DataFlavor.javaFileListFlavor.equals(flavors[i])) {
                return true;
            }
        }
        return false;
    }
    /**
     * Returns the exportEnabled.
     * @return boolean
     */
    public boolean isExportEnabled() {
        return exportEnabled;
    }

    /**
     * Returns the importEnabled.
     * @return boolean
     */
    public boolean isImportEnabled() {
        return importEnabled;
    }

    /**
     * Sets the exportEnabled.
     * @param exportEnabled The exportEnabled to set
     */
    public void setExportEnabled(boolean exportEnabled) {
        this.exportEnabled = exportEnabled;
    }

    /**
     * Sets the importEnabled.
     * @param importEnabled The importEnabled to set
     */
    public void setImportEnabled(boolean importEnabled) {
        this.importEnabled = importEnabled;
    }

}
