package org.eso.dfs.services.gui;

import java.awt.Component;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eso.dfs.services.ServiceException;

/**
 * @author Nick Kornweibel
 *
 * This is the singleton object for the FileImageDisplayService. The
 * service maintains a vector of handlers that form a 
 * chain of responsibility pattern. In this model the Service (this singleton)
 * calls the successor of each handler, rather than the handler performing the
 * call to the successor.
 */
public class FileImageDisplayService {

    /**
     * Set of handlers in the chain.
     */
    private List handlerChain = new ArrayList();
    
    /**
     * Singleton instance of file image display service.
     */
    private static FileImageDisplayService instance;
    
    /**
     * Constructor for FileImageDisplayService.
     */
    private FileImageDisplayService() {
        super();
    }
    
    /**
     * Get the singleton instance. It is synchronized to ensure
     * creation of the singleton is atomic.
     * @return FileImageDisplayService singleton instance.
     */
    public static synchronized FileImageDisplayService getInstance() {
        if (instance == null) {
            instance = new FileImageDisplayService();
        }
        return instance;
    }
    
    /**
     * Display the requested files, if there is in error or they cannot
     * be displayed, an exception is thrown.
     * @param fileIds String array of full file paths or ids
     * @param showImage if true, this is a request to show the image, not just a header or summary.
     * @throws ServiceException if not handled or on error.
     */
    public void displayFile(String[] fileIds, boolean showImage) throws ServiceException {
        boolean handled = false;
        Iterator i = handlerChain.iterator();
        while (!handled && i.hasNext()) {
            FileImageDisplayHandler h = (FileImageDisplayHandler)i.next();
            handled = h.displayFileImages(fileIds, showImage);
        }
        if (!handled) {
            throw new ServiceException("No handler found for requested images.");
        }
    }
    
    /**
     * Display the requested files, if there is an error in display or they
     * cannot be displayed, an error is presented to the user using the
     * component passed as the parent for the JOptionPane error message.
     * TODO: improve error message display. - use a graphic message
     * implies adding a depedency to Gasgano.
     * @param comp parent of GUI error message.
     * @param fileIds String ayy of files to display.
     * @param showImage if true this is a request to display the file image.
     */
    public void displayFile(Component comp, String[] fileIds, boolean showImage) {
        try {
            displayFile(fileIds, showImage);
        } catch (ServiceException e) {
            //GUIUtils.userMessage(comp, e.getMessage());
            System.out.println(e.getMessage());
            
        }
    }
    
    /**
     * Add the given display handler to the chain of handlers.
     * If already present, the previous handler is removed and added again.
     * Additions are made to the end of the chain.
     * @param h the display handler to be added to the chain.
     */
    public void addFileImageDisplayHandler(FileImageDisplayHandler h) {
        removeFileImageDisplayHandler(h);
        handlerChain.add(h);
    }
    
    /**
     * Remove the given handler from the handler chain.
     * @param h
     */
    public void removeFileImageDisplayHandler(FileImageDisplayHandler h) {
        handlerChain.remove(h);
    }
}
