/**
 * @author N. Kornweibel
 * @version 1.0
 * An observing program, as represented in Gasgano.
 */

package org.eso.gasgano.datamodel;

import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.SortedHashtable;

public class Program {

    private String id = null;
    private String desc = null;

    // one object is shared amongst all Program objects.
    private static ObservationDescription obsDescFactory = null;

    private SortedHashtable observations = null;
    private SortedHashtable directory = null;
    private SortedHashtable telescope = null;

    /**
     * @param name The id of this Program.
     */
    public Program(String name) {
        id = name;
        observations = new SortedHashtable();
    }

    /**
     * @param p The ObservationDescription object to be used to used to set
     * the description of observations in the DataModel.
     */
    public static void setDefaultObservationDescription(ObservationDescription p) {
        obsDescFactory = p;
    }

    /**
     * @return The ObservationDescription object to be used to used to set
     * the description of observations in the DataModel.
     */
    public static ObservationDescription getDefaultObservationDescription() {
        return obsDescFactory;
    }

    /**
     * @return The id of this Program.
     */
    public String getId() {
        return id;
    }

    /**
     * @return The description of this Program, or the id if the
     * description is not set.
     */
    public String toString() {
        String rc = desc;
        if (rc == null) {
            rc = getId();
        }
        return rc;
    }

    /**
     * @param s The description of this Program.
     */
    public void setDescription(String s) {
        desc = s;
    }

    /**
     * @param id The id of the observation to be added.
     * @param obc The OBComponent requiring the addition of the observation.
     * @return The observation located/created.
     * Add an Observation to the observation set.
     * An observation can only be added once.
     * To find an observation use findObservation().
     */
    public Observation addObservation(
        String id,
        OBComponent obc) {
        Observation rc = findObservation(id);
        if (rc == null) {

            rc = new Observation(id);
            if (PropertyDB.getInstance().isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println(
                    "Observation " + id + " Added to Prog " + this.id);
            }
            if (obsDescFactory != null) {
                obsDescFactory.setDescription(rc, obc);
            }
            // destroy the vector as it is out of date
            observations.put(id, rc);
        }
        return rc;
    }

    /**
     * @param id The id of the observation to be removed.
     */
    public void removeObservation(String id) {
        observations.remove(id);
    }

    /**
     * @param id The id of the observation to be located.
     * @return The observation located, or null if not located.
     */
    public Observation findObservation(String id) {
        return (Observation) observations.get(id);
    }

    /**
     * @return True if there are no observations under this Program.
     */
    public boolean isEmpty() {
        return observations.isEmpty();
    }

    /**
     * @return The SortedHashtable of observations under this Program.
     */
    public SortedHashtable getObservationSet() {
        return observations;
    }

    public Observation getObservationAt(int i) {
        return (Observation) observations.elementAt(i);
    }

    public int getIndexOfObservation(Observation obs) {
        return observations.indexOf(obs);
    }

    public int getObservationCount() {
        return observations.size();
    }
    // MZ 
    public void cleanup() {
        observations.clear();
    }

    public void setTelescope(SortedHashtable telescope) {
        this.telescope = telescope;
    }

    public SortedHashtable getTelescope() {
        return telescope;
    }

    public void setDirectory(SortedHashtable directory) {
        this.directory = directory;
    }

    public SortedHashtable getDirectory() {
        return directory;
    }

}

// end of file Program.java.
