/**
 * @author N. Kornweibel
 * @version 1.0
 * A base class for reading and reporting information about datafiles.
 */
package org.eso.gasgano.datamodel.filesystem;

import java.io.File;
import java.io.IOException;

import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.BaseName;

public abstract class DFSFile {

    private int fileStatus = UNKNOWN_STATE;
    private long lastModified = 0L;

    /**
     * length of the seconds and milli seconds part of an archive name.
     * ie the last 6 characters of the following name:
     * ISAAC.1996-05-11T13:00:00.029
     */
    private static final int SECONDS_AND_MILLI = 6;

    public static final int UNKNOWN_STATE = 0;
    public static final int FILE_NOT_FOUND = 1;
    public static final int FILE_INVALID = 2;
    public static final int FILE_VALID = 3;
    public static final int FILE_FILTERED = 4;
  

  /**
   * used for testing purposes only, tests the extractBaseName method
   */

    public static void main(String[] args) {
        PropertyDB.getInstance().setDebugLevel(PropertyDB.debugBaseName);
        for (int i = 0; i < args.length; i++) {
            System.out.println(args[i]);
            extractBaseName(args[i]);
        }
    }
    /**
     * @return a short string describing the status of the file 
     */
    public String getStringStatus() {
        return getStringStatus(getStatus());
    }

  /**
   * static method to query the status of a different object.
   * @return: a short string describing the status of the file 
   */

    public static String getStringStatus(int status) {
        String rc = "unknown state";
        if (status == FILE_NOT_FOUND) {
            rc = new String("not found");
        } else if (status == FILE_INVALID) {
            rc = new String("invalid");
        } else if (status == FILE_FILTERED) {
            rc = new String("filtered");
        } else if (status == FILE_VALID) {
            rc = new String("valid");
        }
        return rc;
    }

    /**
     * @param s Set the validity status of this DFSFile to: UNKNOWN_STATE,
     * FILE_NOT_FOUND, FILE_INVALID, FILE_VALID.
     */
    public void setStatus(int s) {
        fileStatus = s;
    }

    /**
     * @return The validity status of this DFSFile: UNKNOWN_STATE,
     * FILE_NOT_FOUND, FILE_INVALID, FILE_VALID.
     */
    public int getStatus() {
        return fileStatus;
    }

    /**
     * @return 0L on error, else the last modification time (in seconds).
     */
    public long getLastModified() {
        long rc = 0L;
        File theFile = getPath();
 
        // update the lastModified entry
        try {
            rc = theFile.lastModified();
        } catch (SecurityException e) {
            System.out.println("error getting modification time: "
                +e.toString());
            rc = 0L;
        }
        return rc;
    }
 
 
    /**
     * @return True if the file mod time is newer than the
     * stored mod time. If an error occurs reading the lastModifiedTime,
     *  true is also returned.
     */
 
    public boolean dataRefreshRequired() {
        boolean rc = true;
        long lastMod = getLastModified();

        if (lastMod == 0 || lastModified != lastMod) {
            rc = true;
        } else {
            rc = false;
        }
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugDataModelChange ) ) {
            System.out.println(getFileName()+" last mod time: "+lastMod
                +" refresh required: "+rc);
        }
        return rc;
    }


    /**
     * Used to update the internal time of last modification, indicating that
     * the file has been updated from the disk
     */
    public void dataRefreshed() {
        lastModified = getLastModified();
    }

    /**
     * abstract method used to read the data from the disk. This method should
     * call dataRefreshed() and setStaus()<p>
     * @return The validity of the file: UNKNOWN_STATE,
     * FILE_NOT_FOUND, FILE_INVALID, FILE_VALID.
     * Call dataRefreshed() method in this function.
     */
    public abstract int readData();

    /**
     * @return The DFS 'base name' of the DFSFile. See Design Document, Assumptions and Definitions
     * for  definition of a DFS 'base name'.
     */
    public String getBaseName() {
        return extractBaseName(getPathString());
    }
  
    /**
     * @return The File object for this DFSFile.
     */
    public abstract File getPath();

    /**
     * @return The full path string of this file.
     */
    public String getPathString() {
        return getPath().getPath();
    }

    /**
     * @return The file name of this file.
     */
    public String getFileName() {
        return getPath().getName();
    }

    /**
     * A utility function for removing the path of a file
     * path and returning just the name.
     * @param The full file path.
     * @return The file name without the path.
     */
    public static String getFileName(String name) {
        File f = new File(name);
        return f.getName();
    }
 
    /**
     * @param fullPath The full file path.
     * @return The file path after removal of non-base name suffix.
     */
    public static String removeSuffix(String fullPath) {
        String rc = fullPath;
        int endBaseName = fullPath.lastIndexOf('_');
        // if we found _tpl - then leave it as is.
        if (endBaseName != -1 && fullPath.charAt(endBaseName+1) != 't') {
            try {
                rc = fullPath.substring(0, endBaseName);
            } catch (StringIndexOutOfBoundsException e) {
                System.out.println(e.toString());
                rc = fullPath;
            }
        }
        if (rc == fullPath) {
            endBaseName = fullPath.lastIndexOf('.');
            try {
                rc = fullPath.substring(0, endBaseName);
            } catch (StringIndexOutOfBoundsException e) {
                System.out.println(e.toString());
                rc = fullPath;
            }
        }
        // System.out.println("NO SUFFIX:<"+rc+"> from <"+fullPath+">");
        return rc;
    }

    /**
     * Access the filesystem (possibly) and resolve the file path
     * into its canonical form ( resolving all links, logical drive names, relative paths. etc.)
     * @return this canonical file path, or null if it could not be resolved.
     */
    public static String getCanonicalPath(String fullPath) {
        String rc = null;
        try {
            File file = new File(fullPath);
            rc = file.getCanonicalPath();
        } catch (IOException e) {
            // error resolving name.
            rc = null;
        }
        return rc;
    }

    /**
     * static utility function for extracting the basename from a full path.
     * For a definition of a basenam, see the Assumptions and Definitions chapter.
     * @param fullPath The path of the file.
     * @return The DFS base name of the file, or the fullPath if there was
     * not valid base name found.
     */
    public static String extractBaseName(String fullPath) {
        return BaseName.extractBaseName(fullPath);
    }

} // end of class DFSFile.java

// end of file DFSFile.java
