/**
 * @author N. Kornweibel
 * @version 1.0
 * 
 * Class to display fits data.
 * 
 * The enormous amount of static members is due
 * to a *graphical* memory leak when using HP-UX and
 * HP Xterminals. It is a memory leak in the X-terminal - not in the JVM.
 * 
 * See "HP-XTERM MEMORY LEAK" work-around in code below.
 * 
 */
package org.eso.gasgano.datamodel.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableCellRenderer;

import org.eso.dfs.gui.RecallingComboBox;
import org.eso.dfs.gui.SwingFileChooser;
import org.eso.dfs.services.gui.FileImageDisplayService;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.datamodel.filesystem.FitsFile;
import org.eso.gasgano.gui.KeywordTable;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;

/**
 * @author Nick Kornweibel.
 * 
 * Panel to display the keyword blocks of a FitsFile and permit searching and
 * filtering, etc.
 * 
 */
public class FitsFileDisplay
    extends JPanel
    implements ActionListener, KeyListener {

    /*
     * Panel and button Labels
     */
    private static final String findLabel = "Find in header:";
    private static final String findButtonLabel = "find";
    private static final String filterLabel = "Filter";
    private static final String autoDisplayLabel = "Auto Display";
    private static final String loadFilterLabel = "Load Filter";

    /*
     * Static components due to HP-XTerm memory leak.
     * See "HP-XTERM MEMORY LEAK" below.
     */
    private static JScrollPane tableView = new JScrollPane();
    //private static JTextField findEntry = new JTextField(20);
    private static RecallingComboBox findEntry = new RecallingComboBox(true, 20);
    
    private static JRadioButton filterView = new JRadioButton();
    private static JRadioButton autoDisplay = new JRadioButton();
    private static JPanel upperPanel = new JPanel(new BorderLayout());
    private static JPanel controlPanel = new JPanel();
    private static JScrollPane controlPanelScroller 
        = new JScrollPane(controlPanel, JScrollPane.VERTICAL_SCROLLBAR_NEVER, 
            JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
    private static JPanel findPanel = new JPanel();
    private static JPanel filterPanel = new JPanel();
    private static JPanel autoDisplayPanel = new JPanel();
    private static JPanel extensionPanel = new JPanel();
    private static JButton searchButton = new JButton(findButtonLabel);
    private static JButton loadFilterButton = new JButton(loadFilterLabel);
    private static JLabel titleLabel = new JLabel();
    
    
    private JTable table = null;
    private String fileName = null;
    private FitsTableModel model = null;
    private boolean cltrDown = false;
    
    private static String findKeyEntry = null;

    /**
     * Are the keywords currently filtered or not.
     */
    private static boolean filtered = false;

    /**
     * Combo box to show the currently selected keyword block and select
     * other keyword blocks (extensions) for display. Disabled if extensions
     * are not required to be displayed.
     */
    private static JComboBox extensionSelector = new JComboBox();

    private static FitsFileDisplay formerThis = null;
    
    /**
     * @param title The first/upper text label displayed.
     * @param mdl The FitsTableModel of keywrd names and values.
     */
    public FitsFileDisplay(String title, FitsTableModel mdl) {
        super(new BorderLayout());

        /*
         * START OF HP-XTERM MEMORY LEAK WORK AROUND.
         * 
         * The following cumbersome code is needed to remove
         * components and prevent the listeners remaining active.
         * 
         * All is due to the need to have static members,
         * because of the memory leak problem using Xterminals and HP-UX.
         * 
         * There are 4 steps involved, see below:
         * 
         * Step 1:
         * Remove all components from the static panels. 
         * 
         */
        upperPanel.removeAll();
        controlPanel.removeAll();
        extensionPanel.removeAll();
        filterPanel.removeAll();
        autoDisplayPanel.removeAll();
        findPanel.removeAll();

        titleLabel.setText(title);
        titleLabel.setHorizontalAlignment(SwingConstants.CENTER);
        titleLabel.setBorder(new EmptyBorder(5, 0, 5, 0));
        titleLabel.setForeground(Color.black);
        model = mdl;

        table = new JTable(model);
        setupCellRenderer();
        table.setCellSelectionEnabled(true);
        tableView.setViewportView(table);
        table.addKeyListener(this);

        // remove the 3rd column, as it is not to be displayed.
        table.getColumnModel().removeColumn(
            table.getColumnModel().getColumn(2));

        /* 
         * Step 2:
         * Any widgets that have addActionListener(this) need
         * to have the formerThis explicitly removed. 
         * 
         */
        if (formerThis != null) {
            searchButton.removeActionListener(formerThis);
            loadFilterButton.removeActionListener(formerThis);
            findEntry.removeActionListener(formerThis);
            filterView.removeActionListener(formerThis);
            extensionSelector.removeActionListener(formerThis);
            autoDisplay.removeActionListener(formerThis);
        }
        /*
         * Step 3:
         * Add the new action listeners. 
         */
        searchButton.addActionListener(this);
        loadFilterButton.addActionListener(this);
        findEntry.addActionListener(this);
        filterView.addActionListener(this);
        extensionSelector.addActionListener(this);
        autoDisplay.addActionListener(this);
        /*
         * Step 4:
         * Set the current "this" so we can remove it from the widget
         * listeners when we have a new "this". This is a part of the
         * HP-UX memory leak work-around.
         */
        formerThis = this;
        /*
         * END OF HP-XTERM MEMORY LEAK WORK-AROUND.
         */

        model.setFiltered(filtered);
        filterView.setText(filterLabel);
        filterView.setSelected(model.isFiltered());

        upperPanel.add(titleLabel, BorderLayout.NORTH);

        extensionPanel.add(new JLabel("Extension: "));
        extensionPanel.add(extensionSelector);
        extensionPanel.setBorder(new LineBorder(Color.black));
        // setup the extension selector
        setupExtensionSelector();
        controlPanel.add(extensionPanel);

        findPanel.add(new JLabel(findLabel));
        findPanel.add(findEntry);
        findPanel.add(searchButton);
        findPanel.setBorder(new LineBorder(Color.black));
        controlPanel.add(findPanel);
        
        filterPanel.add(loadFilterButton);
        filterPanel.add(filterView);
        filterPanel.setBorder(new LineBorder(Color.black));
        controlPanel.add(filterPanel);
        
        autoDisplayPanel.add(autoDisplay);
        autoDisplay.setText(autoDisplayLabel);
        autoDisplayPanel.setBorder(new LineBorder(Color.black));
        controlPanel.add(autoDisplayPanel);
        upperPanel.add(controlPanelScroller, BorderLayout.CENTER);
        add(tableView, BorderLayout.CENTER);
        add(upperPanel, BorderLayout.NORTH);
        fileName = model.getFileName();
        /*
         * if the auto display is already set, send the file to the 
         * display.
         */
        if (autoDisplay.isSelected()) {
            sendFileToDisplay();
        }
    }

    /**
     * @return The JTable component of this panel.
     */
    public JTable getTableComponent() {
        return table;
    }

    /**
     * Internal method to set the correct options into the extension selector.
     * Including disabling the extension selector if extensions are not being displayed.
     */
    private void setupExtensionSelector() {
        // remove existing items
        extensionSelector.removeAllItems();
        // add the header item
        extensionSelector.addItem("HEADER");
        // set the selected item
        extensionSelector.setSelectedIndex(0);

        if (model.getExtensionCount() != -1) {
            // enable the extension selector as extensions are to be displayed.
            extensionSelector.setEnabled(true);

            // add the items for each extension of the FitsFile.
            int xCnt = model.getExtensionCount();
            for (int i = 0; i < xCnt; i++) {
                String xName = model.getExtensionName(i);
                if (xName == null || xName.length() == 0) {
                    xName = "EXT " + (i + 1);
                }
                extensionSelector.addItem(xName);
            }

            // add a table listener to update the extension manager
            // if the cell/row selection changes.
            table
                .getSelectionModel()
                .addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent arg0) {
                    // new selection - update the extensionSelector
                    int row = table.getSelectedRow();
                    if (row > 0) {
                        int extension =
                            ((Integer) model.getValueAt(row, 2)).intValue();
                        updateExtensionSelectorSelection(extension);
                    }
                }
            });
        } else {
            // disable the extension selector as header extensions are not being displayed.
            extensionSelector.setEnabled(false);
        }
    }

    /**
     * Set the current selection of the extensionSelector WITHOUT trigger
     * and event to scroll/display the approriate extension in the JTable.
     * @param i
     */
    private void updateExtensionSelectorSelection(int i) {
        //
        // disable the extensionSelector listener first
        // then set the selected index.
        // then set the listener back
        //
        extensionSelector.removeActionListener(this);
        extensionSelector.setSelectedIndex(i);
        extensionSelector.addActionListener(this);
    }
    
    /**
     * Send the current file to the external display, through the 
     * FileImageDisplayService.
     */
    private void sendFileToDisplay() {
        Cursor c = getCursor();
        setCursor(new Cursor(Cursor.WAIT_CURSOR));
        // send the current file to the display.
        FileImageDisplayService.getInstance().displayFile(this,
            new String[] {model.getFileName() }, true);
        setCursor(c);
    }

    /**
     * Implementation of the ActionListener interface.
     * @param a The ActionEvent to be processed.
     */
    public void actionPerformed(ActionEvent a) {
        if (a.getSource() == filterView) {
            model.setFiltered(filterView.isSelected());
            filtered = model.isFiltered();
            filterView.setSelected(filtered);
        } else if (a.getSource() == loadFilterButton) {
            loadFilter();
        } else if (a.getSource() == extensionSelector) {
            displayExtension();
        } else if (a.getSource() == autoDisplay) {
            PropertyDB.getInstance().setProperty(GasProp.AUTO_DISPLAY_FILE,
                autoDisplay.isSelected() ? "true" : "false");
            if (autoDisplay.isSelected()) {
                sendFileToDisplay();
            }
        } else if (a.getSource() == searchButton) {

            if (findEntry.getSelectedIndex() > -1) {
                findKeyEntry = findEntry.getText();
                String findKey = findKeyEntry.toUpperCase();
                if (findKey.length() > 0) {
                    int row = table.getSelectedRow();
                    if ((row < 0) || (row > table.getRowCount())) {
                        row = 0;
                    }
                    int firstSearch = row;
                    int cell = 0;
                    boolean found = false;
                    while (!found) {
                        String cellValue;
                        row++;
                        if (row == table.getRowCount()) {
                            row = 0;
                        }
                        cell = 0;
                        cellValue =
                            ((String) table.getValueAt(row, cell)).toUpperCase();
                        if (cellValue.indexOf(findKey) != -1) {
                            found = true;
                            break;
                        }
                        cell = 1;
                        cellValue =
                            ((String) table.getValueAt(row, cell)).toUpperCase();
                        if (cellValue.indexOf(findKey) != -1) {
                            found = true;
                            break;
                        }
                        if (firstSearch == row) {
                            break;
                        }
                    }
                    if (found) {
                        showCell(row, cell);
                    } else {
                        JOptionPane.showMessageDialog(
                            this,
                            "\'" + findEntry.getText() + "\' not found.");
                    }
                }
            }
        }
    }

    /**
     * Scroll the table to the specified row and col. 
     * @param row
     * @param cell
     */
    public void showCell(int row, int col) {
        table.setRowSelectionInterval(row, row);
        table.setColumnSelectionInterval(col, col);
        Rectangle r = table.getCellRect(row, col, true);
        table.scrollRectToVisible(r);
    }

    private void loadFilter() {
        JFileChooser kwChooser = new SwingFileChooser();
        DFSDataModel dm = DFSDataModel.getDataModel();
        kwChooser.setCurrentDirectory(
            new File(PropertyDB.getInstance().getProperty("KEYWORDLIST_DIR")));
        kwChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        kwChooser.setDialogType(JFileChooser.OPEN_DIALOG);
        kwChooser.setDialogTitle("Select Keyword List");
        kwChooser.setApproveButtonText("Load");
        kwChooser.setMultiSelectionEnabled(false);
        if (kwChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
            if (kwChooser.getSelectedFile() != null) {
                String selectedFile = kwChooser.getSelectedFile().getPath();
                KeywordTable defaultKWList = new KeywordTable();
                if (!defaultKWList.readFile(selectedFile)) {
                    JOptionPane.showMessageDialog(
                        this,
                        "Warning: error reading: " + selectedFile);
                }
                FitsFile.setKeywordFilter(defaultKWList.getKeywordList());
            }
        }
    }

    /**
     * Scroll the JTable to display the first row of the specified keyword
     * block (header or extension).
     * @param i the extension to be displayed, or -1 for the primary extension.
     * Does nothing if the specified extension does not exist in the model
     */
    public void displayExtension(int i) {
        int row = model.getExtensionStartRow(i);
        if (row >= 0) {
            showCell(row, 0);
        }
    }

    /**
     * Display the extension that is currently selected in the extensionSelector.
     */
    private void displayExtension() {
        int selectIdx = extensionSelector.getSelectedIndex();
        if (selectIdx < 0) {
            // if none selected or some other problem, -1 is returned
            // in this case we default to 0 (the primary header).
            selectIdx =  0;
        }
        displayExtension(selectIdx -1 );
    }

    // KeyListener interface implementation.

    public void keyPressed(KeyEvent k) {
        if (k.getKeyCode() == KeyEvent.VK_CONTROL) {
            cltrDown = true;
        } else if (k.getKeyCode() == KeyEvent.VK_C && cltrDown) {
            int row = -1;
            int col = -1;
            int[] rows = table.getSelectedRows();
            int[] cols = table.getSelectedColumns();
            if (rows.length > 0 && cols.length > 0) {
                row = rows[0];
                col = cols[0];
                Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
                cb.setContents(
                    new StringSelection((String) table.getValueAt(row, col)),
                    null);
            }
        }
    }

    public void keyReleased(KeyEvent k) {
        if (k.getKeyCode() == KeyEvent.VK_CONTROL) {
            cltrDown = false;
        }
    }

    public void keyTyped(KeyEvent k) {
    }

    /**
     * Returns the showExtensions.
     * @return boolean
     */
    public static boolean isShowExtensions() {
        return PropertyDB.getInstance().getProperty(
            GasProp.SHOW_EXTENSIONS_KEY).equalsIgnoreCase(
            "true");
    }

    /**
     * Set up the cell renderer to highlight the start of each fits extension.
     */
    private void setupCellRenderer() {
        TableCellRenderer tcr = table.getDefaultRenderer(Object.class);
        table.setDefaultRenderer(
            Object.class,
            new FitsKeywordDisplayRenderer(tcr));
    }

} // end of class FitsFileDisplay

class FitsKeywordDisplayRenderer implements TableCellRenderer {

    private TableCellRenderer defaultRenderer;

    public FitsKeywordDisplayRenderer(TableCellRenderer r) {
        defaultRenderer = r;
    }

    /**
     * @see javax.swing.table.TableCellRenderer#getTableCellRendererComponent(javax.swing.JTable, java.lang.Object, boolean, boolean, int, int)
     */
    public Component getTableCellRendererComponent(
        JTable table,
        Object value,
        boolean isSelected,
        boolean hasFocus,
        int row,
        int column) {
        Component rc =
            defaultRenderer.getTableCellRendererComponent(
                table,
                value,
                isSelected,
                hasFocus,
                row,
                column);

        if (table.getValueAt(row, 0).equals(Keyword.EXTENSION_METAKEYWORD)) {
            if (isSelected) {
                rc.setBackground(Color.CYAN);
            } else {
                rc.setBackground(Color.GRAY);
            }
        } else {
            if (isSelected) {
                rc.setBackground(Color.CYAN);
            } else {
                rc.setBackground(Color.WHITE);
            }
        }
        return rc;
    }
}

// end of file FitsFileDisplay.java
