/**
 * @author N. Kornweibel
 * @version 1.0
 * This Model mapps a DataModel into an interface usable by JTree.
 * The root of the tree is a String, with Program children.
 * String->Program->Observation->OBComponent.
 * The class of a node is used often to determine the level in the tree.
 */

package org.eso.gasgano.gui;

import java.awt.Cursor;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.tree.TreePath;

import org.eso.gasgano.datamodel.DataModelChangeListener;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.OBComponentFilter;
import org.eso.gasgano.datamodel.database.DBContainer;
import org.eso.gasgano.datamodel.database.DBDataModel;
import org.eso.gasgano.datamodel.gui.DBOBComponent;
import org.eso.gasgano.properties.GasProp;

public class DBTreeTableModel
    extends AbstractTreeTableModel
    implements DataModelChangeListener, TreeExpansionListener {

    private Vector treeListeners = null;
    private DBDataModel dm = null;
    private RootNode root = new RootNode();
    private Vector displayCols = null;
    private int groupDataByLevel = 0;
    protected Vector allTheLeaves = new Vector();

    private JTree theTree = null;

    public DBTreeTableModel(DBDataModel d) {
        init(d);
    }

    private void init(DBDataModel d) {
        dm = d;
        treeListeners = new Vector();
        setRootLabel(null);
        dm.addDataModelChangeListener(this);
        displayCols = GasProp.getProperties().dbDisplayColumns;
    }

    // I so much would like not to need this..
    public void setTree(JTree tree) {
        theTree = tree;
    }

    public int getGroupDataByLevel() {
        return groupDataByLevel;
    }

    public void setGroupDataByLevel(int gbLevel) {
        if (groupDataByLevel != gbLevel) {
            groupDataByLevel = gbLevel;
            setRootLabel(null);
            updateTree();
        }
    }

    public void setRootLabel(String name) {
        if (name == null) {
            String suffix = null;
            String filterStr = null;
            OBComponentFilter filter = dm.getFileFilter();
            if (filter != null) {
                filterStr = new String("Filter: " + filter.getDescription());
            } else {
                filterStr = new String("Unfiltered");
            }
            if (groupDataByLevel == 0) {
                suffix = new String("");
            }
            if (groupDataByLevel == 2) {
                suffix = new String(" grouped by telescope.");
            }
            name =
                new String(
                    "Displaying "
                        + dm.getProgramCount()
                        + " programs "
                        + suffix
                        + " "
                        + filterStr
                        + ".");
        }
        root.setLabel(name);
    }

    public String getRootLabel() {
        String rc = null;
        if (root != null) {
            root.getLabel();
        }
        return rc;
    }

    public Object getChild(Object parent, int index) {
        Object rc = null;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getProgramAt(index);
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().elementAt(index);
            }
        } else if (parent instanceof DBContainer) {
            DBContainer tmpContainer = (DBContainer) parent;
            /*
            if ( tmpContainer.isObservation() ) { 
              dm.fillObservation(tmpContainer.getFather().toString() , tmpContainer.toString()  ) ;
            }
            */
            rc = tmpContainer.elementAt(index);
        } else {
            rc = null;
        }
        return rc;
    }

    public int getChildCount(Object parent) {
        int rc = 0;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getProgramCount();
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().size();
            }
        } else if (parent instanceof DBContainer) {
            rc = ((DBContainer) parent).virtualSize();
        } else {
            rc = 0;
        }
        return rc;
    }

    public int getIndexOfChild(Object parent, Object child) {
        int rc = -1;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getIndexOfProgram((DBContainer) child);
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().indexOf(child);
            }
        } else if (parent instanceof DBContainer) {
            rc = ((DBContainer) parent).indexOf(child);
        } else {
            rc = -1;
        }
        return rc;
    }

    public int getColumnCount() {
        // the file entry plus the entries in DISPLAY_COLUMNS.
        return displayCols.size() + 1;
    }

    public String getColumnName(int column) {
        String rc = null;
        if (column == 0) {
            rc = "File";
        } else {
            rc = (String) displayCols.elementAt(column - 1);
        }
        return rc;
    }

    public Object getValueAt(Object node, int column) {
        Object rc = null;
        if (column > 0) {
            if (node instanceof DBOBComponent) {
                String colData[] = null;
                colData = ((DBOBComponent) node).getDisplayData();
                if (colData != null && colData.length >= column) {
                    rc = colData[column - 1];
                }
            }
        } else {
            rc = node;
        }
        return rc;
    }

    public void toggleColumn() {
        displayCols = GasProp.getProperties().dbDisplayColumns;
    }

    public Class getColumnClass(int column) {
        Class rc = null;
        if (column == 0) {
            rc = TreeTableModel.class;
        } else {
            rc = String.class;
        }
        return rc;
    }

    public Object getRoot() {
        if (root == null) {
            setRootLabel(null);
        }
        return root;
    }

    public boolean isLeaf(Object node) {
        boolean rc = false;
        if (node instanceof DBOBComponent) {
            rc = true;
        }
        return rc;
    }

    public void valueForPathChanged(TreePath path, Object newValue) {
        // tree is not editable...ignore!
    }

    /**
     * this method is  responsible for triggering 
     * TreeStructureChanged events
     * this is done via the invokeLater in a separate Thread
     *
     * for the moment one event per File added or removed
     * in future we might wait for all the additions to be made and
     * then send one unique event. Same for removals
     * 
     * most of the code is dedicated to debugging output
     * an issue which has to be solved is that events are sent
     * even when the tree representation is not suitable for them
     *
     * @author: Michele Zamparelli
     * @param DataModelEvent e    the event received by the DataModel
     * 
     * @exception: none
     */

    public void dataModelChanged(DataModelEvent e) {
        setRootLabel(null);
        boolean debug = false;

        int eType = 0;
        Object path[] = null;
        Object ancestors[] = null;

        Object child = e.getFile();
        eType = e.getEvent();
        if (child == null) {
            System.err.println(
                "DataModelChange Event sent with no OBComponent/Program/Observatio");
            return;
        }
        path = getPathToRoot(child);
        ancestors = new Object[path.length - 1];
        for (int z = 0; z < ancestors.length; z++) {
            ancestors[z] = path[z];
        }

        // screen off uninteresting events for this tree visualization mode
        if ((eType == DataModelEvent.TELESCOPE_DELETED
            || eType == DataModelEvent.TELESCOPE_ADDED)
            && getGroupDataByLevel() != 3) {
            return;
        }

        if ((eType == DataModelEvent.DIRECTORY_DELETED
            || eType == DataModelEvent.DIRECTORY_ADDED)
            && getGroupDataByLevel() != 2) {
            return;
        }

        if (debug) {
            System.err.print(">>> firing ");
            switch (eType) {
                case DataModelEvent.PROGRAM_DELETED :
                    System.err.print("PROGRAM_DELETED");
                    break;

                case DataModelEvent.PROGRAM_ADDED :
                    System.err.print("PROGRAM_ADDED");
                    break;

                case DataModelEvent.OBSERVATION_DELETED :
                    System.err.print("OBSERVATION_DELETED");
                    break;

                case DataModelEvent.OBSERVATION_ADDED :
                    System.err.print("OBSERVATION_ADDED");
                    break;

                case DataModelEvent.DIRECTORY_DELETED :
                    System.err.print("DIRECTORY_DELETED");
                    break;

                case DataModelEvent.DIRECTORY_ADDED :
                    System.err.print("DIRECTORY_ADDED");
                    break;

                case DataModelEvent.TELESCOPE_ADDED :
                    System.err.print("TELESCOPE_ADDED");
                    break;

                case DataModelEvent.TELESCOPE_DELETED :
                    System.err.print("TELESCOPE_DELETED");
                    break;

                case DataModelEvent.FILE_ADDED :
                    System.err.print("FILE_ADDED");
                    break;

                case DataModelEvent.FILE_DELETED :
                    System.err.print("FILE_DELETED");
                    break;
            }
        }

        if (SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Reloader(ancestors));
        } else {
            try {
                SwingUtilities.invokeAndWait(new Reloader(ancestors));
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        if (debug) {
            System.err.flush();
            System.err.print(" [ ");
            for (int i = 0; i < path.length; i++) {
                System.err.print(path[i] + ", ");
                if (path[i] == null) {
                    System.err.println(
                        "\nChild " + child + " has a null element in path");
                }
            }
            System.err.print("]\n");
        }
    }

    /**
     * Simple way to update the entire tree.
     */
    public void updateTree() {
        setRootLabel(null);
        Object path[] = new Object[1];
        path[0] = root;
        fireTreeStructureChanged(this, path, null, null);
        theTree.expandPath(new TreePath(root));
    }

    /**
     * this is just a convenience class to
     * execute the fireTreeStructureChanged method
     * in a separate Thread
     *
     */

    class Reloader implements Runnable {
        Object[] path = null;
        public Reloader(Object[] path) {
            this.path = path;
        }
        public void run() {
            fireTreeStructureChanged(DBTreeTableModel.this, path, null, null);
        }
    }

    protected Enumeration getChildren(Object parent) {
        Object rc = null;
        if (parent == root) {
            if (groupDataByLevel == 0) {
                rc = dm.getProgramSet();
            }
            if (groupDataByLevel == 2) {
                rc = dm.getTelescopesSet();
            }
        } else if (parent instanceof DBContainer) {
            rc = ((DBContainer) parent).elements();
            if (rc == null) {
                System.err.println("No children");
            }
        } else {
            rc = null;
        }
        return (Enumeration) rc;
    }

    public Object[] getPathToRoot(Object child) {
        Object[] path = null;

        if (child instanceof DBOBComponent) {
            DBOBComponent child2 = (DBOBComponent) child;
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                path = new Object[4];
                path[3] = child2;
                path[2] = child2.getObservation(GasProp.TREE_DEFAULT);
                path[1] = ((DBContainer) path[2]).getFather();
                path[0] = getRoot();
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                path = new Object[5];
                path[4] = child2;
                path[3] = child2.getObservation(GasProp.TREE_TELESCOPE);
                path[2] = ((DBContainer) path[3]).getFather();
                path[1] = ((DBContainer) path[2]).getFather();
                path[0] = getRoot();
            }
        }
        if (child instanceof DBContainer) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                DBContainer child2 = (DBContainer) child;
                if (child2.isProgram()) {
                    path = new Object[2];
                    path[0] = getRoot();
                    path[1] = child;
                } else {
                    path = new Object[3];
                    path[0] = getRoot();
                    path[2] = child;
                    path[1] = child2.getFather();
                }
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                DBContainer child2 = (DBContainer) child;
                if (child2.isTelescope()) {
                    path = new Object[2];
                    path[0] = getRoot();
                    path[1] = child;
                } else if (child2.isProgram()) {
                    path = new Object[3];
                    path[0] = getRoot();
                    path[2] = child;
                    path[1] = child2.getFather();
                } else {
                    path = new Object[4];
                    path[0] = getRoot();
                    path[3] = child;
                    path[2] = child2.getFather();
                    path[1] = ((DBContainer) path[2]).getFather();
                }
            }
        }

        if (child instanceof RootNode) {
            path = new Object[1];
            path[0] = child;
        }
        return path;
    }

    public Vector getAllTheLeaves(Object father) {
        allTheLeaves.removeAllElements();
        rec_traverse(father);
        return allTheLeaves;
    }

    public void rec_traverse(Object father) {
        if (isLeaf(father)) {
            allTheLeaves.addElement(father);
            return;
        }

        for (Enumeration e = getChildren(father); e.hasMoreElements();) {
            rec_traverse(e.nextElement());
        }
    }

    /**
     * inner class RootNode
     * pretty much the same as for the DFSFileTreeTableModel
     *
     */

    class RootNode {
        private String label = null;

        public String toString() {
            return label;
        }

        void setLabel(String s) {
            label = s;
        }

        String getLabel() {
            return label;
        }
    }

    //  TreeExpansionListener interface

    public void treeCollapsed(TreeExpansionEvent event) {
        DBContainer tmpCont = null;
        if (event.getPath().getLastPathComponent() instanceof DBContainer) {
            tmpCont = (DBContainer) event.getPath().getLastPathComponent();
            dm.collapseContainer(tmpCont);
        }
    }

    public void treeExpanded(final TreeExpansionEvent event) {
        final Cursor tmpCursor = theTree.getCursor();
        theTree.setCursor(new Cursor(Cursor.WAIT_CURSOR));
        final Runnable expanderRun = new Runnable() {
            public void run() {
                DBContainer tmpCont = null;
                if (event.getPath().getLastPathComponent()
                    instanceof DBContainer) {
                    tmpCont =
                        (DBContainer) event.getPath().getLastPathComponent();
                    dm.expandContainer(tmpCont);
                }
                theTree.setCursor(tmpCursor);
            }
        };

        expanderRun.run();
    }

    ////// RECURSIVE PROCEDURE ////////////////////

    public void expandAllTheNodes(int row) {
        final TreePath tp = theTree.getPathForRow(row);
        final Runnable recExpander = new Runnable() {
            public void run() {
                Cursor tmpCursor = theTree.getCursor();
                recExpand(tp);
                theTree.setCursor(tmpCursor);
            }
        };

        new Thread(recExpander).start();
        //recExpander.run();
    }

    private void recExpand(TreePath tp) {
        if (tp == null) {
            return;
        }
        Object parent = tp.getLastPathComponent();
        theTree.expandPath(tp);
        if (isLeaf(parent)) {
            return;
        }
        for (Enumeration e = getChildren(parent); e.hasMoreElements();) {
            Object tmpObj = e.nextElement();
            recExpand(tp.pathByAddingChild(tmpObj));
        }
    }

    ///////////////////////////////////////////////////////

    public boolean isPopupSuitable(Object container) {
        if (!(container instanceof DBContainer)) {
            return false;
        }
        if (((DBContainer) container).isProgram()) {
            return true;
        }
        return false;
    }

    public boolean periodSelectionChanged() {
        return dm.init();
    }

    public void addSingleProgram(String program) {
        Cursor tmpCursor = theTree.getCursor();
        dm.addSingleProgram(program);
        theTree.setCursor(tmpCursor);
    }
}
// end of file DBTreeModel.java
