/**
 * @author N. Kornweibel
 * @version 1.0
 * This Model mapps a DataModel into an interface usable by JTree.
 * The root of the tree is a String, with Program children.
 * String->Program->Observation->OBComponent.
 * The class of a node is used often to determine the level in the tree.
 */

package org.eso.gasgano.gui;

import java.util.Enumeration;
import java.util.Vector;

import javax.swing.SwingUtilities;
import javax.swing.tree.TreePath;

import org.eso.gasgano.datamodel.DataModelChangeListener;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.InterContainer;
import org.eso.gasgano.datamodel.OBComponentFilter;
import org.eso.gasgano.datamodel.Observation;
import org.eso.gasgano.datamodel.Program;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.datamodel.gui.KeywordContainerOBComponent;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.SortedHashtable;

public class DFSFileTreeTableModel
    extends AbstractTreeTableModel
    implements DataModelChangeListener {

    private Vector treeListeners = null;
    private DFSDataModel dm = null;
    private RootNode root = new RootNode();
    private Vector displayCols = null;
    private int groupDataByLevel = 0;

    protected Vector allTheLeaves = new Vector();

    public DFSFileTreeTableModel(DFSDataModel d) {
        init(d);
    }

    private void init(DFSDataModel d) {
        dm = d;
        treeListeners = new Vector();
        setRootLabel(null);
        dm.addDataModelChangeListener(this);
        displayCols = GasProp.getProperties().fsDisplayColumns;
    }

    public int getGroupDataByLevel() {
        return groupDataByLevel;
    }

    public void setGroupDataByLevel(int gbLevel) {
        if (groupDataByLevel != gbLevel) {
            groupDataByLevel = gbLevel;
            setRootLabel(null);
            updateTree();
        }
    }

    public void setRootLabel(String name) {
        if (name == null) {
            String suffix = null;
            String filterStr = null;
            OBComponentFilter filter = dm.getFileFilter();
            if (filter != null) {
                filterStr = new String("Filter: " + filter.getDescription());
            } else {
                filterStr = new String("Unfiltered");
            }
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                suffix = new String("");
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                suffix = new String(" grouped by directory.");
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                suffix = new String(" grouped by telescope.");
            }
            name =
                new String(
                    "Displaying "
                        + dm.getFilesByName().size()
                        + " files"
                        + suffix
                        + " "
                        + filterStr
                        + ".");
        }
        root.setLabel(name);
    }

    public String getRootLabel() {
        String rc = null;
        if (root != null) {
            root.getLabel();
        }
        return rc;
    }

    public Object getChild(Object parent, int index) {
        Object rc = null;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getProgramAt(index);
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                rc = dm.getDirectoryAt(index);
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().elementAt(index);
            }
        } else if (parent instanceof Program) {
            Program prog = (Program) parent;
            rc = prog.getObservationAt(index);
        } else if (parent instanceof Observation) {
            Observation obs = (Observation) parent;
            int ic = obs.getIntermediateCount();
            if (index >= ic) {
                rc = obs.getComponentAt(index - ic);
            } else {
                rc = obs.getIntermediateAt(index);
            }
        } else if (parent instanceof InterContainer) {
            InterContainer inter = (InterContainer) parent;
            rc = inter.getComponentAt(index);
        } else if (parent instanceof SortedHashtable) {
            SortedHashtable dir = (SortedHashtable) parent;
            rc = dir.elementAt(index);
        } else {
            rc = null;
        }
        return rc;
    }

    public int getChildCount(Object parent) {
        int rc = 0;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getProgramCount();
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                rc = dm.getDirectoryCount();
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().size();
            }
        } else if (parent instanceof Program) {
            Program prog = (Program) parent;
            rc = prog.getObservationCount();
        } else if (parent instanceof InterContainer) {
            InterContainer inter = (InterContainer) parent;
            rc = inter.getComponentCount();
        } else if (parent instanceof Observation) {
            Observation obs = (Observation) parent;
            rc = obs.getComponentCount() + obs.getIntermediateCount();
        } else if (parent instanceof SortedHashtable) {
            SortedHashtable dir = (SortedHashtable) parent;
            rc = dir.size();
        } else if (parent instanceof OBComponent) {
            rc = 0;
        }
        return rc;
    }

    public int getIndexOfChild(Object parent, Object child) {
        int rc = -1;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getIndexOfProgram((Program) child);
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                rc = dm.getDirectoryIndex(child);
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().indexOf(child);
            }
        } else if (parent instanceof Program) {
            Program prog = (Program) parent;
            rc = prog.getIndexOfObservation((Observation) child);
        } else if (parent instanceof Observation) {
            Observation obs = (Observation) parent;
            if (child instanceof OBComponent) {
                rc =
                    obs.getIndexOfComponent((OBComponent) child)
                        + obs.getIntermediateCount();
            }
            if (child instanceof InterContainer) {
                rc = obs.getIndexOfIntermediate((InterContainer) child);
            }
        } else if (parent instanceof InterContainer) {
            InterContainer inter = (InterContainer) parent;
            rc = inter.getIndexOfComponent((OBComponent) child);
        } else if (parent instanceof SortedHashtable) {
            SortedHashtable dir = (SortedHashtable) parent;
            rc = dir.indexOf(child);
        } else {
            rc = -1;
        }
        return rc;
    }

    public int getColumnCount() {
        // the file entry plus the entries in DISPLAY_COLUMNS.
        return displayCols.size() + 1;
    }

    public String getColumnName(int column) {
        String rc = null;
        if (column == 0) {
            rc = "File";
        } else {
            rc = (String) displayCols.elementAt(column - 1);
        }
        return rc;
    }

    public Object getValueAt(Object node, int column) {
        Object rc = null;
        if (column > 0) {
            if (node instanceof KeywordContainerOBComponent) {
                String colData[] = null;
                colData = ((KeywordContainerOBComponent) node).getDisplayData();
                if (colData != null && colData.length >= column) {
                    rc = colData[column - 1];
                }
            }
        } else {
            rc = node;
        }
        return rc;
    }

    public void toggleColumn() {
        displayCols = GasProp.getProperties().fsDisplayColumns;
    }

    public Class getColumnClass(int column) {
        Class rc = null;
        if (column == 0) {
            rc = TreeTableModel.class;
        } else {
            rc = String.class;
        }
        return rc;
    }

    public Object getRoot() {
        if (root == null) {
            setRootLabel(null);
        }
        return root;
    }

    public boolean isLeaf(Object node) {
        boolean rc = false;
        if (node instanceof OBComponent) {
            rc = true;
        }
        return rc;
    }

    public void valueForPathChanged(TreePath path, Object newValue) {
        // tree is not editable...ignore!
    }

    /**
     * this method is  responsible for triggering 
     * TreeStructureChanged events
     * this is done via the invokeLater in a separate Thread
     *
     * the decision on whether to send or not cumulated
     * events is left to the DataModel class<p>
     * 
     * most of the code is dedicated to debugging output
     * The DataModelEvent is a carrier of all information needed
     * to handle the events. Some of them are ignored
     * when the tree representation is not suitable for them<p>
     *
     * For all the DELETE events, the temporal sequence is crucial:
     * first a getRootToPath has to be called, *then* the DataModel
     * must be changed, and *then* the Event can be fired.<p>
     *
     * To guarantee this, the modification of the DataModel is
     * handled personally by this method, in the cases 
     * FILE_DELETED and OBSERVATION_DELETED
     *
     * @author: Michele Zamparelli
     * @param DataModelEvent e    the event received by the DataModel
     * @see: DataModel 
     * @exception: none
     */

    public void dataModelChanged(DataModelEvent e) {
        boolean debug = false;

        int delay = 10; // veeery processor dependent
        int eType = 0;
        Object path[] = null;
        Object ancestors[] = null;
        Object child = e.getFile();

        if (PropertyDB.getInstance().isDebug(PropertyDB.debugDataModelEvent)) {
            debug = true;
        }

        eType = e.getEvent();
        if (child == null) {
            if (debug) {
                System.err.println(
                    "DataModelChange Event sent with no OBComponent/Program/Observation Type:"
                        + eType);
            }
            return;
        }
        path = getPathToRoot(child);

        ancestors = new Object[path.length - 1];
        for (int z = 0; z < ancestors.length; z++) {
            ancestors[z] = path[z];
        }

        // here we remove single files if we have to
        if (eType == DataModelEvent.FILE_DELETED) {
            dm.removeObComponent((OBComponent) child);
        }

        if ((e.getTreeType() != getGroupDataByLevel())
            && (e.getTreeType() != DataModelEvent.TREE_GENERAL)) {
            return;
        }

        if (debug) {
            System.out.print(">>> firing ");
            switch (eType) {
                case DataModelEvent.PROGRAM_DELETED :
                    System.out.print("PROGRAM_DELETED");
                    break;

                case DataModelEvent.PROGRAM_ADDED :
                    System.out.print("PROGRAM_ADDED");
                    break;

                case DataModelEvent.OBSERVATION_DELETED :
                    System.out.print("OBSERVATION_DELETED");
                    break;

                case DataModelEvent.OBSERVATION_ADDED :
                    System.out.print("OBSERVATION_ADDED");
                    break;

                case DataModelEvent.INTERMEDIATE_DELETED :
                    System.out.print("INTERMEDIATE_DELETED");
                    break;

                case DataModelEvent.INTERMEDIATE_ADDED :
                    System.out.print("INTERMEDIATE_ADDED");
                    break;

                case DataModelEvent.DIRECTORY_DELETED :
                    System.out.print("DIRECTORY_DELETED");
                    break;

                case DataModelEvent.DIRECTORY_ADDED :
                    System.out.print("DIRECTORY_ADDED");
                    break;

                case DataModelEvent.TELESCOPE_ADDED :
                    System.out.print("TELESCOPE_ADDED");
                    break;

                case DataModelEvent.TELESCOPE_DELETED :
                    System.out.print("TELESCOPE_DELETED");
                    break;

                case DataModelEvent.FILE_ADDED :
                    System.out.print("FILE_ADDED");
                    break;

                case DataModelEvent.FILE_DELETED :
                    System.out.print("FILE_DELETED");
                    break;
            }

            System.out.print(" Level " + e.getTreeType());
        }

        if (debug) {
            System.out.flush();
            System.out.print(" [ ");
            for (int i = 0; i < path.length; i++) {
                System.out.print(path[i] + ", ");
                if (path[i] == null) {
                    System.out.println(
                        "\nChild " + child + " has a null element in path");
                }
            }
            System.out.print("]\n");
        }

        if (SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Reloader(ancestors));
        } else {
            try {
                SwingUtilities.invokeAndWait(new Reloader(ancestors));
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        // for DFS03623 :
        // setRootLabel has been moved here because it was refreshing
        // the treeview label BEFORE the file was removed from
        // indexes[NAME_INDEX], thus displaying always the previous files count.
        setRootLabel(null);
    }

    /**
     * Simple way to update the entire tree.
     */
    public void updateTree() {
        Object path[] = new Object[1];
        path[0] = root;

        // fireTreeModelEvent(new TreeModelEvent(this, o));
        fireTreeStructureChanged(this, path, null, null);
    }

    /**
     * this is just a convenience class to
     * execute the fireTreeStructureChanged method
     * in a separate Thread
     *
     */

    class Reloader implements Runnable {
        Object[] path = null;
        public Reloader(Object[] path) {
            this.path = path;
        }
        public void run() {
            // TODO: why is there a yield() here ???
            Thread.yield();
            
            fireTreeStructureChanged(
                DFSFileTreeTableModel.this,
                path,
                null,
                null);
        }
    }

    class RootNode {
        private String label = null;

        public String toString() {
            return label;
        }

        void setLabel(String s) {
            label = s;
        }

        String getLabel() {
            return label;
        }
    }

    protected Enumeration getChildren(Object parent) {
        Object rc = null;
        if (parent == root) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                rc = dm.getProgramSet().elements();
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                rc = dm.getDirectories();
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                rc = dm.getTelescopesSet().elements();
            }
        } else if (parent instanceof Program) {
            Program prog = (Program) parent;
            rc = prog.getObservationSet().elements();
        } else if (parent instanceof InterContainer) {
            InterContainer inter = (InterContainer) parent;
            rc = inter.getComponents().elements();
        } else if (parent instanceof Observation) {
            Observation obs = (Observation) parent;
            rc = obs.getDescendants().elements();
        } else if (parent instanceof SortedHashtable) {
            SortedHashtable dir = (SortedHashtable) parent;
            rc = dir.elements();
        } else {
            rc = null;
        }
        return (Enumeration) rc;
    }

    /**
     * Get the directory for the given node, or null if not found.
     * this is guaranteed to be called upon a non leaf
     */
    public String getDirectory(Object parent) {
        String rc = null;
        if (parent == root) {
            return rc;
        }
        rc = parent.toString();
        if (parent instanceof Program) {
            Program prog = (Program) parent;
            if (prog.getObservationCount() > 0
                && prog.getObservationAt(0).getComponentCount() > 0) {
                // a bit tedious, but safe!
                rc =
                    prog
                        .getObservationAt(0)
                        .getComponentAt(0)
                        .getObservation(GasProp.TREE_DIRECTORY)
                        .getProgram()
                        .getDirectory()
                        .toString();
            }
        }
        if (parent instanceof Observation) {
            Observation tmpObs =
                (Observation) parent;
            if (tmpObs.getComponentCount() > 0) {
                rc =
                    tmpObs
                        .getComponentAt(0)
                        .getObservation(GasProp.TREE_DIRECTORY)
                        .getProgram()
                        .getDirectory()
                        .toString();
            }
        }
        return rc;
    }

    public Object[] getPathToRoot(Object child) {
        Object[] path = null;
        boolean interMediate = false;
        int interOffset = 0;

        if (child instanceof OBComponent) {
            OBComponent child2 = (OBComponent) child;
            if (child2.getIntermediateGrouping() != null) {
                interMediate = true;
                interOffset = 1;

                if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                    path = new Object[5];
                    path[4] = child2;
                    path[3] = child2.getObservation(GasProp.TREE_DEFAULT);
                    path[2] = ((InterContainer) path[3]).getObservation();
                    path[1] = ((Observation) path[2]).getProgram();
                    path[0] = getRoot();
                }
                if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                    path = new Object[6];
                    path[5] = child2;
                    path[4] = child2.getObservation(GasProp.TREE_DIRECTORY);
                    path[3] = ((InterContainer) path[4]).getObservation();
                    path[2] = ((Observation) path[3]).getProgram();
                    path[1] =
                        ((Program) path[2]).getDirectory();
                    path[0] = getRoot();
                }
                if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                    path = new Object[6];
                    path[5] = child2;
                    path[4] = child2.getObservation(GasProp.TREE_TELESCOPE);
                    path[3] = ((InterContainer) path[4]).getObservation();
                    path[2] = ((Observation) path[3]).getProgram();
                    path[1] =
                        ((Program) path[2]).getTelescope();
                    path[0] = getRoot();
                }
            } else {
                if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                    path = new Object[4];
                    path[3] = child2;
                    path[2] = child2.getObservation(GasProp.TREE_DEFAULT);
                    path[1] = ((Observation) path[2]).getProgram();
                    path[0] = getRoot();
                }
                if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                    path = new Object[5];
                    path[4] = child2;
                    path[3] = child2.getObservation(GasProp.TREE_DIRECTORY);
                    path[2] = ((Observation) path[3]).getProgram();
                    path[1] =
                        ((Program) path[2]).getDirectory();
                    path[0] = getRoot();
                }
                if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                    path = new Object[5];
                    path[4] = child2;
                    path[3] = child2.getObservation(GasProp.TREE_TELESCOPE);
                    path[2] = ((Observation) path[3]).getProgram();
                    path[1] =
                        ((Program) path[2]).getTelescope();
                    path[0] = getRoot();
                }
            }
        }
        if (child instanceof Program) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                path = new Object[2];
                path[0] = getRoot();
                path[1] = child;
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                path = new Object[3];
                path[0] = getRoot();
                path[1] = ((Program) child).getDirectory();
                path[2] = child;
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                path = new Object[3];
                path[0] = getRoot();
                path[1] = ((Program) child).getTelescope();
                path[2] = child;
            }
        }

        if (child instanceof InterContainer) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                path = new Object[4];
                path[3] = child;
                path[2] = ((InterContainer) child).getObservation();
                path[1] = ((Observation) path[2]).getProgram();
                path[0] = getRoot();
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                path = new Object[5];
                path[4] = child;
                path[3] = ((InterContainer) child).getObservation();
                path[2] = ((Observation) path[3]).getProgram();
                path[1] = ((Program) path[2]).getDirectory();
                path[0] = getRoot();

            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                path = new Object[5];
                path[4] = child;
                path[3] = ((InterContainer) child).getObservation();
                path[2] = ((Observation) path[3]).getProgram();
                path[1] = ((Program) path[2]).getTelescope();
                path[0] = getRoot();
            }
        } else if (child instanceof Observation) {
            if (groupDataByLevel == GasProp.TREE_DEFAULT) {
                path = new Object[3];
                path[0] = getRoot();
                path[1] = ((Observation) child).getProgram();
                path[2] = child;
            }
            if (groupDataByLevel == GasProp.TREE_DIRECTORY) {
                path = new Object[4];
                path[0] = getRoot();
                path[1] = ((Observation) child).getProgram().getDirectory();
                path[2] = ((Observation) child).getProgram();
                path[3] = child;
            }
            if (groupDataByLevel == GasProp.TREE_TELESCOPE) {
                path = new Object[4];
                path[0] = getRoot();
                path[1] = ((Observation) child).getProgram().getTelescope();
                path[2] = ((Observation) child).getProgram();
                path[3] = child;
            }
        }

        if (child instanceof SortedHashtable) {
            path = new Object[2];
            path[0] = getRoot();
            path[1] = child;
        }

        if (child instanceof RootNode) {
            path = new Object[1];
            path[0] = child;
        }
        return path;
    }

    ///////////////// RECURSIVE PROCEDURES /////////////////

    public Vector getAllTheLeaves(Object father) {
        allTheLeaves.removeAllElements();
        rec_traverse(father);
        return allTheLeaves;
    }

    public void rec_traverse(Object father) {
        if (isLeaf(father)) {
            allTheLeaves.addElement(father);
            return;
        }

        for (Enumeration e = getChildren(father); e.hasMoreElements();) {
            rec_traverse(e.nextElement());
        }
    }
}
