/**
 * @author N. Kornweibel
 * @version 1.0
 * A simple Frame for viewing, editing, saving and printing the output
 * from objects implementing the Report interface.
 *
 * The class uses a Report interface to query report related
 * information (only column widths at the moment).
 */

package org.eso.gasgano.gui;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.PrintJob;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Properties;
import java.util.Random;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import org.eso.dfs.gui.SwingFileChooser;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.Report;
import org.eso.gasgano.tools.TestReport;

public class ReportViewer extends JFrame implements ActionListener {

    private String fileName;
    private JTextArea ta;
    private Font printerFont = null;

    private int columnWidths[] = null;

    private static final String selectFontLabel = "Select Font";
    private static final String saveAsLabel = "Save As...";
    private static final String quitLabel = "Quit";
    private static final String printLabel = "Print";

    public static void main(String args[]) {

        ReportViewer rep = new ReportViewer("Test Report", new TestReport());
        rep.pack();
        rep.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });
        rep.setSize(1100, 500);
        // move it away from the top LH corner.
        rep.setLocation(100, 300);
        rep.setVisible(true);
    }

    private void setColumnWidths(int c[]) {
        columnWidths = new int[c.length];
        int total = 0;
        for (int i = 0; i < c.length; i++) {
            total = total + c[i] + 2;
            columnWidths[i] = total;
        }
    }

    /**
     * Constructor creates a temporary file with a PrintWriter,
     * generates the report and reads the report from the
     * temporary file.
     */

    public ReportViewer(String title, Report rep) {
        super(title);
        Random rnd = new Random();
        fileName = new String("/tmp/dfsreport" + rnd.nextInt());
        String errorString = null;
        boolean reportCancelled = false;
        BufferedReader in;
        String tmpString = null;
        // set the font
        DFSDataModel dm = DFSDataModel.getDataModel();
        if ((tmpString = (String) PropertyDB.getInstance().getProperty("PRINT_FONT"))
            != null) {
            // tmpString has the format Family:size
            int colonPost = tmpString.indexOf(':');
            printerFont =
                new Font(
                    tmpString.substring(0, colonPost),
                    Font.PLAIN,
                    Integer.parseInt(
                        tmpString.substring(
                            colonPost + 1,
                            tmpString.length())));
        } else {
            printerFont = new Font("Dialog", Font.PLAIN, 5);
        }

        try {
            FileWriter fw = new FileWriter(fileName);
            PrintWriter pw = new PrintWriter(fw, true);
            rep.setPrintWriter(pw);

            ReportRunner runner = new ReportRunner(rep);

            ReportDialog runReport =
                new ReportDialog(
                    this,
                    "Report",
                    "Generating Report...",
                    true,
                    runner);
            runReport.run();
            reportCancelled = runReport.cancelSelected();
            if (reportCancelled) {
                errorString = new String("Report cancelled.");
            } else {
                if (runner.reportStatus() == false) {
                    errorString = new String("Error generating report.");
                } else {
                    setColumnWidths(rep.getColumnWidths());
                }
                pw.flush();
                pw.close();
            }
        } catch (IOException e) {
            errorString =
                new String("Error generating report: " + e.toString());
        }
        if (errorString != null) {
            JOptionPane.showMessageDialog(this, errorString);
        }

        // open the file and read in the contents.
        // this section will produce memory problems when attempting to 
        // show very large reports.
        // Although it is unusual to catch an OutOfMemory error, we will
        // in this case to display an error message and correctly clean
        // up the report.

        try {
            in = new BufferedReader(new FileReader(fileName));
            StringBuffer text = new StringBuffer();
            String s = "";
            while ((s = in.readLine()) != null) {
                text.append("\n" + s);
            }
            ta = new JTextArea(text.toString());
            ta.setFont(printerFont);
            ta.setEditable(false);
            getContentPane().add(new JScrollPane(ta));
        } catch (OutOfMemoryError e) {
            JOptionPane.showMessageDialog(
                this,
                "Error reading file: "
                    + fileName
                    + "\n the file is too large for display.");
        } catch (IOException e) {
            JOptionPane.showMessageDialog(
                this,
                "Error reading file: " + fileName);
        }
        setupMenu();
    }

    private void setupMenu() {
        // Setup the menu items....
        MenuBar mb = new MenuBar();
        Menu fileMenu = new Menu("File");
        MenuItem printMenuItem = new MenuItem(printLabel);
        printMenuItem.addActionListener(this);
        // printMenuItem.setEnabled(false);
        fileMenu.add(printMenuItem);
        MenuItem saveAsMenuItem = new MenuItem(saveAsLabel);
        saveAsMenuItem.addActionListener(this);
        fileMenu.add(saveAsMenuItem);
        MenuItem quitMenuItem = new MenuItem(quitLabel);
        quitMenuItem.addActionListener(this);
        fileMenu.add(quitMenuItem);
        mb.add(fileMenu);

        Menu fontMenu = new Menu("Options");
        MenuItem fontItem = new MenuItem(selectFontLabel);
        fontItem.addActionListener(this);
        fontMenu.add(fontItem);

        mb.add(fontMenu);
        setMenuBar(mb);
    }

    // implementation of the ActionListener interface
    public void actionPerformed(ActionEvent a) {
        if (a.getActionCommand() == quitLabel) {
            File f = new File(fileName);
            boolean deleted = true;
            try {
                deleted = f.delete();
            } catch (SecurityException e) {
                deleted = false;
            }
            if (!deleted) {
                JOptionPane.showMessageDialog(
                    this,
                    "Failed to delete temporary report file:\n" + fileName);
            }
            setVisible(false);
            // dispose();
        } else if (a.getActionCommand() == saveAsLabel) {
            JFileChooser fc = new SwingFileChooser();
            fc.setDialogTitle("Enter report file name");
            fc.setDialogType(JFileChooser.SAVE_DIALOG);
            fc.setMultiSelectionEnabled(false);
            String selectedFile = null;
            if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
                selectedFile = fc.getSelectedFile().getPath();
            }
            if (selectedFile != null && saveFile(selectedFile)) {
                JOptionPane.showMessageDialog(
                    this,
                    "Report saved to: " + selectedFile);
            }
        } else if (a.getActionCommand() == printLabel) {
            Properties p = new Properties();
            String orientPrint = new String("landscape");
            String printerName = new String("ppbla3");
            String tmpPrint = null;
            DFSDataModel dm = DFSDataModel.getDataModel();
            if ((tmpPrint =
                (String) PropertyDB.getInstance().getProperty("PRINT_ORIENTATION"))
                != null) {
                if (tmpPrint.equals("p")) {
                    orientPrint = "portrait";
                } else {
                    orientPrint = "landscape";
                }
            }

            if ((tmpPrint = (String) PropertyDB.getInstance().getProperty("PRINTER_NAME"))
                != null) {
                printerName = tmpPrint;
            }

            // the following properties apply only to UNIX java VM.
            // need to put these in general properties box.
            p.put("awt.print.numCopies", "1");
            p.put("awt.print.paperSize", "A4");
            p.put("awt.print.destination", "printer");
            p.put("awt.print.orientation", orientPrint);
            p.put("awt.print.printer", printerName);
            PrintJob pjob = null;
            try {
                pjob = getToolkit().getPrintJob(this, "Fits File Report", p);
            } catch (NullPointerException e) {
            	JOptionPane.showMessageDialog(
                        this,
                        "Unable to open print dialog: please check the printer configuration and your local print server",
                        "Error",
                        JOptionPane.ERROR_MESSAGE);
            }
 
            if (pjob != null) {
                Graphics pg = pjob.getGraphics();
                if (pg != null) {
                    pg.setFont(printerFont);
                    int lineHeight =
                        pg.getFontMetrics().getAscent()
                            + pg.getFontMetrics().getDescent();
                    BufferedReader in;
                    int charWidth = pg.getFontMetrics().charWidth('X');
                    try {
                        in = new BufferedReader(new FileReader(fileName));
                        String s = "";
                        int lines = 2;
                        int pageNum = 1;
                        int height = pjob.getPageDimension().height;
                        int pageWidth =
                            pjob.getPageDimension().width - charWidth;
                        while ((s = in.readLine()) != null) {
                            lines++;
                            if ((lines + 4) * lineHeight > height) {
                                pageNum++;
                                pg.dispose();
                                pg = pjob.getGraphics();
                                pg.setFont(printerFont);
                                lines = 6;
                                printString(
                                    "Page: " + pageNum,
                                    pg,
                                    4 * lineHeight,
                                    charWidth,
                                    lineHeight,
                                    pageWidth,
                                    -1);
                            }
                            int linesPrinted =
                                printString(
                                    s,
                                    pg,
                                    lines * lineHeight,
                                    charWidth,
                                    lineHeight,
                                    pageWidth,
                                    -1);
                            lines += (linesPrinted - 1);
                        }
                    } catch (IOException e) {
                        JOptionPane.showMessageDialog(
                            this,
                            "Error reading file: " + fileName);
                    }
                    pg.dispose();
                }
                pjob.end();
            }
        } else if (a.getActionCommand() == selectFontLabel) {
            FontChooser fc = new FontChooser(printerFont);
            if (JOptionPane
                .showConfirmDialog(
                    this,
                    fc,
                    "Edit fonts",
                    JOptionPane.OK_CANCEL_OPTION,
                    JOptionPane.PLAIN_MESSAGE)
                == JOptionPane.OK_OPTION) {
                printerFont = fc.getNewFont();
                ta.setFont(printerFont);
                ta.repaint();
            }
        }
    }

    /** 
     * The printString() routine prints a line of text - it will
     * wrap the text if the line exceeds maxWidth.
     * The method returns the number of lines printed.
     */

    private int printString(
        String s,
        Graphics pg,
        int height,
        int width,
        int lineHeight,
        int pageWidth,
        int colIndex) {
        int rc = 1;
        int fromIndex = 0;
        int toIndex = 0;
        do {
            toIndex = s.indexOf('\t', fromIndex);
            String field;
            try {
                if (toIndex == -1) {
                    field = s.substring(fromIndex);
                } else {
                    field = s.substring(fromIndex, toIndex);
                }
            } catch (StringIndexOutOfBoundsException e) {
                field = null;
            }
            if (field != null) {
                int offset = 30;
                if (colIndex >= 0
                    && columnWidths != null
                    && colIndex < columnWidths.length) {
                    offset = offset + (columnWidths[colIndex] * width);
                }
                if (offset + (field.length() * width) > pageWidth) {
                    // wrap the line
                    String toBePrinted = null;
                    int lastSpaceInField = (pageWidth - offset) / width;
                    if (lastSpaceInField >= field.length()) {
                        lastSpaceInField = field.length() - 1;
                    }
                    int endOfField = lastSpaceInField;
                    while (lastSpaceInField > 0) {
                        if (field.charAt(lastSpaceInField) == ' ') {
                            break;
                        } else {
                            lastSpaceInField--;
                        }
                    }
                    if (lastSpaceInField == 0) {
                        lastSpaceInField = endOfField;
                    }
                    toBePrinted = field.substring(0, lastSpaceInField);
                    pg.drawString(toBePrinted, offset, height);
                    if (lastSpaceInField != field.length() - 1) {
                        String remainder =
                            s.substring(fromIndex + lastSpaceInField + 1);
                        rc
                            += printString(
                                remainder,
                                pg,
                                height + lineHeight,
                                width,
                                lineHeight,
                                pageWidth,
                                colIndex);
                    }
                    break;
                } else {
                    pg.drawString(field, offset, height);
                }
            }
            fromIndex = toIndex + 1;
            colIndex++;
        }
        while (toIndex != -1);
        return rc;
    }

    private int[] setupColumns(String s, FontMetrics fm) {
        int columns[] = null;
        int nCols = 1;
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == '\t') {
                nCols++;
            }
        }
        columns = new int[nCols];
        int from = 0;
        int colIndex = 0;
        int total = 0;
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == '\t') {
                int columnWidth = fm.stringWidth("X") * ((i - from));
                columns[colIndex] = total + columnWidth;
                total = total + columnWidth;
                colIndex++;
                from = i;
            }
        }
        return columns;
    }

    boolean saveFile(String name) {
        boolean rc = true;
        if (name == null) {
            return false;
        }
        try {
            FileWriter fw = new FileWriter(name);
            PrintWriter pw = new PrintWriter(fw);
            FileReader fr = new FileReader(fileName);
            BufferedReader br = new BufferedReader(fr);
            String s = br.readLine();
            while (s != null) {
                pw.println(s);
                s = br.readLine();
            }
            pw.flush();
            pw.close();
        } catch (IOException e) {
            rc = false;
            JOptionPane.showMessageDialog(
                this,
                "Error saving report: " + e.toString());
        }
        return rc;
    }
}
// end of file ReportViewer.java.
