/**
 * @author N. Kornweibel
 * @version 1.0
 * The BinaryKeywordExpr, an abstract class which is used as the
 * base for building binary operators: =, <, >, <=, etc.
 */

package org.eso.gasgano.keyword;

public abstract class BinaryKeywordExpr extends DefaultBoolKeywordExpr {

    static final long serialVersionUID = 4273027700686012321L;
    
    private Object lhs = null;
    private Object rhs = null;

    // two members to optimize evaluation of the expression.
    // An expression containing a String can never be evaluated as a
    // float expression and an expression containing a Float can never
    // be evaluated as a String. All other combinations are possible.
    private boolean neverEvalAsFloatExpr = false;
    private boolean neverEvalAsStringExpr = false;

    /**
     * @param l The left hand value.
     * @param r The right hand value.
     */
    public BinaryKeywordExpr(Object l, Object r) {
        lhs = l;
        rhs = r;
        setAllowsChildren(false);
    }

    public boolean isStringCompatible(Object arg) {
        boolean rc = false;
        if (arg instanceof String) {
            neverEvalAsFloatExpr = true;
            rc = true;
        } else if (arg instanceof Keyword) {
            rc = true;
        }
        return rc;
    }

    public boolean isFloatCompatible(Object arg) {
        boolean rc = false;
        if (arg instanceof Float) {
            neverEvalAsStringExpr = true;
            rc = true;
        } else if (arg instanceof Keyword) {
            Keyword k = (Keyword)arg;
            try {
                Float f = k.getNumericValue();
                rc = true;
            } catch (NumberFormatException e) {
                rc = false;
            }
        } 
        return rc;
    }

    /**
     * @param arg The object whose Float value is required.
     * @return The Float object value for the argument.
     */
    public Float getFloat(Object arg) 
            throws ExpressionEvaluationException {
        Float rc = null;
        if (arg instanceof Float) {
            rc = (Float)arg;
        } else if (arg instanceof Keyword) {
            Keyword k = (Keyword)arg;
            try {
                rc = k.getNumericValue();
            } catch (NumberFormatException e) {
                throw new ExpressionEvaluationException(e.getMessage());
            }
        } else {
            throw new ExpressionEvaluationException(
                "argument is not a Float: "+arg.toString());
        }
        return rc;
    }

    /**
     * @param arg The object whose String value is required.
     * @return The String object value for the argument.
     */
    public static String getString(Object arg) 
            throws ExpressionEvaluationException {
        String rc = null;
        if (arg instanceof String) {
            rc = (String)arg;
            // remove case sensitivity.
            rc = rc.toUpperCase();
        } else if (arg instanceof Keyword) {
            Keyword k = (Keyword)arg;
            rc = k.getValue();
        } else {
            throw new ExpressionEvaluationException(
                "argument is not a String: "+arg.toString());
        }
        return rc;
    }

    public void resolveKeyword(Object arg, KeywordSet s) {
        if (arg instanceof Keyword && s != null) {
            Keyword k = (Keyword)arg;
            Keyword val = s.getKeyword(k.getName());
            if (val == null) {
                k.setValue("");
            } else {
                k.setValue(val.getValue().toUpperCase());
            }
        }
    }

    public abstract boolean eval(float l, float r);

    public abstract boolean eval(String l, String r);

    public boolean evaluate(KeywordSet s) 
            throws ExpressionEvaluationException {
        boolean rc = false;
        resolveKeyword(lhs, s);
        resolveKeyword(rhs, s);
        if (!neverEvalAsFloatExpr && isFloatCompatible(lhs)
                && isFloatCompatible(rhs)) {
            // evaluate the expression as a float.
            rc = eval(getFloat(lhs).floatValue(),
                getFloat(rhs).floatValue());
        } else if (!neverEvalAsStringExpr 
                && isStringCompatible(lhs) && isStringCompatible(rhs)) {
            // evaluate the expression as a String.
            rc = eval(getString(lhs), getString(rhs));
        } else {
	  StringBuffer tmpString = new StringBuffer("Expression arguments are not compatible");

	  if ( lhs instanceof Keyword  && ((Keyword)lhs).getValue() == "") { 
	    tmpString = new StringBuffer("Keyword "+((Keyword)lhs).getName()+" not included");
	    return false;
	  } else { 
	    if ( lhs instanceof Keyword) { 
	      Keyword tmpKey = (Keyword)lhs;
	      tmpString.append(tmpKey.getName() + " ("+tmpKey.getValue()+")");
	    } else { 
	      tmpString.append(lhs);
	    }

	    if ( rhs instanceof Keyword) { 
	      Keyword tmpKey = (Keyword)rhs;
	      tmpString.append(" "+tmpKey.getName() + " ("+tmpKey.getValue()+")");
	    } else { 
	      tmpString.append(rhs);
	    }
	  }
            throw new ExpressionEvaluationException( tmpString.toString() );
        }
        return rc;
    }

    public abstract String getOpString();

    public String toString() {
        String rc = new String(KeywordExpression.valueToString(lhs)
            +" "+getOpString()+" "+KeywordExpression.valueToString(rhs));
        return rc;
    }

    public Object getLeftValue() {
        return lhs;
    }

    public Object getRightValue() {
        return rhs;
    }
}
// end of file BinaryKeywordExpr.java
