/**
 * @author N. Kornweibel
 * @version 1.0
 * A data object for storing an expression based on keywords.
 */

package org.eso.gasgano.keyword;

import java.io.Serializable;
import java.util.Enumeration;

public class KeywordExpression implements Serializable {

    private String asString = null;
    private DefaultBoolKeywordExpr expressionTree = null;
    static final long serialVersionUID = 7804384734795496229L;

    public KeywordExpression() {
    }

    /**
     * @param exp The tree data structure representation of the expression.
     */
    public KeywordExpression(DefaultBoolKeywordExpr exp) {
        setExpressionTree(exp);
    }

    /**
     * @param n The tree data structure representation of the expression.
     */
    public void setExpressionTree(DefaultBoolKeywordExpr n) {
        expressionTree = n;
        asString = null;
    }
    
    /**
     * Create and return a deep copy of this KeywordExpression.
     * @return KeywordExpression
     */
    public KeywordExpression copy() {
        return new KeywordExpression(getExpressionTree().copy());
    }
    

    /**
     * @return The tree data structure representation of the expression.
     */
    public DefaultBoolKeywordExpr getExpressionTree() {
        return expressionTree;
    }

    /**
     * @param s The StringBuffer used to build the expression string.
     * @param n The tree data structure representation of the expression.
     * Warning: This is a recursive method.
     */
    private void resolveExpressionToString(StringBuffer s,
            DefaultBoolKeywordExpr n) {
        s.append("(");
        if (n.getChildCount() > 0) {
            Enumeration e = n.children();
            boolean firstPrinted = false;
            while (e.hasMoreElements()) {
                DefaultBoolKeywordExpr expr 
                    = (DefaultBoolKeywordExpr)e.nextElement();
                if (firstPrinted) {
                    s.append(" "+n.getOpString()+" ");
                } else {
                    firstPrinted = true;
                }
                resolveExpressionToString(s, expr);
            }
        } else {
            s.append(n.toString());
        }
        s.append(")");
    }

    /**
     * @return The human readable String representation of this expression.
     */
    public String toString() {
        if (asString == null) {
            StringBuffer s = new StringBuffer();
            if (expressionTree != null) {
                resolveExpressionToString(s, expressionTree);
            }
            asString = s.toString();
        }
        return asString;
    }


    /**
     * @param kws The KeywordSet representing the data for the expression.
     * @return True if this expression evaluates to true in the context of
     * the KeywordSet kws.
     */
    public boolean evaluate(KeywordSet kws)
            throws ExpressionEvaluationException {
        return expressionTree.evaluate(kws);
    }

    public static Object stringToValue(String arg) {
        Object rc = null;
        if (arg.startsWith("\"")) {
            // assume that the first and last characters are '"'.
            try {
                rc = new String(arg.substring(1, arg.lastIndexOf("\"")));
            } catch (StringIndexOutOfBoundsException e) {
                rc = null;
            }
        } else {
            try {
                rc = Float.valueOf(arg);
            } catch (NumberFormatException e) {
                // it must be a Keyword.
                rc = new Keyword(arg, null);
            }
        }
        return rc;
    }

    public static String valueToString(Object arg) {
        String rc = null;
        if (arg instanceof String) {
            rc = new String("\""+(String)arg+"\"");
        } else if (arg instanceof Keyword) {
            Keyword k = (Keyword)arg;
            rc = k.getName();
        } else {
            rc = arg.toString();
        }
        return rc;
    }
} // end of class KeywordExpression

// end of file KeywordExpression.java
