package org.eso.gasgano.keyword;

import java.util.HashMap;
import java.util.Map;

import org.eso.dfs.util.SharedList;

/**
 * @author Nick Kornweibel
 * 
 * This class has two functions, which perhaps should have been
 * split between two classes, but is not:
 * 
 * 1. The class is an extension of the SharedList specialized
 * for managing keywords. The list does not store keywords,
 * but the entries in the list are valid keyword names
 * (according to Keyword.validName()).
 * 
 * 2. The class maintains a static  (i.e. singleton) registry
 * of SharedKeywordList objects, that can be access by name. There
 * are also some names defined that refer to shared keyword lists
 * in the registry. SharedKeywordLists objects do not have to be
 * taken from the registry or exist in the registry, it is simply
 * a means to share keyword lists by name.
 * 
 * The registry access or SharedKeywordList objects are NOT thread safe.
 */
public class SharedKeywordList extends SharedList {

    // the (singleton) registry of keyword lists.
    private static Map keywordListRegistry = new HashMap();
    
    /**
     * Name of the default or standard keyword list in the registry
     */
    public static final String DefaultKeywordList = "DefaultKeywordList";
    
    static {
        SharedKeywordList l = new SharedKeywordList(40);
        l.setSortList(true);
        putListToRegistry(l, DefaultKeywordList);
    }
    
    /**
     * Get the named shared keyword list from the registry, or null if not
     * present.
     * 
     * @param name
     * @return SharedKeywordList the named shared keyword list from the 
     * registry, or null if not present
     */
    public static SharedKeywordList getListFromRegistry(String name) {
        return (SharedKeywordList)keywordListRegistry.get(name);
    }

    /**
     * Add the given keyword list to the registry, replacing the one
     * there of that name, if present.
     * @param l SharedKeywordList to be added
     * @param name the name of the list
     */
    public static void putListToRegistry(SharedKeywordList l, String name) {
        keywordListRegistry.put(name, l);
    }

    /**
     * Constructor for SharedKeywordList.
     * @param initialCapacity
     */
    public SharedKeywordList(int initialCapacity) {
        super(initialCapacity);
    }

    /**
     * Constructor for SharedKeywordList.
     */
    public SharedKeywordList() {
        super();
    }

    /**
     * This method will only add the object to the list if it is
     * a valid keyword name (per Kerword.validName()).
     * If the entry is not a valid keyword name, the entry is
     * (silenty) not added.
     * @see org.eso.dfs.util.SharedList#add(java.lang.Object)
     */
    public void add(Object o) {
        String name = o == null? "" : o.toString();
        if (Keyword.validName(name)) {
            super.add(o);
        }
    }
}
