package org.eso.cpl;

import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.filechooser.FileFilter;
import org.eso.cpl.gui.ExtensionFileFilter;

/**
 * Represents a sequence of commands in a scripting language.
 * The syntax of the language itself is defined by a 
 * {@link ScriptSyntax} object owned by this sequence; 
 * the syntax translates abstract ideas like setting the value of a macro
 * into text strings which can be executed by an interpreter.
 * <p>
 * An alternative design would have the CommandSequence keep track of
 * the abstract commands from which the script is composed and generate
 * a complete script from a syntax object on demand; this would be more
 * flexible, but quite a bit more effort to implement.
 * 
 * @author   Mark Taylor
 */
public class CommandSequence {

    private final ScriptSyntax syntax_;
    private Map macros_ = new HashMap();
    private StringBuffer buffer_ = new StringBuffer();

    /**
     * Constructs a new CommandSequence with a given concrete syntax.
     *
     * @param  syntax  syntax object which does actual script text building
     */
    public CommandSequence(ScriptSyntax syntax) {
        syntax_ = syntax;
        append(syntax_.startScript());
        append(syntax_.blankLine());
        String header = new StringBuffer()
            .append("Written by ")
            .append(getClass().getName())
            .append(" for ")
            .append(System.getProperty("user.name"))
            .append(" at ")
            .append(new Date())
            .toString();
        append(syntax_.commentLine(header));
    }

    /**
     * Appends text to this sequence which will exec a given command string.
     *
     * @param  text  the command line to execute (words separated by spaces,
     *               no line breaks)
     */
    public void invokeCommand(String text) {
        append(syntax_.invokeCommand(text));
    }
 
    /**
     * Appends text to this sequence which will define a short macro.
     * The expansion <tt>value</tt> should be 'short' (one-line);
     * for potentially long macro values use
     * {@link #defineMacro(String,String[])}.
     * <p>
     * The exact form of <tt>name</tt> may be modified to make it 
     * acceptable for the syntax.
     *
     * @param  name  macro name
     * @param  value  macro expansion (as returned by {@link #useMacro}),
     *                no newlines
     */
    public void defineMacro(String name, String value) {
        name = syntax_.namify(name);
        if (macros_.containsKey(name) &&
            value.equals(macros_.get(name))) {
            // no action required, the macro already has this value
        }
        else {
            macros_.put(name, value);
            append(syntax_.defineMacro(name, value));
        }
    }

    /**
     * Appends text to this macro which will define a potentially long macro.
     * The macro is set to a space-separated concatenation of the
     * elements of a list <tt>valueLines</tt>.
     * <p>
     * The exact form of <tt>name</tt> may be modified to make it 
     * acceptable for the syntax.
     *
     * @param  name  macro name
     * @param  words  array of words forming the macro expansion;
     *                no newline characters should appear in the elements
     */
    public void defineMacro(String name, List valueLines) {
        defineMacro(name, (String[]) valueLines.toArray(new String[0]));
    }

    /**
     * Appends text to this macro which will define a potentially long macro.
     * The macro is set to a space-separated concatenation of the
     * elements of an array <tt>valueLines</tt>.
     * <p>
     * The exact form of <tt>name</tt> may be modified to make it 
     * acceptable for the syntax.
     *
     * @param  name  macro name
     * @param  words  array of words forming the macro expansion;
     *                no newline characters should appear in the elements
     */
    public void defineMacro(String name, String[] valueLines) {
        name = syntax_.namify(name);
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < valueLines.length; i++) {
            if (i > 0) {
                buf.append(' ');
            }
            buf.append(valueLines[i]);
        }
        String effectiveValue = buf.toString();
        if (macros_.containsKey(name) &&
            effectiveValue.equals(macros_.get(name))) {
            // no action required
        }
        else {
            macros_.put(name, effectiveValue);
            append(syntax_.defineMacro(name, valueLines));
        }
    }

    /**
     * Dereferences a macro previously defined by one of the
     * <tt>defineMacro</tt> calls.  Returns a single word which can form
     * part of a line.
     *
     * @param  name  the name of the macro
     * @return  a word which expands to the value defined for
     *          macro <tt>name</tt>
     */
    public String useMacro(String name) {
        name = syntax_.namify(name);
        return syntax_.useMacro(name);
    }

    /**
     * Quotes a word which may contain special characters.
     * Returns a string which within the script can function as a literal
     * meaning the string <tt>text</tt>.
     *
     * @param  text  text to quote
     * @return  a word which is the quoted version of <tt>text</tt>
     */
    public String quoteWord(String text) {
        return syntax_.quoteWord(text);
    }

    /**
     * Appends text to this script which includes comment text.
     *
     * @param  text  comment text (no line breaks)
     */
    public void commentLine(String text) {
        append(syntax_.commentLine(text));
    }

    /**
     * Appends a blank line to this script.
     */
    public void blankLine() {
        append(syntax_.blankLine());
    }

    /**
     * Returns the buffer which contains all the text written to this
     * script so far.
     *
     * @return  the script buffer
     */
    public StringBuffer getBuffer() {
        return buffer_;
    }

    /**
     * Returns a file filter object suitable for use when loading/saving 
     * this script.
     *
     * @return  a file filter
     */
    public FileFilter getFileFilter() {
        return new ExtensionFileFilter(syntax_.getExtension(),
                                       syntax_.getName() + " files");
    }

    /**
     * Returns the name of this sequence's concrete syntax.
     *
     * @return   syntax name
     */
    public String getSyntaxName() {
        return syntax_.getName();
    }

    /**
     * Appends text directly to this script.
     *
     * @param  text  text to append
     */
    private void append(String text) {
        buffer_.append(text);
    }
}
