package org.eso.cpl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * A FrameList with additional characteristics - each frame has 
 * additional attributes 'retained' and 'active';
 *
 * @author   Mark Taylor
 */
public class MarkedFrameList extends FrameList {

    private final Set isActive_ = new HashSet();
    private final Set isRetained_ = new HashSet();

    /** Initial state for each frame of the 'active' attribute. */
    final public boolean activeDefault_ = true;
 
    /** Initial state for each frame of the 'retained' attribute. */
    final public boolean retainedDefault_ = false;

    /**
     * Constructs a new empty list.
     */
    public MarkedFrameList() {
        super();
    }

    /**
     * Constructs a new list initialised from the elements of an existing
     * collection of frames.
     *
     * @param   frames  a collection of Frame objects whose contents form
     *          the initial contents of this list
     * @throws  ClassCastException  if any element of <tt>frames</tt> is
     *          a non-Frame object
     */
    public MarkedFrameList(Collection frames) {
        this();
        addAll(frames);
    }

    /**
     * Indicates whether a frame in this list is marked active or not.
     *
     * @param    frm  frame to test
     * @return   <tt>true</tt> iff <tt>frm</tt> is a member of this list
     *           and is marked active
     */
    public boolean isActive(Frame frm) {
        return isActive_.contains(getID(frm));
    }

    /**
     * Indicates whether a frame in this list is marked retained or not.
     *
     * @param   frm  frame to test
     * @return  <tt>true</tt> iff <tt>frm</tt> is a member of this list
     *          and is marked retained
     */
    public boolean isRetained(Frame frm) {
        return isRetained_.contains(getID(frm));
    }

    /**
     * Sets/unsets the active flag for a frame in this list.
     *
     * @param  frm  the frame to mark
     * @param  isActive  new active status for <tt>frm</tt>
     */
    public void setActive(Frame frm, boolean isActive) {
        Object id = getID(frm);
        if (isActive) {
            isActive_.add(id);
        }
        else {
            isActive_.remove(id);
        }
    }

    /**
     * Sets/unsets the retained flag for a frame in this list.
     *
     * @param  frm  the frame to mark
     * @param  isRetained  new retained status for <tt>frm</tt>
     */
    public void setRetained(Frame frm, boolean isRetained) {
        Object id = getID(frm);
        if (isRetained) {
            isRetained_.add(id);
        }
        else {
            isRetained_.remove(id);
        }
    }

    /**
     * Returns a new FrameList containing only those members of this list
     * marked as active.
     *
     * @return   list of active frames
     */
    public FrameList selectActive() {
        FrameList flist = new FrameList();
        for (Iterator it = iterator(); it.hasNext(); ) {
            Frame frm = (Frame) it.next();
            if (isActive(frm)) {
                flist.add(frm);
            }
        }
        return flist;
    }
      
    /**
     * Returns a new FrameList containing only those members of this list
     * marked as retained.
     *
     * @return   list of retained frames
     */
    public FrameList selectRetained() {
        FrameList flist = new FrameList();
        for (Iterator it = iterator(); it.hasNext(); ) {
            Frame frm = (Frame) it.next();
            if (isRetained(frm)) {
                flist.add(frm);
            }
        }
        return flist;
    }

    public Object set(int i, Object element) {
        Object old = super.set(i, element);
        forget(old);
        introduce(element);
        return old;
    }

    public void add(int i, Object element) {
        super.add(i, element);
        introduce(element);
    }

    public Object remove(int i) {
        Object element = super.remove(i);
        forget(element);
        return element;
    }

    /**
     * Sets up state to keep track of a newly added member of this list.
     *
     * @param  item  member being added
     */
    private void introduce(Object item) {
        Object id = getID(item);
        if (activeDefault_) {
            isActive_.add(id);
        }
        if (retainedDefault_) {
            isRetained_.add(id);
        }
    }

    /**
     * Discards state which is keeping track of a member being removed 
     * from this list.
     *
     * @param  item  member being removed
     */
    private void forget(Object item) {
        Object id = getID(item);
        isActive_.remove(id);
        isRetained_.remove(id);
    }

    /**
     * Gets an ID value corresponding to a member of this list.  Since this
     * ID will be used as the key for a Set, we need to ensure that it
     * is unique for each element of the list.  We do this by returning
     * the identity hash key; the Frame itself wouldn't do, because we
     * wish distinct but identical Frames to be distinguished for this
     * purpose.
     */
    private Object getID(Object item) {
        return new Integer(System.identityHashCode(item));
    }
}
