package org.eso.cpl;

import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Defines a naming scheme for naming sets of product files.
 *
 * @author   Mark Taylor
 */
public abstract class NamingScheme {

    /** Scheme which appends successive integers to the base filename. */
    public static final NamingScheme NUMERIC = new NamingScheme("Numeric") {
        public File transform(File raw) {
            String path = raw.getPath();
            int extIdx = path.lastIndexOf('.');
            String extension;
            if (extIdx > 0) {
                extension = path.substring(extIdx);
                path = path.substring(0, extIdx);
            } else {
                extension = "";
            }
            File file;
            int index = 0;
            do {
                
                file = new File(path +"_"+ pad(""+(index++), '0', 4) + extension);
            } while (file.exists());
            return file;
        }
    };
    
    /**
     * Prefix the padCh to the in String until the length is
     * len, then return it.
     * @param in
     * @param padCh
     * @param len
     * @return String
     */
    private static String pad(String in, char padCh, int len) {
        while (in.length() < len) {
            in = padCh+in;
        }
        return in;
    }

    /** Scheme which does nothing to the base filename. */
    public static final NamingScheme OVERWRITE = 
            new NamingScheme("Overwrite") {
        public File transform(File raw) {
            return raw;
        }
    };

    /** Predefined naming schemes. */
    public static final List STANDARD_SCHEMES =
        Collections.unmodifiableList(Arrays.asList(new NamingScheme[] {
            OVERWRITE,
            NUMERIC,
        }));

    /** Default naming scheme. */
    public static final NamingScheme DEFAULT = NUMERIC;

    private final String name_;

    /**
     * Constructs a new naming scheme with a given scheme name.
     *
     * @param  name  name of the scheme
     */
    protected NamingScheme(String name) {
        name_ = name;
    }

    /**
     * Transforms a file name from its raw state to that defined by 
     * this scheme.
     *
     * @param  raw  the original file name
     * @return  the transformed file name
     */
    public abstract File transform(File raw);

    /**
     * Returns the name of this scheme.
     *
     * @return  scheme name
     */
    public String getName() {
        return name_;
    }

    public String toString() {
        return name_;
    }
}
