package org.eso.cpl;

/**
 * A parameter which affects execution of a recipe.
 * Instances of this class contain 
 * <ul>
 * <li>metadata about the parameter (name, help information, initial value etc) 
 * <li>constraints on its value (data type, range etc)
 * <li>the value itself
 * </ul>
 * <p>
 * In the CPL execution model, a parameter is one of a list of parameters
 * owned by a recipe; it can be queried for its metadata (name,
 * type, context etc) but preparing the recipe for execution involves
 * setting the values of the recipe's parameters in place.
 * This functionality is implemented here using the Bridge 
 * pattern, to clarify the separation between the roles played by 
 * the <tt>Parameter</tt> object itself (conversion, validation and 
 * maintenance of the parameter's value) 
 * and that played by the implementation-specific 
 * {@link ParameterImp} object
 * (provision of metadata and constraints, and transmitting new values 
 * to the recipe's parameter list).
 *
 * @author   Mark Taylor
 */
public class Parameter {

    private final ParameterImp impl_;
    private final String context_;
    private final String help_;
    private final String name_;
    private final String tag_;
    private final ParameterType type_;
    private final ParameterConstraint constraint_;
    private final Object default_;
    private Object value_;

    /**
     * Constructs a new Parameter whose behaviour is defined by an
     * implementation object.
     *
     * @param  impl  implementation object providing services to this parameter
     * @throws ParameterValueException  if the initial value and the 
     *         parameter constraints provided by <tt>impl</tt> are inconsistent
     */
    public Parameter(ParameterImp impl) throws ParameterValueException {
        impl_ = impl;
        context_ = impl.getContext();
        help_ = impl.getHelp();
        name_ = impl.getName();
        tag_ = impl.getTag();
        type_ = impl.getType();
        constraint_ = impl.getConstraint();
        default_ = impl.getInitialValue();
        setValue(default_);
    }

    /**
     * Retuns the implementation object from which this parameter derives
     * its behaviour.
     *
     * @return  parameter service object
     */
    public ParameterImp getImp() {
        return impl_;
    }
 
    public String getContext() {
        return context_;
    }

    public Object getDefault() {
        return default_;
    }

    public String getHelp() {
        return help_;
    }

    public String getName() {
        return name_;
    }

    public String getTag() {
        return tag_;
    }

    public ParameterType getType() {
        return type_;
    }

    public ParameterConstraint getConstraint() {
        return constraint_;
    }

    /**
     * Returns the value of this parameter as an object.  The class of
     * the returned value is given by 
     * <tt>getParameterType().getValueClass()</tt>;
     * it will typically be a String or a primitive wrapper object.
     *
     * @return   the actual value of this parameter
     */
    public Object getValue() {
        return value_;
    }

    /**
     * Sets the value of this parameter from an object.
     * If the supplied object <tt>val</tt> is not of the correct 
     * type for this parameter
     * (is not an instance of <tt>getParameterType().getValueClass()</tt>)
     * or violates its constraints, 
     * a <tt>ParameterValueException</tt> is thrown.
     * A corollary of this is that the value cannot be set to <tt>null</tt> - 
     * is that right?
     *
     * @param  val  the new value
     * @throws  ParameterValueException if <tt>val</tt> is not a legal value
     *          for this parameter
     */
    public void setValue(Object val) throws ParameterValueException {

        // Validate the submitted value.
        Class vclass = type_.getValueClass();
        if (val == null && vclass == String.class) {
            // String types support null values, which we map to empty
            // strings until a more firm policy is decided and how to support
            // null values.
            val = "";
        }
        if (!vclass.isAssignableFrom(val.getClass())) {
            String msg = "Attempt to set " + this + " to non-" + 
                         vclass.getName() + " value " + val + 
                         " (" + val.getClass().getName() + ")";
            throw new ParameterValueException(msg);
        }
        if (constraint_ != null) {
            constraint_.checkValue(val);
        }

        // Set our copy of the value accordingly.
        value_ = val;

        // Instruct the implementation to store the value.
        impl_.setValue(val);
    }

    /**
     * Sets the value of this parameter from a string representation.
     * An attempt is made to interpret the given string <tt>str</tt>
     * into an object of the correct type for this parameter, and
     * to set the parameter value accordingly.  If either the 
     * conversion fails or the resulting value violates this parameter's
     * constraints, a <tt>ParameterValueException</tt> is thrown.
     * <p>
     * As far as possible this method should be capable of restoring 
     * a parameter's value using the result of the value's <tt>toString</tt>
     * method.
     *
     * @param  str  a String representation of the new value
     * @throws  ParameterValueException if <tt>str</tt> does not represent a 
     *          legal value of this parameter
     */
    public void setValueFromString(String str) throws ParameterValueException {
        setValue(type_.fromString(str));
    }

    public String toString() {
        return new StringBuffer()
            .append(getName())
            .append(" (")
            .append(getType())
            .append(getConstraint() == null ? "" : (": " + getConstraint()))
            .append(") = ")
            .append(getValue())
            .toString();
    }
}
