package org.eso.cpl;

/**
 * A constraint on a value which requires it to be between two (inclusive)
 * limits.  This constraint can only be applied to a type which is
 * {@link java.lang.Comparable}.
 *
 * @author   Mark Taylor
 */
public class RangeConstraint implements ParameterConstraint {

    private ParameterType type_;
    private Comparable min_;
    private Comparable max_;
    private String descrip_;

    /**
     * Constructs a new RangeConstraint.
     * If <tt>tclass=type.getValueClass()</tt>,
     * then tclass must implement {@link java.lang.Comparable},
     * and <tt>min</tt> and <tt>max</tt> must both be instances of
     * <tt>tclass</tt>.
     * No check is made that <tt>min&lt;=max</tt>, but if this is not
     * the case, no value will count as legal.
     *
     * @param  type  the parameter type for the constraint
     * @param  min   the minimum value in the permissible range (inclusive)
     * @param  max   the maximum value in the permissible range (inclusive)
     * @throws  IllegalArgumentException  if a RangeConstraint cannot be
     *          constructed from these parameters
     */
    public RangeConstraint(ParameterType type, Object min, Object max) 
            throws ParameterValueException {
        Class tclass = type.getValueClass();
        if (!Comparable.class.isAssignableFrom(tclass)) {
            throw new IllegalArgumentException("Type " + type + 
                                               " not comparable");
        }
        if (!tclass.isAssignableFrom(min.getClass())) {
            throw new IllegalArgumentException("min not of class " + tclass);
        }
        if (!tclass.isAssignableFrom(max.getClass())) {
            throw new IllegalArgumentException("max not of class " + tclass);
        }
        type_ = type;
        min_ = (Comparable) min;
        max_ = (Comparable) max;
        descrip_ = min_ + " <= val <= " + max_;
    }

    /**
     * Throws a <tt>ParameterValueException</tt> if <tt>val&lt;min</tt> 
     * or <tt>val&gt;max</tt>.
     *
     * @param  the value to check
     */
    public void checkValue( Object val ) throws ParameterValueException {
        Class tclass = type_.getValueClass();
        if (!tclass.isAssignableFrom(val.getClass())) {
            throw new IllegalArgumentException("value not of class " + tclass);
        }
        Comparable cval = (Comparable) val;
        if (cval.compareTo(min_) < 0) {
            throw new ParameterValueException("Value " + cval + " < " + min_);
        }
        if (cval.compareTo(max_) > 0) {
            throw new ParameterValueException("Value " + cval + " > " + max_);
        }
    }

    /**
     * Returns the minimum legal value.
     *
     * @return  minimum
     */
    public Comparable getMinimum() {
        return min_;
    }

    /**
     * Returns the maximum legal value.
     *
     * @return  maximum
     */
    public Comparable getMaximum() {
        return max_;
    }

    public String getDescription() {
        return descrip_;
    }
}
