package org.eso.cpl;

import java.io.File;

/**
 * A data reduction recipe which can be used to operate on a set of 
 * <tt>Frame</tt>s.  Each instance of this class has its own parameter list.
 * To invoke the recipe, the sequence is:
 * <ol>
 * <li>Construct/obtain the Recipe object
 * <li>Get a list of parameters using {@link #getDefaultParameters} and set
 *     their values as required
 * <li>Assemble an array of Frames on which the recipe should operate
 * <li>Call {@link #execute}
 * <li>Examine the array of Frames returned by <tt>execute</tt> for the result
 * </ol>
 *
 * @author   Mark Taylor
 */
public interface Recipe {

    /**
     * Returns an array of parameters which can control the execution
     * of this recipe.  The parameters in the returned array will have
     * their default values (these must also be legal values).
     * Callers may change the values of elements of this array 
     * to configure how the recipe will be run, but no change should be 
     * made to the array itself.
     * A new array is returned every time this method is called.
     *
     * @return  a new set of parameters for this recipe
     * @throws  CPLException if some error occurs
     */
    Parameter[] getDefaultParameters();

    /**
     * Invokes this recipe with given parameter values on a list of frames,
     * and returns a list of frames which were produced by the operation.
     * The <tt>params</tt> array should be the result of a call to 
     * {@link #getDefaultParameters}
     *
     * @param  params   an array of parameters for this recipe
     * @param  inFrames  the input list of frames on which to perform the recipe
     * @param  productRoot  the directory into which files produced by the
     *         execution should be written.  If <tt>null</tt> some default
     *         location (perhaps the current directory) will be used
     * @return  the list of frames following the invocation
     * @throws  CPLException  if an error occurs during invocation
     */
    FrameList execute(Parameter[] params, FrameList inFrames,
                      File productRoot) throws CPLException;

    /**
     * Returns the author of this recipe.
     *
     * @return  author's name, or <tt>null</tt> if unknown
     */
    String getAuthor();

    /**
     * Returns a copyright and license string applying to this recipe.
     *
     * @return  copyright string, or <tt>null</tt> if not available
     */
    String getCopyright();

    /**
     * Returns a detailed description for this recipe.
     * The description is the detailed help for the recipe. 
     * For formatting the output the characters '\n', '\t' may be 
     * embedded in the returned string.  If no help is available,
     * <tt>null</tt> may be returned.
     *
     * @return  recipe help string
     */
    String getDescription();

    /**
     * Returns an email address for this recipe's author.
     *
     * @return  author's email, or <tt>null</tt> if not available
     */
    String getEmail();

    /**
     * Returns the name of this recipe.
     *
     * @return  recipe name
     */
    String getName();

    /**
     * Returns a short help string for this recipe.
     * The synopsis should summarize the plugin's purpose in not more 
     * than a few lines. It may contain '\n' characters.
     *
     * @return  synopsis string, or <tt>null</tt> if not available
     */
    String getSynopsis();

    /**
     * Returns the version of this Recipe.  This should be an encoded version
     * of the usual MAJOR.MINOR.MICRO form for version numbers.
     *
     * @return  recipe version number
     */
    long getVersion();

    /**
     * Returns the library from which this recipe originated.
     * What kind of object this is will depend on the recipe implementation;
     * <tt>null</tt> may be returned if there is no appropriate object.
     *
     * @return  the source of this recipe
     */
    RecipeLibrary getLibrary();
}
