package org.eso.cpl.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.TransferHandler;
import javax.swing.border.Border;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Segment;

import org.eso.cpl.AssociationUtils;
import org.eso.cpl.BourneSyntax;
import org.eso.cpl.CPLException;
import org.eso.cpl.CommandSequence;
import org.eso.cpl.CshSyntax;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameList;
import org.eso.cpl.MarkedFrameList;
import org.eso.cpl.NamingScheme;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterValueException;
import org.eso.cpl.PerlSyntax;
import org.eso.cpl.Recipe;
import org.eso.cpl.Request;
import org.eso.cpl.ScriptSyntax;
import org.eso.cpl.jni.CPLControl;
import org.eso.cpl.jni.LibraryLoader;
import org.eso.cpl.jni.MessageListener;
import org.eso.cpl.jni.StreamMessageListener;
import org.eso.dfs.datatransfer.FileListTransferHandler;
import org.eso.dfs.gui.SwingFileChooser;
import org.eso.dfs.services.ServiceException;
import org.eso.dfs.services.data.AssociationService;
import org.eso.dfs.services.data.FileAssociationNode;

/**
 * A panel which provides the main graphical front end for executing
 * a single recipe or a queue of them.  As well as execution itself,
 * there is provision for displaying and editing execution parameters
 * and data files, loading and saving configuration, displaying, results, etc.
 * <p>
 * Execution is done asynchronously to avoid locking up the GUI for long
 * recipes.  Hooks are present for interrupting this execution, but
 * are not currently exposed in the public interface (I don't know
 * if there's any good way to interrupt C CPL recipes mid-execution).
 * <p>
 * Message (debug, log, etc) are sent through a set of methods:
 * void errorLog(String msg)
 * void commandLog(String msg)
 * void controlLog(String msg)
 * void messageLog(String msg)
 * <p>
 * Changing recipes is handled by CPLFrame - such as through loading a config file,
 * but is currently disabled. The reason for this is that Gasgano assumes that a CPLFrame
 * always contains the recipe it was initially loaded with. There is no means to
 * notify Gasgano that a recipe in a CPLFrame has changed - this notification would
 * be required by Gasgano in order to update its menu items that show
 * the set of loaded CPLFrames (and their recipes). For this reason the recipe
 * is set in the constructor only and setRecipe() has been commented out.
 *
 * @author   Mark Taylor
 */
public class CPLFrame extends JFrame implements ExecutionListener {

    private final RequestPanel singlePanel_;
    private final RequestPanel queuedPanel_;
    private final RequestPanel executingPanel_;
    private final JTabbedPane tabber_;
    private final JTable outFrameTable_;
    private final JScrollPane outFrameScroller_;
    private final LogArea logArea_;
    private final JComboBox syntaxChooser_;
    private final RequestPoolDisplay poolDisplay_;
    
    // the log message writers
    private PrintWriter errorWriter_;
    private PrintWriter messageWriter_;
    private PrintWriter controlWriter_;
    private PrintWriter commandWriter_;
    
    /**
     * Writer used to record the execution of a recipe.
     * The logFileWriter_ is non-null only during execution of a recipe.
     */ 
    private PrintWriter logFileWriter_;
    
    /**
     * Date formatter used to create the execution log file names.
     */
    private static DateFormat dateFormatter_ = new SimpleDateFormat("yyyy-MM-dd_HH:mm:ss");

    private final Action executeSingleAction_;
    private final Action executeQueueAction_;
    private final Action enqueueAction_;
    private final Action loadAction_;
    private final Action saveAction_;
    private final Action closeAction_;
    private final Action aboutAction_;
    private final Action clearPanelAction_;
    private final Action logClearAction_;
    private final Action logSaveAction_;
    // not implemented yet: 
    // private final Action scriptSaveAction_;
    private final Action outClearAction_;
    private JFileChooser configChooser_;
    private JFileChooser logChooser_;
    private JFileChooser scriptChooser_;
    private RequestPool requestPool_;
    private Recipe recipe_;
    private FrameList outFrames_;
    private LibraryLoader libLoader_;
    private CommandSequence script_;
    private int scriptPos_;
    private RequestExecutor activeExecutor_;
    private int nextQueued_;
    
    private static final String XmlSuffix = ".xml";
    
    /**
     * The name of the panel, as appears in the title bar.
     */
    private String panelName_;

    private static final Cursor BUSY_CURSOR = new Cursor(Cursor.WAIT_CURSOR);
    private static final String SH_SYNTAX = "Bourne shell";
    private static final String CSH_SYNTAX = "C shell";
    private static final String PERL_SYNTAX = "Perl";

    /**
     * Constructs a new blank CPL Frame;
     */
    public CPLFrame(Recipe recipe) {
        this(recipe, null);
    }

    /**
     * Constructs a new blank CPL Frame, the title name will be the name
     * parameter given.
     */    
    // DS FEB 2007 - SPR DFS02892 
    // added a constructor in overload, to pass the product dir 
    public CPLFrame(Recipe recipe, String name) { this(recipe, name, null); }

    
    public CPLFrame(Recipe recipe, String name, String productDirPath) {
        // set the title for this frame.
        if (name == null) {
            name = recipe == null ? "No Recipe" : recipe.getName();
        }
        panelName_ = name;
        setTitle(panelName_);
        recipe_ = recipe;
        JPanel contentPane = (JPanel) getContentPane();
        contentPane.setLayout(new BorderLayout());
        JSplitPane mainBox = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        mainBox.setOneTouchExpandable(true);
        JComponent inputBox = new Box(BoxLayout.Y_AXIS);
        JComponent outputBox = new Box(BoxLayout.Y_AXIS);
        mainBox.setTopComponent(inputBox);
        mainBox.setBottomComponent(outputBox);
        contentPane.add(mainBox, BorderLayout.CENTER);
        Border gapBorder = BorderFactory.createEmptyBorder(5, 5, 5, 5);
        mainBox.setBorder(gapBorder);
        final Component parent = this;
        requestPool_ = new RequestPool();

        // Set up actions.
        executeSingleAction_ = new AbstractAction("Execute") {
            public void actionPerformed(ActionEvent evt) {
                RequestQueue q1 = new RequestQueue();
                q1.add(new NamedRequest(singlePanel_.getRequest(), "Current"));
                executeQueue(q1);
            }
        };
        executeQueueAction_ = new AbstractAction("Execute Selected") {
            public void actionPerformed(ActionEvent evt) {
                RequestQueue q = requestPool_.getSelectedRequests();
                controlLog(
                    "Executing " + q.size() + "-element queue");
                executeQueue(q);
            }
        };
        enqueueAction_ = new AbstractAction("Add to pool") {
            public void actionPerformed(ActionEvent evt) {
                enqueueRequest(
                    new NamedRequest(
                        singlePanel_.getRequest(),
                        nextQueuedRequestName()));
            }
        };
        loadAction_ = new AbstractAction("Load Current Settings") {
            public void actionPerformed(ActionEvent evt) {
                Throwable error;
                try {
                    Request req = loadConfiguration(singlePanel_.getRequest());
                    if (req != null) {
                        setSingleRequest(new NamedRequest(req, "Current"));
                        showRequestPanel(singlePanel_);
                    }
                    error = null;
                } catch (IOException e) {
                    error = e;
                } catch (CPLException e) {
                    error = e;
                } catch (ParameterValueException e) {
                    error = e;
                }
                if (error != null) {
                    ErrorDialog.showError(
                        error,
                        "Error loading config file",
                        parent);
                }
            }
        };
        saveAction_ = new AbstractAction("Save Current Settings") {
            public void actionPerformed(ActionEvent evt) {
                Throwable error;
                try {
                    saveConfiguration(singlePanel_.getRequest());
                    error = null;
                } catch (IOException e) {
                    error = e;
                }
                if (error != null) {
                    ErrorDialog.showError(
                        error,
                        "Error saving config file",
                        parent);
                }
            }
        };
        clearPanelAction_ = new AbstractAction("Clear Current") {
            public void actionPerformed(ActionEvent evt) {
                if (JOptionPane
                    .showConfirmDialog(
                        CPLFrame.this, 
                        "Are you sure you want to clear all frames and logs?",
                        "Clear Recipe Panel",
                        JOptionPane.YES_NO_OPTION)
                    == JOptionPane.YES_OPTION)
                clearPanel();
            }
        };
        closeAction_ = new AbstractAction("Close") {
            public void actionPerformed(ActionEvent evt) {   
                dispose();
            }
        };
        aboutAction_ = new AbstractAction("About") {
            public void actionPerformed(ActionEvent evt) {
                
                // DS JAN 2007 -  DFS02615 (CPL help window is modal)
                // // final JDialog frame = new JDialog(CPLFrame.this, panelName_, true);
                final JDialog frame = new JDialog(CPLFrame.this, panelName_, false);
                JPanel pan = new JPanel(new BorderLayout());
                RecipeDisplay rdisp = new RecipeDisplay();
                
                rdisp.setRecipe(CPLFrame.this.getRecipe());
                pan.add(rdisp, BorderLayout.CENTER);
                frame.setTitle(panelName_);
                frame.getContentPane().add(pan);
                frame.pack();
                frame.setLocationRelativeTo(CPLFrame.this);
                frame.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        frame.dispose();
                    }
                });
                frame.setVisible(true);
            }
        };
        logClearAction_ = new AbstractAction("Clear Messages") {
            public void actionPerformed(ActionEvent evt) {
                clearLogMessages();
            }
        };
        logSaveAction_ = new AbstractAction("Save Messages") {
            public void actionPerformed(ActionEvent evt) {
                JFileChooser saver = getLogSaver();
                if (saver.showSaveDialog(parent)
                    == JFileChooser.APPROVE_OPTION) {
                    try {
                        writeDocument(
                            logArea_.getDocument(),
                            saver.getSelectedFile());
                    } catch (IOException e) {
                        ErrorDialog.showError(
                            e,
                            "Error writing log file",
                            parent);
                    }
                }
            }
        };
        /*
         * Not implemented for now...
         *
        scriptSaveAction_ = new AbstractAction("Save Script") {
            public void actionPerformed(ActionEvent evt) {
                JFileChooser saver = getScriptSaver(script_);
                if (saver.showSaveDialog(parent)
                    == JFileChooser.APPROVE_OPTION) {
                    Writer writer = null;
                    try {
                        writer = new FileWriter(saver.getSelectedFile());
                        writer.write(script_.getBuffer().toString());
                    } catch (IOException e) {
                        ErrorDialog.showError(
                            e,
                            "Error writing script file",
                            parent);
                    } finally {
                        if (writer != null) {
                            try {
                                writer.close();
                            } catch (IOException e) {
                            }
                        }
                    }
                }
            }
        };
         */
        outClearAction_ = new AbstractAction("Clear Output Frames") {
            public void actionPerformed(ActionEvent evt) {
                clearOutputFrames();
            }
        };

        // Add menus.
        JMenuBar mbar = new JMenuBar();
        setJMenuBar(mbar);
        JMenu fileMenu = new JMenu("File");
        mbar.add(fileMenu);
        fileMenu.add(saveAction_);
        fileMenu.add(loadAction_);
        fileMenu.add(clearPanelAction_);
        // not implemented yet: 
        // fileMenu.add(scriptSaveAction_);
        fileMenu.add(closeAction_);
        
        JMenu helpMenu = new JMenu("Help");
        helpMenu.add(aboutAction_);
        mbar.add(helpMenu);

        // Set up the chooser for script syntax.
        syntaxChooser_ =
            new JComboBox(new String[] { SH_SYNTAX, CSH_SYNTAX, PERL_SYNTAX });
        syntaxChooser_.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent evt) {
                if (evt.getStateChange() == ItemEvent.SELECTED) {
                    initScript();
                }
            }
        });
        syntaxChooser_.setSelectedItem(SH_SYNTAX);
        syntaxChooser_.revalidate();

        // Construct a new box for all input stuff (request panel plus queue).
        Box inBox = new Box(BoxLayout.X_AXIS);
        inputBox.add(inBox);

        // Construct request display/edit panel for new requests, queued
        // requests, and the currently executing request.
        
        // DS FEB 2007 - SPR DFS02892 
        // // singlePanel_ = new RequestPanel();
        // // queuedPanel_ = new RequestPanel();
        // // executingPanel_ = new RequestPanel();
        singlePanel_ = new RequestPanel(productDirPath);
        queuedPanel_ = new RequestPanel(productDirPath);
        executingPanel_ = new RequestPanel(productDirPath);
        
        executingPanel_.setLocked(true);

        // Place the two request display/edit panels in a tabbed pane.
        tabber_ = new JTabbedPane(JTabbedPane.TOP);
        tabber_.add("Current", singlePanel_);
        tabber_.add("Queued", queuedPanel_);
        tabber_.add("Executing", executingPanel_);
        showRequestPanel(singlePanel_);
        tabber_.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
        tabber_.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                configureAvailableActions();
            }
        });

        // Set up and place a panel to hold the single/current request 
        // representation and associated controls.
        JComponent singleBox = new JPanel(new BorderLayout());
        Box smid = new Box(BoxLayout.X_AXIS);
        Box sbot = new Box(BoxLayout.X_AXIS);
        singleBox.add(sbot, BorderLayout.SOUTH);
        singleBox.add(smid, BorderLayout.CENTER);
        smid.add(tabber_);
        sbot.add(Box.createHorizontalGlue());
        sbot.add(center(new JLabel(Icons.ARROW_SE)));
        sbot.add(Box.createHorizontalStrut(5));
        sbot.add(center(new JButton(executeSingleAction_)));
        sbot.add(Box.createHorizontalStrut(5));
        sbot.add(center(new JLabel(Icons.ARROW_ES)));
        sbot.add(Box.createHorizontalGlue());
        inBox.add(singleBox);

        // Set up and place a panel to hold the queue representation and
        // associated controls.
        JComponent queueBox = new JPanel(new BorderLayout()) {
            public Dimension getMaximumSize() {
                return new Dimension(
                    getPreferredSize().width,
                    super.getMaximumSize().height);
            }
        };
        Box qtop = new Box(BoxLayout.Y_AXIS);
        Box qmid = new Box(BoxLayout.Y_AXIS);
        Box qbot = new Box(BoxLayout.Y_AXIS);
        queueBox.add(qtop, BorderLayout.NORTH);
        queueBox.add(qmid, BorderLayout.CENTER);
        queueBox.add(qbot, BorderLayout.SOUTH);
        poolDisplay_ = new RequestPoolDisplay(requestPool_);
        poolDisplay_.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        JScrollPane qscroller = poolDisplay_.wrapInScroller();
        GUIUtils.wrapInBorder(qscroller, "Request Pool");
        qtop.add(Box.createVerticalStrut(32));
        qtop.add(wrapIcon(Icons.ARROW_EES));
        JButton enqueueButton = new JButton(enqueueAction_);
        qscroller.setTransferHandler(getFileListDropHandler());
        qtop.add(center(enqueueButton));
        qtop.add(wrapIcon(Icons.ARROW_S));
        qmid.add(center(qscroller));
        qbot.add(wrapIcon(Icons.ARROW_S));
        qbot.add(center(new JButton(executeQueueAction_)));
        qbot.add(wrapIcon(Icons.ARROW_S));
        inBox.add(queueBox);

        // Set up a table for displaying output frames.
        outFrameTable_ = new DataColumnTable();
        GUIUtils.configureTable(outFrameTable_);
        GUIUtils.enableFileListDragDrop(outFrameTable_);
        outFrameTable_.addMouseListener(new TableComponentMouseListener());

        // Add output frame table to frame.
        outFrameScroller_ = GUIUtils.wrapInScroller(outFrameTable_);
        Box outBox = new Box(BoxLayout.Y_AXIS);
        Box outControlBox = new Box(BoxLayout.X_AXIS);
        outBox.add(outControlBox);
        outBox.add(Box.createVerticalStrut(5));
        outBox.add(outFrameScroller_);
        outputBox.add(GUIUtils.wrapInBorder(outBox, "Output Frames"));
        outControlBox.add(Box.createHorizontalGlue());
        JButton outClearButton = new JButton(outClearAction_);
        outClearButton.setText("Clear");
        outControlBox.add(outClearButton);

        // Place log message panel.
        outputBox.add(Box.createVerticalStrut(5));
        logArea_ = new LogArea();
        Box logBox = new Box(BoxLayout.Y_AXIS);
        Box logControlBox = new Box(BoxLayout.X_AXIS);
        logBox.add(logControlBox);
        logBox.add(Box.createVerticalStrut(5));
        logBox.add(GUIUtils.wrapInScroller(logArea_));
        outputBox.add(GUIUtils.wrapInBorder(logBox, "Log Messages"));
        // logControlBox.add(new JLabel("Script format: "));
        // logControlBox.add(syntaxChooser_);
        JButton logSaveButton = new JButton(logSaveAction_);
        JButton logClearButton = new JButton(logClearAction_);
        logSaveButton.setText("Save");
        logClearButton.setText("Clear");
        logControlBox.add(Box.createHorizontalGlue());
        logControlBox.add(logSaveButton);
        logControlBox.add(Box.createHorizontalStrut(5));
        logControlBox.add(logClearButton);

        // Set up logging machinery.
        messageWriter_ =
            new PrintWriter(
                new TextAreaWriter(logArea_, LogArea.MESSAGE_STYLE));
        controlWriter_ =
            new PrintWriter(
                new TextAreaWriter(logArea_, LogArea.CONTROL_STYLE));
        errorWriter_ =
            new PrintWriter(new TextAreaWriter(logArea_, LogArea.ERROR_STYLE));
        commandWriter_ =
            new PrintWriter(
                new TextAreaWriter(logArea_, LogArea.COMMAND_STYLE));

        // Respond to queue change events.
        requestPool_.items_.addListDataListener(new ListDataListener() {
            public void contentsChanged(ListDataEvent evt) {
            }
            public void intervalAdded(ListDataEvent evt) {
                configureAvailableActions();
            }
            public void intervalRemoved(ListDataEvent evt) {
                configureAvailableActions();
            }
        });

        // Respond to queue display selection events.
        poolDisplay_.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent evt) {
                if (!evt.getValueIsAdjusting() && activeExecutor_ == null) {
                    Object[] selected = poolDisplay_.getSelectedValues();
                    if (selected.length == 1) {
                        Request request = (Request) selected[0];
                        if (request != queuedPanel_.getRequest()) {
                            queuedPanel_.setRequest((Request) selected[0]);
                            showRequestPanel(queuedPanel_);
                        }
                    } else {
                        showRequestPanel(singlePanel_);
                    }
                    configureAvailableActions();
                }
            }
        });

        // Additional component configuration.
        initScript();
        configureAvailableActions();
        pack();
    }

    /*
     * Sets the recipe which this panel allows the user to execute.
     * This will clear any current or queued requests.
     * This method is currently commented out for reasons
     * given in the header comment.
     *
     * @param  recipe  the new recipe
     * @throws  CPLException  if there is a CPL problem
     */
    /*
    public void setRecipe(Recipe recipe) {
        recipe_ = recipe;
        setSingleRequest(new NamedRequest(new Request(recipe), "Current"));
        requestQueue_.clear();
        if (recipe != null) {
            controlLog("Configuring recipe: " + recipe);
        }
    }
    */

    /**
     * Returns the recipe which this panel currently allows the user to execute.
     *
     * @return  recipe
     */
    public Recipe getRecipe() {
        return recipe_;
    }

    /**
     * Sets the output frame list.
     * Called by this object when the recipe is changed or executed.
     *
     * @param  output frames
     */
    private void setOutputFrames(FrameList outFrames) {
        outFrames_ = outFrames;
        FrameList flist = (outFrames == null) ? new FrameList() : outFrames;
        outFrameTable_.setModel(new FrameTableModel(flist, false));

        // Ensure that the end of the table is visible (so that most
        // recently added frames will be obvious).
        JScrollBar vbar = outFrameScroller_.getVerticalScrollBar();
        vbar.setValue(vbar.getMaximum());
    }

    /**
     * Returns the frame list which has resulted from request executions.
     *
     * @return  output frame list
     */
    public FrameList getOutputFrames() {
        return outFrames_;
    }

    /**
     * Sets the object which is used to locate JNI recipe libraries.
     * It is important that this is set for the state deserialization
     * from XML to work properly.
     *
     * @param libLoader the loader to use
     */
    public void setLibraryLoader(LibraryLoader libLoader) {
        libLoader_ = libLoader;
    }

    /**
     * Returns the object which is used to locate JNI recipe libraries.
     *
     * @return  library loader
     */
    public LibraryLoader getLibraryLoader() {
        return libLoader_;
    }

    /**
     * Sets the request displayed in the current request panel. 
     * NOTE: the request must have the same recipe (and version)
     * as the CPLFrame getRecipe() or the request will not be set.
     *
     * @param  req  request to install
     * @return boolean true if the request was set, else false.
     */
    public boolean setSingleRequest(Request req) {
        Recipe recipe = req.getRecipe();
        /*
         * can not/does not set recipe - see header comment.
        if (recipe == null
            || recipe_ == null
            || !recipe.getName().equals(recipe_.getName())
            || recipe.getVersion() != recipe_.getVersion()) {
            setRecipe(req.getRecipe());
        }
        */
        boolean rc;
        if (!req.getRecipe().equals(getRecipe())) {
            rc = false;
        } else {
            singlePanel_.setRequest(req);
            rc = true;
        }
        return rc;
    }

    /**
     * Adds a new request to the end of the execution queue.
     * NOTE: the request must have the same recipe (and version)
     * as the CPLFrame getRecipe() or the request will not be set. 
     * 
     * @param  req  request to enqueue
     * @return boolean true if the request was set, else false.
     */
    public void enqueueRequest(Request req) {
        boolean rc;
        if (!req.getRecipe().equals(getRecipe())) {
            rc = false;
        } else {
            requestPool_.add(req);
            rc = true;
        }
    }

    /**
     * Queue the given request, adding the frames in the addFrames frame list
     * to the set of input frames.
     * NOTE: the request must have the same recipe (and version)
     * as the CPLFrame getRecipe() or the request will not be set. 
     * 
     * @param req
     * @param addFrames
     * @return boolean true if the request was set, else false.
     */
    private boolean enqueueRequest(Request req, FrameList addFrames) {
        boolean rc;
        if (!req.getRecipe().equals(getRecipe())) {
            rc = false;
        } else {
            req.getInputFrames().addAll(addFrames);
            requestPool_.add(req);
            rc = true;
        }
        return rc;
    }
    
    /**
     * Process a List of File objects into Frame objects, group them
     * according to the grouping value and then return each group as
     * a FrameList element in a List.
     * @param files
     * @return List
     */
    private List getFrameGroups(List files) {
        List rc = new ArrayList();
        return rc;
    }

    /**
     * Initiates execution of a queue of recipe requests.
     *
     * @param  queue of Request objects to execute
     */
    public void executeQueue(final RequestQueue queue) {
        final RequestExecutor rex = new RequestExecutor();
        rex.addExecutionListener(this);
        final MessageListener msgListener = new LogAreaMessageListener();
        Thread runner = new Thread() {
            public void run() {
                setBusyLater(true);
                CPLControl.addMessageListener(msgListener);
                while (activeExecutor_ == rex && queue.hasNext()) {
                    final Request req = queue.next();
                    SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            executingPanel_.setRequest(req);
                        }
                    });
                    rex.execute(req);
                }
                CPLControl.removeMessageListener(msgListener);
                if (activeExecutor_ == rex) {
                	setActiveExecutor(null);
                    setBusyLater(false);
                    if (getShownRequestPanel() == executingPanel_) {
                        final RequestPanel rp =
                            (queue == requestPool_ && queue.hasNext())
                                ? queuedPanel_
                                : singlePanel_;
                        SwingUtilities.invokeLater(new Runnable() {
                            public void run() {
                                showRequestPanel(rp);
                            }
                        });
                    }
                }
            }
        };
        setActiveExecutor(rex);
        showRequestPanel(executingPanel_);
        runner.start();
    }

    /**
     * Updates the GUI as the results of the start of a recipe execution.
     * The execution log file is created and opened setting a valid
     * value in logFileWriter_). If the file exists, it is
     * removed and created. The log file will be created in the
     * product directory using the naming scheme selected in the
     * request. 
     * @param  evt  event describing the execution
     */
    public void executionStarted(ExecutionEvent evt) {
        Request request = evt.getRequest();
        controlLog("Executing request: " + request);
        // create the execution log file.
        NamingScheme namer = request.getNamingScheme();
        String logFileName = request.getRecipe().getName()+"_"+dateFormatter_.format(new Date())+".log";
        File log = new File(request.getProductRoot().getAbsolutePath()+File.separator+logFileName);
        if (log.exists() && !log.delete()) {
            errorLog("Error (1) creating execution log file : "+log.getPath());
            log = null;
        } else {
            try {
                if (!log.createNewFile()) {
                    errorLog("Error (2) creating execution log file: "+log.getPath());
                    log = null; 
                }
            } catch (IOException e) {
                e.printStackTrace();
                errorLog("Error (3) creating execution log file: "+log.getPath());
                log = null; 
            }
        }
        if (log != null) {
            // we have created a valid log.
            try {
                FileWriter logWriter = new FileWriter(log);
                logFileWriter_ = new PrintWriter(new BufferedWriter(logWriter));
            } catch (IOException e) {
                errorLog("Error opening execution log file: "+log.getPath());
            }
        }
        request.appendToScript(script_);
        
        StringBuffer sbuf = script_.getBuffer();
        // log a string report of the request to the message window.
        commandLog(request.toReportString());
        scriptPos_ = sbuf.length();
    }

    /**
     * Updates the GUI as the result of the end of a recipe execution.
     * The execution log file is flushed and closed, setting the
     * logFileWriter_ member to null).
     * @param  evt  event describing the execution result
     */
    public void executionEnded(ExecutionEvent evt) {

        // Add any new output frames to the end of the displayed frame list.
        // Remove any duplicates from earlier in the list at the same time.
        FrameList outFrames0 = getOutputFrames();
        if (outFrames0 == null) {
            outFrames0 = new FrameList();
        }
        FrameList outFrames1 = evt.getOutputFrames();
        List newProductFiles = new ArrayList();
        if (outFrames1 != null) {
            if (outFrames1.size() > 0) {
                controlLog("Product frames:");
            }
            for (Iterator it = outFrames1.iterator(); it.hasNext();) {
                Frame frm = (Frame) it.next();
                controlLog("   " + frm.getFile());
                if (outFrames0.contains(frm)) {
                    outFrames0.remove(frm);
                }
                newProductFiles.add(frm.getFile());
                outFrames0.add(frm);

            }
        }
        File[] newFiles = new File[newProductFiles.size()];
        newProductFiles.toArray(newFiles);
        try {
            for (int i = 0; i < newFiles.length; i++) {
                newFiles[i] = newFiles[i].getCanonicalFile();
            }
            AssociationService.getInstance().addFiles(newFiles);
        } catch (IOException e) {
            e.printStackTrace();
            errorLog(
                "Error reading product files: " + e.getMessage());
        } catch (ServiceException e) {
            e.printStackTrace();
            errorLog(
                "Error adding files to data model: " + e.getMessage());
        }
        setOutputFrames(outFrames0);

        // Write messages to the log according to recipe execution.
        controlLog("Completion status: " + evt.getStatus());
        Throwable error = evt.getError();
        if (error == null) {
            // no further output
        } else if (error instanceof CPLException) {
            errorLog("Execution error: " + error.getMessage());
            Throwable cause = error.getCause();
            if (cause != null) {
                StringWriter wr = new StringWriter();
                cause.printStackTrace(new PrintWriter(wr));
                errorLog("caused by\n"+wr.toString());
            }
        } else {
            StringWriter wr = new StringWriter();
            error.printStackTrace(new PrintWriter(wr));
            errorLog("caused by\n"+wr.toString());
            errorLog(wr.toString());
        }
        // flush, close and null the log file
        logFileWriter_.flush();
        logFileWriter_.close();
        logFileWriter_ = null;
        
        // update the poolDisplay_ as this request may be a part of the queue and
        // therefore its execution status should be reflected in the poolDisplay.
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                poolDisplay_.repaint();
            }});
    }

    /**
     * Installs a RequestExecutor object as the active executor of this panel.
     * This means that the executor is currently engaged in executing
     * a recipe on behalf of this panel.
     * It also deinstalls any previous incumbent of the role.
     * It does not start the executor.
     * An executor can be deactivated by calling this method with
     * <tt>ex==null</tt>.
     *
     * @param  ex  executor to install as this object's active executor
     */
    private void setActiveExecutor(RequestExecutor ex) {

        activeExecutor_ = ex;

    	// this is for DFS06600 : configureAvailableActions() is invoked
        // indirectly by code in the main thread, and therefore needs to be
        // "invoked later" to avoid nasty problems with the Swing GUI thread.
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                configureAvailableActions();
            }
        });
    }

    /**
     * Called when something has changed to ensure that actions are
     * enabled or disabled appropriately.
     */
    private void configureAvailableActions() {
        boolean isExecuting = activeExecutor_ != null;
        boolean singleShowing =
            singlePanel_.equals(tabber_.getSelectedComponent());
        boolean hasQueue = requestPool_.getNumberSelectedRequests() > 0;
        boolean isSelected = poolDisplay_.getSelectedIndices().length == 1;

        executeQueueAction_.setEnabled(hasQueue && !isExecuting);
        executeSingleAction_.setEnabled(singleShowing && !isExecuting);
        singlePanel_.setLocked(isExecuting);
        queuedPanel_.setLocked(isExecuting);
        enqueueAction_.setEnabled(singleShowing);
        tabber_.setEnabledAt(
            tabber_.indexOfComponent(queuedPanel_),
            isSelected);
        tabber_.setEnabledAt(
            tabber_.indexOfComponent(executingPanel_),
            isExecuting);
        loadAction_.setEnabled(singleShowing);
        saveAction_.setEnabled(singleShowing);
        clearPanelAction_.setEnabled(singleShowing);
        
    }

    /**
     * Returns a file chooser object for selecting a configuration file
     * to save to.
     *
     * @return   file chooser
     */
    private JFileChooser getConfigSaver() {
        if (configChooser_ == null) {
            configChooser_ = new SwingFileChooser();
            FileFilter xmlFilter = new ExtensionFileFilter(XmlSuffix, "XML files");
            configChooser_.setFileFilter(xmlFilter);
        }
        return configChooser_;
    }

    /**
     * Returns a file chooser object for selecting a configuration file
     * to load from.
     *
     * @return  file chooser
     */
    private JFileChooser getConfigLoader() {
        return getConfigSaver();
    }

    /**
     * Returns the file chooser used for saving log files.
     *
     * @return  log file chooser
     */
    private JFileChooser getLogSaver() {
        if (logChooser_ == null) {
            logChooser_ = new SwingFileChooser();
            FileFilter logFilter = new ExtensionFileFilter(".log", "Log files");
            logChooser_.setFileFilter(logFilter);
        }
        return logChooser_;
    }

    /**
     * Returns the file chooser used for saving script files.
     *
     * @return   script file chooser
     */
    private JFileChooser getScriptSaver(CommandSequence script) {
        if (scriptChooser_ == null) {
            scriptChooser_ = new SwingFileChooser();
        }
        scriptChooser_.setFileFilter(script_.getFileFilter());
        return scriptChooser_;
    }

    /**
     * Clear the log output area and restart the
     * CommandSequence object.
     */
    public void clearLogMessages() {
        logArea_.clear();
        initScript();
    }
    
    /**
     * Clear all frames in the output/product panel
     */
    public void clearOutputFrames() {
        setOutputFrames(new FrameList());
    }
    
    /**
     * Clear all frames in the input panel.
     */
    public void clearInputFrames() {
        Request req = singlePanel_.getRequest();
        req.getInputFrames().clear();
        setSingleRequest(req);
    }
    /**
     * Clear all frames and log messages from the
     * Recipe Panel.
     */
    public void clearPanel() {
        // clear log messages
        clearLogMessages();
        // get the current request, clear the input frames and set the request.
        clearInputFrames();
        // clear the output frames
        clearOutputFrames();
    }
    
    /**
     * Constructs a new CommandSequence object to receive information
     * about the commands that have been executed.
     *
     * @return  new script object
     */
    private void initScript() {
        ScriptSyntax syntax;
        Object sname = syntaxChooser_.getSelectedItem();
        if (sname.equals(CSH_SYNTAX)) {
            syntax = new CshSyntax();
        } else if (sname.equals(PERL_SYNTAX)) {
            syntax = new PerlSyntax();
        } else {
            syntax = new BourneSyntax();
        }
        script_ = new CommandSequence(syntax);
        scriptPos_ = 0;
    }

    /**
     * Invites the user to select a configuration file, and returns a
     * Request based on it.  A template request may be provided to 
     * fill in parts of the config file which are not present or not
     * selected.
     * 
     * @return Request the loaded request or null if a configuration was not loaded.
     * 
     * A configuration is not loaded on file error, if the user
     * cancels or if the configuration is for a different recipe. See
     * the CPLFrame header comment for why changing recipes after loading
     * the CPLFrame is currently disabled.
     *
     * @param  template  request object to fill in gaps
     */
    private Request loadConfiguration(Request template)
        throws IOException, CPLException, ParameterValueException {
        JFileChooser loader = getConfigLoader();
        int result = loader.showDialog(this, "Load Configuration");
        if (result == JFileChooser.APPROVE_OPTION) {
            File file = loader.getSelectedFile();
            if (file.exists()) {
                InputStream istrm = new FileInputStream(file);
                try {
                    // Deserialize the stream to create an Request.
                    Request request = new Request(istrm, getLibraryLoader());
                    boolean sameRecipe = request.getRecipe().equals(getRecipe());
                    if (!sameRecipe) {
                        String [] msg = {
                            "The selected file contain a different recipe, ("
                            +request.getRecipe().getName()+" v"+request.getRecipe().getVersion()
                            +")",
                            "The recipe and parameters will be ignored when loading.",
                            "Do you wish to continue?"};
                        if (JOptionPane.showConfirmDialog(this, msg, "Load Settings",
                            JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE)
                            == JOptionPane.CANCEL_OPTION) {
                                return null;
                        }
                    }

                    // Extract the items of information held by the request.
                    
                    Recipe recipe = request.getRecipe();
                    Parameter[] params =
                        recipe != null ? request.getParameters() : null;
                    FrameList frames = request.getInputFrames();

                    // Construct and show a selector panel which will allow
                    // the user to choose which items to load.
                    RequestItemSelector select = new RequestItemSelector();
                    
                    // removed from RequestItemSelector class.
                    // select.recipeBox_.setEnabled(recipe != null);
                    
                    select.parametersBox_.setSelected(sameRecipe && params != null);
                    select.parametersBox_.setEnabled(sameRecipe && params != null);
                    if (frames == null) {
                        select.calibFramesBox_.setSelected(false);
                        select.calibFramesBox_.setEnabled(false);
                        select.rawFramesBox_.setSelected(false);
                        select.rawFramesBox_.setEnabled(false);
                    }
                    String[] msg =
                        {
                            "Choose the items you wish to load",
                            "from the selected file" };
                    int sret = select.showDialog(this, "Load Settings", msg);

                    // According to the user's input, modify the state of this
                    // panel from parts of the deserialized request.
                    if (sret == JOptionPane.OK_OPTION) {
                        boolean useRecipe = false; // select.recipeBox_.isSelected();
                        boolean useParams = select.parametersBox_.isSelected();
                        boolean useCalib = select.calibFramesBox_.isSelected();
                        boolean useRaw = select.rawFramesBox_.isSelected();
                        boolean useRoot = select.productRootBox_.isSelected();
                        boolean useNaming =
                            select.namingSchemeBox_.isSelected();
                        Recipe recipe1;
                        Parameter[] params1;
                        MarkedFrameList frames1;
                        /* can not set recipe, see class header for details. 
                        if (recipe != null) {
                            recipe1 = recipe;
                            params1 =
                                useParams
                                    ? request.getParameters()
                                    : recipe.getDefaultParameters();
                        } else {
                            recipe1 = template.getRecipe();
                            params1 = template.getParameters();
                        }
                        */
                        recipe1 = template.getRecipe();
                        params1 = useParams ? request.getParameters() : template.getParameters();
                        if (useCalib || useRaw) {
                            frames1 = new MarkedFrameList();
                            if (useCalib) {
                                frames1.addAll(
                                    frames.selectByGroup(FrameGroup.CALIB));
                            }
                            if (useRaw) {
                                frames1.addAll(
                                    frames.selectByGroup(FrameGroup.RAW));
                            }
                        } else {
                            frames1 = template.getInputFrames();
                        }
                        File root1 =
                            useRoot
                                ? request.getProductRoot()
                                : template.getProductRoot();
                        ;
                        NamingScheme namer1 =
                            useNaming
                                ? request.getNamingScheme()
                                : template.getNamingScheme();
                        return new Request(
                            recipe1,
                            params1,
                            frames1,
                            root1,
                            namer1);
                    }
                } finally {
                    istrm.close();
                }
            } else {
                throw new FileNotFoundException("No such file: " + file);
            }
        }
        return null;
    }

    /**
     * Invites the user to name a new configuration file, and writes
     * a given request to it.
     *
     * @param  req  the request panel whose state is to be stored
     */
    private void saveConfiguration(Request request) throws IOException {
        JFileChooser saver = getConfigSaver();
        int result = saver.showDialog(this, "Save Configuration");
        if (result == JFileChooser.APPROVE_OPTION) {

            // Obtain and configure a Request object to write.  The user
            // can specify exactly which bits of the config are to be saved.
            RequestItemSelector select = new RequestItemSelector();
            /*
            if (request.getRecipe() == null) {
                select.recipeBox_.setSelected(false);
                select.recipeBox_.setEnabled(false);
            }
            */
            if (request.getInputFrames() == null) {
                select.calibFramesBox_.setSelected(false);
                select.calibFramesBox_.setEnabled(false);
                select.rawFramesBox_.setSelected(false);
                select.rawFramesBox_.setEnabled(false);
            } else {
                FrameList frames = request.getInputFrames();
                FrameList raw = frames.selectByGroup(FrameGroup.RAW);
                FrameList calib = frames.selectByGroup(FrameGroup.CALIB);
                if (raw.size() == 0) {
                    select.rawFramesBox_.setSelected(false);
                    select.rawFramesBox_.setEnabled(false);
                }
                if (calib.size() == 0) {
                    select.calibFramesBox_.setSelected(false);
                    select.calibFramesBox_.setEnabled(false);
                }
            }
            String[] msg =
                {
                    "Choose the items you wish to save",
                    "to the selected configuration file" };
            int sret = select.showDialog(this, "Save options", msg);
            if (sret == JOptionPane.OK_OPTION) {
                boolean useRecipe = true; // select.recipeBox_.isSelected();
                boolean useParams = select.parametersBox_.isSelected();
                boolean useCalib = select.calibFramesBox_.isSelected();
                boolean useRaw = select.rawFramesBox_.isSelected();
                boolean useRoot = select.productRootBox_.isSelected();
                boolean useNaming = select.namingSchemeBox_.isSelected();
                MarkedFrameList frames = new MarkedFrameList();
                if (useCalib || useRaw) {
                    MarkedFrameList frms = request.getInputFrames();
                    if (useCalib) {
                        frames.addAll(frms.selectByGroup(FrameGroup.CALIB));
                    }
                    if (useRaw) {
                        frames.addAll(frms.selectByGroup(FrameGroup.RAW));
                    }
                }
                Request req1 =
                    new Request(
                        useRecipe ? request.getRecipe() : null,
                        useParams ? request.getParameters() : null,
                        frames,
                        useRoot ? request.getProductRoot() : null,
                        useNaming ? request.getNamingScheme() : null);

                /* Do the saving. */
                File file = saver.getSelectedFile();
                // Append XML to the selected file is not already present
                if (!file.getName().endsWith(XmlSuffix)) {
                    file = new File(file.getAbsolutePath()+XmlSuffix);
                }
                OutputStream ostrm = new FileOutputStream(file);
                try {
                    req1.write(ostrm);
                } finally {
                    ostrm.close();
                }
            }
        }
    }

    /**
     * Tries to look busy.
     *
     * @param  isBusy  whether this panel should look busy or not
     */
    private void setBusy(boolean isBusy) {
        setCursor(isBusy ? BUSY_CURSOR : null);
    }

    /**
     * Asynchronously invokes {@link #setBusy} in the event dispatch thread.
     *
     * @param  isBusy  whether this panel should look busy or not
     */
    private void setBusyLater(final boolean isBusy) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                setBusy(isBusy);
            }
        });
    }

    /**
     * Ensures that the given request panel is the one which is visible in 
     * the request panel tabbed display.
     *
     * @param  panel  request panel to display
     */
    private void showRequestPanel(RequestPanel panel) {
        tabber_.setSelectedComponent(panel);
        tabber_.setEnabled(panel != executingPanel_);
    }

    /**
     * Returns the RequestPanel which is currently on view.
     *
     * @param  displayed request panel
     */
    private RequestPanel getShownRequestPanel() {
        return (RequestPanel) tabber_.getSelectedComponent();
    }

    /**
     * Returns a new (not previously used) name for a request in the queue.
     * 
     * @return  request name
     */
    private String nextQueuedRequestName() {
        return "Request " + ++nextQueued_;
    }

    private static JComponent center(JComponent comp) {
        comp.setAlignmentX(0.5f);
        comp.setAlignmentY(0.5f);
        return comp;
    }

    private static JComponent wrapIcon(Icon icon) {
        JLabel label = new JLabel(icon);
        label.setAlignmentX(0.5f);
        label.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        return label;
    }
    
    /**
     * This is one of the key message logging methods.
     * Send a message to the control writer.
     * @param msg
     */
    public void controlLog(String msg) {
        if (!msg.endsWith("\n")) {
            msg = msg+"\n";
        }
        controlWriter_.print(msg);
        controlWriter_.flush();
        if (logFileWriter_ != null) {
            logFileWriter_.print(msg);
            logFileWriter_.flush();
        }
    }
    
    /**
     * This is one of the key message logging methods.
     * Send a message to the message writer.
     * @param msg
     */
    public void messageLog(String msg) {
        if (!msg.endsWith("\n")) {
            msg = msg+"\n";
        }
        System.out.println(msg);
        messageWriter_.print(msg);
        messageWriter_.flush();
        if (logFileWriter_ != null) {
            logFileWriter_.print(msg);
            logFileWriter_.flush();
        }
    }
    
    /**
     * This is one of the key message logging methods.
     * Send a message to the error writer.
     * @param msg
     */
    public void errorLog(String msg) {
        if (!msg.endsWith("\n")) {
            msg = msg+"\n";
        }
        System.out.println(msg);
        errorWriter_.print(msg);
        errorWriter_.flush();
        if (logFileWriter_ != null) {
            logFileWriter_.print(msg);
            logFileWriter_.flush();
        }
    }
    
    /**
     * This is one of the key message logging methods.
     * Send a message to the command writer.
     * @param msg
     */
    public void commandLog(String msg) {
        if (!msg.endsWith("\n")) {
            msg = msg+"\n";
        }
        System.out.println(msg);
        commandWriter_.print(msg);
        commandWriter_.flush();
        if (logFileWriter_ != null) {
            logFileWriter_.print(msg);
            logFileWriter_.flush();
        }
    }

    /**
     * Writes the contents of a Document to a given file.
     *
     * @param  doc  the document to write
     * @param  file  the destination file
     */
    private static void writeDocument(Document doc, File file)
        throws IOException {
        Writer writer = null;
        try {
            writer = new FileWriter(file);
            writer = new BufferedWriter(writer);

            // Write the contents of the document.  This boilerplate lifted
            // from the Document.getText javadocs.
            int nleft = doc.getLength();
            Segment text = new Segment();
            text.setPartialReturn(true);
            int offs = 0;
            while (nleft > 0) {
                try {
                    doc.getText(offs, nleft, text);
                } catch (BadLocationException e) {
                    throw (IOException) new IOException(
                        e.getMessage()).initCause(
                        e);
                }
                for (char c = text.first();
                    c != Segment.DONE;
                    c = text.next()) {
                    writer.write(c);
                }
                nleft -= text.count;
                offs += text.count;
            }
        } finally {
            if (writer != null) {
                writer.close();
            }
        }
    }

    /*
     * Implementation of the MessageListener class to display CPL messages
     * in the text area.
     */
    private class LogAreaMessageListener implements MessageListener {
        private String terminate(String msg) {
            return (msg.charAt(msg.length() - 1) == '\n') ? msg : (msg + '\n');
        }
        public void errMessage(String msg) {
            errorLog(terminate("ERROR: " + msg));
        }
        public void outMessage(String msg) {
            messageLog(terminate(msg));
        }
        public void logMessage(String domain, String msg, int flags) {
            outMessage(StreamMessageListener.formatMessage(domain, msg, flags));
        }
    }

    /**
     * Helper class which implements PrintWriter so as to dump output in
     * a given text component.
     */
    private static class TextAreaWriter extends Writer {
        private LogArea target_;
        private String style_;
        public TextAreaWriter(LogArea target, String style) {
            target_ = target;
            style_ = style;
        }
        public void write(char[] cbuf, int off, int len) {
            target_.append(new String(cbuf, off, len), style_);
            moveToEnd();
        }
        public void write(String str) {
            target_.appendMessage(str);
            moveToEnd();
        }
        public void flush() {
        }
        public void close() {
        }
        private void moveToEnd() {
            target_.setCaretPosition(target_.getDocument().getLength());
        }
    }

    /**
     * Like a Request, but has an additional name attribute.  This is 
     * returned by the toString method, which in turn is used by cell
     * renderers.
     */
    private static class NamedRequest extends Request {
        final String name_;
        public NamedRequest(Request req, String name) {
            super(req);
            name_ = name;
        }
        public String toString() {
            return name_;
        }
    }

    /**
     * Get a drop handler that will association and group files dropped into a
     * set of requests to the request queue.
     * @return TransferHandler
     */
    private TransferHandler getFileListDropHandler() {
        return new FileListTransferHandler() {
            protected List exportFileList(JComponent arg0) {
                return null;
            }
            protected void importFileList(JComponent arg0, List files) {
                addInputFramesToQueue(files);
            }
        };
    }
    
    /**
     * Add the list of input frames to the queue, combining the input list
     * with the frames in the current panel.
     * The FrameList must be converted to a List of File objects, which is a little
     * wasteful as the FrameList contains classification info which is lost, 
     * and must be gathered again by call the AssociationUtils function.
     * The only information that is missing from the Frames in the FrameList (and
     * thus the call to AssociationUtils) is the grouping information.
     * @param input
     */
    public void addInputFramesToQueue(FrameList input) {
        List files = new ArrayList();
        Iterator fileIter = input.iterator();
        while (fileIter.hasNext()) {
            files.add(((Frame)fileIter.next()).getFile());
        }
        addInputFramesToQueue(files);
    }
    
    /**
     * Add the set of files to the queue, after classification, etc.
     * @param files
     */    
    public void addInputFramesToQueue(List files) {
        FileAssociationNode[] nodes =
            AssociationUtils.classifyFiles(files);
        Map groups = new HashMap();
        for (int i = 0; i < nodes.length; i++) {
            Frame frm = AssociationUtils.getFrameFromNode(nodes[i]);
            if (frm.getTag() == null) {
                // not processed, forget it..
                continue;
            }
            String groupName = nodes[i].getGrouping()[0];
            FrameList group = (FrameList) groups.get(groupName);
            if (group == null) {
                group = new FrameList();
                groups.put(groupName, group);
            }
            group.add(frm);
        }
        Iterator groupI = groups.values().iterator();
        while (groupI.hasNext()) {
            enqueueRequest(
                new NamedRequest(
                    singlePanel_.getRequest(),
                    nextQueuedRequestName()),
                (FrameList) groupI.next());
        }
    }
    
    /**
     * Add the list of input frames to the currently displayed request, unless
     * it is the execution panel, in which case nothing happens.
     * @param input
     */
    public void addInputFramesToShownRequest(FrameList input) {
        RequestPanel p = getShownRequestPanel();
        if (p != this.executingPanel_) {
            p.addInputFrames(input);
        }
    }
    
    /**
     * Get the panel name as appears in the title bar.
     * @return String
     */
    public String getPanelName() {
        return panelName_;
    }

}
