package org.eso.cpl.gui;

import java.awt.event.ActionEvent;
import java.io.File;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import org.eso.dfs.gui.SwingFileChooser;

/**
 * Component which allows the user to select a directory.
 * The component is suitable for putting in a window, but allows a 
 * file chooser dialog to be popped up on request.
 *
 * @author   Mark Taylor
 */
public class DirectorySelector extends JPanel {

    private final JFileChooser chooser_;
    private final JTextField dirField_ = new JTextField();
    private Action browseAction_;
    
    /**
     * The default selection - which is always updated to the last selection
     * made - if setDirectory() is called with null, the default selection
     * is set. 
     */
    private static File defaultDirectory_;

    
    /**
     * Constructs a new directory selector, setting explicitly the initial path.
     *
     * @param  labelText  text labelling the component
     */
    public DirectorySelector(String labelText, File initial) {
      
      // DS Jan 2007 - SPR DFS02892 (Product directory from initialization property) 
      // Added this constructor so that the parent object can read the initialisation 
      // property and instantiates this object with its value as initial directory. 
      
      this(labelText);
      
      // Use the input File only if it is a directory, otherwhise leave 
      // the current directory as is defined by DirectorySelector(String):  
      if(initial == null) { return; }
      try
      {
        if (initial.isDirectory())
        {
          // Now, here I could directly set the path withouth touching the 
          // default, like: dirField_.setText(initial.getPath());
          // but it is probably better to set defaultDirectory_ as well calling 
          // the proper set method, as long as it does not interfere with other 
          // default directories.
          setDirectory(initial);
        }
      }
      catch(Exception ignored) {}
    }    
    
    
    /**
     * Constructs a new directory selector.
     *
     * @param  labelText  text labelling the component
     */
    public DirectorySelector(String labelText) {
        setLayout(new BoxLayout(this, BoxLayout.X_AXIS));

        // Construct and configure a file chooser.
        chooser_ = new SwingFileChooser();
        chooser_.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

        // Set up an action for interactive browsing of directory tree.
        browseAction_ = new AbstractAction("Browse") {
            public void actionPerformed(ActionEvent evt) {
                File dir = getDirectory();
                if (dir != null && dir.exists() && dir.isDirectory()) {
                    chooser_.setSelectedFile(getDirectory());
                }
                int result = chooser_.showDialog(DirectorySelector.this,
                                                 "Select Directory");
                if (result == JFileChooser.APPROVE_OPTION) {
                    setDirectory(chooser_.getSelectedFile());
                }
            }
        };
        // try and set the default dir field to display the cwd, but silently
        // fail on error.
        if (defaultDirectory_ == null) {
            String cwd = System.getProperty("user.dir");
            if (cwd != null && cwd.length() > 0) {
                defaultDirectory_ = new File(cwd);
                
            }
        }
        dirField_.setText(defaultDirectory_.getPath());
        
        // Place components.
        add(new JLabel(labelText + ":"));
        add(Box.createHorizontalStrut(5));
        add(dirField_);
        add(Box.createHorizontalStrut(5));
        add(new JButton(browseAction_));
        
        /*
         * document listener to update the defaultDirectory_
         * setting when the text field is modified.
         */
        dirField_.getDocument().addDocumentListener(new DocumentListener() {
            public void insertUpdate(DocumentEvent evt) {
                textChanged();
            }
            public void removeUpdate(DocumentEvent evt) {
                textChanged();
            }
            public void changedUpdate(DocumentEvent evt) {
            }
            void textChanged() {
                defaultDirectory_ = getDirectory();
            }
        });
    }

    /**
     * Returns the selected directory, or <tt>null</tt> if none has been
     * selected.
     *
     * @return  the directory
     */
    public File getDirectory() {
        String dir = dirField_.getText();
        return (dir == null || dir.length() == 0) ? null
                                                  : new File(dir);
    }

    /**
     * Sets the selected directory.
     *
     * @param  dir  the directory
     */
    public void setDirectory(File dir) {
        dirField_.setText(dir == null ? defaultDirectory_.getPath() : dir.getPath());
        if (dir != null) {
            // update the default directory with this setting.
            defaultDirectory_ = dir;
        }
    }
    /**
     * @see java.awt.Component#setEnabled(boolean)
     */
    public void setEnabled(boolean b) {
        super.setEnabled(b);
        dirField_.setEnabled(b);
        browseAction_.setEnabled(b);
    }

    /**
     * Returns the dirField_.
     * @return JTextField
     */
    public JTextField getDirField() {
        return dirField_;
    }

    /**
     * Returns the defaultDirectory.
     * @return File
     */
    public static File getDefaultDirectory() {
        return defaultDirectory_;
    }

    /**
     * Sets the defaultDirectory.
     * @param defaultDirectory The defaultDirectory to set
     */
    public static void setDefaultDirectory(File defaultDirectory) {
        DirectorySelector.defaultDirectory_ = defaultDirectory;
    }

}
