package org.eso.cpl.gui;

import java.util.EventObject;
import org.eso.cpl.Request;
import org.eso.cpl.FrameList;
import org.eso.cpl.Recipe;

/**
 * Event describing a stage in the life of a recipe execution.
 *
 * @author   Mark Taylor
 */
public class ExecutionEvent extends EventObject {

    private final Request request_;
    private final FrameList outFrames_;
    private final Status status_;
    private final Throwable error_;

    /**
     * Constructs a new ExecutionEvent.
     *
     * @param  source the source of this event
     * @param  request  the Request which characterises the execution request
     * @param  outFrames  the output frame list (results of the execution)
     * @param  status   completion status
     * @param  th  an error that occurred during execution (may be null)
     */
    public ExecutionEvent(Object source, Request request, FrameList outFrames,
                          Status status, Throwable th ) {
        super(source);
        request_ = request;
        outFrames_ = outFrames;
        status_ = status;
        error_ = th;
    }

    /** 
     * Get the Request which this event refers to.
     *
     * @return  the request
     */
    public Request getRequest() {
        return request_;
    }

    /**
     * Gets the output frame list (result of execution).
     *
     * @return  output frame list
     */
    public FrameList getOutputFrames() {
        return outFrames_;
    }

    /**
     * Gets the execution status.
     *
     * @return  status
     */
    public Status getStatus() {
        return status_;
    }

    /**
     * Returns any exception associated with execution.
     *
     * @return  a throwable, or <tt>null</tt>
     */
    public Throwable getError() {
        return error_;
    }

    /**
     * Enumeration class enumerating the possible outcomes of an execution.
     */
    public static class Status {
        private final String name_;
        private Status(String name) {
            name_ = name;
        }
        public String toString() {
            return name_;
        }

        /** Execution started. */
        public static final Status STARTED = new Status("STARTED");

        /** Successful execution. */
        public static final Status SUCCESS = new Status("SUCCESS");

        /** Failed execution. */
        public static final Status FAILURE = new Status("FAILURE");
    }
}
