package org.eso.cpl.gui;

import java.awt.Color;
import javax.swing.JTextPane;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

/**
 * Text component which can display and append text with different styles.
 * Most of the work done by this convenience class is to restrict 
 * the possibilities of JTextPane and provide the useful parts of it 
 * in an easily digestible way.
 *
 * @author   Mark Taylor
 */
public class LogArea extends JTextPane {

    private StyledDocument doc_;

    /** String representing text style for ordinary messages. */
    public static final String MESSAGE_STYLE = "message";

    /** String representing text style for error text. */
    public static final String ERROR_STYLE = "error";

    /** String representing text style for commands issued. */
    public static final String COMMAND_STYLE = "command";

    /** String representing text style for control messages. */
    public static final String CONTROL_STYLE = "control";

    /**
     * Constructs a new LogArea.
     */
    public LogArea() {
        doc_ = getStyledDocument();
        setEditable(false);
        Style defStyle = StyleContext.getDefaultStyleContext()
                                     .getStyle(StyleContext.DEFAULT_STYLE);
        makeStyle(CONTROL_STYLE, defStyle, new Color(139, 35, 35)); // brown
        makeStyle(MESSAGE_STYLE, defStyle, Color.black);
        makeStyle(ERROR_STYLE, defStyle, Color.red);
        makeStyle(COMMAND_STYLE, defStyle, Color.blue);
    }

    /**
     * Creates a new text style.
     *
     * @param  name  name by which the new style will be referred to
     * @param  base  style on which the new one will be based on
     * @param  color  color of the new style
     * @return   new style 
     */
    private void makeStyle(String name, Style base, Color color) {
        Style style = addStyle(name, base);
        StyleConstants.setForeground(style, color);
    }

    /**
     * Removes all the text in this area.
     */
    public void clear() {
        try {
            doc_.remove(0, doc_.getLength());
        }
        catch (BadLocationException e) {
            throw new AssertionError(e);
        }
    }

    /**
     * Appends text with a given style.
     *
     * @param  string  the text to append
     * @param  styleName, one of 
     *         MESSAGE_STYLE, ERROR_STYLE, COMMAND_STYLE, CONTROL_STYLE
     */
    public void append(String string, String styleName) {
        try {
            doc_.insertString(doc_.getLength(), string, getStyle(styleName));
        }
        catch (BadLocationException e) {
            throw new AssertionError(e);
        }
    }

    /**
     * Appends text in message style.
     *
     * @param  text the text to append
     */
    public void appendMessage(String text) {
        append(text, MESSAGE_STYLE);
    }

    /**
     * Appends text in error style.
     *
     * @param  text the text to append
     */
    public void appendError(String text) {
        append(text, ERROR_STYLE);
    }

    /**
     * Appends text in command style.
     *
     * @param  text the text to append
     */
    public void appendCommand(String text) {
        append(text, COMMAND_STYLE);
    }

    /**
     * Appends text in control style.
     *
     * @param  text the text to append
     */
    public void appendControl(String text) {
        append(text, CONTROL_STYLE);
    }

    /**
     * Prevent line wrapping.
     */
    public boolean getScrollableTracksViewportWidth() {
        return false;
    }
}
