package org.eso.cpl.gui;

import java.awt.Dimension;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.util.Vector;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import org.eso.cpl.NamingScheme;
import org.eso.cpl.Request;

/**
 * Panel which allows the user to select naming scheme specifics for
 * a recipe execution request.
 *
 * @author   Mark Taylor
 */
public class NamingPanel extends JPanel {

  private final JComboBox namingSelector_;
  private final DirectorySelector rootSelector_;
  private Request request_;
  /**
   * True if the panel is in read-only mode.
   */
  private boolean locked_ = false;


  /**
   * Constructs a new naming panel.
   *
   * @param  schemes  naming schemes to permit for selection
   */
  public NamingPanel() 
  { 
    // DS FEB 2007 - SPR DFS02892 
    // The original code is now part of the overloaded constructor.
    this(null); 
  }

  public NamingPanel(String productDirPath) 
  {

    // DS FEB 2007 - SPR DFS02892 
    // added this constructor in overload, because the Gasgano 
    // Properties should not be access directly from the javacpl package
    File productDir = null;
    if(productDirPath != null)
    {
      try
      { 
        productDir = new File(productDirPath);
        if(! productDir.isDirectory())
        {
          productDir = null;
        }
      }
      catch(Exception ignored)
      {
        productDir = null;
      }
    }

    if( productDir != null )
    { 
      rootSelector_ = new DirectorySelector("Product Root Directory", productDir);
    }
    else
    {
      rootSelector_ = new DirectorySelector("Product Root Directory");
    }

    JTextField dirField = rootSelector_.getDirField();
    dirField.getDocument().addDocumentListener(new DocumentListener() {
      public void insertUpdate(DocumentEvent evt) {
        textChanged();
      }
      public void removeUpdate(DocumentEvent evt) {
        textChanged();
      }
      public void changedUpdate(DocumentEvent evt) {
      }
      void textChanged() {
        request_.setProductRoot(rootSelector_.getDirectory());
      }
    });

    // Construct a component to do naming scheme selection.
    namingSelector_ =
      new JComboBox(new Vector(NamingScheme.STANDARD_SCHEMES));
    namingSelector_.setSelectedIndex(0);
    namingSelector_.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent evt) {
        if (evt.getStateChange() == ItemEvent.SELECTED) {
          Object selected = namingSelector_.getSelectedItem();
          request_.setNamingScheme((NamingScheme) selected);
        }
      }
    });

    // Do layout of the panel.
    setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
    add(rootSelector_);
    add(Box.createHorizontalStrut(20));
    add(new JLabel("Naming Scheme: "));
    add(namingSelector_);
    GUIUtils.wrapInBorder(this, "Product Naming");

    // Make sure we are displaying something sensible.
    setRequest(new Request(null, null, null, null, NamingScheme.NUMERIC));
  }

  /**
   * Sets the Request object which will hold state about this panel.
   *
   * @param  req  the request
   */
  public void setRequest(Request req) {
    request_ = req;
    // set the product root to the default if there is not one
    // already set.
    if (request_.getProductRoot() == null) {
      request_.setProductRoot(getDefaultProductRoot());
    }
    namingSelector_.setSelectedItem(req.getNamingScheme());
    rootSelector_.setDirectory(req.getProductRoot());
  }

  public Dimension getMaximumSize() {
    return new Dimension(super.getMaximumSize().width,
        super.getPreferredSize().height);
  }
  /**
   * Returns the locked_.
   * @return boolean
   */
  public boolean isLocked() {
    return locked_;
  }

  /**
   * Sets the locked_.
   * @param locked_ The locked_ to set
   */
  public void setLocked(boolean locked) {
    locked_ = locked;
    namingSelector_.setEnabled(!locked);
    rootSelector_.setEnabled(!locked);
  }

  /**
   * Get the directory that is the current default product root.
   * @return File
   */
  public static File getDefaultProductRoot() {
    return DirectorySelector.getDefaultDirectory();
  }

}
