package org.eso.cpl.gui;

import java.awt.Component;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import org.eso.cpl.RangeConstraint;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterConstraint;
import org.eso.cpl.ParameterValueException;

/**
 * A TableModel which represents all known information about a set of
 * {@link org.eso.cpl.Parameter} objects.
 *
 * @author   Mark Taylor
 */
public class ParameterTableModel extends DataColumnTableModel {

    private final Parameter[] params_;

    /**
     * Constructs a new ParameterTableModel from an array of parameters.
     *
     * @param  params  the array of parameters which the new table will display
     */
    public ParameterTableModel(Parameter[] params) {
        params_ = params;
        final int nrow = params.length;

        // Set up cell renderers and editors tailored to the data types
        // of each parameter.  These are used for rendering/editing the
        // actual and default value fields.
        final TableCellRenderer[] valueRenderers = new TableCellRenderer[nrow];
        final TableCellEditor[] valueEditors = new TableCellEditor[nrow];
        for (int irow = 0; irow < nrow; irow++) {
            Parameter param = getParameter(irow);
            valueRenderers[irow] = Renderers.getParameterValueRenderer(param);
            valueEditors[irow] = Renderers.getParameterValueEditor(param);
        }

        // All the columns in this table will display help text for the 
        // relevant parameter as a tool tip.
        abstract class ParameterDataColumn extends DataColumn {
            
            public ParameterDataColumn(String name, boolean isEditable, int prefWidth) {
                super(name, isEditable, prefWidth);
            }
            public ParameterDataColumn(String name, boolean isEditable) {
                super(name, isEditable);
            }
            public String getRendererTip(int irow) {
                return getParameter(irow).getHelp();
            }
        }

        // Set up DataColumn objects for each of the various immutable fields
        // in the parameters.  Each of these columns knows how to display
        // its bit of information about each parameter in the list.
        DataColumn contextColumn = new ParameterDataColumn("Context", false) {
            public Object getValue(int irow) {
                return getParameter(irow).getContext();
            }
        };
        DataColumn defaultColumn = new ParameterDataColumn("Default", false) {
            public Object getValue(int irow) {
                return getParameter(irow).getDefault();
            }
            public TableCellRenderer getCellRenderer(int irow) {
                return valueRenderers[irow];
            }
        };
        DataColumn helpColumn = new ParameterDataColumn("Description", false) {
            public Object getValue(int irow) {
                return getParameter(irow).getHelp();
            }
        };
        DataColumn nameColumn = new ParameterDataColumn("Name", false, 200) {
            public Object getValue(int irow) {
                return getParameter(irow).getName();
            }
        };
        DataColumn rangeColumn = new ParameterDataColumn("Range", false) {
            public Object getValue(int irow) {
                ParameterConstraint cons = getParameter(irow).getConstraint();
                if (cons instanceof RangeConstraint) {
                    RangeConstraint range = (RangeConstraint) cons;
                    return range.getMinimum() + ".." + range.getMaximum();
                }
                else {
                    return null;
                }
            }
        };
        DataColumn tagColumn = new ParameterDataColumn("Tag", false) {
            public Object getValue(int irow) {
                return getParameter(irow).getTag();
            }
        };
        DataColumn typeColumn = new ParameterDataColumn("Type", false) {
            public Object getValue(int irow) {
                return getParameter(irow).getType();
            }
        };

        // Set up a DataColumn for displaying the actual value of the 
        // parameters.  As well as display, it also needs to be able to
        // set the value if a new one is supplied by an editor.
        DataColumn valueColumn = new ParameterDataColumn("Value", true) {
            public Object getValue(int irow) {
                return getParameter(irow).getValue();
            }
            public void setValue(int irow, Object value) {
                Parameter param = getParameter(irow);
                try {
                    if (value instanceof String) { 
                        param.setValueFromString((String) value);
                    }
                    else {
                        param.setValue(value);
                    }
                }
                catch (ParameterValueException e) {
                    // This can be triggered by the JTable even though it's
                    // about to block the edit as an illegal operation
                    // (because the editor's stopCellEditing method has
                    // returned false).  So ignore it.
                }
            }
            public TableCellRenderer getCellRenderer(int irow) {
                return valueRenderers[irow];
            }
            public TableCellEditor getCellEditor(int irow) {
                return valueEditors[irow];
            }
        };

        // Set up a list of DataColumns in the order in which they should
        // appear in the table.
        List columns = new ArrayList();
        columns.add(nameColumn);
        columns.add(valueColumn);
        columns.add(defaultColumn);
        columns.add(rangeColumn);
        // dont show Description - it is not that helpful for existing CPL
        // recipes and appears in the tool tips.
        // columns.add(helpColumn);
        // columns.add(contextColumn);
        // columns.add(tagColumn);

        // Don't need to show types, since it's clear from the rendering.
        // columns.add(typeColumn); 

        // Use the list as the data for this table model.
        setDataColumns(columns);
    }

    public int getRowCount() {
        return params_.length;
    }

    /**
     * Returns the Parameter which is described by a given row of this model.
     *
     * @param  irow  the table row 
     * @return  the Parameter associated with <tt>irow</tt>
     */
    public Parameter getParameter(int irow) {
        return params_[irow];
    }
}
