package org.eso.cpl.gui;

import java.awt.Color;
import java.awt.Component;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.DefaultTreeCellRenderer;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;


/**
 * A JTree which displays recipe libraries.
 * Nodes are all {@link javax.swing.tree.DefaultMutableTreeNode} objects,
 * for which the user object may currently be one of:
 * <ul>
 * <li>{@link org.eso.cpl.RecipeLibrary},
 * <li>{@link org.eso.cpl.Recipe}.
 * </ul>
 *
 * @author   Mark Taylor
 */
public class RecipeTree extends JTree {

    private final DefaultTreeModel model_;
    private final MutableTreeNode root_;

    /**
     * Constructs a new RecipeTree.
     */
    public RecipeTree() {
        super(new DefaultTreeModel(new DefaultMutableTreeNode(), true));
        model_ = (DefaultTreeModel) getModel();
        root_ = (MutableTreeNode) model_.getRoot();
        setCellRenderer(new RecipeCellRenderer());
        setRootVisible(false);
        setShowsRootHandles(true);
    }

    /**
     * Adds a recipe library to the set from which recipes can be selected.
     *
     * @param  lib  the library to add
     */
    public void addLibrary(RecipeLibrary lib) {

        // Create a new node based representing the library.
        DefaultMutableTreeNode libNode = new DefaultMutableTreeNode(lib, true);

        // Populate it with children representing its recipes.
        Recipe[] recipes = lib.getRecipes();
        for (int i = 0; i < recipes.length; i++) {
            libNode.add(new DefaultMutableTreeNode(recipes[i], false));
        }

        // Populate it with non-recipe children if it has any.
        Object[] others = lib.getOtherItems();
        if (others != null) {
            for (int i = 0; i < others.length; i++) {
                libNode.add(new DefaultMutableTreeNode(others[i], false));
            }
        }

        // Add the new node to the tree.
        model_.insertNodeInto(libNode, root_, root_.getChildCount());

        // Expand the library node so its children are visible.
        expandPath(new TreePath(libNode.getPath()));
    }

    /**
     * Convenience method to return the user object from a tree model node.
     * The returned value should be a non-null object, but if something
     * strange has happened to the tree it might return null (without error).
     *
     * @param  node  the tree node to query
     * @return  user object at the node; should be a 
     *          {@link org.eso.cpl.Recipe} or {@link org.eso.cpl.RecipeLibrary}
     */
    public Object getUserObject(Object node) {
        return (node instanceof DefaultMutableTreeNode)
             ? ((DefaultMutableTreeNode) node).getUserObject()
             : null;
    }

    /**
     * Indicates whether the given model node is useless to expected 
     * users of this tree.
     *
     * @param  node  the tree node to query
     * @return  true iff <tt>node</tt> is duff
     */
    private boolean isDuff(Object node) {
        Object uo = getUserObject(node);
        return ! ((uo instanceof RecipeLibrary) ||
                  (uo instanceof Recipe));
    }

    /**
     * Custom cell rendering.
     */
    private class RecipeCellRenderer extends DefaultTreeCellRenderer {
 
        Color color0;
        Color color1; 

        public Component getTreeCellRendererComponent(JTree tree, Object node,
                                                      boolean selected,
                                                      boolean expanded,
                                                      boolean leaf, int row,
                                                      boolean hasFocus) {

            // Superclass renderer does the hard work.
            super.getTreeCellRendererComponent(tree, node, selected, expanded,
                                               leaf, row, hasFocus);
                   
            // Get the object from inside the tree node.
            Object uo = getUserObject(node);
            if (node != null) {

                // Custom rendering for RecipeLibrary.
                if (uo instanceof RecipeLibrary) {
                    RecipeLibrary lib = (RecipeLibrary) uo;
                    setText(lib.getName());
                }

                // Custom rendering for Recipe.
                else if (uo instanceof Recipe) {
                    Recipe rec = (Recipe) uo;
                    setText(rec.getName());
                }
            }

            // Provide visual information about whether this is a selectable
            // item or not.
            if (color1 == null) {
                color1 = getForeground();
                color0 = new Color(color1.getColorSpace(),
                                   color1.getColorComponents(null), 0.5f);
            }
            setForeground(isDuff(node) ? color0 : color1);

            // Return this as the rendered cell.
            return this;
        }
    }
}
