package org.eso.cpl.gui;

import java.awt.Component;
import java.awt.Toolkit;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.EventObject;
import java.util.List;
import java.util.regex.Pattern;

import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.event.CellEditorListener;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import org.eso.cpl.EnumConstraint;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterConstraint;
import org.eso.cpl.ParameterType;
import org.eso.cpl.ParameterValueException;

/**
 * One stop shop for TableCellRenderers and TableCellEditors used by
 * classes in this package.
 * <p>
 * There is quite a lot of complication here to achieve not very dramatic
 * effects on rendering.  It may not be the most elegant or maintainable
 * code I have ever written.  The good news is that if you don't like it
 * you can rewrite it or replace it by something much simpler which will
 * work almost as well.
 *
 * @author   Mark Taylor
 */
class Renderers {

    private static JTable dummyJTable_ = new JTable();
    private static TableCellRenderer floatRenderer_ =
        new FloatingPointRenderer(dummyJTable_
                                 .getDefaultRenderer(Float.class));
    private static TableCellRenderer doubleRenderer_ = 
        new FloatingPointRenderer(dummyJTable_
                                 .getDefaultRenderer(Double.class));
    private static TableCellRenderer trailingStringRenderer_ =
        new TrailingRenderer(dummyJTable_.getDefaultRenderer(String.class));
    private static TableCellEditor trailingStringEditor_ =
        new TrailingEditor(getDefaultEditor(String.class));
    private static TableCellRenderer componentRenderer_ =
        new ComponentRenderer();

    /**
     * Provides a TableCellRenderer object suitable for use with a given
     * Parameter.
     *
     * @param  param a parameter
     * @return  a cell renderer to render <tt>param</tt> values
     */
    public static TableCellRenderer getParameterValueRenderer(Parameter param) {
        return getDefaultRenderer(param.getType().getValueClass());
    }

    /**
     * Provides a TableCellEditor object suitable for use with a given
     * Parameter.
     *
     * @param  param  a parameter
     * @return  a cell editor to edit <tt>param</tt> values
     */
    public static TableCellEditor getParameterValueEditor(Parameter param) {
        ParameterConstraint constraint = param.getConstraint();
        ParameterType ptype = param.getType();

        // EnumConstraint gets an EnumEditor.  No validation is required,
        // since the combobox can only return one of the given options.
        if (constraint instanceof EnumConstraint) {
            return getOptionEditor(((EnumConstraint) constraint).getOptions());
        }

        // Otherwise use a default editor but modify its validation method to
        // give the parameter validation a chance to reject values.
        else {
            final Parameter param1 = param;
            Class clazz = ptype.getValueClass();
            return new DelegateTableCellEditor(getDefaultEditor(clazz)) {
                public boolean stopCellEditing() {
                    super.stopCellEditing();
                    boolean ok = attemptSetValue(param1, getCellEditorValue());
                    if (ok) {
                        return super.stopCellEditing();
                    }
                    else {
                        super.cancelCellEditing();
                        return false;
                    }
                }
            };
        }
    }

    /**
     * Returns an editor from which can be selected one of a given
     * disjunction of options.
     *
     * @param  options  the options which form legal input values for the
     *         returned editor
     * @return  a suitable editor object
     */
    public static TableCellEditor getOptionEditor(List options) {
        Object[] opts = options.toArray();
        JComboBox optSelector = new JComboBox(opts);
        final TableCellEditor ed = new DefaultCellEditor(optSelector);
        optSelector.addPopupMenuListener(new PopupMenuListener() {
            public void popupMenuCanceled(PopupMenuEvent evt) {
            }
            public void popupMenuWillBecomeInvisible(PopupMenuEvent evt) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        ed.stopCellEditing();
                    }
                });
            }
            public void popupMenuWillBecomeVisible(PopupMenuEvent evt) {
            }
        });
        return ed;
    }

    /**
     * Returns a cell renderer which renders oversize strings so that
     * the right hand end is visible.
     * <p>
     * <strong>Note: this method doesn't work.  It just gives you a normal
     * renderer.  Needs to be written properly.</strong>
     *
     * @return  trailing renderer
     */
    public static TableCellRenderer getTrailingStringRenderer() {
        return trailingStringRenderer_;
    }

    /**
     * Returns a cell editor which displays oversize strings so that
     * initially the right hand end is visible.
     * <p>
     * <strong>Note: this method doesn't work.  It just gives you a normal
     * renderer.  Needs to be written properly.</strong>
     *
     * @return  trailing editor
     */
    public static TableCellEditor getTrailingStringEditor() {
        return trailingStringEditor_;
    }

    /**
     * Returns a standard kind of renderer component for rendering objects
     * of a given class.
     *
     * @param  clazz  the class to render
     * @return  suitable renderer for <tt>clazz</tt> objects
     */
    public static TableCellRenderer getDefaultRenderer(Class clazz) {
        if (clazz == Float.class) {
            return floatRenderer_;
        }
        else if (clazz == Double.class) {
            return doubleRenderer_;
        }
        else {
            TableCellRenderer rend = dummyJTable_.getDefaultRenderer(clazz);
            return (rend == null) ? new DefaultTableCellRenderer() : rend;
        }
    }

    /**
     * Returns a cell renderer which operates on a model values which
     * are themselves AWT Component objects.  The rendering just consists
     * of returning the object in question.
     *
     * @return  component renderer
     */
    public static TableCellRenderer getComponentRenderer() {
        return componentRenderer_;
    }
    
    /**
     * Follows pattern set by ComponentRenderer, with the difference that
     * a new editor is returned for each call to getComponentEditor() (i.e.
     * one editor per column).
     * 
     * Returns a cell editor which operates on a model values which
     * are themselves AWT Component objects.  The editor just consists
     * of returning the object in question.
     *
     * @return  component editor
     */
    public static TableCellEditor getComponentEditor() {
        return new ComponentEditor();
    }

    /**
     * Returns a standard kind of editor component for editing objects 
     * of a given class.
     *
     * @param  clazz  the class to edit
     * @return   suitable editor for <tt>clazz</tt> objects
     */
    public static TableCellEditor getDefaultEditor(Class clazz) {

        // Get the editor that a JTable would supply.  This is generally 
        // nicer than what you'd get by constructing a new DefaultCellEditor.
        TableCellEditor ed = dummyJTable_.getDefaultEditor(clazz);
        if (ed == null) {
            ed = new DefaultCellEditor(new JTextField());
        }

        // In all cases, we want one click to start editing.  This is because
        // selection is currently disabled in all our tables, so single
        // clicks have no other purpose.  If selection becomes required,
        // this policy may have to change.
        if (ed instanceof DefaultCellEditor) {
            ((DefaultCellEditor) ed).setClickCountToStart(1);
        }

        // Don't select cells, there's no point and it leaves the focus in
        // the wrong place.
        ed = new DelegateTableCellEditor(ed) {
            public boolean shouldSelectCell(EventObject evt) {
                return false;
            }
        };

        // Return the configured editor.
        return ed;
    }

    /**
     * Attempts to set the value of a parameter to a given value, indicating
     * whether it succeeded or not. 
     * 
     * @param  param  the parameter whose value is to be set
     * @param  val   the value to set <tt>param</tt> to
     * @return  <tt>true</tt> if <tt>param</tt> was set successully to 
     *          <tt>val</tt>, or
     *          <tt>false</tt> if it failed because <tt>val</tt> was not
     *          a valid value for <tt>param</tt>
     */
    private static boolean attemptSetValue(Parameter param, Object val) {
        if (val == null) {
            return false;
        }
        try {  
            if (val instanceof String) {
                param.setValueFromString((String) val);
            }
            else {
                param.setValue(val);
            }
            return true;
        }
        catch (ParameterValueException e) {
            Toolkit.getDefaultToolkit().beep();
            return false;
        }
    }

    /**
     * Cell renderer which renders component objects directly. 
     */
    private static class ComponentRenderer implements TableCellRenderer {
        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int irow, int icol) {
            return (value instanceof Component) 
                 ? (Component) value
                 : null;
        }
    }

    /**
     * Cell renderer which renders component objects directly. 
     */
    private static class ComponentEditor implements TableCellEditor {

        private Component comp;
        
        /**
         * @see javax.swing.table.TableCellEditor#getTableCellEditorComponent(javax.swing.JTable, java.lang.Object, boolean, int, int)
         */
        public Component getTableCellEditorComponent(
            JTable table,
            Object value,
            boolean isSelected,
            int row,
            int column) {
                if (value instanceof Component) {
                    comp = (Component)value;
                }
                return comp;
        }
        /**
         * @see javax.swing.CellEditor#getCellEditorValue()
         */
        public Object getCellEditorValue() {
            return comp;
        }
        /**
         * @see javax.swing.CellEditor#isCellEditable(java.util.EventObject)
         */
        public boolean isCellEditable(EventObject anEvent) {
            return true;
        }
        /**
         * @see javax.swing.CellEditor#shouldSelectCell(java.util.EventObject)
         */
        public boolean shouldSelectCell(EventObject anEvent) {
            return false;
        }
        /**
         * @see javax.swing.CellEditor#stopCellEditing()
         */
        public boolean stopCellEditing() {
            return true;
        }
        /**
         * @see javax.swing.CellEditor#cancelCellEditing()
         */
        public void cancelCellEditing() {
        }
        /**
         * @see javax.swing.CellEditor#addCellEditorListener(javax.swing.event.CellEditorListener)
         */
        public void addCellEditorListener(CellEditorListener l) {
        }
        /**
         * @see javax.swing.CellEditor#removeCellEditorListener(javax.swing.event.CellEditorListener)
         */
        public void removeCellEditorListener(CellEditorListener l) {
        }
    }

    /**
     * Cell renderer which renders floating point values.
     * This derives most of its behaviour from a given renderer instance,
     * but makes sure the displayed value doesn't look like an integer 
     * (it sticks a ".0" after it if necessary).
     */
    private static class FloatingPointRenderer implements TableCellRenderer {
        private final TableCellRenderer base_;
        private static Pattern intPattern_ = Pattern.compile("[\\+\\-]?[0-9]*");
        public FloatingPointRenderer(TableCellRenderer base) {
            base_ = base;
        }
        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int irow, int icol) {
            Component comp = 
                base_.getTableCellRendererComponent(table, value, isSelected,
                                                    hasFocus, irow, icol);
            if (comp instanceof JLabel) {
                JLabel label = (JLabel) comp;
                String text = label.getText();
                if (intPattern_.matcher(text).matches()) {
                    label.setText(text + ".0");
                }
            }
            return comp;
        }
    }

    /**
     * A cell editor in which the right hand end of an oversize content
     * string will be displayed (rather than the left hand end, which
     * is usually the default).
     * <p>
     * <strong>Note: This class doesn't do anything.  It just gives you
     * a normal editor.  It needs to be written so it works.</strong>
     */
    private static class TrailingEditor extends DelegateTableCellEditor {
        public TrailingEditor(TableCellEditor base) {
            super(base);
        }
        public Component getTableCellEditorComponent(JTable table, Object value,
                                                     boolean isSelected,
                                                     int irow, int icol) {
            Component comp =
                super.getTableCellEditorComponent(table, value, isSelected,
                                                  irow, icol);
            return comp;
        }
    }

    /**
     * A cell renderer in which the right hand end of an oversize content
     * string will be displayed (rather than the left hand end, which
     * is usually the default).
     * <p>
     * <strong>Note: This class doesn't do anything.  It just gives you
     * a normal renderer.  It needs to be written so it works.</strong>
     */
    private static class TrailingRenderer implements TableCellRenderer {
        private final TableCellRenderer base_;
        public TrailingRenderer(TableCellRenderer base) {
            base_ = base;
        }
        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int irow, int icol) {
            Component comp =
                base_.getTableCellRendererComponent(table, value, isSelected, 
                                                    hasFocus, irow, icol);
            return comp;
        }
    }

    /**
     * Helper class which implements a TableCellEditor by delegating all
     * its methods to a base object.  This is used so that we can
     * 'inherit' behaviour from a TableCellEditor obtained at runtime,
     * replacing behaviour of a few of its methods.
     */
    private static class DelegateTableCellEditor implements TableCellEditor {
        final TableCellEditor base_;
        DelegateTableCellEditor(TableCellEditor base) {
            base_ = base;
        }
        public void addCellEditorListener(CellEditorListener listener) {
            base_.addCellEditorListener(listener);
        }
        public void cancelCellEditing() {
            base_.cancelCellEditing();
        }
        public Object getCellEditorValue() {
            return base_.getCellEditorValue();
        }
        public boolean isCellEditable(EventObject evt) {
            return base_.isCellEditable(evt);
        }
        public void removeCellEditorListener(CellEditorListener listener) {
            base_.removeCellEditorListener(listener);
        }
        public boolean shouldSelectCell(EventObject evt) {
            return base_.shouldSelectCell(evt);
        }
        public boolean stopCellEditing() {
            return base_.stopCellEditing();
        }
        public Component getTableCellEditorComponent(JTable table, Object value,
                                                     boolean isSelected,
                                                     int irow, int icol) {
            return base_.getTableCellEditorComponent(table, value, isSelected,
                                                     irow, icol);
        }
    }
}
