package org.eso.cpl.jni;

import org.eso.cpl.ParameterConstraint;
import org.eso.cpl.ParameterImp;
import org.eso.cpl.ParameterType;

/**
 * Provides parameter implementation services based on a cpl_parameter
 * structure in a cpl_parameter_list.
 */
public class JNIParameterImp implements ParameterImp {
    private final byte[] nativeState_;
    private final JNIRecipe recipe_;
    private final ParameterType type_;
    private final ParameterConstraint constraint_;
    private final Object initialValue_;
    private String context_;
    private String help_;
    private String name_;
    private String tag_;
    private boolean isActive_ = true;

    /**
     * Private constructor - only invoked by native code.
     *
     * @param  nativeState  opaque byte array used by native code
     * @param  recipe  the JNIRecipe with which this parameter is associated -
     *         reference kept since the recipe's finalization would
     *         invalidate this parameter
     * @param  type  parameter type object
     * @param  constraint  parameter constraint object
     * @param  initialValue  parameter's initial value
     */
    private JNIParameterImp(byte[] nativeState, JNIRecipe recipe,
                            ParameterType type, ParameterConstraint constraint, 
                            Object initialValue) {
        nativeState_ = nativeState;
        recipe_ = recipe;
        type_ = type;
        constraint_ = constraint;
        initialValue_ = initialValue;
    }

    public String getContext() {
        return context_;
    }

    public String getHelp() {
        return help_;
    }

    public String getName() {
        return name_;
    }

    public String getTag() {
        return tag_;
    }

    public ParameterType getType() {
        return type_;
    }

    public ParameterConstraint getConstraint() {
        return constraint_;
    }

    public Object getInitialValue() {
        return initialValue_;
    }

    /**
     * Sets the value of this parameter.  If called following deactivation,
     * an IllegalStateException will result.
     *
     * @param   val  the new object value
     */
    public void setValue(Object val) {
        if (isActive_) {
            synchronized (CPLControl.LOCK) {
                nativeSetValue(val);
            }
        }
        else {
            throw new IllegalStateException("Parameter " + getName() + 
                                            " is no longer active");
        }
    }

    /**
     * Mandates that the native methods of this parameter may
     * no longer be used.  This method is called when the JNIRecipe 
     * which spawned this parameter object is deinitialized, since 
     * such deinitialize will invalidate native resources associated
     * with this object.  A deactivated object cannot be reactivated.
     */
    void deactivate() {
        isActive_ = false;
    }

    private native void nativeSetValue(Object val);
}
