package org.eso.cpl.jni;

import java.io.File;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;

/**
 * An object representing a DLL which may contain one or more recipes.
 * <p>
 * This object knows which DLL it corresponds to (it keeps a copy of
 * the handle at the native level), but it does not itself keep track
 * of whether that library is loaded.  Whether it is loaded depends on
 * whether any recipes are registered with it.  This is controlled by
 * the <tt>register</tt> and <tt>deregister</tt> methods. 
 * A recipe will normally register with the library on construction,
 * and deregister with it on finalization (or possibly before), so
 * when the last one deregisters the DLL can be unloaded.
 *
 * @author   Mark Taylor
 */
public class PluginLibrary implements RecipeLibrary {

    private final byte[] nativeState_;
    private final String location_;
    private final String name_;
    private final JNIRecipe[] recipes_;
    private final OtherPlugin[] otherPlugins_;
    private int nReg_ = 0;

    /**
     * Constructs a new PluginLibrary.  This constructor is only called
     * from native code.
     *
     * @param  nativeState  a buffer containing opaque bytes used by 
     *         native code
     * @param  location  the name of the library file
     * @param  recipes  the recipes held by this library
     * @param  otherPlugins  the non-recipe plugins held by this library
     */
    private PluginLibrary(byte[] nativeState, String location,
                          JNIRecipe[] recipes, OtherPlugin[] otherPlugins) {
        nativeState_ = nativeState;
        location_ = location;
        recipes_ = recipes;
        otherPlugins_ = otherPlugins;
        name_ = new File(location).getName();
    }

    /**
     * Returns an array of recipes which are available from this library.
     * The return value may be a zero-element array if this does appear
     * to be a CPL plugin library but contains no recipes (it may contain
     * other types of plugins), or <tt>null</tt> if the loadable module
     * represented by this PluginLibrary does not appear to be a CPL
     * plugin library.
     *
     * @return  an array of the usable recipes in this library
     */
    public Recipe[] getRecipes() {
        return recipes_;
    }

    /**
     * Returns an array of plugin objects found in the module which
     * do not represent usable recipes.  The results of this method
     * may be useful in the case that a module does not appear to 
     * contain the recipes that it was believed to.
     *
     * @return  an array of type {@link OtherPlugin}[] objects
     */
    public Object[] getOtherItems() {
        return otherPlugins_;
    }

    /**
     * Returns the filename of the DLL.
     *
     * @return  name
     */
    public String getName() {
        return name_;
    }

    /**
     * Returns the pathname of the DLL.
     *
     * @return  location
     */
    public String getLocation() {
        return location_;
    }

    public String toString() {
        return name_;
    }

    /**
     * Registers a user of this library.  
     * This serves to keep a reference count of library users so that 
     * the DLL can be unloaded when it is no longer needed.
     */
    void register() {
        nReg_++;
    }

    /**
     * Deregisters a user of this library;
     * This serves to keep a reference count of library users so that 
     * the DLL can be unloaded when it is no longer needed.
     */
    void deregister() throws LTDLException {
        nReg_--;
        if (nReg_ <= 0) {
            synchronized (CPLControl.LOCK) {
                nativeUnload();
            }
        }
    }

    private native void nativeUnload() throws LTDLException;
}
