package org.eso.cpl.jni;

import java.io.PrintStream;

/**
 * Basic implementation of the <tt>MessageListener</tt> interface.
 * Messages are handled by sending them to print streams.
 *
 * @author   Mark Taylor
 */
public class StreamMessageListener implements MessageListener {

    private PrintStream out_;
    private PrintStream err_;

    /**
     * Constructs a new message listener which sends messages to specified
     * print streams.
     *
     * @param  out  the stream for normal (out and log) messages
     * @param  err  the stream for error messages
     */
    public StreamMessageListener(PrintStream out, PrintStream err) {
        out_ = out;
        err_ = err;
    }

    /**
     * Constructs a new message listener which sends messages to 
     * <tt>System.out</tt> and <tt>System.err</tt>.
     */
    public StreamMessageListener() {
        this(System.out, System.err);
    }

    /**
     * Sets the stream to which normal messages, including log messages,
     * are output.
     * If set to <tt>null</tt>, such messages will be discared.
     *
     * @param  out  output stream
     */
    public void setOutputStream(PrintStream out) {
        out_ = out;
    }

    /**
     * Sets the stream to which error messages are output.
     * If set to <tt>null</tt>, such messages will be discarded.
     *
     * @param  err  error stream
     */
    public void setErrorStream(PrintStream err) {
        err_ = err;
    }

    /**
     * Returns the stream to which normal messages are output.
     *
     * @return  output stream
     */
    public PrintStream getOutputStream() {
        return out_;
    }

    /**
     * Returns the stream to which error messages are output.
     *
     * @return  error stream
     */
    public PrintStream getErrorStream() {
        return err_;
    }

    /*
     * Implementation of the MessageListener interface.
     */

    public void outMessage(String msg) {
        PrintStream out = getOutputStream();
        if (out != null) {
            out.print(msg);
        }
    }

    public void errMessage(String msg) {
        PrintStream err = getErrorStream();
        if (err != null) {
            err.print(msg);
        }
    }

    public void logMessage(String domain, String message, int flags) {
        PrintStream out = getOutputStream();
        if (out != null) {
            out.print(formatMessage(domain, message, flags));
        }
    }

    /**
     * Formats a log message for output.
     * 
     * @param   domain  the domain of the message
     * @param   msg     the message to log
     * @param   flags   an integer flag pertaining to this message; 
     *          the format of this is currently undefined
     */
    public static String formatMessage(String domain, String message,
                                       int flags) {
        return domain + ": " + flags + "\t" + message;
    }
}
