package org.eso.cpl.test;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.SequenceInputStream;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import junit.framework.AssertionFailedError;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

import org.eso.cpl.CPLException;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameList;
import org.eso.cpl.MarkedFrameList;
import org.eso.cpl.NamingScheme;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterType;
import org.eso.cpl.ParameterValueException;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;
import org.eso.cpl.Request;
import org.eso.cpl.jni.LTDLException;
import org.eso.cpl.jni.LibraryLoader;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Test for serialisation and deserialisation of a Request object.
 * <p>
 * $Id: RequestTest.java,v 1.7 2008/10/01 15:49:11 mpadovan Exp $
 * @author   Mark Taylor
 */
public class RequestTest extends TestCase {

    public RequestTest(String name) {
        super(name);
    }

    static Request getRequest() throws LTDLException, CPLException {
        RecipeLibrary pluglib = JNICPLTest.getGiraffeLibrary();
        Recipe mbRecipe = pluglib.getRecipes()[0];
        MarkedFrameList inFrames = JNICPLTest.getMasterBiasInputFrames();
        return new Request(mbRecipe, mbRecipe.getDefaultParameters(), 
                           inFrames, new File("."), NamingScheme.OVERWRITE);
    }

    public void testXMLSerialization()
            throws LTDLException, CPLException, ParameterValueException,
                   IOException, SAXException {
        Request req1 = getRequest();
        req1.getRecipe();
        req1.setNamingScheme(NamingScheme.NUMERIC);
        req1.setProductRoot(new File("/some/dir"));
        Parameter par = req1.getParameters()[3];
        assertEquals(ParameterType.INT, par.getType());
        par.setValue(new Integer(9999));
        Request req2 = serializeDeserialize(req1);
        assertRequestEquals(req1, req2);

        Parameter p2 = req2.getParameters()[1];
        assertEquals(ParameterType.DOUBLE, p2.getType());
        p2.setValue(new Double(((Double)p2.getValue()).doubleValue() + 1.0));
        assertRequestNotEquals(req1, req2);
        p2.setValue(p2.getDefault());
        assertRequestEquals(req1, req2);

        FrameList fl2 = req2.getInputFrames();
        fl2.getFrame(0).setGroup(FrameGroup.PRODUCT);
        assertRequestNotEquals(req1, req2);
        fl2.getFrame(0).setGroup(FrameGroup.RAW);
        assertRequestEquals(req1, req2);
        Frame frm = (Frame) fl2.remove(fl2.size() - 1);
        assertRequestNotEquals(req1, req2);
        fl2.add(frm);
        assertRequestEquals(req1, req2);

        req2.setNamingScheme(NamingScheme.OVERWRITE);
        assertRequestNotEquals(req1, req2);
        req2.setNamingScheme(req1.getNamingScheme());
        assertRequestEquals(req1, req2);

        req2.setProductRoot(new File("/some/where/else"));
        assertRequestNotEquals(req1, req2);
        req2.setProductRoot(req1.getProductRoot());
        assertRequestEquals(req1, req2);
    }

    private Request serializeDeserialize(Request req) 
            throws LTDLException, CPLException, ParameterValueException,
                   SAXException, IOException {
        try {
            ByteArrayOutputStream ostrm = new ByteArrayOutputStream();
            req.write(ostrm);
            ostrm.close();
            byte[] buffer = ostrm.toByteArray();
            validateConfig(new ByteArrayInputStream(buffer));
            InputStream istrm = new ByteArrayInputStream(buffer);
            return new Request(istrm, new LibraryLoader());
        }
        catch (IOException e) {
            throw new AssertionError();
        }
    }

    private void assertRequestEquals(Request req1, Request req2) 
            throws CPLException {

        Recipe rec1 = req1.getRecipe();
        Recipe rec2 = req2.getRecipe();
        assertTrue(rec1.getName() != null);
        assertEquals(rec1.getName(), rec2.getName());
        assertEquals(rec1.getAuthor(), rec2.getAuthor());
        assertEquals(rec1.getVersion(), rec2.getVersion());

        Parameter[] par1 = req1.getParameters();
        Parameter[] par2 = req2.getParameters();
        int npar = par1.length;
        assertTrue(npar > 0);
        assertEquals(npar, par2.length);
        for (int i = 0; i < npar; i++) {
            Parameter p1 = par1[i];
            Parameter p2 = par2[i];
            assertEquals(p1.getName(), p2.getName());
            assertEquals(p1.getValue(), p2.getValue());
        }

        FrameList fl1 = req1.getInputFrames();
        FrameList fl2 = req2.getInputFrames();
        int nf = fl1.size();
        assertEquals(nf, fl2.size());
        for (int i = 0; i < nf; i++) {
            assertEquals(fl1.get(i), fl2.get(i));
        }

        File root1 = req1.getProductRoot();
        File root2 = req2.getProductRoot();
        assertEquals(root1 == null ? new File(".") : root1,
                     root2 == null ? new File(".") : root2);

        NamingScheme namer1 = req1.getNamingScheme();
        NamingScheme namer2 = req2.getNamingScheme();
        assertEquals(namer1, namer2);
    }

    private void assertRequestNotEquals(Request req1, Request req2) 
            throws CPLException {
        boolean equal;
        try {
            assertRequestEquals(req1, req2);
            equal = true;
        }
        catch (AssertionFailedError e) {
            equal = false;
        }
        if (equal) {
            fail("Requests shouldn't have been equal, but were");
        }
    }

    private void validateConfig(InputStream strm)
            throws SAXException, IOException {
        String decl = new StringBuffer()
            .append("<?xml version='1.0'?>\n")
            .append("<!DOCTYPE request SYSTEM '")
            .append(getClass().getResource("/org/eso/cpl/cpl-state.dtd"))
            .append("'>\n")
            .toString();
        InputStream dstrm = new SequenceInputStream(
                                new ByteArrayInputStream(decl.getBytes()),
                                strm);
        // dstrm = new TeeInputStream(dstrm);
        SAXParser parser;
        try {
            SAXParserFactory sfact = SAXParserFactory.newInstance();
            sfact.setValidating(true);
            parser = sfact.newSAXParser();
        }
        catch (ParserConfigurationException e) {
            throw new AssertionFailedError(e.getMessage());
        }
        assertTrue(parser.isValidating());
        DefaultHandler handler = new DefaultHandler() {
            public void warning(SAXParseException e) {fail(e.getMessage());}
            public void error(SAXParseException e) {fail(e.getMessage());}
            public void fatalError(SAXParseException e) {fail(e.getMessage());}
        };

        // Do the parsing.  Note that we need to override the error handling
        // methods since by default the one(s) that handles validation errors
        // is a no-op.
        parser.parse(dstrm, new DefaultHandler() {
            public void warning( SAXParseException e ) throws SAXException {
                rethrow( e );
            }
            public void error( SAXParseException e ) throws SAXException {
                rethrow( e );
            }
            public void fatalError( SAXParseException e ) throws SAXException {
                rethrow( e );
            }
            private void rethrow( SAXParseException e ) throws SAXException {
                throw new SAXException( "Parse error in " + e.getSystemId()
                                      + " at line " + e.getLineNumber()
                                      + " column " + e.getColumnNumber(), e );
            }
        });
    }

    public static TestSuite suite() {
        return new TestSuite(RequestTest.class);
    }

    public static void main(String[] args) {
        TestRunner.run(suite());
    }

    class TeeInputStream extends FilterInputStream {
        public TeeInputStream(InputStream strm) {
            super(strm);
        }
        public int read() throws IOException {
            int i = super.read(); 
            System.out.print((char)i);
            return i;
        }
        public int read(byte[] b) throws IOException {
            int i = super.read(b);
            System.out.print(new String(b));
            return i;
        }
        public int read(byte[] b, int off, int len) throws IOException {
            int i = super.read(b,off,len);
            System.out.print(new String(b,off,len));
            return i;
        }
        public boolean markSupported() {
            return false;
        }
    };
}
